/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

// Package util contains some functions for format check
package util

import (
	"regexp"

	"github.com/pkg/errors"
)

const (
	// ChainIDReg is the chain id regex format.
	ChainIDReg = "^[A-Za-z0-9_][A-Za-z0-9_-]{0,99}$"

	// ChainAddrReg is the chain addr regex format.
	ChainAddrReg = "^[A-Za-z0-9+/=]{1,100}$"

	// ContractNameReg is the contract name regex format.
	ContractNameReg = "^[A-Za-z0-9_][A-Za-z0-9_-]{0,99}$"

	// ContractVersionReg is the contract version regex format.
	ContractVersionReg = "^[A-Za-z0-9_][A-Za-z0-9_\\.]{0,99}$"

	// TransactionIDReg is the transaction id regex format.
	TransactionIDReg = "^[A-Za-z0-9]{1,100}$"

	// maxDescLen is max description length.
	maxDescLen = 256

	// max policy length
	maxPolicyLen = 10000
)

// CheckChainID check the chain ID format
func CheckChainID(chainID string) error {
	ok, err := checkString(ChainIDReg, chainID)
	if err != nil {
		return errors.WithMessage(err, "check string error")
	}
	if !ok {
		return errors.Errorf("check chain id not match format: %v", ChainIDReg)
	}
	return nil
}

// CheckChainAddr check the chain address format
func CheckChainAddr(chainAddr string) error {
	ok, err := checkString(ChainAddrReg, chainAddr)
	if err != nil {
		return errors.WithMessage(err, "check string error")
	}
	if !ok {
		return errors.Errorf("check chain addr not match format: %v", ChainAddrReg)
	}
	return nil
}

// CheckContractName check the contract name format
func CheckContractName(name string) error {
	ok, err := checkString(ContractNameReg, name)
	if err != nil {
		return errors.WithMessage(err, "check string error")
	}
	if !ok {
		return errors.Errorf("check contract name not match format: %v", ContractNameReg)
	}
	return nil
}

// CheckContractVersion check the contract version format
func CheckContractVersion(version string) error {
	ok, err := checkString(ContractVersionReg, version)
	if err != nil {
		return errors.WithMessage(err, "check string error")
	}
	if !ok {
		return errors.Errorf("check contract version not match format: %v", ContractVersionReg)
	}
	return nil
}

// CheckTransactionID check transaction ID format
func CheckTransactionID(txID string) error {
	ok, err := checkString(TransactionIDReg, txID)
	if err != nil {
		return errors.WithMessage(err, "check string error")
	}
	if !ok {
		return errors.Errorf("check transaction id not match format: %v", TransactionIDReg)
	}
	return nil
}

// CheckPolicy check policy format
func CheckPolicy(policy string) error {
	if policy == "" {
		return errors.New("policy string cannot be empty")
	}
	if len(policy) > maxPolicyLen {
		return errors.Errorf("supported max policy string length: %v, but got: %v", maxPolicyLen, len(policy))
	}
	return nil
}

func checkDescription(desc string) error {
	if len(desc) > maxDescLen {
		return errors.Errorf("expected description length less than %v, but got %v", maxDescLen, len(desc))
	}
	return nil
}

// CheckParams check the format of contract parameters
func CheckParams(chainID string, contract string, version string) error {
	if err := CheckChainID(chainID); err != nil {
		return errors.WithMessage(err, "check chain id error")
	}
	if err := CheckContractName(contract); err != nil {
		return errors.WithMessage(err, "check contract name error")
	}
	if err := CheckContractVersion(version); err != nil {
		return errors.WithMessage(err, "check contract version error")
	}
	return nil
}

// CheckVoteParams check the format of vote parameters
func CheckVoteParams(chainID string, contract string, version string, policy string, desc string) error {
	if err := CheckParams(chainID, contract, version); err != nil {
		return errors.WithMessage(err, "check params error")
	}
	if err := CheckPolicy(policy); err != nil {
		return errors.WithMessage(err, "check policy error")
	}
	if err := checkDescription(desc); err != nil {
		return errors.WithMessage(err, "check description error")
	}
	return nil
}

func checkString(reg string, s string) (bool, error) {
	if reg == "" {
		return false, errors.New("string is empty")
	}
	r, err := regexp.Compile(reg)
	if err != nil {
		return false, errors.WithMessage(err, "regexp compile error")
	}
	return r.MatchString(s), nil
}
