/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

package crypto

import (
	"bytes"
	"crypto/aes"
	"crypto/cipher"
	"crypto/rand"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/cryptomgr"
)

type aes128 struct {
	key []byte
}

// NewAes128KeyWithRandNum new aes 128 key with key bytes
func NewAes128KeyWithRandNum(key []byte) (cryptomgr.SymmetricKey, error) {
	if len(key) != cryptomgr.SymmetricKeyLen {
		return nil, errors.New("the key len is not 128 bit")
	}
	aesKey := aes128{key: key}
	return &aesKey, nil
}

// NewAes128Key new aes 128 key
func NewAes128Key() (cryptomgr.SymmetricKey, error) {
	key, err := generateAES128Key()
	if err != nil {
		return nil, err
	}
	aesKey := aes128{key: key}
	return &aesKey, nil
}

// GetKeyBytes get bytes format key
func (a *aes128) GetKeyBytes() []byte {
	return a.key
}

// Encrypt encrypt plain txt
func (a *aes128) Encrypt(plainTxt []byte) ([]byte, error) {
	iv, err := generateAES128IV()
	if err != nil {
		return nil, err
	}

	cipherMsg, err := encWithAES128IV(a.key, iv, plainTxt)
	if err != nil {
		return nil, err
	}
	cipherMsg = append(cipherMsg, iv...)
	return cipherMsg, nil
}

// Decrypt decrypt message with cipher txt and iv
func (a *aes128) Decrypt(cipherTxtAndIv []byte) ([]byte, error) {
	length := len(cipherTxtAndIv)
	if length < cryptomgr.IvLen+cryptomgr.SymmetricKeyLen || length%cryptomgr.IvLen != 0 {
		return nil, errors.Errorf("the cipher msg length:%d is not correct", length)
	}

	iv := cipherTxtAndIv[length-cryptomgr.IvLen:]
	encMgs := cipherTxtAndIv[:length-cryptomgr.IvLen]
	return decWithAES128IV(a.key, iv, encMgs)
}

// generateAES128Key generate secure random number(128bit) as AES 128 key
func generateAES128Key() ([]byte, error) {
	aesKey := make([]byte, cryptomgr.SymmetricKeyLen)
	n, err := rand.Read(aesKey)
	if err != nil {
		return nil, err
	}
	if n != cryptomgr.SymmetricKeyLen {
		return nil, errors.Errorf("the aes 128 key len:%d is not correct", n)
	}
	return aesKey, nil
}

func generateAES128IV() ([]byte, error) {
	aes128IV := make([]byte, cryptomgr.IvLen)
	n, err := rand.Read(aes128IV)
	if err != nil {
		return nil, err
	}
	if n != cryptomgr.IvLen {
		return nil, errors.Errorf("the aes 128 iv len:%d is not correct", n)
	}
	return aes128IV, nil
}

// encWithAES128IV enc with aes 128 CBC mode, the enc result length is len(result)%16==0 && len(result)-len(text)<16
func encWithAES128IV(key []byte, iv []byte, text []byte) ([]byte, error) {
	// generate key data block
	block, err := aes.NewCipher(key)
	if err != nil {
		return nil, err
	}
	// pad to be 16x
	blockSize := block.BlockSize()
	originData := pad(text, blockSize)
	// get the cbc block mode
	blockMode := cipher.NewCBCEncrypter(block, iv)
	// crypto and output to encrypted
	encrypted := make([]byte, len(originData))
	blockMode.CryptBlocks(encrypted, originData)
	return encrypted, nil
}

// PKCS7 padding mode
func pad(ciphertext []byte, blockSize int) []byte {
	if blockSize == 0 {
		return ciphertext
	}
	padding := blockSize - len(ciphertext)%blockSize
	padtext := bytes.Repeat([]byte{byte(padding)}, padding)
	return append(ciphertext, padtext...)
}

func decWithAES128IV(key []byte, iv []byte, text []byte) ([]byte, error) {
	// get key block
	block, err := aes.NewCipher(key)
	if err != nil {
		return nil, err
	}
	// get the cbc mode
	blockMode := cipher.NewCBCDecrypter(block, iv)
	// output to the originData
	originData := make([]byte, len(text))
	blockMode.CryptBlocks(originData, text)
	// delete the padding
	return unpad(originData), nil
}

func unpad(ciphertext []byte) []byte {
	length := len(ciphertext)
	// delete the last time padding
	unpadding := int(ciphertext[length-1])
	return ciphertext[:(length - unpadding)]
}
