/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2021. All rights reserved.
 */

// Package crypto the international cryptographic algorithm for generation of certificate/key/hashFunc
package crypto

import (
	"crypto/ecdsa"
	"crypto/elliptic"
	"crypto/rand"
	"crypto/x509"
	"encoding/pem"
	"fmt"
	"math/big"
	"strings"
	"time"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/cryptomgr"
	"huawei.com/huaweichain/common/logger"
)

var log = logger.GetModuleLogger("crypto", "ecdsaalg")

const daysInOneYear = 365
const hoursInOneDay = 24

type ecdsacert struct {
	pemCert []byte
	cert    *x509.Certificate
	pubKey  *ecdsa.PublicKey
}

func generateCertBasic(info *cryptomgr.CertBasicInfo) (*x509.Certificate, error) {
	if info == nil {
		return nil, errors.New("failed to create cert basic, info is nil")
	}
	numberLimit := new(big.Int).Lsh(big.NewInt(1), 128)
	serialNumber, err := rand.Int(rand.Reader, numberLimit)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to create cert basic")
	}

	currentTime := time.Now()
	cert := &x509.Certificate{
		SerialNumber: serialNumber,
		NotBefore:    currentTime,
		NotAfter: currentTime.Add(time.Hour * daysInOneYear * hoursInOneDay *
			time.Duration(info.ValidationYears)),
		BasicConstraintsValid: true,
	}

	cert.Subject.Country = []string{info.Country}

	cert.Subject.Province = []string{info.Province}

	cert.Subject.Locality = []string{info.Locality}

	cert.Subject.Organization = []string{info.Organization}

	cert.Subject.OrganizationalUnit = []string{info.OrganizationUnit}

	cert.Subject.CommonName = info.CommonName

	cert.DNSNames = []string{info.CommonName}

	return cert, nil
}

func generateEcdsaCert(template *x509.Certificate, pubKey *ecdsa.PublicKey, caCert *x509.Certificate,
	caPriKey *ecdsa.PrivateKey) (*ecdsacert, error) {
	cert, err := x509.CreateCertificate(rand.Reader, template, caCert, pubKey, caPriKey)
	if err != nil {
		log.Errorf("Create X509 self sign cert failed:%s", err.Error())
		return nil, errors.WithMessage(err, "create X509 self sign cert failed")
	}
	certificate, err := x509.ParseCertificate(cert)
	if err != nil {
		log.Errorf("Get x509 certificate failed:%s", err)
		return nil, errors.WithMessage(err, "get x509 certificate failed")
	}

	memory := pem.EncodeToMemory(&pem.Block{Type: "CERTIFICATE", Bytes: cert})

	ecdsaPubKey, ok := certificate.PublicKey.(*ecdsa.PublicKey)
	if !ok {
		return nil, errors.New("the public key is not ecdsa key")
	}

	ecdsaCert := &ecdsacert{
		pemCert: memory,
		cert:    certificate,
		pubKey:  ecdsaPubKey,
	}

	return ecdsaCert, nil
}

// GenerateMiddleCACert generate middle ca cert
func GenerateMiddleCACert(info *cryptomgr.CertBasicInfo, pubKey cryptomgr.Key, caPriKey cryptomgr.Key,
	caCert cryptomgr.Cert) (cryptomgr.Cert, error) {
	if info == nil || pubKey == nil || caPriKey == nil || caCert == nil {
		return nil, errors.New("input parameter is not correct")
	}
	certBasic, err := generateCertBasic(info)
	if err != nil {
		return nil, errors.WithMessage(err, "create cert basic failed")
	}

	certBasic.IsCA = true
	keyUsage := x509.KeyUsageDigitalSignature |
		x509.KeyUsageKeyEncipherment |
		x509.KeyUsageCertSign |
		x509.KeyUsageCRLSign
	certBasic.KeyUsage = keyUsage
	certBasic.ExtKeyUsage = []x509.ExtKeyUsage{x509.ExtKeyUsageServerAuth, x509.ExtKeyUsageClientAuth}

	ecdsaCaPriKey, ok := caPriKey.(*ecdsaKey)
	if !ok {
		log.Error("This is not a ecdsa private key")
		return nil, errors.New("priKey is not a ecdsa private key")
	}
	ecdsaCaCert, ok := caCert.(*ecdsacert)
	if !ok {
		return nil, errors.New("the cert is not ecdsa cert")
	}

	ecdsaPubKey, ok := pubKey.(*ecdsaKey)
	if !ok {
		return nil, errors.New("the pub key is not ecdsa key")
	}
	if ecdsaPubKey.pubKey == nil {
		return nil, errors.New("the pubKey is nil")
	}
	certBasic.SubjectKeyId = Hash(elliptic.Marshal(ecdsaPubKey.curve, ecdsaPubKey.pubKey.X, ecdsaPubKey.pubKey.Y))

	return generateEcdsaCert(certBasic, ecdsaPubKey.pubKey, ecdsaCaCert.cert, ecdsaCaPriKey.priKey)
}

// GenerateCert create self signed ca cert.
func GenerateCert(info *cryptomgr.CertBasicInfo, pubKey cryptomgr.Key, caPriKey cryptomgr.Key,
	caCert cryptomgr.Cert) (cryptomgr.Cert, error) {
	if info == nil || pubKey == nil || caPriKey == nil || caCert == nil {
		return nil, errors.New("input parameter is not correct")
	}

	certBasic, err := generateCertBasic(info)
	if err != nil {
		return nil, errors.WithMessage(err, "generate cert basic template failed")
	}
	certBasic.IsCA = false
	certBasic.KeyUsage = x509.KeyUsageDigitalSignature |
		x509.KeyUsageKeyEncipherment

	ecdsaCaCert, ok := caCert.(*ecdsacert)
	if !ok {
		return nil, errors.New("this is not a ecdsa certification")
	}
	caPrivateKey, ok := caPriKey.(*ecdsaKey)
	if !ok {
		return nil, errors.New("this is not a ecdsa private key")
	}
	ecdsaPubKey, ok := pubKey.(*ecdsaKey)
	if !ok {
		return nil, errors.New("this is not a ecdsa pub key")
	}

	return generateEcdsaCert(certBasic, ecdsaPubKey.pubKey, ecdsaCaCert.cert, caPrivateKey.priKey)
}

// GenerateSelfSignCert create self signed ca cert.
func GenerateSelfSignCert(info *cryptomgr.CertBasicInfo,
	key cryptomgr.Key) (cryptomgr.Cert, error) {
	if info == nil || key == nil {
		return nil, errors.New("input parameter is not correct")
	}
	certBasic, err := generateCertBasic(info)
	if err != nil {
		return nil, errors.WithMessage(err, "generate cert basic template failed")
	}
	certBasic.IsCA = true
	keyUsage := x509.KeyUsageDigitalSignature |
		x509.KeyUsageKeyEncipherment |
		x509.KeyUsageCertSign |
		x509.KeyUsageCRLSign
	certBasic.KeyUsage = keyUsage
	certBasic.ExtKeyUsage = []x509.ExtKeyUsage{x509.ExtKeyUsageServerAuth, x509.ExtKeyUsageClientAuth}

	privateKey, ok := key.(*ecdsaKey)
	if !ok {
		log.Error("PriKey is not a ecdsa private key")
		return nil, errors.New("private key is not ecdsa key")
	}

	if privateKey.priKey == nil {
		return nil, errors.New("the priKey is nil")
	}
	certBasic.SubjectKeyId = Hash(elliptic.Marshal(privateKey.curve, privateKey.priKey.X, privateKey.priKey.Y))

	return generateEcdsaCert(certBasic, privateKey.pubKey, certBasic, privateKey.priKey)
}

// GetCert get certificate from pem bytes.
func GetCert(pemCert []byte) (cryptomgr.Cert, error) {
	if len(pemCert) == 0 {
		log.Infof("the pem cert is nil")
		return nil, errors.New("the pem cert is nil")
	}

	block, _ := pem.Decode(pemCert)

	if block == nil {
		return nil, errors.New("decode the pem cert failed")
	}
	x509cert, err := x509.ParseCertificate(block.Bytes)
	if err != nil {
		return nil, errors.Errorf("parse certificate failed: %s", err.Error())
	}
	pubKey := x509cert.PublicKey
	ecdsaPubKey, ok := pubKey.(*ecdsa.PublicKey)
	if !ok {
		return nil, errors.New("the key is not ecdsa public key")
	}
	if x509cert.SignatureAlgorithm != x509.ECDSAWithSHA256 {
		log.Errorf("the sign algorithm of certificate is not ecdsa_with_sha256:%d", x509cert.SignatureAlgorithm)
		return nil, errors.New("the cert signature algorithm is not ecdsa_with_sha256")
	}

	eccert := &ecdsacert{
		pemCert: pemCert,
		cert:    x509cert,
		pubKey:  ecdsaPubKey,
	}

	return eccert, nil
}

// CheckEcdsaCert check certificate interface implemented with ecdsa cert struct
func CheckEcdsaCert(cert cryptomgr.Cert) error {
	if cert == nil {
		return errors.New("the cert is nil")
	}
	ecdsaCert, ok := cert.(*ecdsacert)
	if !ok {
		return errors.New("the cert is not ecdsa cert")
	}
	if ecdsaCert.pemCert == nil {
		return errors.New("the pem format cert is nil")
	}
	if ecdsaCert.pubKey == nil {
		return errors.New("the ecdsa public key is nil")
	}
	if ecdsaCert.cert == nil {
		return errors.New("the ecdsa certificate is nil")
	}
	return nil
}

// GetFingerPrint get the ginger print of this cert.
func (e *ecdsacert) GetFingerPrint() string {
	hashInfo := Hash(e.cert.Raw)
	fingerStr := fmt.Sprintf("%x", hashInfo)
	return fingerStr
}

// GetSerialNumber get serial number of cert
func (e *ecdsacert) GetSerialNumber() string {
	if e.cert == nil {
		return ""
	}
	const hexStrType = 16
	return strings.ToUpper(e.cert.SerialNumber.Text(hexStrType))
}

// GetCommonName get common name of this cert.
func (e *ecdsacert) GetCommonName() string {
	return e.cert.Subject.CommonName
}

// GetSignAlg get sign algorithm of certificate
func (e *ecdsacert) GetSignAlg() (string, error) {
	if e.cert == nil {
		return "", errors.New("the cert is empty")
	}
	if e.cert.SignatureAlgorithm != x509.ECDSAWithSHA256 {
		log.Errorf("the sign algorithm is %s, not EcdsaWithSha256", e.cert.SignatureAlgorithm.String())
		return "", errors.Errorf("the sign algorithm is not EcdsaWithSha256:%s", e.cert.SignatureAlgorithm.String())
	}
	return cryptomgr.EcdsaWithSha256, nil
}

// GetDerPublicKey get der public key from cert
func (e *ecdsacert) GetDerPublicKey() ([]byte, error) {
	if e.cert == nil {
		return nil, errors.New("the x509 cert is nil")
	}
	pubKey, ok := e.cert.PublicKey.(*ecdsa.PublicKey)
	if !ok {
		return nil, errors.New("the public key is not ecdsa public key")
	}
	pubKeyDer, err := x509.MarshalPKIXPublicKey(pubKey)

	if err != nil {
		return nil, errors.New("get der format public key failed")
	}
	return pubKeyDer, nil
}

// GetExpireTime get expire time for this cert.
func (e *ecdsacert) GetExpireTime() time.Time {
	return e.cert.NotAfter
}

// GetOrganizationalUnit get organization unit of this cert.
func (e *ecdsacert) GetOrganizationalUnit() []string {
	return e.cert.Subject.OrganizationalUnit
}

// GetOrganization get organization.
func (e *ecdsacert) GetOrganization() []string {
	return e.cert.Subject.Organization
}

// Verify verify the msg signature.
func (e *ecdsacert) Verify(msg []byte, signature []byte, hashOpt string) error {
	pubKey, ok := e.cert.PublicKey.(*ecdsa.PublicKey)
	if ok {
		return verifyWithPubKey(pubKey, msg, signature)
	}

	return errors.New("verify failed the public key is not ecdsa publickey")
}

func (e *ecdsacert) checkSigWithRoots(rootCerts []cryptomgr.Cert) error {
	for _, root := range rootCerts {
		ecdsaCert, ok := root.(*ecdsacert)
		if !ok {
			continue
		}

		err := e.cert.CheckSignatureFrom(ecdsaCert.cert)
		if err != nil {
			continue
		}
		return nil
	}
	return errors.New("no CA has signed the certificate")
}

// CheckValidation check whether this cert is signed by the key corresponding to one of the root certificates
func (e *ecdsacert) CheckValidation(rootCerts []cryptomgr.Cert) (string, error) {
	err := e.checkSigWithRoots(rootCerts)
	if err != nil {
		return "", err
	}

	var opts x509.VerifyOptions
	certPool := x509.NewCertPool()

	// Get bytes
	for _, v := range rootCerts {
		ecdsaCert, ok := v.(*ecdsacert)
		if !ok {
			continue
		}
		certPool.AddCert(ecdsaCert.cert)
	}
	opts.Roots = certPool

	chains, err := e.cert.Verify(opts)
	if err != nil {
		return "", err
	}
	return e.rootOrg(chains[0])
}

func (e *ecdsacert) rootOrg(chains []*x509.Certificate) (string, error) {
	if len(chains) == 0 {
		return "", errors.New("len of cert chains is 0")
	}
	rootOrgs := chains[len(chains)-1].Subject.Organization
	if len(rootOrgs) == 0 {
		return "", nil
	}
	return rootOrgs[0], nil
}

// GetPemCertBytes get the pem cert bytes of this cert.
func (e *ecdsacert) GetPemCertBytes() []byte {
	return e.pemCert
}

func (e *ecdsacert) GetBasicInfo() *cryptomgr.CertBasicInfo {
	organization := ""
	if len(e.cert.Subject.Organization) > 0 {
		organization = e.cert.Subject.Organization[0]
	}
	organizationalUnit := ""
	if len(e.cert.Subject.OrganizationalUnit) > 0 {
		organizationalUnit = e.cert.Subject.OrganizationalUnit[0]
	}
	country := ""
	if len(e.cert.Subject.Country) > 0 {
		country = e.cert.Subject.Country[0]
	}
	province := ""
	if len(e.cert.Subject.Province) > 0 {
		province = e.cert.Subject.Province[0]
	}
	locality := ""
	if len(e.cert.Subject.Locality) > 0 {
		locality = e.cert.Subject.Locality[0]
	}
	validateYears := e.cert.NotAfter.Year() - e.cert.NotBefore.Year()

	return &cryptomgr.CertBasicInfo{
		Organization:     organization,
		OrganizationUnit: organizationalUnit,
		Country:          country,
		Province:         province,
		Locality:         locality,
		CommonName:       e.cert.Subject.CommonName,
		ValidationYears:  uint(validateYears),
	}
}

// CheckPublicKeyLength check whether the public key length meets the security requirements
func (e *ecdsacert) CheckPublicKeyLength() error {
	if e.cert == nil {
		return errors.New("the x509 cert is nil")
	}
	pubKey, ok := e.cert.PublicKey.(*ecdsa.PublicKey)
	if !ok {
		return errors.New("the public key is not ecdsa public key")
	}
	publicKeyLength := len(elliptic.Marshal(pubKey.Curve, pubKey.X, pubKey.Y))
	if publicKeyLength < cryptomgr.ECCPublicKeyLength {
		return errors.Errorf("the length: %d do not meets the security requirements", publicKeyLength)
	}
	return nil
}

// IsCaCert check if the cert is ca cert
func (e *ecdsacert) IsCaCert() (bool, error) {
	if e.cert == nil {
		return false, errors.New("invalid x509 cert")
	}
	return e.cert.IsCA, nil
}
