/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2020. All rights reserved.
 */

// Package crypto is the international cryptographic algorithms implementation
package crypto

import (
	"crypto/elliptic"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/cryptomgr"
	"huawei.com/huaweichain/common/cryptomgr/cryptoimpl"
)

func init() {
	cryptoimpl.RegisterCertFactory(&EcdsaCertFactory{name: cryptomgr.EcdsaWithSha256})
	// for ecdsa_with_sha256 config, use the secp256r1 elliptic curve
	cryptoimpl.RegisterAsKeyFactory(&EcdsaP256KeyFactory{name: cryptomgr.EcdsaWithSha256})
	cryptoimpl.RegisterHashFunc(cryptomgr.Sha256, Hash)
	cryptoimpl.RegisterHashForFileFunc(cryptomgr.Sha256, HashForFile)
	cryptoimpl.RegisterSymmetricKeyFactory(&AESKeyFactory{name: cryptomgr.Aes128})
}

// AESKeyFactory AES key factory
type AESKeyFactory struct {
	name string
}

// Name get AES key factory name
func (ak *AESKeyFactory) Name() string {
	return ak.name
}

// NewSymmetricKeyWithRandNum new symmetric key with base64 key string
func (ak *AESKeyFactory) NewSymmetricKeyWithRandNum(key []byte) (cryptomgr.SymmetricKey, error) {
	return NewAes128KeyWithRandNum(key)
}

// NewSymmetricKey new symmetric key
func (ak *AESKeyFactory) NewSymmetricKey() (cryptomgr.SymmetricKey, error) {
	return NewAes128Key()
}

// EcdsaP256KeyFactory ecdsa p256 key factory
type EcdsaP256KeyFactory struct {
	name string
}

// Name get the ecdsa p256 key factory name
func (ek *EcdsaP256KeyFactory) Name() string {
	return ek.name
}

// GenerateKey create key.
func (ek *EcdsaP256KeyFactory) GenerateKey() (cryptomgr.Key, error) {
	curve := elliptic.P256()
	return CreateKeyWithCurve(curve)
}

// GetKeyFromPem get key from pem bytes.
func (ek *EcdsaP256KeyFactory) GetKeyFromPem(pemKey []byte) (cryptomgr.Key, error) {
	keyLen := len(pemKey)
	if keyLen == 0 || keyLen > cryptomgr.MaxPEMKeyLen {
		return nil, errors.Errorf("the key len is %d, which is not correct", keyLen)
	}
	return GetKeyFromPem(pemKey)
}

// GetPubKeyFromDer get key from der bytes.
func (ek *EcdsaP256KeyFactory) GetPubKeyFromDer(derKey []byte) (cryptomgr.Key, error) {
	keyLen := len(derKey)
	if keyLen == 0 || keyLen > cryptomgr.MaxDerKeyLen {
		return nil, errors.Errorf("the key len is %d, which is not correct", keyLen)
	}
	return GetPubKeyFromDer(derKey)
}

// CreateKeyFromBuffer create key from tee buffer
func (ek *EcdsaP256KeyFactory) CreateKeyFromBuffer(buffer []byte) (cryptomgr.Key, error) {
	return CreateKeyFromBuffer(buffer)
}

// EcdsaCertFactory ecdsa certificate factory
type EcdsaCertFactory struct {
	name string
}

// Name get ecdsa cert factory name
func (ec *EcdsaCertFactory) Name() string {
	return ec.name
}

// GenerateSelfSignCert create self sign CA certificate
func (ec *EcdsaCertFactory) GenerateSelfSignCert(info *cryptomgr.CertBasicInfo,
	key cryptomgr.Key) (cryptomgr.Cert, error) {
	err := cryptoimpl.CheckCertInfo(info)
	if err != nil {
		return nil, err
	}

	err = CheckEcdsaPriKey(key)
	if err != nil {
		return nil, err
	}

	return GenerateSelfSignCert(info, key)
}

// GenerateMiddleCACert generate middle ca cert
func (ec *EcdsaCertFactory) GenerateMiddleCACert(info *cryptomgr.CertBasicInfo, pubKey cryptomgr.Key,
	caPriKey cryptomgr.Key, caCert cryptomgr.Cert) (cryptomgr.Cert, error) {
	err := checkEcdsaCertPara(info, pubKey, caPriKey, caCert)
	if err != nil {
		return nil, err
	}
	return GenerateMiddleCACert(info, pubKey, caPriKey, caCert)
}

// GenerateCert create certificate
func (ec *EcdsaCertFactory) GenerateCert(info *cryptomgr.CertBasicInfo, pubKey cryptomgr.Key, caPriKey cryptomgr.Key,
	caCert cryptomgr.Cert) (cryptomgr.Cert, error) {
	err := checkEcdsaCertPara(info, pubKey, caPriKey, caCert)
	if err != nil {
		return nil, err
	}
	return GenerateCert(info, pubKey, caPriKey, caCert)
}

func checkEcdsaCertPara(info *cryptomgr.CertBasicInfo, pubKey cryptomgr.Key, caPriKey cryptomgr.Key,
	caCert cryptomgr.Cert) error {
	err := cryptoimpl.CheckCertInfo(info)
	if err != nil {
		return err
	}
	err = CheckEcdsaPubKey(pubKey)
	if err != nil {
		return err
	}
	err = CheckEcdsaPriKey(caPriKey)
	if err != nil {
		return err
	}
	err = CheckEcdsaCert(caCert)
	if err != nil {
		return err
	}
	return nil
}

// GetCertFromPem get certificate from pem.
func (ec *EcdsaCertFactory) GetCertFromPem(pemCert []byte) (cryptomgr.Cert, error) {
	certLen := len(pemCert)
	if certLen == 0 {
		return nil, errors.New("the cert len is 0, which is not correct")
	}
	return GetCert(pemCert)
}
