/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package cryptoimpl

import (
	"math/big"
	"strings"
	"unsafe"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/cryptomgr"
)

func checkWhiteSpaceOfCN(commonName string) error {
	for _, v := range commonName {
		if v != ' ' {
			return nil
		}
	}
	return errors.New("the common name should not be whitespace or array of whitespace")
}

// CheckCertInfo check the info for cert generation
func CheckCertInfo(info *cryptomgr.CertBasicInfo) error {
	if info == nil {
		return errors.New("the cert info is nil")
	}

	if len(info.CommonName) == 0 || len(info.CommonName) > cryptomgr.MaxCommonNameLen {
		return errors.Errorf("the length of common name is %d, which exceeds max common name length %d",
			len(info.CommonName), cryptomgr.MaxCommonNameLen)
	}
	// the common name should be used as dns name, which must not be " "(rfc 5280:In addition,
	// while the string " " is a legal domain name, subjectAltName extensions with a dNSName of " " MUST NOT be used. ).
	err := checkWhiteSpaceOfCN(info.CommonName)
	if err != nil {
		return err
	}
	if len(info.Locality) > cryptomgr.MaxLocalityLen {
		return errors.Errorf("the locality length is %d, which exceeds max locality length %d",
			len(info.Locality), cryptomgr.MaxLocalityLen)
	}
	if len(info.Province) > cryptomgr.MaxStateLen {
		return errors.Errorf("the province length is %d, which exceeds max province length %d",
			len(info.Province), cryptomgr.MaxStateLen)
	}
	if len(info.Country) > cryptomgr.MaxCountryLen {
		return errors.Errorf("the country length is %d, which exceeds max country length %d, for example, "+
			"China should be be identified as CN", len(info.Country), cryptomgr.MaxCountryLen)
	}
	if len(info.OrganizationUnit) > cryptomgr.MaxOrgUnitLen {
		return errors.Errorf("the organization unit length is %d, which exceeds max org unit length %d",
			len(info.OrganizationUnit), cryptomgr.MaxOrgUnitLen)
	}
	if len(info.Organization) > cryptomgr.MaxOrgNameLen {
		return errors.Errorf("the organization name length is %d, which exceeds max organization name length %d",
			len(info.Organization), cryptomgr.MaxOrgNameLen)
	}
	if info.ValidationYears < cryptomgr.MinValidationYears || info.ValidationYears > cryptomgr.MaxValidationYears {
		return errors.Errorf("the validation years should be %d <= validation years <= %d",
			cryptomgr.MinValidationYears, cryptomgr.MaxValidationYears)
	}
	return nil
}

// GetCfgHashAlg get the hash option from config
func GetCfgHashAlg(signAlg string) string {
	alg := strings.ToLower(signAlg)
	switch alg {
	case cryptomgr.Sm2WithSm3:
		return cryptomgr.Sm3
	case cryptomgr.EcdsaWithSha256:
		return cryptomgr.Sha256
	case cryptomgr.Ed25519:
		return cryptomgr.Sha256
	default:
		return ""
	}
}

// CleanBytes clean the memory of bytes
func CleanBytes(info []byte) {
	for i := 0; i < len(info); i++ {
		info[i] = 0
	}
}

// CleanBigInt clean the memory of big.Int
func CleanBigInt(a *big.Int) {
	if a == nil {
		return
	}
	bitInfo := a.Bits()
	for i := 0; i < len(bitInfo); i++ {
		bitInfo[i] = '0'
	}
}

// CleanString clear string sensitive info
func CleanString(s string) {
	if len(s) <= 1 {
		return
	}
	bs := *(*[]byte)(unsafe.Pointer(&s))
	for i := 0; i < len(bs); i++ {
		bs[i] = 0
	}
}
