/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2020. All rights reserved.
 */

// Package cryptomgr crypto management.
package cryptomgr

import (
	"math/big"
	"time"
)

// 对于签名，我们需要用wnode节点本地的私钥进行签名
// 对于验签，如果是链外请求（比如创建链，查询链），需要用wnode节点本地的CA证书验证消息中证书的可靠性，保证请求者是本节点所属组织的管理员
// 对于验签，如果是链内请求，需要用链上的各个组织内的CA验证消息中证书的可靠性
// Symmetric key type
const (
	AES256 string = "aes_256"
	SM4    string = "sm4"
)

const (
	// GmSSLSm2WithSm3 gmssl sm2withsm3 algorithm flag
	GmSSLSm2WithSm3 = "sm2sign-with-sm3"
	// Sm2WithSm3  gm sign algorithm
	Sm2WithSm3 = "sm2_with_sm3"
	// Secp256r1Key is asymmetric key with ECC curve secp256r1
	Secp256r1Key = "secp256r1_key"

	// Sm2p256v1Key is asymmetric key with ECC curve sm2p256v1
	Sm2p256v1Key = "sm2p256v1_key"

	// EcdsaWithSha256 ecdsa sign algorithm
	EcdsaWithSha256 = "ecdsa_with_sha256"
	// Ed25519 ed25519 sign algorithm
	Ed25519 = "ed25519"
	// Sm3 sm3 hash algorithm
	Sm3 = "sm3"
	// Sha256 sha256 hash algorithm
	Sha256 = "sha256"

	// Sm4 symmetric encryption algorithm
	Sm4 = "sm4"

	// Aes128 symmetric encryption algorithm
	Aes128 = "aes128"

	// ECCPublicKeyLength meets the security requirements of public key length
	ECCPublicKeyLength = 65

	// SymmetricKeyLen symmetric key len
	SymmetricKeyLen = 16

	// IvLen iv len
	IvLen = 16
)

// ECDSASignature ecdsa signature struct with R and S.
type ECDSASignature struct {
	R, S *big.Int
}

// Key describe the key
type Key interface {
	GetPemBytes() []byte

	GetDerBytes() ([]byte, error)

	IsSymmetric() bool

	IsPrivate() bool

	GetPublicKey() Key

	Signature

	Verification

	CleanKeySensitiveInfo()
}

// Signature sign interface.
type Signature interface {
	Sign(msg []byte, hashAlg string) ([]byte, error)
}

// Verification verify interface.
type Verification interface {
	Verify(msg []byte, signature []byte, hashAlg string) error
}

// SymmetricKey symmetric key interface
type SymmetricKey interface {
	Encrypt(plainTxt []byte) ([]byte, error)
	Decrypt(cipherTxtAndIv []byte) ([]byte, error)
	GetKeyBytes() []byte
}

// CertBasicInfo the certificate basic information
type CertBasicInfo struct {
	Organization     string
	OrganizationUnit string
	Country          string
	Province         string
	Locality         string
	CommonName       string
	ValidationYears  uint
}

const (
	// MaxValidationYears the max validation years for certificate generation
	MaxValidationYears = 30
	// MinValidationYears the min validation years for certificate generation
	MinValidationYears = 1
	// MaxCommonNameLen the max common name length for certificate generation
	MaxCommonNameLen = 64
	// MaxOrgNameLen the max organization name length for certificate generation
	MaxOrgNameLen = 64
	// MaxOrgUnitLen the max organization unit name length for certificate generation
	MaxOrgUnitLen = 64
	// MaxCountryLen the max country length for certificate generation
	MaxCountryLen = 3
	// MaxStateLen the max state length for certificate generation
	MaxStateLen = 128
	// MaxLocalityLen the max locality length for certificate generation
	MaxLocalityLen = 128
	// MaxPEMKeyLen the max supported pem key length
	MaxPEMKeyLen = 2048
	// MaxDerKeyLen the max supported der key length
	MaxDerKeyLen = 2048
)

// Cert describe the certificate
type Cert interface {
	GetCommonName() string
	GetDerPublicKey() ([]byte, error)
	GetExpireTime() time.Time
	GetOrganizationalUnit() []string
	GetOrganization() []string
	Verification
	CheckValidation(rootCerts []Cert) (string, error)
	GetPemCertBytes() []byte
	GetFingerPrint() string
	GetSerialNumber() string
	GetSignAlg() (string, error)
	GetBasicInfo() *CertBasicInfo
	CheckPublicKeyLength() error
	IsCaCert() (bool, error)
}
