/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

// Package credentials implementation "google.golang.org/grpc/credentials"
package credentials

import (
	"context"
	"net"

	"google.golang.org/grpc/credentials"

	"huawei.com/huaweichain/gmssl/tls"
)

var (
	// alpnProtoStr are the specified application level protocols for gRPC.
	alpnProtoStr = []string{"h2"}
)

// TLSInfo contains the auth information for a TLS authenticated connection.
// It implements the AuthInfo interface.
type TLSInfo struct {
	State tls.ConnectionState
}

// AuthType returns the type of TLSInfo as a string.
func (t TLSInfo) AuthType() string {
	return "tls"
}

type tlsCreds struct {
	// TLS configuration
	config *tls.Config
}

// NewTLS uses c to construct a TransportCredentials based on TLS.
func NewTLS(c *tls.Config) credentials.TransportCredentials {
	tc := &tlsCreds{config: cloneTLSConfig(c)}
	tc.config.NextProtos = alpnProtoStr
	return tc
}

// ClientHandshake implements the client side handshake protocol.
func (c *tlsCreds) ClientHandshake(ctx context.Context, authority string,
	rawConn net.Conn) (net.Conn, credentials.AuthInfo, error) {
	cfg := cloneTLSConfig(c.config)
	if cfg.ServerName == "" {
		serverName, _, err := net.SplitHostPort(authority)
		if err != nil {
			// If the authority had no host port or if the authority cannot be parsed, use it as-is.
			serverName = authority
		}
		cfg.ServerName = serverName
	}
	conn := tls.Client(rawConn, cfg)
	if err := conn.Handshake(); err != nil {
		return nil, nil, err
	}
	return conn, TLSInfo{State: conn.ConnectionState()}, nil
}

// ServerHandshake implements the server side ALTS handshaker.
func (c *tlsCreds) ServerHandshake(rawConn net.Conn) (net.Conn, credentials.AuthInfo, error) {
	cfg := cloneTLSConfig(c.config)
	conn := tls.Server(rawConn, cfg)
	if err := conn.Handshake(); err != nil {
		return nil, nil, err
	}
	return conn, TLSInfo{State: conn.ConnectionState()}, nil
}

// Info provides the ProtocolInfo of this TransportCredentials.
func (c *tlsCreds) Info() credentials.ProtocolInfo {
	return credentials.ProtocolInfo{
		SecurityProtocol: "tls",
		SecurityVersion:  "1.2",
		ServerName:       c.config.ServerName,
	}
}

// Clone makes a copy of this TransportCredentials.
func (c *tlsCreds) Clone() credentials.TransportCredentials {
	return nil
}

// OverrideServerName overrides the server name used to verify the hostname
// on the returned certificates from the server.
func (c *tlsCreds) OverrideServerName(serverNameOverride string) error {
	c.config.ServerName = serverNameOverride
	return nil
}

func cloneTLSConfig(cfg *tls.Config) *tls.Config {
	if cfg == nil {
		return &tls.Config{}
	}
	return cfg.Clone()
}
