/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package credentials

import (
	"context"
	"io/ioutil"
	"log"
	"net"
	"sync"
	"testing"

	"google.golang.org/grpc"
	"huawei.com/huaweichain/gmssl/credentials/testdata/echo"
	"huawei.com/huaweichain/gmssl/tls"
)

const (
	port    = ":50051"
	address = "localhost:50051"
)

var end chan bool

type server struct{}

func (s *server) Echo(ctx context.Context, req *echo.EchoRequest) (*echo.EchoResponse, error) {
	return &echo.EchoResponse{Result: req.Req}, nil
}

func (s *server) StreamEcho(stream echo.Echo_StreamEchoServer) error {
	for {
		req, err := stream.Recv()
		if err != nil {
			log.Fatal(err)
		}
		for i := 0; i < 10; i++ {
			resp := &echo.EchoResponse{Result: "stream accept: " + req.Req}
			stream.Send(resp)
		}
	}
	return nil
}

const rootFile = "testdata/root.cert.pem"
const certFile = "testdata/CLIENT.cert.pem"
const keyFile = "testdata/CLIENT.private.pkcs8.key.pem"

func serverRun() {
	cfg := getCfg()
	cfg.ClientAuth = tls.RequireAnyClientCert
	creds := NewTLS(cfg)
	s := grpc.NewServer(grpc.Creds(creds))
	echo.RegisterEchoServer(s, &server{})

	lis, err := net.Listen("tcp", port)
	if err != nil {
		log.Fatalf("fail to listen: %v", err)
	}
	err = s.Serve(lis)
	if err != nil {
		log.Fatalf("Serve: %v", err)
	}
}

func streamClient() {
	cfg := getCfg()
	creds := NewTLS(cfg)
	conn, err := grpc.Dial(address, grpc.WithTransportCredentials(creds))
	if err != nil {
		log.Fatalf("cannot to connect: %v", err)
	}

	c := echo.NewEchoClient(conn)
	stream, err := c.StreamEcho(context.Background())
	if err != nil {
		log.Fatal(err)
	}

	for i := 0; i < 100; i++ {
		group := sync.WaitGroup{}
		group.Add(10)
		go func() {
			for i := 0; i < 10; i++ {
				res, err := stream.Recv()
				if err != nil {
					log.Fatal(err)
				}
				if res.Result != "stream accept: aaa" {
					log.Fatalf("expect 'stream accept: aaa', real '%v'", res.Result)
				}
				group.Done()
			}
		}()
		if err := stream.Send(&echo.EchoRequest{Req: "aaa"}); err != nil {
			log.Fatal(err)
		}
		group.Wait()
	}
}

func clientRun() {
	cfg := getCfg()
	creds := NewTLS(cfg)
	conn, err := grpc.Dial(address, grpc.WithTransportCredentials(creds))
	if err != nil {
		log.Fatalf("cannot to connect: %v", err)
	}

	c := echo.NewEchoClient(conn)
	for i := 0; i < 100; i++ {
		res, err := c.Echo(context.Background(), &echo.EchoRequest{Req: "hello"})
		if err != nil {
			log.Fatalf("failed to echo: %v", err)
		}
		if res.Result != "hello" {
			log.Fatalf("expect 'hello', real '%v'", res.Result)
		}
	}
	end <- true
}

func getCfg() *tls.Config {
	key, err := ioutil.ReadFile(keyFile)
	if err != nil {
		log.Fatal(err)
	}
	cert, err := ioutil.ReadFile(certFile)
	if err != nil {
		log.Fatal(err)
	}
	root, err := ioutil.ReadFile(rootFile)
	if err != nil {
		log.Fatal(err)
	}

	return &tls.Config{
		Key:        key,
		Certs:      [][]byte{cert},
		Roots:      [][]byte{root},
		ServerName: "oss.PMS.com",
	}
}

func Test(t *testing.T) {
	end = make(chan bool, 64)
	go serverRun()
	go clientRun()
	streamClient()
	<-end
}
