//go:build !nogm
// +build !nogm

/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2021. All rights reserved.
 */

// Package gmcrypto the gm algorithm for generate of certificate and key
package gmcrypto

import (
	"strings"
	"time"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/cryptomgr"
	"huawei.com/huaweichain/common/logger"
	"huawei.com/huaweichain/gmssl/wrapper"
)

type gmcert struct {
	pemCert           []byte
	gmsslCert         *wrapper.Certificate
	pubKey            *wrapper.PublicKey
	derPubKey         []byte
	serialNumber      string
	fingerPrint       string
	expireTime        time.Time
	subjectCN         string
	subjectOU         string
	subjectO          string
	subjectPostalCode string
	subjectStreet     string
	subjectLocality   string
	subjectCountry    string
}

const defaultSm2ID = "1234567812345678"

var log = logger.GetDefaultLogger()

// GenerateMiddleCACert generate middle CA cert
func GenerateMiddleCACert(info *cryptomgr.CertBasicInfo, pubKey cryptomgr.Key, caPriKey cryptomgr.Key,
	caCert cryptomgr.Cert) (cryptomgr.Cert, error) {
	return geneCert(info, pubKey, caPriKey, caCert, true)
}

// GenerateCert generate cert
func GenerateCert(info *cryptomgr.CertBasicInfo, pubKey cryptomgr.Key, caPriKey cryptomgr.Key,
	caCert cryptomgr.Cert) (cryptomgr.Cert, error) {
	return geneCert(info, pubKey, caPriKey, caCert, false)
}

func geneCert(info *cryptomgr.CertBasicInfo, pubKey cryptomgr.Key, caPriKey cryptomgr.Key,
	caCert cryptomgr.Cert, isMiddleCert bool) (cryptomgr.Cert, error) {
	if info == nil || pubKey == nil || caPriKey == nil || caCert == nil {
		return nil, errors.New("input parameter is not correct")
	}
	gmPubKey, ok := pubKey.(*gmKey)
	if !ok {
		return nil, errors.New("the public key is not gm key")
	}
	gmCaPriKey, ok := caPriKey.(*gmKey)
	if !ok {
		return nil, errors.New("the ca private key is not gm key")
	}
	gmCaCert, ok := caCert.(*gmcert)
	if !ok {
		return nil, errors.New("the ca cert is not gm cert")
	}

	var gmCert *wrapper.Certificate
	var err error
	gmCertInfo := &wrapper.GmCertInfo{
		Country:         info.Country,
		Province:        info.Province,
		Locality:        info.Locality,
		Org:             info.Organization,
		OU:              info.OrganizationUnit,
		CN:              info.CommonName,
		ValidationYears: info.ValidationYears,
	}
	if isMiddleCert {
		gmCert, err = wrapper.GenerateMiddleCACert(gmCertInfo, gmPubKey.pubKey, gmCaPriKey.priKey, gmCaCert.gmsslCert)
		if err != nil {
			return nil, errors.WithMessage(err, "generate middle cert failed")
		}
	} else {
		gmCert, err = wrapper.GenerateCert(gmCertInfo, gmPubKey.pubKey, gmCaPriKey.priKey, gmCaCert.gmsslCert)
		if err != nil {
			return nil, errors.WithMessage(err, "generate leaf cert failed")
		}
	}
	pemCert := wrapper.GetPemFromCert(gmCert)
	return GetCert(pemCert)
}

// GenerateSelfSignCert create self signed certificate
func GenerateSelfSignCert(info *cryptomgr.CertBasicInfo,
	key cryptomgr.Key) (cryptomgr.Cert, error) {
	if info == nil || key == nil {
		return nil, errors.New("input parameter is not correct")
	}
	gk, ok := key.(*gmKey)
	if !ok {
		return nil, errors.New("the key is not gm key")
	}
	gmCertInfo := &wrapper.GmCertInfo{
		Country:         info.Country,
		Province:        info.Province,
		Locality:        info.Locality,
		Org:             info.Organization,
		OU:              info.OrganizationUnit,
		CN:              info.CommonName,
		ValidationYears: info.ValidationYears,
	}
	certificate, err := wrapper.GenerateSelfSignCert(gmCertInfo, gk.priKey)
	if err != nil {
		return nil, err
	}
	pemCert := wrapper.GetPemFromCert(certificate)
	return GetCert(pemCert)
}

// GetCert create certificate from pem
func GetCert(pem []byte) (cryptomgr.Cert, error) {
	if len(pem) == 0 {
		return nil, errors.New("the pem cert is nil")
	}
	certificate, err := wrapper.NewCertificateFromPEM(string(pem), "")
	if err != nil {
		return nil, errors.WithMessage(err, "change gm pem to certificate failed")
	}
	alg := certificate.GetSignAlgorithm()
	if alg != cryptomgr.GmSSLSm2WithSm3 {
		return nil, errors.Errorf("the sign algorithm of certificate is %s, not sm2_with_sm3", alg)
	}
	publicKey, err := certificate.GetPublicKey()
	if err != nil {
		return nil, errors.Errorf("get public key failed from gm certificate:%s", err.Error())
	}
	derPubKey, err := publicKey.GetDer()
	if err != nil {
		return nil, errors.WithMessage(err, "get der public key failed")
	}
	serialNumber, err := certificate.GetSerialNumber()
	if err != nil {
		return nil, errors.WithMessage(err, "get serial number failed")
	}
	fingerPrint := certificate.GetFingerPrint()
	expireTime := certificate.GetExpireTime()

	cert := gmcert{
		pemCert:      pem,
		gmsslCert:    certificate,
		pubKey:       publicKey,
		derPubKey:    derPubKey,
		serialNumber: serialNumber,
		fingerPrint:  fingerPrint,
		expireTime:   expireTime,
	}

	err = cert.setSubjects()
	if err != nil {
		return nil, err
	}

	return &cert, nil
}

// CheckGmCertPara check certificate interface implemented by gm cert struct
func CheckGmCertPara(cert cryptomgr.Cert) error {
	if cert == nil {
		return errors.New("the cert is nil")
	}
	gmCert, ok := cert.(*gmcert)
	if !ok {
		return errors.New("the cert is not gm cert")
	}
	return checkGmCertInfo(gmCert)
}

func checkGmCertInfo(gmCert *gmcert) error {
	if gmCert.gmsslCert == nil {
		return errors.New("the gm cert struct pointer is nil")
	}
	if gmCert.pubKey == nil {
		return errors.New("the public key in gm cert is nil")
	}
	if gmCert.pemCert == nil {
		return errors.New("the pem format gm cert is nil")
	}
	if gmCert.derPubKey == nil {
		return errors.New("the der format gm cert is nil")
	}

	if gmCert.serialNumber == "" {
		return errors.New("the gm cert serial number is empty")
	}
	if gmCert.fingerPrint == "" {
		return errors.New("the gm cert finger print is empty")
	}
	if gmCert.subjectCN == "" {
		return errors.New("the gm cert common name is empty")
	}
	if gmCert.subjectO == "" {
		return errors.New("the gm cert organization is empty")
	}
	if gmCert.subjectCountry == "" {
		return errors.New("the gm cert country is empty")
	}
	return nil
}

func (g *gmcert) setSubjects() error {
	subject, err := g.gmsslCert.GetSubject()
	if err != nil {
		return errors.WithMessage(err, "get subject failed from cert")
	}
	subjects := strings.Split(subject, "/")
	for _, v := range subjects {
		index := strings.Index(v, "=")
		if index == -1 {
			continue
		}
		subjectLabel := v[:index]
		resStr := v[index+1:]
		switch subjectLabel {
		case "CN":
			g.subjectCN = resStr
		case "OU":
			g.subjectOU = resStr
		case "O":
			g.subjectO = resStr
		case "postalCode":
			g.subjectPostalCode = resStr
		case "street":
			g.subjectStreet = resStr
		case "L":
			g.subjectLocality = resStr
		case "ST":
			g.subjectStreet = resStr
		case "C":
			g.subjectCountry = resStr
		default:
			log.Infof("Unknown subject")
		}
	}
	return nil
}

// GetCommonName get common name of cert
func (g *gmcert) GetCommonName() string {
	if g.gmsslCert == nil {
		log.Warn("the cert is empty")
		return ""
	}
	return g.subjectCN
}

// GetSignAlg get sign algorithm of certificate
func (g *gmcert) GetSignAlg() (string, error) {
	if g.gmsslCert == nil {
		return "", errors.New("the cert is empty")
	}
	alg := g.gmsslCert.GetSignAlgorithm()
	if alg != cryptomgr.GmSSLSm2WithSm3 {
		log.Errorf("the sign algorithm of certificate is %s, not Sm2WithSm3", alg)
		return "", errors.New("get sign algorithm failed")
	}

	return cryptomgr.Sm2WithSm3, nil
}

// GetExpireTime get expire time of cert
func (g *gmcert) GetExpireTime() time.Time {
	return g.expireTime
}

// GetOrganizationalUnit get organization unit of cert
func (g *gmcert) GetOrganizationalUnit() []string {
	if g.gmsslCert == nil {
		return nil
	}
	return []string{g.subjectOU}
}

// GetOrganization get organizations of cert
func (g *gmcert) GetOrganization() []string {
	if g.gmsslCert == nil {
		return nil
	}
	return []string{g.subjectO}
}

// Verify verify signature
func (g *gmcert) Verify(msg []byte, signature []byte, hashOpt string) error {
	digest, err := calcSm2SignDigest(g.pubKey, msg)
	if err != nil {
		return errors.WithMessage(err, "calc digest failed in verify")
	}
	return g.pubKey.Verify(digest, signature)
}

// GetDerPublicKey get public key from cert
func (g *gmcert) GetDerPublicKey() ([]byte, error) {
	return g.derPubKey, nil
}

// CheckPublicKeyLength check whether the public key length meets the security requirements
func (g *gmcert) CheckPublicKeyLength() error {
	return nil
}

// CheckValidation check whether this cert is signed by the key corresponding to one of the root certificates
func (g *gmcert) CheckValidation(rootCerts []cryptomgr.Cert) (string, error) {
	if len(rootCerts) == 0 {
		return "", errors.New("there is no root certs")
	}
	for _, rootCert := range rootCerts {
		gmRootCert, ok := rootCert.(*gmcert)
		if !ok {
			continue
		}
		verifyRes := g.gmsslCert.VerifyCert(gmRootCert.gmsslCert)
		if verifyRes {
			return gmRootCert.subjectO, nil
		}
	}

	return "", errors.Errorf("verify certificate %s failed by root certs %s", g.GetCommonName(),
		rootCerts[0].GetCommonName())
}

// GetPemCertBytes get pem cert bytes
func (g *gmcert) GetPemCertBytes() []byte {
	return g.pemCert
}

// GetFingerPrint get finger print of cert
func (g *gmcert) GetFingerPrint() string {
	return g.fingerPrint
}

// GetSerialNumber get serial number of cert
func (g *gmcert) GetSerialNumber() string {
	return g.serialNumber
}

// IsCaCert check if the cert is ca cert
func (g *gmcert) IsCaCert() (bool, error) {
	if g.gmsslCert == nil {
		return false, errors.New("gm cert is nil")
	}
	return g.gmsslCert.IsCaCert()
}

func (g *gmcert) GetBasicInfo() *cryptomgr.CertBasicInfo {
	return &cryptomgr.CertBasicInfo{
		Organization:     g.subjectO,
		OrganizationUnit: g.subjectOU,
		Country:          g.subjectCountry,
		Province:         g.subjectStreet,
		Locality:         g.subjectLocality,
		CommonName:       g.subjectCN,
	}
}
