/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2020. All rights reserved.
 */

// Package gmcrypto is the gm crypto implementation
package gmcrypto

import (
	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/cryptomgr"
	"huawei.com/huaweichain/common/cryptomgr/cryptoimpl"
)

func init() {
	cryptoimpl.RegisterCertFactory(&GmCertFactory{name: cryptomgr.Sm2WithSm3})
	// for sm2_with_sm3 config, use the sm2p256v1 elliptic curve
	cryptoimpl.RegisterAsKeyFactory(&GmKeyFactory{name: cryptomgr.Sm2WithSm3})
	cryptoimpl.RegisterHashFunc(cryptomgr.Sm3, Hash)
	cryptoimpl.RegisterHashForFileFunc(cryptomgr.Sm3, HashForFile)
	cryptoimpl.RegisterSymmetricKeyFactory(&GmSymmetricKeyFactory{name: cryptomgr.SM4})
}

// GmSymmetricKeyFactory gm symmetric key factory
type GmSymmetricKeyFactory struct {
	name string
}

// Name get gm symmetric key factory name
func (gs *GmSymmetricKeyFactory) Name() string {
	return gs.name
}

// NewSymmetricKeyWithRandNum new symmetric key with random number
func (gs *GmSymmetricKeyFactory) NewSymmetricKeyWithRandNum(key []byte) (cryptomgr.SymmetricKey, error) {
	return NewSm4KeyWithRandNum(key)
}

// NewSymmetricKey new symmetric key
func (gs *GmSymmetricKeyFactory) NewSymmetricKey() (cryptomgr.SymmetricKey, error) {
	return NewSm4Key()
}

// GmKeyFactory gm key factory
type GmKeyFactory struct {
	name string
}

// Name get gm key factory
func (g *GmKeyFactory) Name() string {
	return g.name
}

// GenerateKey create key.
func (g *GmKeyFactory) GenerateKey() (cryptomgr.Key, error) {
	return GeneratePriKey()
}

// GetKeyFromPem get key from pem.
func (g *GmKeyFactory) GetKeyFromPem(pemKey []byte) (cryptomgr.Key, error) {
	keyLen := len(pemKey)
	if keyLen == 0 || keyLen > cryptomgr.MaxPEMKeyLen {
		return nil, errors.Errorf("the key len is %d, which is not correct", keyLen)
	}
	return GetPriKey(pemKey)
}

// GetPubKeyFromDer get key from der.
func (g *GmKeyFactory) GetPubKeyFromDer(derKey []byte) (cryptomgr.Key, error) {
	keyLen := len(derKey)
	if keyLen == 0 || keyLen > cryptomgr.MaxDerKeyLen {
		return nil, errors.Errorf("the key len is %d, which is not correct", keyLen)
	}
	return GetPubKeyFromDer(derKey)
}

// CreateKeyFromBuffer create key from buffer
func (g *GmKeyFactory) CreateKeyFromBuffer(buffer []byte) (cryptomgr.Key, error) {
	return nil, errors.New("not support to create key from buffer")
}

// GmCertFactory gm cert factory
type GmCertFactory struct {
	name string
}

// Name get gm cert factory name
func (g *GmCertFactory) Name() string {
	return g.name
}

// GenerateSelfSignCert create self sign CA cert
func (g *GmCertFactory) GenerateSelfSignCert(info *cryptomgr.CertBasicInfo,
	key cryptomgr.Key) (cryptomgr.Cert, error) {
	err := cryptoimpl.CheckCertInfo(info)
	if err != nil {
		return nil, err
	}
	err = CheckGmPriKey(key)
	if err != nil {
		return nil, err
	}
	return GenerateSelfSignCert(info, key)
}

// GenerateMiddleCACert generate middle ca cert
func (g *GmCertFactory) GenerateMiddleCACert(info *cryptomgr.CertBasicInfo, pubKey cryptomgr.Key,
	caPriKey cryptomgr.Key, caCert cryptomgr.Cert) (cryptomgr.Cert, error) {
	err := checkGmCertPara(info, pubKey, caPriKey, caCert)
	if err != nil {
		return nil, err
	}
	return GenerateMiddleCACert(info, pubKey, caPriKey, caCert)
}

// GenerateCert create cert by gm cert factory.
func (g *GmCertFactory) GenerateCert(info *cryptomgr.CertBasicInfo, pubKey cryptomgr.Key, caPriKey cryptomgr.Key,
	caCert cryptomgr.Cert) (cryptomgr.Cert, error) {
	err := checkGmCertPara(info, pubKey, caPriKey, caCert)
	if err != nil {
		return nil, err
	}
	return GenerateCert(info, pubKey, caPriKey, caCert)
}

func checkGmCertPara(info *cryptomgr.CertBasicInfo, pubKey cryptomgr.Key, caPriKey cryptomgr.Key,
	caCert cryptomgr.Cert) error {
	err := cryptoimpl.CheckCertInfo(info)
	if err != nil {
		return err
	}
	err = CheckGmPubKey(pubKey)
	if err != nil {
		return err
	}
	err = CheckGmPriKey(caPriKey)
	if err != nil {
		return err
	}
	err = CheckGmCertPara(caCert)
	if err != nil {
		return err
	}
	return nil
}

// GetCertFromPem Get gm cert from pem bytes.
func (g *GmCertFactory) GetCertFromPem(pemCert []byte) (cryptomgr.Cert, error) {
	certLen := len(pemCert)
	if certLen == 0 {
		return nil, errors.New("the cert len is 0, which is not correct")
	}
	return GetCert(pemCert)
}
