//go:build !nogm
// +build !nogm

/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2021. All rights reserved.
 */

// Package gmcrypto the gm algorithm for generate of certificate and key
package gmcrypto

import (
	"encoding/pem"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/cryptomgr"
	"huawei.com/huaweichain/common/cryptomgr/cryptoimpl"
	"huawei.com/huaweichain/gmssl/wrapper"
)

type gmKey struct {
	isPrivate bool
	priKey    *wrapper.PrivateKey
	pemPriKey []byte
	derPriKey []byte
	pubKey    *wrapper.PublicKey
	pemPubKey []byte
	derPubKey []byte
}

const (
	ecParamgenCurve = "ec_paramgen_curve"
	sm2Curve        = "sm2p256v1"
	ecParamEnc      = "ec_param_enc"
	namedCurve      = "named_curve"
	ecAlg           = "EC"
)

// CheckGmPubKey check key interface implemented by gm public key struct
func CheckGmPubKey(key cryptomgr.Key) error {
	if key == nil {
		return errors.New("the gm public key is nil")
	}
	gk, ok := key.(*gmKey)
	if !ok {
		return errors.New("the public key is not gm key")
	}
	return checkGmPubKey(gk)
}

func checkGmPubKey(pubKey *gmKey) error {
	if pubKey == nil {
		return errors.New("the public key is nil")
	}
	if pubKey.pemPubKey == nil {
		return errors.New("the pem format public key is nil")
	}
	if pubKey.derPubKey == nil {
		return errors.New("the der format public key is nil")
	}
	if pubKey.pubKey == nil {
		return errors.New("the gm ssl public key is nil")
	}
	return nil
}

// CheckGmPriKey check key interface implemented by gm private key struct
func CheckGmPriKey(key cryptomgr.Key) error {
	if key == nil {
		return errors.New("the key is nil")
	}
	gk, ok := key.(*gmKey)
	if !ok {
		return errors.New("the key is not gm key")
	}
	return checkGmPriKey(gk)
}

func checkGmPriKey(priKey *gmKey) error {
	if priKey == nil {
		return errors.New("the gm private key is nil")
	}
	if priKey.isPrivate != true {
		return errors.New("this is not private key")
	}
	if priKey.pemPriKey == nil {
		return errors.New("the pem format gm key is nil")
	}
	if priKey.pemPubKey == nil {
		return errors.New("the pem public key is nil")
	}
	if priKey.derPriKey == nil {
		return errors.New("the der format gm key is nil")
	}
	if priKey.derPubKey == nil {
		return errors.New("the der public key is nil")
	}
	if priKey.priKey == nil {
		return errors.New("the gmssl private key is nil")
	}
	if priKey.pubKey == nil {
		return errors.New("the gmssl public key is nil")
	}
	return nil
}

// GeneratePriKey generate private key
func GeneratePriKey() (cryptomgr.Key, error) {
	const curveNum = 2
	sm2KeyArgs := [][curveNum]string{
		{ecParamgenCurve, sm2Curve},
		{ecParamEnc, namedCurve},
	}
	sm2Key, err := wrapper.GeneratePrivateKey(ecAlg, sm2KeyArgs)
	if err != nil {
		log.Errorf("gmssl generate private key failed:%s", err.Error())
		return nil, errors.Errorf("gmssl generate private key failed:%s", err.Error())
	}
	pemSm2Key, err := sm2Key.GetPemWithoutEnc()
	if err != nil {
		log.Errorf("get pem format gm key failed:%s", err.Error())
		return nil, errors.Errorf("get pem format gm key failed:%s", err.Error())
	}
	return GetPriKey([]byte(pemSm2Key))
}

// GetPubKey generate public key from pem
func GetPubKey(pemKey []byte) (cryptomgr.Key, error) {
	if len(pemKey) == 0 {
		log.Warn("the pem public key is nil")
		return nil, errors.New("the pem public key is nil")
	}
	key := &gmKey{}
	publicKey, err := wrapper.NewPublicKeyFromPEM(string(pemKey))
	if err != nil {
		log.Errorf("This is not a gm public key, err is %s.", err.Error())
		return nil, errors.Errorf("this is not a gm public key, err is %s", err.Error())
	}

	key.pubKey = publicKey
	der, err := publicKey.GetDer()
	if err != nil {
		return nil, err
	}
	key.derPubKey = der
	key.pemPubKey = pemKey
	key.isPrivate = false
	return key, nil
}

// GetPubKeyFromDer generate public key from der
func GetPubKeyFromDer(derKey []byte) (cryptomgr.Key, error) {
	if len(derKey) == 0 {
		return nil, errors.New("the der public key is nil")
	}
	return nil, nil // TODO implement
}

// GetPriKey generate private key from pem
func GetPriKey(pemKey []byte) (cryptomgr.Key, error) {
	if len(pemKey) == 0 {
		log.Warn("the pem private key is nil")
		return nil, errors.New("the pem private key is nil")
	}
	key := &gmKey{}
	privateKey, err := wrapper.NewPrivateKeyFromPEM(string(pemKey), "test")
	if err != nil {
		return nil, err
	}
	key.priKey = privateKey
	key.pemPriKey = pemKey
	derBlock, _ := pem.Decode(pemKey)
	derPriKey, err := ExtractPrivateKeyFromPKCS8(derBlock.Bytes)
	if err != nil {
		return nil, errors.WithMessagef(err, "get der private key failed")
	}
	key.derPriKey = derPriKey

	publicKeyPEM, err := privateKey.GetPublicKeyPEM()
	if err != nil {
		log.Errorf("Get gm public key from private key failed, err is %s", err.Error())
		return nil, err
	}
	key.pemPubKey = []byte(publicKeyPEM)
	publicKey, err := wrapper.NewPublicKeyFromPEM(string(key.pemPubKey))
	if err != nil {
		log.Errorf("Generate gm public key from pem public key failed, err is %s", err.Error())
		return nil, err
	}
	derPubKey, err := publicKey.GetDer()
	if err != nil {
		return nil, err
	}
	key.derPubKey = derPubKey
	key.pubKey = publicKey
	key.isPrivate = true
	return key, nil
}

// GetPemBytes get pem format key
func (g *gmKey) GetPemBytes() []byte {
	if g.isPrivate {
		return g.pemPriKey
	}
	return g.pemPubKey
}

// GetDerBytes get der format key
func (g *gmKey) GetDerBytes() ([]byte, error) {
	if g.isPrivate {
		return g.derPriKey, nil
	}
	return g.derPubKey, nil
}

// IsSymmetric is symmetric key
func (g *gmKey) IsSymmetric() bool {
	return false
}

// IsPrivate is private key
func (g *gmKey) IsPrivate() bool {
	return g.isPrivate
}

// GetPublicKey get public key object from private key
func (g *gmKey) GetPublicKey() cryptomgr.Key {
	pubKey, err := GetPubKey(g.pemPubKey)
	if err != nil {
		return nil
	}
	return pubKey
}

// GetPrivateKey get private key
func (g *gmKey) GetPrivateKey() (interface{}, error) {
	return g.priKey, nil
}

// Sign get signature
func (g *gmKey) Sign(msg []byte, hashAlg string) ([]byte, error) {
	if !g.isPrivate {
		return nil, errors.New("this is not private key")
	}
	digest, err := calcSm2SignDigest(g.pubKey, msg)
	if err != nil {
		return nil, errors.WithMessage(err, "calc digest failed in sign")
	}

	sig, err := g.priKey.Sign(digest)
	if err != nil {
		return nil, errors.Errorf("sign error is %s", err.Error())
	}
	return sig, nil
}

// Verify verify signature
func (g *gmKey) Verify(msg []byte, signature []byte, hashAlg string) error {
	digest, err := calcSm2SignDigest(g.pubKey, msg)
	if err != nil {
		return errors.WithMessage(err, "calc digest failed when public verify")
	}
	return g.pubKey.Verify(digest, signature)
}

// CleanKeySensitiveInfo clean sensitive info in key object
func (g *gmKey) CleanKeySensitiveInfo() {
	cryptoimpl.CleanBytes(g.pemPriKey)
	cryptoimpl.CleanBytes(g.derPriKey)
	g.pemPriKey = nil
	g.derPriKey = nil
}

func calcSm2SignDigest(sm2pk *wrapper.PublicKey, message []byte) ([]byte, error) {
	sm2zid, err := sm2pk.ComputeSM2IDDigest(defaultSm2ID)
	if err != nil {
		return nil, err
	}

	sm3ctx, err := wrapper.NewDigestContext(sm3HashFlag)
	defer sm3ctx.Free()
	if err != nil {
		return nil, err
	}

	err = sm3ctx.Update(sm2zid)
	if err != nil {
		return nil, err
	}
	err = sm3ctx.Update(message)
	if err != nil {
		return nil, err
	}

	hashRes, err := sm3ctx.Final()
	if err != nil {
		return nil, err
	}
	return hashRes, nil
}
