//go:build !nogm
// +build !nogm

/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

package gmcrypto

import (
	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/cryptomgr"
	"huawei.com/huaweichain/gmssl/wrapper"
)

type gmSm4 struct {
	key []byte
}

// NewSm4KeyWithRandNum new sm4 key with random number
func NewSm4KeyWithRandNum(key []byte) (cryptomgr.SymmetricKey, error) {
	sm4 := &gmSm4{key: key}
	return sm4, nil
}

// NewSm4Key new sm4 key
func NewSm4Key() (cryptomgr.SymmetricKey, error) {
	key, err := generateSm4Key()
	if err != nil {
		return nil, err
	}
	sm4 := &gmSm4{key: key}
	return sm4, nil
}

// GetKeyBytes get bytes format key
func (g *gmSm4) GetKeyBytes() []byte {
	return g.key
}

// Encrypt encrypt plain txt
func (g *gmSm4) Encrypt(plainTxt []byte) ([]byte, error) {
	iv, err := generateSm4IV()
	if err != nil {
		return nil, err
	}

	cipherMsg, err := encWithSm4IV(g.key, iv, plainTxt)
	if err != nil {
		return nil, err
	}
	cipherMsg = append(cipherMsg, iv...)
	return cipherMsg, nil
}

// Decrypt decrypt cipher txt and iv
func (g *gmSm4) Decrypt(cipherTxtAndIv []byte) ([]byte, error) {
	length := len(cipherTxtAndIv)
	if length < cryptomgr.IvLen+cryptomgr.SymmetricKeyLen {
		return nil, errors.Errorf("the cipher msg length:%d is not correct", length)
	}
	iv := cipherTxtAndIv[length-cryptomgr.IvLen:]
	encMgs := cipherTxtAndIv[:length-cryptomgr.IvLen]
	return decWithSm4IV(g.key, iv, encMgs)
}

// generateSm4Key generate random key
func generateSm4Key() ([]byte, error) {
	keyLength, err := wrapper.GetCipherKeyLength("SMS4")
	if err != nil {
		return nil, err
	}
	key, err := wrapper.GenerateRandom(keyLength)
	if err != nil {
		return nil, err
	}
	return key, nil
}

// generateSm4IV generate random iv
func generateSm4IV() ([]byte, error) {
	ivLength, err := wrapper.GetCipherIVLength("SMS4")
	if err != nil {
		return nil, err
	}
	key, err := wrapper.GenerateRandom(ivLength)
	if err != nil {
		return nil, err
	}
	return key, nil
}

// encWithSm4IV enc with sm4
func encWithSm4IV(key []byte, iv []byte, msg []byte) ([]byte, error) {
	return procWithSm4IV(key, iv, msg, true)
}

func procWithSm4IV(key []byte, iv []byte, msg []byte, isEncrypt bool) ([]byte, error) {
	encryptor, err := wrapper.NewCipherContext("SMS4", key, iv, isEncrypt)
	if err != nil {
		return nil, err
	}
	text1, err := encryptor.Update(msg)
	if err != nil {
		return nil, err
	}
	text2, err := encryptor.Final()
	if err != nil {
		return nil, err
	}
	text := make([]byte, 0, len(text1)+len(text2))
	text = append(text, text1...)
	text = append(text, text2...)
	return text, nil
}

// decWithSm4IV dec with sm4
func decWithSm4IV(key []byte, iv []byte, encMsg []byte) ([]byte, error) {
	return procWithSm4IV(key, iv, encMsg, false)
}
