/**
 * @file ylong_tls.h
 * @copyright Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 * @brief ylong_tls 对外头文件
 * @details ylong_tls 对外头文件
 * @author w00620137
 * @date 2022-07-14
 * @version v0.1.0
 * *******************************************************************************************
 * @par 修改日志：
 * <table>
 * <tr><th>Date        <th>Version  <th>Author       <th>Description
 * <tr><td>2022-07-14  <td>0.1.0    <td>w00620137    <td>创建初始版本
 * </table>
 * *******************************************************************************************
 */

/**
 * @defgroup    ylong_tls TLS C 封装模块
 * @ingroup     ylong_tls
 */

#ifndef YLONG_TLS_H
#define YLONG_TLS_H

#ifdef __cplusplus
#if __cplusplus
extern "C" {
#endif
#endif

/**
* @ingroup     ylong_tls
* @brief       ylong_tls_stream 底层使用的 IO。
* @attention   返回的 ylong_tls_bio* 成员句柄的内存分配来自 Rust 侧，需要调用 ylong_tls_bio_free 释放。
*/
typedef void ylong_tls_bio;

/**
* @ingroup     ylong_tls
* @brief       ylong_tls_stream 底层使用的 IO 使用的上下文，由用户自定义内容。
* @param       write    [IN]    写函数，传入 ylong_tls_bio，buffer 以及 buffer 大小，将 buffer 内容写入底层 IO。如不需要可传入
*                               `NULL`。对于非阻塞 IO，当底层 IO 返回 WouldBlock 错误时，函数应返回
*                               ERRNO_BIO_WRITE_WOULD_BLOCK。
* @param       read     [IN]    读函数，传入 ylong_tls_bio，buffer 以及 buffer 大小，将底层 IO 读取内容写入 buffer。如不需要可
*                               传入 `NULL`。对于非阻塞 IO，当底层 IO 返回 WouldBlock 错误时，函数应返回
*                               ERRNO_BIO_READ_WOULD_BLOCK。
* @param       create   [IN]    data 创建函数，传入 ylong_tls_bio，创建 bio 且函数非空时，调用该函数创建其中与 IO 资源相关的成员
*                               `ptr`。如果用户自行管理 IO 资源或不需要这一函数，可传入 `NULL`。
* @param       destroy  [IN]    data 析构函数。传入 ylong_tls_bio，析构其中与 IO 资源相关的成员 `ptr`。如果用户自行管理 IO 资源或
*                               不需要这一函数，可传入 `NULL`。当非空时，将在调用 `ylong_tls_bio_free` 时调用，析构 bio 中的 IO
*                               资源。
* @attention   用户应当保证设置的函数指针以及文件指针在当前 TLS 链接流程中有效，如果指针失效，TLS 握手或传输应用数据中可能出现
*              未定义行为。
*/
typedef struct ylong_tls_bio_method {
    int (*write) (ylong_tls_bio *, const char *, unsigned int);
    int (*read) (ylong_tls_bio *, char *, unsigned int);
    int (*create) (ylong_tls_bio *);
    int (*destroy) (ylong_tls_bio *);
} ylong_tls_bio_method;

/**
* @ingroup     ylong_tls
* @brief       ylong_tls 使用的密钥结构体。
* @attention   返回的 ylong_tls_pkey* 成员句柄的内存分配来自 Rust 侧，需要调用 ylong_tls_pkey_free 释放。
*/
typedef void ylong_tls_pkey;

/**
* @ingroup     ylong_tls
* @brief       ylong_tls 使用的证书结构体。
* @attention   返回的 ylong_tls_cert_chain* 成员句柄的内存分配来自 Rust 侧，需要调用 ylong_tls_cert_chain_free 释放。
*/
typedef void ylong_tls_cert_chain;

/**
 * @ingroup     ylong_tls
 * @brief       用于指向 Rust 侧 ConfigContext 的数据结构。传递给 C 侧使用的指针都会统一转换成 ylong_tls_ctx*。
 * @attention   返回的 ylong_tls_ctx* 成员句柄的内存分配来自 Rust 侧，需要调用 ylong_tls_ctx_free 释放。
 */
typedef void ylong_tls_ctx;

/**
 * @ingroup     ylong_tls
 * @brief       用于指向 Rust 侧 TlsStream<T> 的数据结构。传递给 C 侧使用的指针都会统一转换成 ylong_tls_stream*。
 * @attention   返回的 ylong_tls_stream* 成员句柄的内存分配来自 Rust 侧，需要调用 ylong_tls_stream_free 释放。
 */
typedef void ylong_tls_stream;

// 通用返回
#define YLONG_TLS_SUCCESS 1
#define ERRNO_SETTING_FAILURE 0
#define ERRNO_NULL -1

// 设置 ylong_tls_ctx 时产生的错误
#define ERRNO_PRIVKEY_OR_CERT_IO -2
#define ERRNO_INVALID_PRIVKEY_OR_CERT -3

// 使用 ylong_tls_ctx 生成 ylong_tls_stream 时产生的错误
#define YLONG_TLS_BUILD_CTX_SUCCESS 0
#define ERRNO_INVALID_KEY 1
#define ERRNO_INVALID_CERT 2
#define ERRNO_ROOT_CERT_UNSPECIFIED 3
#define ERRNO_UNSUPPORTED_VERSION 4
#define ERRNO_UNSUPPORTED_CIPHER_SUITE 5
#define ERRNO_EMPTY_ALPN 6
#define ERRNO_UNCLASSIFIED_BUILD_CTX_FAILURE 7

// 使用 ylong_tls_stream 链接时产生的错误
#define ERRNO_STREAM_END_MISMATCH 0
#define ERRNO_HS_WANTS_READ -2
#define ERRNO_HS_WANTS_WRITE -3
#define ERRNO_TLS_TIMEOUT -4
#define ERRNO_INVALID_HOSTNAME -5
#define ERRNO_NO_RW_BIO -6
#define ERRNO_BIO_CONNECTION_FAILURE -7
#define ERRNO_EOF -8
#define ERRNO_OUT_OF_MEMORY -9
#define ERRNO_UNCLASSIFIED_HS_FAILURE -10
#define ERRNO_UNCLASSIFIED_TLS_CONN_FAILURE -11

// 使用 ylong_tls_stream / ylong_tls_bio 传输数据时产生的错误
#define ERRNO_BIO_READ_WOULD_BLOCK -2        // 对于非阻塞 IO，函数应返回该值表示读函数 Would Block 错误
#define ERRNO_BIO_WRITE_WOULD_BLOCK -3       // 对于非阻塞 IO，函数应返回该值表示写函数 Would Block 错误
#define ERRNO_BIO_TIMEOUT -4                 // 读写函数应返回该值表示超时错误
#define ERRNO_BIO_EOF -5                     // 读写函数应返回该值表示 EOF 错误
#define ERRNO_NO_TLS_CONN -6
#define ERRNO_HS_INCOMPLETE -7
#define ERRNO_UNCLASSIFIED_RW_FAILURE -8

// TLS shutdown 状态
#define YLONG_TLS_NOT_SHUTDOWN 0
#define YLONG_TLS_RECV_SHUTDOWN 0b01
#define YLONG_TLS_SEND_SHUTDOWN 0b10

// TLS 消息头长度
// 5: TlsCipherText content_type + version + length
// 8: nonce (GCM_EXPLICIT_NONCE_LEN)
// 1: tls 1.3 type
// 16: tag (AEAD_TAG_LEN)
//
// TLS 消息:
//                                  1 byte     16 bytes
// --------------------------------------------------------
// |           payload             | type |      tag      |
// --------------------------------------------------------
// TLS 1.2 没有 type 位。
#define YLONG_TLS_MAX_SEAL_OVERHEAD_LENGTH (5 + 8 + 1 + 16)

/**
 * @ingroup     ylong_tls
 * @brief       根据上下文中包含的方法，创建对应的 ylong_tls_bio，并以一个 ylong_tls_bio* 成员句柄形式返回。
 * @param       method      [IN]    需要配置的 bio 方法。
 * @attention   返回的 ylong_tls_bio* 成员句柄的内存分配来自 Rust 侧，需要调用 ylong_tls_bio_free 释放。
 * @retval      != NULL，合法的 ylong_tls_bio* 成员句柄，包含 ctx 中定义的方法和文件指针。
 * @retval      NULL，生成 bio 错误，ctx 中可能没有包含读函数或写函数指针。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
ylong_tls_bio *ylong_tls_bio_new(struct ylong_tls_bio_method *method);

/**
 * @ingroup     ylong_tls
 * @brief       创建读写 vector 的 ylong_tls_bio，并以一个 ylong_tls_bio* 成员句柄形式返回。可用于生成默认读取证书私钥或证书的 bio。
 *              将证书私钥 / 证书 / 指定对象添加到 bio 中应调用 `ylong_tls_bio_write`，读取则应调用 `ylong_tls_bio_read`。
 *              使用 ylong_tls_bio_read 读取 bio 中内容会导致文件指针后移，在读尽当前内容后，无法再从该 bio 中循环读取，除非写入新内容。
 * @attention   返回的 ylong_tls_bio* 成员句柄的内存分配来自 Rust 侧，需要调用 ylong_tls_bio_free 释放。
 * @retval      != NULL，合法的 ylong_tls_bio* 成员句柄，包含 ctx 中定义的方法和文件指针。当该 bio 读写遇到错误时，返回以下错误：
 *                -
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
ylong_tls_bio *ylong_tls_vector_bio_new();

/**
 * @ingroup     ylong_tls
 * @brief       设置 ylong_tls_bio 中包含的数据指针。
 * @param       bio         [IN]    需要设置的 ylong_tls_bio。
 * @param       data        [IN]    需要设置的数据指针。
 * @retval      YLONG_TLS_SUCCESS，设置成功。
 * @retval      ERRNO_NULL，传入 bio 为空。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_bio_set_data(ylong_tls_bio *bio, void *data);

/**
 * @ingroup     ylong_tls
 * @brief       返回 ylong_tls_bio 中包含的数据指针。
 * @param       bio         [IN]    需要读取的 ylong_tls_bio。
 * @retval      != NULL，返回 ylong_tls_bio 中包含的指针。
 * @retval      NULL，ylong_tls_bio 中包含的指针为空，或传入 bio 为空。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
void *ylong_tls_bio_get_data(ylong_tls_bio *bio);

/**
 * @ingroup     ylong_tls
 * @brief       ylong_tls_bio 写函数，传入 ylong_tls_bio，buffer 以及 buffer 大小，将 buffer 内容写入底层 IO。
 * @param       bio         [IN]    需要写入的 ylong_tls_bio。
 * @param       data        [IN]    写入的数据指针。
 * @param       len         [IN]    data 长度。
 * @retval      >= 0，成功，返回实际读取的字符个数。
 * @retval      ERRNO_NULL，入参 bio、data 为 NULL，或 bio 未配置写函数。
 * @retval      ERRNO_UNCLASSIFIED_RW_FAILURE，其他错误。可以调用 `ylong_tls_get_bio_err` 获取具体错误。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_bio_write(ylong_tls_bio *bio, const void *data, unsigned int len);

/**
 * @ingroup     ylong_tls
 * @brief       ylong_tls_bio 读函数，传入 ylong_tls_bio，buffer 以及 buffer 大小，从底层 IO 读取内容至 buffer。
 * @param       bio         [IN]    需要读取的 ylong_tls_bio。
 * @param       data        [OUT]   读取存入的数据指针。
 * @param       len         [IN]    data 长度。
 * @retval      >= 0，成功，返回实际读取的字符个数。
 * @retval      ERRNO_NULL，入参 bio、buf 为 NULL，或 bio 未配置读函数。
 * @retval      ERRNO_UNCLASSIFIED_RW_FAILURE，其他错误。可以调用 `ylong_tls_get_bio_err` 获取具体错误。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_bio_read(ylong_tls_bio *bio, void *buf, unsigned int num);

/**
 * @ingroup     ylong_tls
 * @brief       创建对应的 ConfigContext，并以一个 ylong_tls_ctx* 成员句柄形式返回。
 * @attention   返回的 ylong_tls_ctx* 成员句柄的内存分配来自 Rust 侧，需要调用 ylong_tls_ctx_free 释放。
 * @retval      != NULL，合法的 ylong_tls_ctx* 成员句柄，包含对于版本、密码套的默认配置。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
ylong_tls_ctx *ylong_tls_ctx_new_client();

/**
 * @ingroup     ylong_tls
 * @brief       创建对应的 ConfigContext，并以一个 ylong_tls_ctx* 成员句柄形式返回。
 * @attention   返回的 ylong_tls_ctx* 成员句柄的内存分配来自 Rust 侧，需要调用 ylong_tls_ctx_free 释放。
 * @retval      != NULL，合法的 ylong_tls_ctx* 成员句柄，包含对于版本、密码套的默认配置。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
ylong_tls_ctx *ylong_tls_ctx_new_server();

/**
 * @ingroup     ylong_tls
 * @brief       设置使用版本，返回对应状态码。
 * @param       ctx         [IN]    需要配置的目标 ctx。
 * @param       versions    [IN]    配置的版本号，以 int[] 表示。目前支持如下版本：
 *                                    - TLCP (GB/T 38686-2020), 0x0101
 *                                    - TLS 1.2 (RFC5246), 0x0303
 *                                    - TLS 1.3 (RFC8446), 0x0304
 * @param       len         [IN]    版本号列表长度，以 size_t 表示。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 versions 或 ctx 为 NULL。
 * @retval      ERRNO_SETTING_FAILURE，配置失败，不支持的版本。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_set_versions(ylong_tls_ctx *ctx, int versions[], size_t len);

/**
 * @ingroup     ylong_tls
 * @brief       设置使用密码套，返回对应状态码。
 * @param       ctx    [IN]    需要配置的目标 ctx。
 * @param       str    [IN]    配置的密码套，以字符串表示，用 `:` 分隔不同的密码套。不设置则使用默认密码套。支持密码套如下：
 *                               通用密码套：
 *                               - ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,
 *                               - ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,
 *                               - ECDHE_RSA_WITH_AES_256_GCM_SHA384,
 *                               - ECDHE_RSA_WITH_AES_128_GCM_SHA256,
 *                               - ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256,
 *                               - ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256,
 *                               - AES_128_GCM_SHA256,
 *                               - AES_256_GCM_SHA384,
 *                               - CHACHA20_POLY1305_SHA256,
 *                               国密密码套：
 *                               - SM4_GCM_SM3,
 *                               - SM_ECDHE_SM4_GCM_SM3,
 *                               - SM_ECC_SM4_GCM_SM3
 *                             通用模式默认开启所有通用密码套，国密模式默认开启所有国密密码套，混合模式则默认开启所有密码套。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 或 str 为 NULL。
 * @retval      ERRNO_SETTING_FAILURE，可能包含未知的密码套，或使用错误格式输入。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_set_ciphersuites(ylong_tls_ctx *ctx, const char *str);

/**
 * @ingroup     ylong_tls
 * @brief       设置是否在协商使用国密密码对称加密密码套的情况下，使用一整套国密密码套。使用模式为非混合模式 TLS 时，该设置无效。
 * @param       ctx         [IN]    需要配置的目标 ctx。
 * @param       use_full    [IN]    是否使用一整套国密密码套，>= 1 表示使用，<= 0 表示可以使用国密 + 非国密的密码套组合。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 为 NULL。
 * @retval      ERRNO_SETTING_FAILURE，其他错误，链接库当前编译 feature 下不支持该设置
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_use_sm_full_suites(ylong_tls_ctx *ctx, int use_full);

/**
 * @ingroup     ylong_tls
 * @brief       设置使用密钥交换算法，返回对应状态码。
 * @param       ctx    [IN]    需要配置的目标 ctx。
 * @param       str    [IN]    配置的密钥交换算法，以字符串表示，用 `:` 分隔不同的密码套。不设置则使用默认密码套。支持密码套如下：
 *                               通用密码套：
 *                               - secp256r1,
 *                               - secp384r1,
 *                               - secp521r1,
 *                               - x25519,
 *                               - x448
*                                国密密码套：
 *                               - sm2
 *                             通用模式默认开启所有通用密码套，国密模式默认开启所有国密密码套，混合模式则默认开启所有密码套。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 或 str 为 NULL。
 * @retval      ERRNO_SETTING_FAILURE，可能包含未知的密码套，或使用错误格式输入。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_set_groups_list(ylong_tls_ctx *ctx, const char *str);

/**
 * @ingroup     ylong_tls
 * @brief       使用 buffer 解析 pkcs8 格式私钥 PEM 流，返回对应的私钥结构体。
 * @param       pem    [IN]    pkcs8 格式私钥 PEM 流。
 * @param       len    [IN]    PEM 流长度。
 * @retval      != NULL，读取成功。
 * @retval      NULL，配置失败，传入 pem 为 NULL，pem 长度与 len 不匹配，或是使用错误 / 不支持的格式输入。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
ylong_tls_pkey* ylong_tls_pem_read_privatekey(const void *pem, unsigned int len);

/**
 * @ingroup     ylong_tls
 * @brief       使用 bio 解析 pkcs8 格式私钥 PEM 流，返回对应的私钥结构体。
 * @param       bio    [IN]    读取 pkcs8 格式 PEM 私钥的 bio，解析私钥时会调用 bio 中注册的 read 函数。
 * @retval      != NULL，读取成功。
 * @retval      NULL，配置失败，传入 bio 为 NULL，或是使用错误 / 不支持的格式输入。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
ylong_tls_pkey* ylong_tls_pem_read_privatekey_bio(ylong_tls_bio *bio);

/**
 * @ingroup     ylong_tls
 * @brief       设置使用终端单证书的私钥，返回对应状态码。不会返回 ERRNO_SETTING_FAILURE (在当前 ylong_tls 编译
 *              feature 中不支持该功能)，在通用 TLS、国密 TLS、混合 TLS，设置单证书时都应使用该函数设置。
 * @param       ctx    [IN]    需要配置的目标 ctx。
 * @param       file   [IN]    终端证书私钥的 #pkcs8 文件路径地址。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 为 NULL。
 * @retval      ERRNO_PRIVKEY_OR_CERT_IO，配置失败，发生 IO 错误。
 * @retval      ERRNO_INVALID_PRIVKEY_OR_CERT，配置失败，可能包含未知或错误的密钥对，或使用错误或不支持的格式输入。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_use_privatekey_file(ylong_tls_ctx *ctx, const char *file);

/**
 * @ingroup     ylong_tls
 * @brief       设置使用终端单证书的私钥，返回对应状态码。不会返回 ERRNO_SETTING_FAILURE (在当前 ylong_tls 编译
 *              feature 中不支持该功能)，在通用 TLS、国密 TLS、混合 TLS，设置单证书时都应使用该函数设置。
 * @param       ctx    [IN]    需要配置的目标 ctx。
 * @param       pkey   [IN]    终端证书私钥结构体，可以通过 ylong_tls_pem_read_privatekey 解析 pkcs8 格式私钥 PEM 流得到私钥结构体。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 或 pkey 为 NULL。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_use_privatekey(ylong_tls_ctx *ctx, ylong_tls_pkey *pkey);

/**
 * @ingroup     ylong_tls
 * @brief       设置使用终端国密双证书的私钥，返回对应状态码。
 * @param       ctx            [IN]    需要配置的目标 ctx。
 * @param       sig_key_file   [IN]    终端国密签名证书私钥的文件路径地址。
 * @param       enc_key_file   [IN]    终端国密加密证书私钥的文件路径地址。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 为 NULL。
 * @retval      ERRNO_SETTING_FAILURE，在当前 ylong_tls 编译 feature 中不支持该功能。
 * @retval      ERRNO_PRIVKEY_OR_CERT_IO，配置失败，读取证书密钥文件发生 IO 错误。
 * @retval      ERRNO_INVALID_PRIVKEY_OR_CERT，配置失败，证书密钥对设置失败。可能包含未知或错误的密钥对，或使用错误或不支持的格式输入。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_use_sm_privkey_file(ylong_tls_ctx *ctx, const char *sig_key_file, const char *enc_key_file);

/**e
 * @ingroup     ylong_tls
 * @brief       设置使用终端国密双证书的私钥，返回对应状态码。可以通过 ylong_tls_pem_read_privatekey 解析 pkcs8 格式私钥 PEM 流得到
 *              私钥结构体。
 * @param       ctx            [IN]    需要配置的目标 ctx。
 * @param       sig_key        [IN]    终端国密签名证书私钥。
 * @param       enc_key        [IN]    终端国密加密证书私钥。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 或密钥为 NULL。
 * @retval      ERRNO_SETTING_FAILURE，在当前 ylong_tls 编译 feature 中不支持该功能。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_use_sm_privkey(ylong_tls_ctx *ctx, ylong_tls_pkey *sig_key, ylong_tls_pkey *enc_key);

/**
 * @ingroup     ylong_tls
 * @brief       设置使用混合终端证书(通用TLS单证书 + 国密双证书)的私钥，返回对应状态码。
 * @param       ctx                 [IN]    需要配置的目标 ctx。
 * @param       general_key_file    [IN]    终端通用TLS证书私钥的文件路径地址。
 * @param       sig_key_file        [IN]    终端国密签名证书私钥的文件路径地址。
 * @param       enc_key_file        [IN]    终端国密加密证书私钥的文件路径地址。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 为 NULL。
 * @retval      ERRNO_SETTING_FAILURE，在当前 ylong_tls 编译 feature 中不支持该功能。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_use_hybrid_privkey_file(ylong_tls_ctx *ctx, const char *general_key_file, const char *sig_key_file, const char *enc_key_file);

/**
 * @ingroup     ylong_tls
 * @brief       设置使用混合终端证书(通用TLS单证书 + 国密双证书)的私钥，返回对应状态码。可以通过 ylong_tls_pem_read_privatekey 解析
 *              pkcs8 格式私钥 PEM 流得到私钥结构体。
 * @param       ctx                 [IN]    需要配置的目标 ctx。
 * @param       general_key         [IN]    终端通用TLS证书私钥。
 * @param       sig_key             [IN]    终端国密签名证书私钥。
 * @param       enc_key             [IN]    终端国密加密证书私钥。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 为 NULL。
 * @retval      ERRNO_SETTING_FAILURE，在当前 ylong_tls 编译 feature 中不支持该功能。
 * @retval      ERRNO_PRIVKEY_OR_CERT_IO，配置失败，读取证书密钥文件发生 IO 错误。
 * @retval      ERRNO_INVALID_PRIVKEY_OR_CERT，配置失败，证书密钥对设置失败。可能包含未知或错误的密钥对，或使用错误或不支持的格式输入。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_use_hybrid_privkey(ylong_tls_ctx *ctx, ylong_tls_pkey *general_key, ylong_tls_pkey *sig_key, ylong_tls_pkey *enc_key);

/**
 * @ingroup     ylong_tls
 * @brief       解析 pkcs8 格式证书 PEM 流，返回对应的证书结构体。
 * @param       pem         [IN]    pkcs8 格式私钥 PEM 流。
 * @param       len         [IN]    PEM 流长度。
 * @retval      != NULL，读取成功。
 * @retval      NULL，配置失败，传入 pem 为 NULL，pem 长度与 len 不匹配，或是使用错误 / 不支持的格式输入。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
ylong_tls_cert_chain* ylong_tls_pem_read_x509(const void *pem, unsigned int len);

/**
 * @ingroup     ylong_tls
 * @brief       解析 pkcs8 格式证书 PEM 流，返回对应的证书结构体。
 * @param       bio    [IN]    读取 pkcs8 格式 PEM 证书的 bio，解析证书时会调用 bio 中注册的 read 函数。
 * @retval      != NULL，读取成功。
 * @retval      NULL，配置失败，传入 bio 为 NULL，或是使用错误 / 不支持的格式输入。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
ylong_tls_cert_chain* ylong_tls_pem_read_x509_bio(ylong_tls_bio *bio);

/**
 * @ingroup     ylong_tls
 * @brief       添加证书到指定证书链。
 * @param       dst         [IN]    目标证书链。
 * @param       dst_idx     [IN]    在目标证书链添加证书的索引位置。如果索引大于等于目标证书链长度，将添加在证书链尾端。
 * @param       src         [IN]    待添加证书的证书链。
 * @param       src_idx     [IN]    待添加证书的索引位置。如果索引大于等于源证书链长度，将报错。
 * @retval      YLONG_TLS_SUCCESS，添加成功。
 * @retval      ERRNO_NULL，添加失败，传入证书链为 NULL。
 * @retval      ERRNO_SETTING_FAILURE，添加失败，待添加证书的索引位置大于等于源证书链长度，或索引小于0。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_add_cert_to_cert_chain(ylong_tls_cert_chain *dst, unsigned int dst_idx, ylong_tls_cert_chain *src, unsigned int src_idx);

/**
 * @ingroup     ylong_tls
 * @brief       将解析得出的单独的国密签名证书链和国密加密证书链合并成一条证书链，便于调用 `ylong_tls_ctx_use_cert_chain` 进行设置。
 *              会改变 `sig_cert` 的值，不改变 `enc_cert` 的值。按照 TLCP 规范，合并会将加密叶子证书，即第一个证书，添加到签名叶子证
 *              书之后。
 * @param       sig_cert         [IN]    国密签名证书链。
 * @param       enc_cert         [IN]    国密加密证书链。
 * @retval      YLONG_TLS_SUCCESS，添加成功。
 * @retval      ERRNO_NULL，添加失败，传入证书链为 NULL。
 * @retval      ERRNO_SETTING_FAILURE，添加失败，签名证书链或加密证书链中不包含证书。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_combine_sm_cert_chain(ylong_tls_cert_chain *sig_cert, ylong_tls_cert_chain *enc_cert);

/**
 * @ingroup     ylong_tls
 * @brief       设置使用证书链，返回对应状态码。
 * @param       ctx    [IN]    需要配置的目标 ctx。
 * @param       file   [IN]    证书链的PEM文件路径地址，如有多个文件，以 `:` 分隔不同的证书文件，证书应当以终端证书-中间证书-根证书的顺序
 *                             按序排列。如果是国密，且使用 TLCP 版本或是使用 ECC 密钥交换算法，这一参数应当包含双证书链。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 为 NULL。
 * @retval      ERRNO_SETTING_FAILURE，在当前 ylong_tls 编译 feature 中不支持该功能，或当前设置中不应当使用该接口。请参阅注释。
 * @retval      ERRNO_PRIVKEY_OR_CERT_IO，配置失败，读取文件时发生 IO 错误。
 * @retval      ERRNO_INVALID_PRIVKEY_OR_CERT，配置失败，不支持的证书。可能包含未知或错误的证书链，或使用错误或不支持的格式输入。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_use_cert_chain_file(ylong_tls_ctx *ctx, const char *file);

/**
 * @ingroup     ylong_tls
 * @brief       设置使用证书链，返回对应状态码。可以通过 ylong_tls_pem_read_x509 解析pkcs8 格式证书 PEM 流得到证书结构体。如果是国密，且使用
 *              TLCP 版本或是使用 ECC 密钥交换算法，这一参数应当包含双证书链。
 * @param       ctx         [IN]    需要配置的目标 ctx。
 * @param       certs       [IN]    以 `ylong_tls_cert_chain` 表示的证书链，证书应当以终端证书-中间证书-根证书的顺序按序排列。如果是国密，且使用
 *                                  TLCP 版本或是使用 ECC 密钥交换算法，这一参数应当包含双证书链。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 或证书为 NULL。
 * @retval      ERRNO_SETTING_FAILURE，在当前 ylong_tls 编译 feature 中不支持该功能，或当前设置中不应当使用该接口。请参阅注释。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_use_cert_chain(ylong_tls_ctx *ctx, ylong_tls_cert_chain *certs);

/**
 * @ingroup     ylong_tls
 * @brief       设置使用混合TLS时，需要设置通用证书链以及国密双证书链。返回对应状态码。
 * @param       ctx                 [IN]    需要配置的目标 ctx。
 * @param       general_cert_file   [IN]    通用证书链的文件路径地址，如有多个文件，以 `:` 分隔不同的证书文件，证书应当以终端证书-
 *                                          中间证书-根证书的顺序按序排列。
 * @param       sm_cert_file        [IN]    国密证书链的文件路径地址，如果使用 TLCP 版本或是使用 ECC 密钥交换算法，应当包含双证书链。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 为 NULL。
 * @retval      ERRNO_SETTING_FAILURE，在当前 ylong_tls 编译 feature 中不支持该功能。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_use_hybrid_cert_chain_file(ylong_tls_ctx *ctx, const char *general_cert_file, const char *sm_cert_file);

/**
 * @ingroup     ylong_tls
 * @brief       设置使用证书链，返回对应状态码。可以通过 ylong_tls_pem_read_x509 解析pkcs8 格式证书 PEM 流得到证书结构体。
 * @param       ctx                 [IN]    需要配置的目标 ctx。
 * @param       general_certs       [IN]    以 `ylong_tls_cert_chain` 表示的通用证书链，证书应当以终端证书-中间证书-根证书的顺序按序排列。如果是国密，且使用
 *                                          TLCP 版本或是使用 ECC 密钥交换算法，这一参数应当包含双证书链。
 * @param       sm_certs            [IN]   国密证书链的文件路径地址，如果使用 TLCP 版本或是使用 ECC 密钥交换算法，应当包含双证书链。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 为 NULL。
 * @retval      ERRNO_PRIVKEY_OR_CERT_IO，配置失败，读取文件时发生 IO 错误。
 * @retval      ERRNO_INVALID_PRIVKEY_OR_CERT，配置失败，不支持的证书。可能包含未知或错误的证书链，或使用错误或不支持的格式输入。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_use_hybrid_cert_chain(ylong_tls_ctx *ctx, ylong_tls_cert_chain *general_certs, ylong_tls_cert_chain *sm_certs);

/**
 * @ingroup     ylong_tls
 * @brief       设置使用根证书，返回对应状态码。根证书文件格式为 PEM。
 * @param       ctx       [IN]    需要配置的目标 ctx。
 * @param       ca_file   [IN]    根证书的文件路径地址，如有多个文件，以 `:` 分隔不同的证书文件。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 为 NULL。
 * @retval      PRIVKEY_OR_CERT_IO_ERR，配置失败，读取文件时发生 IO 错误。
 * @retval      ERRNO_INVALID_PRIVKEY_OR_CERT，配置失败，不支持的证书。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_load_ca_files(ylong_tls_ctx *ctx, const char *ca_file);

/**
 * @ingroup     ylong_tls
 * @brief       设置使用根证书，返回对应状态码。根证书文件格式为 PEM。可以通过 ylong_tls_pem_read_x509 解析 pkcs8 格式证书 PEM 流得
 *              到证书结构体。如果根证书已经被设置，设置的根证书会被覆盖。
 * @param       ctx      [IN]    需要配置的目标 ctx。
 * @param       ca       [IN]    设置的根证书。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 或证书为 NULL。
 * @retval      PRIVKEY_OR_CERT_IO_ERR，配置失败，解析证书文件时发生 IO 错误。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_load_verify_ca(ylong_tls_ctx *ctx, ylong_tls_cert_chain *ca);

/**
 * @ingroup     ylong_tls
 * @brief       设置客户端验证服务端证书时，是否验证 domain，选择验证 IP 地址。该配置对于服务端无效。
 * @param       ctx         [IN]    需要配置的目标 ctx。
 * @param       ignore      [IN]    是否验证对方证书 domain，>= 1 表示验证，<= 0 表示不验证。默认为验证。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 为 NULL。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_set_verify_ignore_domain(ylong_tls_ctx *ctx, int ignore);

/**
 * @ingroup     ylong_tls
 * @brief       设置服务端是否强制验证客户端身份，该配置对于客户端无效，客户端必须验证服务端证书。服务端对客户端证书的验证有三种形式：
 *                1. 不验证，即不发送证书请求；使用这种模式应当设置 mandatory == 0，并且不使用 `ylong_tls_ctx_load_ca_files`
 *                    导入根证书；
 *                2. 可选验证，即发送证书请求，但是允许客户端不回复证书，如果回复则验证证书；使用这种模式应当设置 mandatory == 0，并且使用
 *                    `ylong_tls_ctx_load_ca_files` 导入根证书；
 *                3. 强制验证，设置要求强制验证，或所选密码套要求强制验证，即发送证书请求，客户端必须回复证书，且证书必须合法。使用这种模式应当
 *                    设置 mandatory == 1，并且使用 `ylong_tls_ctx_load_ca_files` 导入根证书。
 * @param       ctx         [IN]    需要配置的目标 ctx。
 * @param       mandatory   [IN]    是否强制验证对方证书，== 1 表示强制验证，== 0 表示非强制验证。如果使用密码套要求验证，即使设置非强制验证，
 *                                  服务端也会向客户端发送证书请求。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 ctx 为 NULL。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_set_verify(ylong_tls_ctx *ctx, int mandatory);

/**
 * @ingroup     ylong_tls
 * @brief       设置使用缓存大小，返回之前的配置大小。
 * @param       ctx      [IN]    需要配置的目标 ctx。
 * @param       size     [IN]    设置缓存大小。
 *                                 - 在客户端使用，如果大小为0，则关闭缓存，不允许客户端恢复握手。
 *                                 - 在服务端使用，如果服务端允许发送 Ticket，那么设置大小为0，会只允许 Stateless Resumption，
 *                                    即只使用 SessionID 恢复握手；如果不允许服务端发送 Ticket，则不允许服务端恢复握手。
 * @retval      ERRNO_NULL，传入 ctx 为 NULL。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_set_session_storage(ylong_tls_ctx *ctx, long size);

/**
 * @ingroup     ylong_tls
 * @brief       设置是否允许使用 Ticket。
 * @param       ctx        [IN]    需要配置的目标 ctx。
 * @param       enable     [IN]    是否开启 Ticket 允许 TLS 进行恢复握手，>= 1 表示开启，<= 0 表示关闭。
 * @retval      YLONG_TLS_SUCCESS，设置成功。
 * @retval      ERRNO_NULL，传入 ctx 为 NULL。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_set_ticket(ylong_tls_ctx *ctx, int enable);

/**
 * @ingroup     ylong_tls
 * @brief       设置客户端或服务端所支持的 ALPN 扩展，返回对应状态码。
 * @param       ctx            [IN]    需要配置的目标 ctx。
 * @param       protos         [IN]    ALPN protos 字串，以 `:` 分隔支持的 ALPN。如果传入空指针，则清空已有的设置。
 * @param       len            [IN]    ALPN protos 字串长度。
 * @retval      YLONG_TLS_SUCCESS，设置成功。
 * @retval      ERRNO_SETTING_FAILURE，设置失败。
 * @retval      ERRNO_NULL，传入 ctx 为 NULL。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_set_alpn_protos(ylong_tls_ctx *ctx, const unsigned char *protos, unsigned int len);

/**
 * @ingroup     ylong_tls
 * @brief       设置客户端是否要发送 SNI 扩展，返回对应状态码。
 * @param       ctx            [IN]    需要配置的目标 ctx。
 * @param       enabled        [IN]    是否开启客户端 SNI 扩展，>= 1 表示开启，<= 0 表示关闭。
 * @retval      YLONG_TLS_SUCCESS，设置成功。
 * @retval      ERRNO_NULL，传入 ctx 为 NULL。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_ctx_set_sni(ylong_tls_ctx *ctx, int enabled);

/**
 * @ingroup     ylong_tls
 * @brief       根据 ylong_tls_ctx，生成对应的客户端配置，以一个 ylong_tls_stream* 成员句柄形式返回。
 * @param       ctx            [IN]    配置 ctx。
 * @retval      NULL，配置错误。可以调用 ylong_tls_get_ctx_err 获取具体错误原因。
 * @retval      =! NULL，合法的 ylong_tls_stream* 成员句柄，包含对应配置。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
ylong_tls_stream *ylong_tls_stream_new(ylong_tls_ctx *ctx);

/**
 * @ingroup     ylong_tls
 * @brief       判断给定 ylong_tls_stream 是否是客户端 stream。
 * @param       stream            [IN]    判断是否是客户端的 ylong_tls_stream。
 * @retval      1，给定 stream 是客户端连接 stream。
 * @retval      0，给定 stream 是服务端连接 stream。
 * @retval      ERRNO_NULL，给定 stream 为 NULL。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_stream_is_client(ylong_tls_stream *stream);

/**
 * @ingroup     ylong_tls
 * @brief       判断给定 ylong_tls_stream 是否已完成握手。
 * @param       stream            [IN]    目标 ylong_tls_stream。
 * @retval      1，已完成握手。
 * @retval      0，未完成握手。
 * @retval      ERRNO_NULL，给定 stream 为 NULL。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_stream_hs_complete(ylong_tls_stream *stream);

/**
 * @ingroup     ylong_tls
 * @brief       设置底层使用 IO，返回对应状态码。目前只支持阻塞 IO。设置成功之后，ylong_tls_stream 释放时会将 bio 一起释放掉，
 *              不可以再调用 `ylong_tls_bio_free` 进行释放，否则会造成二次释放（double-free）问题。TLS 链接时会调用 bio 中
 *              注册的 read 和 write 函数指针。
 * @param       stream    [IN]    需要配置的目标 stream。
 * @param       rbio      [IN]    读 IO 句柄。
 * @param       wbio      [IN]    写 IO 句柄。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，stream 为 NULL。
 * @retval      ERRNO_SETTING_FAILURE，配置 IO 失败，IO 可能未包含读写函数，函数指针为空。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_set_bio(ylong_tls_stream *stream, ylong_tls_bio *bio);

/**
 * @ingroup     ylong_tls
 * @brief       获取底层使用的 IO，ylong_tls_stream 释放时会将 bio 一起释放掉，不可以再调用 `ylong_tls_bio_free` 进行释放，否则会
 *              造成二次释放（double-free）问题。
 * @param       stream    [IN]    目标 stream。
 * @retval      != NULL，返回对应设置在 ylong_tls_stream 中的 BIO。
 * @retval      NULL，传入 stream 为 NULL，或 stream 还未设置 BIO。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
ylong_tls_bio *ylong_tls_get_bio(ylong_tls_stream *stream);

/**
 * @ingroup     ylong_tls
 * @brief       设置客户端连接的主机域名，该设置对于服务端无效。主机域名允许为空，但是只有在客户端设置了证书校验忽视域名
 *              （即调用 `ylong_tls_ctx_set_verify_ignore_domain`）且关闭 SNI 时，空域名不会导致调用
 *              `ylong_tls_connect` 产生链接错误。
 * @param       stream     [IN]    需要配置的目标 stream。
 * @param       hostname   [IN]    连接的主机域名。
 * @retval      YLONG_TLS_SUCCESS，配置成功。
 * @retval      ERRNO_NULL，配置失败，传入 stream 为 NULL。
 * @retval      ERRNO_INVALID_HOSTNAME，配置失败，主机域名非法，包含非 utf8 字符。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_set_hostname(ylong_tls_stream *stream, const char *hostname);

/**
 * @ingroup     ylong_tls
 * @brief       根据客户端 ylong_tls_stream，连接的主机域名，进行TLS链接。握手成功以后，再次调用 connect 时，客户端会尝试请求新的 TLS 链接，
 *              并重新开始 TLS 握手流程。可以使用 ylong_tls_get_stream_conn_err 获取具体错误信息。
 * @param       stream            [IN]    配置完成的 ylong_tls_stream。
 * @retval      YLONG_TLS_SUCCESS，链接并握手成功，可以进行应用数据传输。
 * @retval      ERRNO_NULL，传入 stream 为 NULL。
 * @retval      ERRNO_HS_WANTS_READ，由于 IO 为非阻塞 IO，且读资源未准备好，握手被中断。
 * @retval      ERRNO_HS_WANTS_WRITE，由于 IO 为非阻塞 IO，且写资源未准备好，握手被中断。
 * @retval      ERRNO_INVALID_HOSTNAME，传入不适配的 hostaname，要求校验 cert domain 或 发送 SNI 时，传入 NULL。
 * @retval      ERRNO_STREAM_END_MISMATCH，使用服务端 Stream 调用 connect。
 * @retval      ERRNO_NO_RW_BIO，BIO 读写函数为空指针。
 * @retval      ERRNO_BIO_CONNECTION_FAILURE，BIO 其他错误。
 * @retval      ERRNO_EOF，传输未完成，UnexpectedEOF 错误。
 * @retval      ERRNO_OUT_OF_MEMORY，内存错误。
 * @retval      ERRNO_UNCLASSIFIED_HS_FAILURE，握手时错误，
 * @retval      ERRNO_UNCLASSIFIED_TLS_CONN_FAILURE，其他错误。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_connect(ylong_tls_stream *stream);

/**
 * @ingroup     ylong_tls
 * @brief       根据服务端 ylong_tls_stream，接收 TLS 链接并开始握手。握手成功以后，再次调用 accept 时，服务端会尝试接收新的 TLS 链接，
 *              并重新开始 TLS 握手流程。可以使用 ylong_tls_get_stream_conn_err 获取具体错误信息。
 * @param       stream            [IN]    配置完成的 ylong_tls_stream。
 * @retval      YLONG_TLS_SUCCESS，链接并握手成功，可以进行应用数据传输。
 * @retval      ERRNO_NULL，传入 stream 为 NULL。
 * @retval      ERRNO_STREAM_END_MISMATCH，使用客户端 Stream 调用 accept。
 * @retval      ERRNO_HS_WANTS_READ，由于 IO 为非阻塞 IO，且读资源未准备好，握手被中断。
 * @retval      ERRNO_HS_WANTS_WRITE，由于 IO 为非阻塞 IO，且写资源未准备好，握手被中断。
 * @retval      ERRNO_BIO_CONNECTION_FAILURE，BIO 错误。
 * @retval      ERRNO_EOF，传输未完成，UnexpectedEOF 错误。
 * @retval      ERRNO_OUT_OF_MEMORY，内存错误。
 * @retval      ERRNO_UNCLASSIFIED_HS_FAILURE，握手时错误，
 * @retval      ERRNO_UNCLASSIFIED_TLS_CONN_FAILURE，其他错误。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_accept(ylong_tls_stream *stream);

/**
 * @ingroup     ylong_tls
 * @brief       给定完成链接以及握手的 ylong_tls_stream，读取协商的 ALPN 扩展字段。
 * @param       stream            [IN]    完成链接以及握手的 ylong_tls_stream。
 * @param       data              [OUT]    存放 ALPN protos 协商字符串指针。如果没有协商 ALPN 或协商失败，data 会被设置为 NULL。
 * @param       len               [OUT]    存放协商的 ALPN 的长度。如果没有协商 ALPN 或协商失败，len 会被设置为 0。
 * @retval      YLONG_TLS_SUCCESS，返回成功。
 * @retval      ERRNO_NULL，传入 stream 为 NULL。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_get_alpn(const ylong_tls_stream *stream, unsigned char **data, unsigned int *len);

/**
 * @ingroup     ylong_tls
 * @brief       给定完成链接以及握手的 ylong_tls_stream，读取传输的应用数据。该接口只能用于串行场景，不可以并发使用。
 * @param       stream            [IN]    完成链接以及握手的 ylong_tls_stream。
 * @param       buf               [OUT]   读取存入的数据指针。
 * @param       num               [IN]    缓存大小。
 * @retval      >= 0，成功，返回实际读取的字符个数。
 * @retval      ERRNO_NULL，入参 stream 为 NULL。
 * @retval      ERRNO_BIO_READ_WOULD_BLOCK，读资源未就绪。
 * @retval      ERRNO_BIO_TIMEOUT, 底层 bio 超时。
 * @retval      ERRNO_BIO_EOF，底层 IO 返回 EOF。
 * @retval      ERRNO_NO_TLS_CONN，入参 stream 未开始 TLS 链接。
 * @retval      ERRNO_HS_INCOMPLETE，入参 stream 未完成 TLS 握手。
 * @retval      ERRNO_UNCLASSIFIED_RW_FAILURE，其他错误。使用 ylong_tls_get_stream_conn_err 获取具体错误信息。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_read(ylong_tls_stream *stream, void *buf, unsigned int num);

/**
 * @ingroup     ylong_tls
 * @brief       给定完成链接以及握手的 ylong_tls_stream，在有并发写操作的情况下，应该使用此接口读取传输的应用数据。不可以用于并发读场景。
 *              如果没有并发写操作，可以使用性能更好的 `ylong_tls_read`。读操作遇到错误时，需要向底层 IO 写入 TLS 消息，与非并发写接口
 *              并发使用时会有数据竞争问题，因此，在该函数中，对于读操作之间的写入加锁。
 * @param       stream            [IN]    完成链接以及握手的 ylong_tls_stream。
 * @param       buf               [OUT]   读取存入的数据指针。
 * @param       num               [IN]    缓存大小。
 * @retval      >= 0，成功，返回实际读取的字符个数。
 * @retval      ERRNO_NULL，入参 stream 为 NULL。
 * @retval      ERRNO_BIO_READ_WOULD_BLOCK，读资源未就绪。
 * @retval      ERRNO_BIO_TIMEOUT, 底层 bio 超时。
 * @retval      ERRNO_BIO_EOF，底层 IO 返回 EOF。
 * @retval      ERRNO_NO_TLS_CONN，入参 stream 未开始 TLS 链接。
 * @retval      ERRNO_HS_INCOMPLETE，入参 stream 未完成 TLS 握手。
 * @retval      ERRNO_UNCLASSIFIED_RW_FAILURE，其他错误。使用 ylong_tls_get_stream_conn_err 获取具体错误信息。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_read_wlocked(ylong_tls_stream *stream, void *buf, unsigned int num);

/**
 * @ingroup     ylong_tls
 * @brief       给定正在握手的 ylong_tls_stream，读取 BIO 中存储的数据。该函数会试图推进握手流程，如果使用同步阻塞 IO，可能会导致函数
 *              阻塞调用。
 * @param       stream            [IN]    完成链接以及握手的 ylong_tls_stream。
 * @param       buf               [OUT]   读取存入的数据指针。
 * @param       num               [IN]    缓存大小。
 * @retval      >= 0，成功，返回实际读取的字符个数。当握手已完成时，也会返回 0。如果需要读取应用数据，应调用 ylong_tls_read。
 * @retval      ERRNO_NULL，入参 stream 为 NULL。
 * @retval      ERRNO_BIO_TIMEOUT, 底层 bio 超时。
 * @retval      ERRNO_BIO_EOF，底层 IO 返回 EOF。
 * @retval      ERRNO_NO_TLS_CONN，入参 stream 未开始 TLS 链接。
 * @retval      ERRNO_UNCLASSIFIED_RW_FAILURE，其他错误。使用 ylong_tls_get_stream_conn_err 获取具体错误信息。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_hs_read(ylong_tls_stream *stream, void *buf, unsigned int num);

/**
 * @ingroup     ylong_tls
 * @brief       给定完成链接以及握手的 ylong_tls_stream，写入传输的应用数据。该接口只能用于串行场景，不可以并发使用。
 * @param       stream            [IN]    完成链接以及握手的 ylong_tls_stream。
 * @param       buf               [IN]    写入源。
 * @param       num               [IN]    缓存大小。
 * @retval      >= 0，成功，返回实际写入的字符个数。
 * @retval      ERRNO_NULL，入参 stream 为 NULL。
 * @retval      ERRNO_BIO_WRITE_WOULD_BLOCK，写资源未就绪。
 * @retval      ERRNO_BIO_TIMEOUT, 底层 bio 超时。
 * @retval      ERRNO_BIO_EOF，底层 IO 返回 EOF。
 * @retval      ERRNO_NO_TLS_CONN，入参 stream 未开始 TLS 链接。
 * @retval      ERRNO_HS_INCOMPLETE，入参 stream 未完成TLS握手。
 * @retval      ERRNO_UNCLASSIFIED_RW_FAILURE，其他错误。使用 ylong_tls_get_stream_conn_err 获取具体错误信息。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_write(ylong_tls_stream *stream, const void *buf, unsigned int num);

/**
 * @ingroup     ylong_tls
 * @brief       给定完成链接以及握手的 ylong_tls_stream，在有并发读操作的情况下，应该使用此接口写入传输的应用数据。不可以用于并发写场景。
 *              如果没有并发读操作，可以使用性能更好的 `ylong_tls_write`。当读操作遇到错误时，需要向底层 IO 写入 TLS 消息，与非并发写接口
 *              并发使用时会有数据竞争问题，因此，在该函数中，对于写操作加锁。
 * @param       stream            [IN]    完成链接以及握手的 ylong_tls_stream。
 * @param       buf               [IN]    写入源。
 * @param       num               [IN]    缓存大小。
 * @retval      >= 0，成功，返回实际写入的字符个数。
 * @retval      ERRNO_NULL，入参 stream 为 NULL。
 * @retval      ERRNO_BIO_WRITE_WOULD_BLOCK，写资源未就绪。
 * @retval      ERRNO_BIO_TIMEOUT, 底层 bio 超时。
 * @retval      ERRNO_BIO_EOF，底层 IO 返回 EOF。
 * @retval      ERRNO_NO_TLS_CONN，入参 stream 未开始 TLS 链接。
 * @retval      ERRNO_HS_INCOMPLETE，入参 stream 未完成TLS握手。
 * @retval      ERRNO_UNCLASSIFIED_RW_FAILURE，其他错误。使用 ylong_tls_get_stream_conn_err 获取具体错误信息。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_write_rlocked(ylong_tls_stream *stream, const void *buf, unsigned int num);

/**
 * @ingroup     ylong_tls
 * @brief       给定正在握手的 ylong_tls_stream，将数据写入 BIO 中。该函数会试图推进握手流程，如果使用同步阻塞 IO，可能会导致函数
 *              阻塞调用。
 * @param       stream            [IN]    完成链接以及握手的 ylong_tls_stream。
 * @param       buf               [IN]    写入源。
 * @param       num               [IN]    缓存大小。
 * @retval      >= 0，成功，返回实际写入的字符个数。当握手已完成时，也会返回 0。如果需要写入应用数据，应调用 ylong_tls_write。
 * @retval      ERRNO_NULL，入参 stream 为 NULL。
 * @retval      ERRNO_BIO_TIMEOUT, 底层 bio 超时。
 * @retval      ERRNO_BIO_EOF，底层 IO 返回 EOF。
 * @retval      ERRNO_NO_TLS_CONN，入参 stream 未开始 TLS 链接。
 * @retval      ERRNO_UNCLASSIFIED_RW_FAILURE，其他错误。使用 ylong_tls_get_stream_conn_err 获取具体错误信息。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_hs_write(ylong_tls_stream *stream, const void *buf, unsigned int num);

/**
 * @ingroup     ylong_tls
 * @brief       获取当前 TLS 链接的终止状态，以 bit mask 的形式展现，
 * @param       stream    [IN]    目标 stream。
 * @retval      ERRNO_NULL，传入 stream 为 NULL。
 * @retval      YLONG_TLS_NOT_SHUTDOWN，stream 未收到也没有发送 close_notify 消息。
 * @retval      YLONG_TLS_RECV_SHUTDOWN，stream 收到对端发来的 close_notify 消息。
 * @retval      YLONG_TLS_SEND_SHUTDOWN，stream 已发送 close_notify 消息。
 * @retval      YLONG_TLS_RECV_SHUTDOWN xor YLONG_TLS_SEND_SHUTDOWN，stream 已发送 close_notify 消息，并收到对端发送的 close_notify。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_get_shutdown(ylong_tls_stream *stream);

/**
 * @ingroup     ylong_tls
 * @brief       终止 TLS 链接，向对端发送 CLOSE_NOTIFY。
 * @param       stream    [IN]    目标 stream。
 * @retval      YLONG_TLS_SUCCESS，发送成功。
 * @retval      ERRNO_NULL，传入 stream 为 NULL。
 * @retval      ERRNO_NO_TLS_CONN，入参 stream 未开始 TLS 链接。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_shutdown(ylong_tls_stream *stream);

/**
 * @ingroup     ylong_tls
 * @brief       获取生成客户端或服务端配置时产生的错误。
 * @param       ctx            [IN]    配置 ctx。
 * @retval      YLONG_TLS_BUILD_CTX_SUCCESS，配置成功。
 * @retval      ERRNO_INVALID_KEY，密钥配置错误。
 * @retval      ERRNO_INVALID_CERT，证书链配置错误。
 * @retval      ERRNO_ROOT_CERT_UNSPECIFIED，要求验证，但是缺少根证书配置。
 * @retval      ERRNO_UNSUPPORTED_VERSION，不支持的版本。
 * @retval      ERRNO_UNSUPPORTED_CIPHER_SUITE，不支持的密码套。
 * @retval      ERRNO_EMPTY_ALPN，ALPN 为空。
 * @retval      ERRNO_UNCLASSIFIED_BUILD_CTX_FAILURE，其他错误。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_get_ctx_err(ylong_tls_ctx *ctx);

/**
 * @ingroup     ylong_tls
 * @brief       获取 ylong_tls_bio 读写时产生的错误。获取到的资源不需要用户自己释放。
 * @param       bio            [IN]     读写使用的 ylong_tls_bio。
 * @param       err            [OUT]    存放错误信息字符串的指针，当链接存在错误时会替换指针，链接成功时替换为 NULL。
 * @param       len            [OUT]    存放错误信息字符串长度。当 bio 为空或如果发生的错误没有错误信息时，输出 0.
 * @retval      YLONG_TLS_SUCCESS，传入 bio 不为 `Null`，进行错误信息读取。
 * @retval      ERRNO_NULL，传入 bio 为 `Null`。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_get_bio_err(ylong_tls_bio *bio, char** err, unsigned int* len);

/**
 * @ingroup     ylong_tls
 * @brief       获取 TLS 握手与数据传输时时产生的错误。用户不需要释放获取到的 err 指针。
 * @param       stream         [IN]    链接以及握手使用的 ylong_tls_stream。
 * @param       err            [OUT]    存放错误信息字符串的指针，当链接存在错误时会替换指针，链接成功时替换为 NULL。
 * @param       len            [OUT]    存放错误信息字符串长度。当 stream 为空或如果发生的错误没有错误信息时，输出 0.
 * @retval      YLONG_TLS_SUCCESS，传入 stream 不为 `Null`，进行错误信息读取。
 * @retval      ERRNO_NULL，传入 stream 为 `Null`。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
int ylong_tls_get_stream_conn_err(ylong_tls_stream *stream, char** err, unsigned int* len);

/**
 * @ingroup     ylong_tls
 * @brief       销毁一个结构实例，并释放内存。如果 destroy 函数非空，调用这一函数指针析构 bio 中的 IO 资源。如指针为空并且需要获取 IO 资
 *              源，用户可以通过 `ylong_tls_bio_get_data` 获取这一指针并进行管理。
 * @param       bio    [IN]    待销毁的 ylong_tls_bio* 成员句柄。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
void ylong_tls_bio_free(ylong_tls_bio *bio);

/**
 * @ingroup     ylong_tls
 * @brief       销毁一个结构实例，并释放内存。
 * @param       pkey    [IN]    待销毁的 ylong_tls_pkey* 成员句柄。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
void ylong_tls_pkey_free(ylong_tls_pkey *pkey);

/**
 * @ingroup     ylong_tls
 * @brief       销毁一个结构实例，并释放内存。
 * @param       cert        [IN]    待销毁的 ylong_tls_cert_chain* 成员句柄。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
void ylong_tls_cert_chain_free(ylong_tls_cert_chain *cert);

/**
 * @ingroup     ylong_tls
 * @brief       销毁一个结构实例，并释放内存。
 * @param       ctx    [IN]    待销毁的 ylong_tls_ctx* 成员句柄。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @see
 */
void ylong_tls_ctx_free(ylong_tls_ctx *ctx);

/**
 * @ingroup     ylong_tls
 * @brief       销毁一个结构实例，并释放内存。
 * @param       stream    [IN]    待销毁的 ylong_tls_stream* 成员句柄。
 * @li          ylong_tls.h：该接口声明所在的文件。
 * @attention   ylong_tls_stream_free 析构时会将绑定的 bio 一起释放，如果 bio 中存在 ptr，且 ptr 的析构函数没有绑定在 bio 中，
 *              需要使用 ylong_tls_bio_get_data 获取指针并析构对应资源，否则会造成内存泄漏。
 * @see
 */
void ylong_tls_stream_free(ylong_tls_stream *stream);

#ifdef __cplusplus
#if __cplusplus
}
#endif
#endif

#endif //YLONG_TLS_H
