/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package tls

import (
	"bytes"
	"io"
	"sync"
	"sync/atomic"
)

const (
	maxRwBufSize = 18437
)

type rwBuf struct {
	mu sync.Mutex
	r  []byte
	rl int32
	w  []byte
	wl int32
	c  *Conn
}

func newBioBuf(c *Conn) *rwBuf {
	return &rwBuf{
		r: make([]byte, maxRwBufSize),
		w: make([]byte, maxRwBufSize),
		c: c,
	}
}

func (b *rwBuf) readBufLen() int {
	return int(atomic.LoadInt32(&b.rl))
}

func (b *rwBuf) writeBufLen() int {
	return int(atomic.LoadInt32(&b.wl))
}

func (b *rwBuf) advanceReadBuf(i int) {
	atomic.AddInt32(&b.rl, int32(i))
}

func (b *rwBuf) advanceWriteBuf(i int) {
	atomic.StoreInt32(&b.wl, int32(i))
}

func (b *rwBuf) resetReadBuf(n int) {
	copy(b.r[:], b.r[n:])
	atomic.AddInt32(&b.rl, int32(-n))
}

func (b *rwBuf) resetWriteBuf() {
	atomic.StoreInt32(&b.wl, 0)
}

func (b *rwBuf) write() error {
	b.mu.Lock()
	size := b.writeBufLen()
	if size == 0 {
		b.mu.Unlock()
		return nil
	}
	reader := bytes.NewReader(b.w[:size])
	_, err := io.Copy(b.c.conn, reader)
	if err != nil {
		b.mu.Unlock()
		return err
	}
	b.resetWriteBuf()
	b.mu.Unlock()
	return nil
}
