/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package tls

// #cgo CFLAGS: -I${SRCDIR}/../packaged/include
// #cgo LDFLAGS: -lylong_tls -ldl -lpthread -lm
//
// #cgo linux,amd64 LDFLAGS: -L${SRCDIR}/../packaged/lib/linux-amd64
// #cgo linux,arm64 LDFLAGS: -L${SRCDIR}/../packaged/lib/linux-aarch64
//
// #include <ylong_tls.h>
import "C"

import (
	"io"
	"net"
	"unsafe"

	_ "huawei.com/huaweichain/gmssl/packaged/include"
	_ "huawei.com/huaweichain/gmssl/packaged/lib"
)

// ReadCallback used for C.ylong_tls_bio_go_conn_read
//
//export ReadCallback
func ReadCallback(connId unsafe.Pointer, out *C.char, len C.uint) C.int {
	b := (*rwBuf)(connId)
	if !b.c.isHandshakeCompleted() {
		return readHandshake(b.c.conn, out, len)
	}
	rl := b.c.buf.readBufLen()
	if rl == 0 {
		return C.ERRNO_BIO_READ_WOULD_BLOCK
	}
	// Use the memory directly without copy
	readBytes := (*[1 << 30]byte)(unsafe.Pointer(out))[:int(len):int(len)]

	src := b.r[:rl]
	n := copy(readBytes, src)
	b.c.buf.resetReadBuf(n)
	return C.int(n)
}

// WriteCallback used for C.ylong_tls_bio_go_conn_write
//
//export WriteCallback
func WriteCallback(connId unsafe.Pointer, in *C.char, len C.uint) C.int {
	b := (*rwBuf)(connId)
	if !b.c.isHandshakeCompleted() {
		return writeHandshake(b.c.conn, in, len)
	}

	writeLen := b.c.buf.writeBufLen()
	if writeLen != 0 {
		return C.ERRNO_BIO_WRITE_WOULD_BLOCK
	}
	// Use the memory directly without copy
	writeBytes := (*[1 << 30]byte)(unsafe.Pointer(in))[:int(len):int(len)]
	n := copy(b.w, writeBytes)
	b.c.buf.advanceWriteBuf(n)
	return C.int(n)
}

// DestroyCallback used for C.ylong_tls_bio_go_conn_destroy
//
//export DestroyCallback
func DestroyCallback(_ unsafe.Pointer) C.int {
	return 1
}

func readHandshake(conn net.Conn, out *C.char, len C.uint) C.int {
	// Use the memory directly without copy
	readBytes := (*[1 << 30]byte)(unsafe.Pointer(out))[:int(len):int(len)]
	readLen, err := conn.Read(readBytes)
	if err != nil {
		netErr, ok := err.(net.Error)
		if ok && netErr.Timeout() {
			return C.ERRNO_BIO_TIMEOUT
		}
		if err == io.EOF {
			return C.ERRNO_BIO_EOF
		}
		return -1
	}
	return C.int(readLen)
}

func writeHandshake(conn net.Conn, in *C.char, len C.uint) C.int {
	// Use the memory directly without copy
	writeBytes := (*[1 << 30]byte)(unsafe.Pointer(in))[:int(len):int(len)]
	writeLen, err := conn.Write(writeBytes)
	if err != nil {
		return -1
	}
	return C.int(writeLen)
}
