/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package tls

/*
#include "ssl.h"
*/
import "C"
import (
	"crypto/tls"
	"io"
	"net"
	"sync"
	"time"
	"unsafe"

	"github.com/pkg/errors"
)

// A Conn represents a secured connection.
// It implements the net.Conn interface.
type Conn struct {
	conn     net.Conn
	isClient bool

	// handshakeStatus is 1 if the connection is currently transferring
	// application data (i.e. is not currently processing a handshake).
	// This field is only to be accessed with sync/atomic.
	handshakeStatus uint32
	handshakeErr    error   // error resulting from handshake
	config          *Config // configuration passed to constructor
	// constant after handshake; protected by handshakeMutex
	handshakeMutex sync.Mutex
	streamMutex    sync.RWMutex
	once           sync.Once

	buf *rwBuf

	stream unsafe.Pointer // C.ylong_tls_stream
}

// Server returns a new TLS server side connection
// using conn as the underlying transport.
// The configuration config must be non-nil and must include
// at least one certificate or else set GetCertificate.
func Server(conn net.Conn, config *Config) *Conn {
	c := &Conn{
		conn:   conn,
		config: config,
	}
	c.buf = newBioBuf(c)
	return c
}

// Client returns a new TLS client side connection
// using conn as the underlying transport.
// The config cannot be nil: users must set either ServerName or
// InsecureSkipVerify in the config.
func Client(conn net.Conn, config *Config) *Conn {
	c := &Conn{
		conn:     conn,
		config:   config,
		isClient: true,
	}
	c.buf = newBioBuf(c)
	return c
}

// Read reads data from the connection.
// Read can be made to time out and return an error after a fixed
// time limit; see SetDeadline and SetReadDeadline.
func (c *Conn) Read(b []byte) (int, error) {
	if err := c.Handshake(); err != nil {
		return 0, errors.WithMessage(err, "handshake error")
	}
	if len(b) == 0 {
		// Put this after Handshake, in case people were calling
		// Read(nil) for the side effect of the Handshake.
		return 0, nil
	}

	c.streamMutex.RLock()
	n, err := c.read(b)
	c.streamMutex.RUnlock()
	return n, err
}

func (c *Conn) read(b []byte) (int, error) {
	if c.stream == nil {
		return 0, errors.New("stream closed when read")
	}
	var res C.int
	for {
		code := C.ylong_tls_bio_set_data(C.ylong_tls_get_bio(c.stream), unsafe.Pointer(c.buf))
		if code != C.YLONG_TLS_SUCCESS {
			return 0, errors.Errorf("failed to set bio for ylong tls with code %d", int32(code))
		}
		res = C.ylong_tls_read_wlocked(c.stream, unsafe.Pointer(&b[0]), C.uint(len(b)))
		if res < 0 {
			switch res {
			case C.ERRNO_BIO_READ_WOULD_BLOCK:
				offset := c.buf.readBufLen()
				n, err := c.conn.Read(c.buf.r[offset:])
				if err != nil {
					return 0, err
				}
				c.buf.advanceReadBuf(n)
				continue
			case C.ERRNO_BIO_WRITE_WOULD_BLOCK:
				if err := c.buf.write(); err != nil {
					return 0, err
				}
				continue
			case C.ERRNO_BIO_TIMEOUT:
				return 0, &net.OpError{Op: "read", Err: &timeoutError{}}
			case C.ERRNO_BIO_EOF:
				return 0, io.EOF
			default:
				result := C.getStreamError(c.stream)
				msg := C.GoStringN(result.err, result.err_len)
				C.freeResultErr(result)
				return 0, errors.Errorf("failed to call ylong tls read with error code %d and message: %v",
					int(res), msg)
			}
		}

		if err := c.buf.write(); err != nil {
			return 0, err
		}
		break
	}

	isShutdown := C.ylong_tls_get_shutdown(c.stream)
	if isShutdown&C.YLONG_TLS_RECV_SHUTDOWN == C.YLONG_TLS_RECV_SHUTDOWN {
		return 0, io.EOF
	}
	return int(res), nil
}

// Write writes data to the connection.
// Write can be made to time out and return an error after a fixed
// time limit; see SetDeadline and SetWriteDeadline.
func (c *Conn) Write(b []byte) (int, error) {
	if err := c.Handshake(); err != nil {
		return 0, errors.WithMessage(err, "handshake error")
	}

	c.streamMutex.RLock()
	n, err := c.write(b)
	c.streamMutex.RUnlock()
	return n, err
}

func (c *Conn) write(b []byte) (int, error) {
	if c.stream == nil {
		return 0, errors.New("stream closed when write")
	}
	code := C.ylong_tls_bio_set_data(C.ylong_tls_get_bio(c.stream), unsafe.Pointer(c.buf))
	if code != C.YLONG_TLS_SUCCESS {
		return 0, errors.Errorf("failed to set bio for ylong tls with code %d", int32(code))
	}
	for {
		res := C.ylong_tls_write_rlocked(c.stream, unsafe.Pointer(&b[0]), C.uint(len(b)))
		if res < 0 {
			if res == C.ERRNO_BIO_WRITE_WOULD_BLOCK {
				if err := c.buf.write(); err != nil {
					return 0, err
				}
				continue
			}
			result := C.getStreamError(c.stream)
			msg := C.GoStringN(result.err, result.err_len)
			C.freeResultErr(result)
			return 0, errors.Errorf("failed to call ylong tls write with error code %d and message: %v",
				int(res), msg)
		}
		if err := c.buf.write(); err != nil {
			return 0, err
		}

		isShutdown := C.ylong_tls_get_shutdown(c.stream)
		if isShutdown&C.YLONG_TLS_SEND_SHUTDOWN == C.YLONG_TLS_SEND_SHUTDOWN {
			return 0, io.EOF
		}
		return int(res), nil
	}
}

// Close closes the connection.
func (c *Conn) Close() error {
	var err error
	c.once.Do(func() {
		err = c.conn.Close()
		c.streamMutex.Lock()
		C.ylong_tls_stream_free(c.stream)
		c.stream = nil
		c.streamMutex.Unlock()
	})
	return err
}

// LocalAddr returns the local network address.
func (c *Conn) LocalAddr() net.Addr {
	return c.conn.LocalAddr()
}

// RemoteAddr returns the remote network address.
func (c *Conn) RemoteAddr() net.Addr {
	return c.conn.RemoteAddr()
}

// SetDeadline sets the read and write deadlines associated with the connection.
// A zero value for t means Read and Write will not time out.
// After a Write has timed out, the TLS state is corrupt and all future writes will return the same error.
func (c *Conn) SetDeadline(t time.Time) error {
	return c.conn.SetDeadline(t)
}

// SetReadDeadline sets the read deadline on the underlying connection.
// A zero value for t means Read will not time out.
func (c *Conn) SetReadDeadline(t time.Time) error {
	return c.conn.SetReadDeadline(t)
}

// SetWriteDeadline sets the write deadline on the underlying connection.
// A zero value for t means Write will not time out.
// After a Write has timed out, the TLS state is corrupt and all future writes will return the same error.
func (c *Conn) SetWriteDeadline(t time.Time) error {
	return c.conn.SetWriteDeadline(t)
}

// ConnectionState returns basic TLS details about the connection.
func (c *Conn) ConnectionState() ConnectionState {
	var state ConnectionState
	state.HandshakeComplete = c.isHandshakeCompleted()
	if c.config.Renegotiation != tls.RenegotiateNever {
		state.ekm = func(label string, context []byte, length int) ([]byte, error) {
			return nil, errors.New("crypto/tls: ExportKeyingMaterial is unavailable when renegotiation is enabled")
		}
	}
	return state
}

type timeoutError struct{}

func (e *timeoutError) Error() string   { return "i/o timeout" }
func (e *timeoutError) Timeout() bool   { return true }
func (e *timeoutError) Temporary() bool { return true }
