/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package tls

// #include "ssl.h"
import "C"

import (
	"strings"
	"sync/atomic"
	"unsafe"

	"github.com/pkg/errors"
)

// Handshake runs the client or server handshake
func (c *Conn) Handshake() error {
	c.handshakeMutex.Lock()
	defer c.handshakeMutex.Unlock()

	if err := c.handshakeErr; err != nil {
		return err
	}
	if c.isHandshakeCompleted() {
		return nil
	}

	if c.isClient {
		c.handshakeErr = c.clientHandshake()
	} else {
		c.handshakeErr = c.serverHandshake()
	}
	return c.handshakeErr
}

func (c *Conn) clientHandshake() error {
	var cert []byte
	for i := range c.config.Certs {
		cert = append(cert, c.config.Certs[i]...)
	}
	root := c.getRoot()
	serverName := C.CString(c.config.ServerName)
	defer C.free(unsafe.Pointer(serverName))
	var builder strings.Builder
	for _, each := range c.config.NextProtos {
		builder.WriteString(each)
		builder.WriteString(":")
	}
	alpnProto := C.CString(strings.TrimSuffix(builder.String(), ":"))
	defer C.free(unsafe.Pointer(alpnProto))

	cfg := C.struct_config{
		conn_id:    unsafe.Pointer(c.buf),
		serverName: serverName,
		alpn_proto: alpnProto,
		key:        unsafe.Pointer(&(c.config.Key[0])),
		key_len:    C.uint(len(c.config.Key)),
		cert:       unsafe.Pointer(&cert[0]),
		cert_len:   C.uint(len(cert)),
		root:       unsafe.Pointer(&root[0]),
		root_len:   C.uint(len(root)),
	}

	result := C.clientHandshake(cfg)
	defer C.freeResultErr(result)
	if result.err_len != 0 {
		return errors.Errorf("client handshake error: %v", C.GoStringN(result.err, result.err_len))
	}
	c.stream = result.stream
	c.handshakeComplete()
	return nil
}

func (c *Conn) handshakeComplete() {
	atomic.StoreUint32(&c.handshakeStatus, 1)
}

func (c *Conn) isHandshakeCompleted() bool {
	return atomic.LoadUint32(&c.handshakeStatus) == 1
}

func (c *Conn) serverHandshake() error {
	var cert []byte
	for i := range c.config.Certs {
		cert = append(cert, c.config.Certs[i]...)
	}
	root := c.getRoot()
	var builder strings.Builder
	for _, each := range c.config.NextProtos {
		builder.WriteString(each)
		builder.WriteString(":")
	}
	alpnProto := C.CString(strings.TrimSuffix(builder.String(), ":"))
	defer C.free(unsafe.Pointer(alpnProto))

	cfg := C.struct_config{
		conn_id:    unsafe.Pointer(c.buf),
		alpn_proto: alpnProto,
		mandatory:  C.int(c.config.ClientAuth),
		key:        unsafe.Pointer(&(c.config.Key[0])),
		key_len:    C.uint(len(c.config.Key)),
		cert:       unsafe.Pointer(&cert[0]),
		cert_len:   C.uint(len(cert)),
		root:       unsafe.Pointer(&root[0]),
		root_len:   C.uint(len(root)),
	}

	result := C.serverHandshake(cfg)
	defer C.freeResultErr(result)
	if result.err_len != 0 {
		return errors.Errorf("server handshake error: %v", C.GoStringN(result.err, result.err_len))
	}
	c.stream = result.stream
	c.handshakeComplete()
	return nil
}

func (c *Conn) getRoot() []byte {
	roots := c.config.Roots
	if c.config.GetRoots != nil {
		roots = c.config.GetRoots()
	}
	var root []byte
	for i := range roots {
		root = append(root, roots[i]...)
	}
	return root
}
