/* *********************************************************************************************************************
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 * Description: GM
 * Create     : 2022-10-11
 * ****************************************************************************************************************** */

#include <ssl.h>

int initContext(ylong_tls_ctx *ctx, struct config cfg, char* err)
{
    int res;
    int versions[] = {0x0304}; // not use double cert
    res = ylong_tls_ctx_set_versions(ctx, versions, 1);
    if (res != YLONG_TLS_SUCCESS) {
        sprintf(err, "set version error: %d", res);
        return res;
    }

    res = ylong_tls_ctx_use_sm_full_suites(ctx, 0);
    if (res != YLONG_TLS_SUCCESS) {
        sprintf(err, "set use_sm_full_suites false: %d", res);
        return res;
    }

    if (strlen(cfg.alpn_proto) != 0) {
        res = ylong_tls_ctx_set_alpn_protos(ctx, cfg.alpn_proto, strlen(cfg.alpn_proto));
        if (res != YLONG_TLS_SUCCESS) {
            sprintf(err, "set alpn protos error: %d", res);
            return res;
        }
    }

    ylong_tls_pkey *pkey = ylong_tls_pem_read_privatekey(cfg.key, cfg.key_len);
    res = ylong_tls_ctx_use_privatekey(ctx, pkey);
    ylong_tls_pkey_free(pkey);
    if (res != YLONG_TLS_SUCCESS) {
        sprintf(err, "read private key error: %d", res);
        return res;
    }

    ylong_tls_cert_chain *cert_chain = ylong_tls_pem_read_x509(cfg.cert, cfg.cert_len);
    res = ylong_tls_ctx_use_cert_chain(ctx, cert_chain);
    ylong_tls_cert_chain_free(cert_chain);
    if (res != YLONG_TLS_SUCCESS) {
        sprintf(err, "read cert pem: %d", res);
        return res;
    }

    ylong_tls_cert_chain *root = ylong_tls_pem_read_x509(cfg.root, cfg.root_len);
    res = ylong_tls_ctx_load_verify_ca(ctx, root);
    ylong_tls_cert_chain_free(root);
    if (res != YLONG_TLS_SUCCESS) {
        sprintf(err, "read root pem: %d", res);
        return res;
    }
    return YLONG_TLS_SUCCESS;
}

static int ylong_tls_bio_go_conn_destroy(ylong_tls_bio* bio) {
    void *id_ptr = ylong_tls_bio_get_data(bio);
    return DestroyCallback(id_ptr);
}

static int ylong_tls_bio_go_conn_write(ylong_tls_bio* bio, const char* in, unsigned int inl) {
    void *conn = ylong_tls_bio_get_data(bio);
    return WriteCallback(conn, (char*) in, inl);
}

static int ylong_tls_bio_go_conn_read(ylong_tls_bio* bio, char* out, unsigned int outl) {
    void *conn = ylong_tls_bio_get_data(bio);
    return ReadCallback(conn, out, outl);
}

int setBio(ylong_tls_stream *stream, void* id_ptr) {
    ylong_tls_bio_method method;
    method.write = ylong_tls_bio_go_conn_write;
    method.read = ylong_tls_bio_go_conn_read;
    method.create = NULL;
    method.destroy = NULL;
    ylong_tls_bio *bio = ylong_tls_bio_new(&method);
    if (bio == NULL) {
        return -1;
    }
    int res = ylong_tls_bio_set_data(bio, id_ptr);
    if (res != YLONG_TLS_SUCCESS) {
        return res;
    }
    return ylong_tls_set_bio(stream, bio);
}

struct result clientHandshake(struct config cfg) {
    struct result res;
    res.err = malloc(sizeof(char) * ERR_LEN);
    int code;

    ylong_tls_ctx *ctx = ylong_tls_ctx_new_client();
    if (ctx == NULL) {
        sprintf(res.err, "new client error");
        res.err_len = strlen(res.err);
        return res;
    }
    code = initContext(ctx, cfg, res.err);
    if (code != YLONG_TLS_SUCCESS) {
        res.err_len = strlen(res.err);
        return res;
    }

    ylong_tls_stream *stream = ylong_tls_stream_new(ctx);
    if (stream == NULL) {
        code = ylong_tls_get_ctx_err(ctx);
        sprintf(res.err, "new client stream error: %d", code);
        res.err_len = strlen(res.err);
        ylong_tls_ctx_free(ctx);
        return res;
    }
    ylong_tls_ctx_free(ctx);
    code = setBio(stream, cfg.conn_id);
    if (code != YLONG_TLS_SUCCESS) {
        sprintf(res.err, "set bio error: %d", code);
        res.err_len = strlen(res.err);
        ylong_tls_stream_free(stream);
        return res;
    }

    code = ylong_tls_set_hostname(stream, cfg.serverName);
    if (code != YLONG_TLS_SUCCESS) {
        sprintf(res.err, "set hostname error: %d", code);
        res.err_len = strlen(res.err);
        ylong_tls_stream_free(stream);
        return res;
    }
    code = ylong_tls_connect(stream);
    if (code != YLONG_TLS_SUCCESS) {
        char *con_err;
        int err_len;
        ylong_tls_get_stream_conn_err(stream, &con_err, &err_len);
        sprintf(res.err, "connect error [%d]: %s", code, con_err);
        res.err_len = strlen(res.err);
        ylong_tls_stream_free(stream);
        return res;
    }
    res.stream = stream;
    return res;
}

struct result serverHandshake(struct config cfg) {
    struct result res;
    res.err = malloc(sizeof(char) * ERR_LEN);
    res.err_len = 0;
    int code;

    ylong_tls_ctx *ctx = ylong_tls_ctx_new_server();
    if (ctx == NULL) {
        sprintf(res.err, "new server error");
        res.err_len = strlen(res.err);
        return res;
    }
    code = initContext(ctx, cfg, res.err);
    if (code != YLONG_TLS_SUCCESS) {
        res.err_len = strlen(res.err);
        return res;
    }
    code = ylong_tls_ctx_set_verify(ctx, cfg.mandatory);
    if (code != YLONG_TLS_SUCCESS) {
        sprintf(res.err, "set verify: %d", code);
        res.err_len = strlen(res.err);
        return res;
    }

    ylong_tls_stream *stream = ylong_tls_stream_new(ctx);
    if (stream == NULL) {
        code = ylong_tls_get_ctx_err(ctx);
        sprintf(res.err, "new server stream error: %d", code);
        res.err_len = strlen(res.err);
        ylong_tls_ctx_free(ctx);
        return res;
    }
    ylong_tls_ctx_free(ctx);
    code = setBio(stream, cfg.conn_id);
    if (code != YLONG_TLS_SUCCESS) {
        sprintf(res.err, "set bio error: %d", code);
        res.err_len = strlen(res.err);
        ylong_tls_stream_free(stream);
        return res;
    }

    code = ylong_tls_accept(stream);
    if (code != YLONG_TLS_SUCCESS) {
        char *con_err;
        int err_len;
        ylong_tls_get_stream_conn_err(stream, &con_err, &err_len);
        sprintf(res.err, "accept error [%d]: %s", code, con_err);
        res.err_len = strlen(res.err);
        ylong_tls_stream_free(stream);
        return res;
    }
    res.stream = stream;
    return res;
}

struct result getStreamError(ylong_tls_stream *stream) {
    char *con_err;
    int err_len;
    int code = ylong_tls_get_stream_conn_err(stream, &con_err, &err_len);

    struct result res;
    res.err = malloc(sizeof(char) * ERR_LEN);
    sprintf(res.err, "stream error : %s", con_err);
    res.err_len = strlen(res.err);
    return res;
}

void freeResultErr(struct result res) {
    free(res.err);
}