/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2022. All rights reserved.
 */

package wrapper

/*
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include "gmcert.h"
#include "gmkey.h"
*/
import "C"

import (
	"runtime"
	"unsafe"

	"github.com/pkg/errors"
)

// Certificate encapsulation the C X509 struct
type Certificate struct {
	x509 *C.X509
}

// NewCertificateFromPEM new certificate from pem
func NewCertificateFromPEM(pem string, pass string) (*Certificate, error) {
	cpem := C.CString(pem)
	defer C.free(unsafe.Pointer(cpem))
	cpass := C.CString(pass)
	defer C.free(unsafe.Pointer(cpass))

	bio := C.BIO_new_mem_buf(unsafe.Pointer(cpem), -1)
	if bio == nil {
		return nil, GetErrors()
	}
	defer C.BIO_free(bio)

	x509 := C.PEM_read_bio_X509(bio, nil, nil, unsafe.Pointer(cpass))
	if x509 == nil {
		return nil, GetErrors()
	}
	ret := &Certificate{x509: x509}
	runtime.SetFinalizer(ret, func(ret *Certificate) {
		C.X509_free(ret.x509)
	})

	return ret, nil
}

// GetSubject get subject info from certificate
func (cert *Certificate) GetSubject() (string, error) {
	name := C.X509_get_subject_name(cert.x509)
	if name == nil {
		return "", GetErrors()
	}
	cSubjectName := C.X509_NAME_oneline(name, nil, 0)
	if cSubjectName == nil {
		return "", GetErrors()
	}
	defer C.OPENSSL_free_func(unsafe.Pointer(cSubjectName))
	return C.GoString(cSubjectName), nil
}

// GetIssuer get issuer
func (cert *Certificate) GetIssuer() (string, error) {
	name := C.X509_get_issuer_name(cert.x509)
	if name == nil {
		return "", GetErrors()
	}
	cIssuerName := C.X509_NAME_oneline(name, nil, 0)
	if cIssuerName == nil {
		return "", GetErrors()
	}
	defer C.OPENSSL_free_func(unsafe.Pointer(cIssuerName))
	return C.GoString(cIssuerName), nil
}

// GetSerialNumber get serial number from certificate
func (cert *Certificate) GetSerialNumber() (string, error) {
	serial := C.X509_get0_serialNumber(cert.x509)
	if serial == nil {
		return "", GetErrors()
	}
	bn := C.ASN1_INTEGER_to_BN(serial, nil)
	if bn == nil {
		return "", GetErrors()
	}
	defer C.BN_free(bn)
	hex := C.BN_bn2hex(bn)
	if hex == nil {
		return "", GetErrors()
	}
	defer C.OPENSSL_free_func(unsafe.Pointer(hex))
	return C.GoString(hex), nil
}

// GetPublicKey get public key from certificate
func (cert *Certificate) GetPublicKey() (*PublicKey, error) {
	pkey := C.X509_get_pubkey(cert.x509)
	if pkey == nil {
		return nil, GetErrors()
	}
	ctx := C.GetPubKeyCtxFromPk(pkey)
	if ctx == nil {
		return nil, errors.New("the public key ctx is nil")
	}
	ret := &PublicKey{pkey: pkey, ctx: ctx}
	runtime.SetFinalizer(ret, func(ret *PublicKey) {
		C.EVP_PKEY_free(ret.pkey)
	})
	return ret, nil
}

// IsCaCert check if the cert is ca cert
func (cert *Certificate) IsCaCert() (bool, error) {
	ret := C.isCaCert(cert.x509)
	if ret == -1 {
		return false, errors.New("invalid x509 cert")
	}
	if ret == 0 {
		return false, nil
	}
	return true, nil
}
