/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2022. All rights reserved.
 */

package wrapper

/*
#include <string.h>
#include <openssl/evp.h>
#include <openssl/crypto.h>
*/
import "C"

import (
	"errors"
	"runtime"
	"unsafe"
)

// GetCipherKeyLength get cipher key length for certain cipher algorithm
func GetCipherKeyLength(name string) (int, error) {
	cname := C.CString(name)
	defer C.free(unsafe.Pointer(cname))
	cipher := C.EVP_get_cipherbyname(cname)
	if cipher == nil {
		return 0, GetErrors()
	}
	return int(C.EVP_CIPHER_key_length(cipher)), nil
}

// GetCipherBlockLength get cipher block lenght for certain cipher algorithm
func GetCipherBlockLength(name string) (int, error) {
	cname := C.CString(name)
	defer C.free(unsafe.Pointer(cname))
	cipher := C.EVP_get_cipherbyname(cname)
	if cipher == nil {
		return 0, GetErrors()
	}
	return int(C.EVP_CIPHER_block_size(cipher)), nil
}

// GetCipherIVLength get cipher IV lenght for certain cipher algorithm
func GetCipherIVLength(name string) (int, error) {
	cname := C.CString(name)
	defer C.free(unsafe.Pointer(cname))
	cipher := C.EVP_get_cipherbyname(cname)
	if cipher == nil {
		return 0, GetErrors()
	}
	return int(C.EVP_CIPHER_iv_length(cipher)), nil
}

// CipherContext cipher context
type CipherContext struct {
	ctx *C.EVP_CIPHER_CTX
}

// NewCipherContext new a cipher context struct
func NewCipherContext(name string, key, iv []byte, encrypt bool) (
	*CipherContext, error) {
	cname := C.CString(name)
	defer C.free(unsafe.Pointer(cname))
	cipher := C.EVP_get_cipherbyname(cname)
	if cipher == nil {
		return nil, GetErrors()
	}
	if len(key) == 0 {
		return nil, errors.New("no key")
	}
	if len(key) != int(C.EVP_CIPHER_key_length(cipher)) {
		return nil, errors.New("invalid key length")
	}
	if int(C.EVP_CIPHER_iv_length(cipher)) != 0 {
		if len(iv) == 0 {
			return nil, errors.New("no IV")
		}
		if len(iv) != int(C.EVP_CIPHER_iv_length(cipher)) {
			return nil, errors.New("invalid IV")
		}
	}
	ctx := C.EVP_CIPHER_CTX_new()
	if ctx == nil {
		return nil, GetErrors()
	}
	ret := &CipherContext{ctx: ctx}
	runtime.SetFinalizer(ret, func(ret *CipherContext) {
		C.EVP_CIPHER_CTX_free(ret.ctx)
	})
	enc := 1
	if encrypt == false {
		enc = 0
	}
	if C.EVP_CipherInit(ctx, cipher, (*C.uchar)(&key[0]),
		(*C.uchar)(&iv[0]), C.int(enc)) != 1 {
		return nil, GetErrors()
	}
	return ret, nil
}

// Update input plaintext and get ciphertext
func (ctx *CipherContext) Update(in []byte) ([]byte, error) {
	outbuf := make([]byte, len(in)+int(C.EVP_CIPHER_CTX_block_size(ctx.ctx)))
	outlen := C.int(len(outbuf))
	if C.EVP_CipherUpdate(ctx.ctx, (*C.uchar)(&outbuf[0]), &outlen,
		(*C.uchar)(&in[0]), C.int(len(in))) != 1 {
		return nil, GetErrors()
	}
	return outbuf[:outlen], nil
}

// Final get final ciphertext
func (ctx *CipherContext) Final() ([]byte, error) {
	outbuf := make([]byte, int(C.EVP_CIPHER_CTX_block_size(ctx.ctx)))
	outlen := C.int(len(outbuf))
	if C.EVP_CipherFinal(ctx.ctx, (*C.uchar)(&outbuf[0]), &outlen) != 1 {
		return nil, GetErrors()
	}
	return outbuf[:outlen], nil
}
