/* *********************************************************************************************************************
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 * Description: GM
 * Create     : 2022-02-21
 * ****************************************************************************************************************** */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <openssl/crypto.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <openssl/x509v3.h>
#include <securec.h>
#include <securectype.h>
#include <stdbool.h>

#include "gmcert.h"

#define MAX_DNS_NAME_LEN 256

static X509 *NewX509(const CertInfo *certInfo, const EVP_PKEY *pubKey);
static int X509NameAdd(X509 *x509, CertInfo *certInfo, EVP_PKEY *pubKey);
static void FreeSnoBtmp(ASN1_INTEGER *sno, BIGNUM *btmp);

int GetSignAlgorithm(X509 *x, char *sigAlgBuf, uint *bufLen)
{
    if ((x == NULL) || (sigAlgBuf == NULL) || (bufLen == NULL)) {
        return -1;
    }
    if (*bufLen == 0) {
        return -1;
    }

    const X509_ALGOR *sig_type = NULL;

    X509_get0_signature(NULL, &sig_type, x);

    BIO *mem = BIO_new(BIO_s_mem());
    if (mem == NULL) {
        printf("new bio failed\n");
        return -1;
    }

    if (i2a_ASN1_OBJECT(mem, sig_type->algorithm) <= 0) {
        BIO_free(mem);
        printf("Error getting the signature algorithm\n");
        return -1;
    }

    int ret = BIO_read(mem, sigAlgBuf, *bufLen);
    if (ret <= 0) {
        BIO_free(mem);
        printf("BIO_read to mem failed\n");
        return -1;
    }
    *bufLen = ret;
    BIO_free(mem);
    return 0;
}

int GetExipreTime(X509 *x, char *timeBuf, uint bufLen)
{
    if ((x == NULL) || (timeBuf == NULL) || (bufLen == 0)) {
        return 0;
    }

    ASN1_TIME *time = X509_get_notAfter(x);
    BIO *mem = BIO_new(BIO_s_mem());
    if (mem == NULL) {
        printf("new bio failed");
        return 0;
    }
    ASN1_TIME_print(mem, time);
    BUF_MEM *bptr;
    BIO_get_mem_ptr(mem, &bptr);

    errno_t rc = strcpy_s(timeBuf, bufLen, bptr->data);
    if (rc != EOK) {
        BIO_free(mem);
        printf("cp timeBuf failed %d\n", rc);
        return 0;
    }

    BIO_free(mem);
    return 1;
}

long BIO_get_mem_data_func(BIO *b, char **pp)
{
    if ((b == NULL) || (pp == NULL)) {
        return -1;
    }
    return BIO_get_mem_data(b, pp);
}

void OPENSSL_free_func(void *addr) {
    if (addr == NULL) {
        return;
    }
	OPENSSL_free(addr);
}

int GetFingerPrintf(const X509 *x, char *fingerPrintBuf, const uint bufLen)
{
    if ((x == NULL) || (fingerPrintBuf == NULL) || (bufLen == 0)) {
        return 0;
    }
    unsigned char md[EVP_MAX_MD_SIZE] = {0};
    int n = 0;
    const EVP_MD *digest;

    digest = EVP_get_digestbyname("sha256");

    if (!X509_digest(x, digest, md, &n)) {
        printf("X509_digest failed\n");
        return 0;
    }

    errno_t rc = memcpy_s(fingerPrintBuf, bufLen, md, n);
    if (rc != EOK) {
        printf("cp public key failed %d\n", rc);
        return 0;
    }

    return n;
}

void FreeMemory(X509_STORE_CTX* ctx, X509_STORE* store)
{
    if (ctx != NULL) {
        X509_STORE_CTX_free(ctx);
    }
    if (store != NULL) {
        X509_STORE_free(store);
    }
    return;
}

int VerifyCert(X509 *cert, X509 *caCert)
{
    if ((cert == NULL) || (caCert == NULL)) {
        return -1;
    }
    X509_STORE *store = NULL;
    X509_STORE_CTX *ctx = NULL;

    store = X509_STORE_new();
    if (store == NULL) {
        return -1;
    }
    if (!X509_STORE_add_cert(store, caCert)) {
        FreeMemory(ctx, store);
        return -1;
    }

    ctx = X509_STORE_CTX_new();
    if (ctx == NULL) {
        return -1;
    }
    if (!X509_STORE_CTX_init(ctx, store, cert, NULL)) {
        FreeMemory(ctx, store);
        return -1;
    }

    int ret = X509_verify_cert(ctx);
    if (ret <= 0) {
        FreeMemory(ctx, store);
        return -1;
    }
    FreeMemory(ctx, store);
    return ret;
}

static int SetExtension(X509 *issuer, X509 *cert, int nid, char *value)
{
    X509_EXTENSION *ext;
    X509V3_CTX ctx;
    unsigned long err;

    X509V3_set_ctx(&ctx, issuer, cert, NULL, NULL, 0);
    ext = X509V3_EXT_conf_nid(NULL, &ctx, nid, value);
    if (!ext) {
        err = ERR_get_error();
        printf("X509V3_EXT_conf_nid set extension failed, err: 0x%lx\n", err);
        return -1;
    }

    X509_add_ext(cert, ext, -1);
    X509_EXTENSION_free(ext);
    return 0;
}

static int X509NameAdd(X509 *x509, CertInfo *certInfo, EVP_PKEY *pubKey)
{
    (void)X509_gmtime_adj(X509_get_notBefore(x509), 0);
    (void)X509_gmtime_adj(X509_get_notAfter(x509), GM_TIME_ADJ_YEAR_ATTRIBUTE * certInfo->years);
    if (!X509_set_pubkey(x509, pubKey)) {
        return -1;
    }
    X509_NAME *name = X509_get_subject_name(x509);
    if (!X509_NAME_add_entry_by_txt(name, "C", MBSTRING_ASC, certInfo->country, -1, -1, 0)) {
        return -1;
    }
    if (!X509_NAME_add_entry_by_txt(name, "ST", MBSTRING_ASC, certInfo->province, -1, -1, 0)) {
        return -1;
    }
    if (!X509_NAME_add_entry_by_txt(name, "L", MBSTRING_ASC, certInfo->locality, -1, -1, 0)) {
        return -1;
    }
    if (!X509_NAME_add_entry_by_txt(name, "O", MBSTRING_ASC, certInfo->org, -1, -1, 0)) {
        return -1;
    }
    if (!X509_NAME_add_entry_by_txt(name, "OU", MBSTRING_ASC, certInfo->ou, -1, -1, 0)) {
        return -1;
    }
    if (!X509_NAME_add_entry_by_txt(name, "CN", MBSTRING_ASC, certInfo->commonName, -1, -1, 0)) {
        return -1;
    }
    return 0;
}

static X509 *NewX509(const CertInfo *certInfo, const EVP_PKEY *pubKey)
{
    if ((certInfo == NULL) || (pubKey == NULL)) {
        return NULL;
    }
    X509 *x509 = X509_new();
    if (x509 == NULL) {
        printf("X509_new failed, err: 0x%lx\n", ERR_get_error());
        return NULL;
    }
    return x509;
}

static void FreeSnoBtmp(ASN1_INTEGER *sno, BIGNUM *btmp)
{
    if (sno != NULL) {
        ASN1_INTEGER_free(sno);
    }
    if (btmp != NULL) {
        BN_free(btmp);
    }
};

X509 *GenerateCertBasic(CertInfo *certInfo, EVP_PKEY *pubKey)
{
    unsigned long err;
    int ret = 0;
    X509 *x509 = NewX509(certInfo, pubKey);
    if (x509 == NULL) {
        return NULL;
    }
    int svRet = X509_set_version(x509, 2);
    if (svRet != 1) {
        err = ERR_get_error();
        ret = -1;
        goto END;
    }
    ASN1_INTEGER *sno = ASN1_INTEGER_new();
    if (!sno) {
        ret = -1;
        goto END;
    }
    BIGNUM *btmp = BN_new();
    if (!btmp) {
        ret = -1;
        goto END;
    }
    if (BN_rand(btmp, BN_RAND_BITS_LENGTH, 0, 0) == 0) {
        ret = -1;
        goto END;
    }
    if (!BN_to_ASN1_INTEGER(btmp, sno)) {
        ret = -1;
        goto END;
    }
    if (X509_set_serialNumber(x509, sno) == 0) {
        ret = -1;
        goto END;
    }
    if (X509NameAdd(x509, certInfo, pubKey) == -1) {
        ret = -1;
        goto END;
    }
END:
    FreeSnoBtmp(sno, btmp);
    if (ret != 0) {
        if (x509 != NULL) {
            X509_free(x509);
            x509 = NULL;
        }
    }
    return x509;
}

X509 *GenerateMiddleCACert(CertInfo *certInfo, EVP_PKEY *pubKey, EVP_PKEY *caPriKey, X509 *caCert)
{
    int ret = 0;
    if ((caPriKey == NULL) || (caCert == NULL)) {
        return NULL;
    }
    X509 *x509 = GenerateCertBasic(certInfo, pubKey);
    if (x509 == NULL) {
        return NULL;
    }

    X509_NAME *caName;
    caName = X509_get_subject_name(caCert);
    X509_set_issuer_name(x509, caName);

    // todo judge return value err process
    ret = SetExtension(caCert, x509, NID_ext_key_usage, "clientAuth, serverAuth");
    if (ret != 0) {
        goto END;
    }
    ret = SetExtension(caCert, x509, NID_subject_key_identifier, "hash");
    if (ret != 0) {
        goto END;
    }
    ret = SetExtension(caCert, x509, NID_authority_key_identifier, "keyid:always");
    if (ret != 0) {
        goto END;
    }
    ret = SetExtension(caCert, x509, NID_key_usage, "digitalSignature, keyEncipherment, keyCertSign, cRLSign");
    if (ret != 0) {
        goto END;
    }
    ret = SetExtension(caCert, x509, NID_basic_constraints, "CA:true");
    if (ret != 0) {
        goto END;
    }

    char *dnsInfo = (char*)malloc(MAX_DNS_NAME_LEN);
    if (dnsInfo == NULL){
        goto END;
    }
    memset_s(dnsInfo, MAX_DNS_NAME_LEN, 0, MAX_DNS_NAME_LEN);
    int usedLen = sprintf_s(dnsInfo, 256, "DNS:%s", certInfo->commonName);
    if (usedLen <= 0) {
        printf("the sprintf_s failed\n");
        goto END;
    }

    ret = SetExtension(x509, x509, NID_subject_alt_name, dnsInfo);
    if (ret != 0) {
        goto END;
    }

    X509_sign(x509, caPriKey, EVP_sm3());
END:
    if (dnsInfo != NULL) {
        free(dnsInfo);
        dnsInfo = NULL;
    }
    if (ret != 0) {
        if (x509 != NULL) {
            X509_free(x509);
            x509 = NULL;
        }
    }
    return x509;
}

// 传入私钥（内含公钥）
X509 *GenerateCert(CertInfo *certInfo, EVP_PKEY *pubKey, EVP_PKEY *caPriKey, X509 *caCert)
{
    int ret = 0;
    if ((certInfo == NULL) || (pubKey == NULL) || (caPriKey == NULL) || (caCert == NULL)) {
        return NULL;
    }
    X509 *x509 = GenerateCertBasic(certInfo, pubKey);
    if (x509 == NULL) {
        return NULL;
    }

    X509_NAME *caName;
    caName = X509_get_subject_name(caCert);
    if (!X509_set_issuer_name(x509, caName)) {
        ret = -1;
        goto END;
    }

    // todo judge return value err process
    ret = SetExtension(caCert, x509, NID_authority_key_identifier, "keyid:always");
    if (ret != 0) {
        goto END;
    }
    ret = SetExtension(caCert, x509, NID_key_usage, "digitalSignature");
    if (ret != 0) {
        goto END;
    }
    ret = SetExtension(caCert, x509, NID_basic_constraints, "CA:false");
    if (ret != 0) {
        goto END;
    }

    char *dnsInfo = (char*)malloc(MAX_DNS_NAME_LEN);
    if (dnsInfo == NULL){
        goto END;
    }
    memset_s(dnsInfo, MAX_DNS_NAME_LEN, 0, MAX_DNS_NAME_LEN);
    int usedLen = sprintf_s(dnsInfo, 256, "DNS:%s", certInfo->commonName);
    if (usedLen <= 0) {
        printf("the sprintf_s failed\n");
        goto END;
    }

    ret = SetExtension(x509, x509, NID_subject_alt_name, dnsInfo);
    if (ret != 0) {
        goto END;
    }

    X509_sign(x509, caPriKey, EVP_sm3());
END:
    if (dnsInfo != NULL) {
        free(dnsInfo);
        dnsInfo = NULL;
    }
    if (ret != 0) {
        if (x509 != NULL) {
            X509_free(x509);
            x509 = NULL;
        }
    }

    return x509;
}
// 传入私钥（内含公钥）
X509 *GenerateSelfSignCert(CertInfo *certInfo, EVP_PKEY *pkey)
{
    int ret = 0;
    if ((certInfo == NULL) || (pkey == NULL)) {
        return NULL;
    }
    X509 *x509 = GenerateCertBasic(certInfo, pkey);
    if (x509 == NULL) {
        return NULL;
    }
    X509_NAME *name;
    name = X509_get_subject_name(x509);
    X509_set_issuer_name(x509, name);

    ret = SetExtension(x509, x509, NID_ext_key_usage, "clientAuth, serverAuth");
    if (ret != 0) {
        goto END;
    }
    // todo hash is the default
    ret = SetExtension(x509, x509, NID_subject_key_identifier, "hash");
    if (ret != 0) {
        goto END;
    }
    ret = SetExtension(x509, x509, NID_authority_key_identifier, "keyid:always");
    if (ret != 0) {
        goto END;
    }
    ret = SetExtension(x509, x509, NID_key_usage, "digitalSignature, keyEncipherment, keyCertSign, cRLSign");
    if (ret != 0) {
        goto END;
    }
    ret = SetExtension(x509, x509, NID_basic_constraints, "CA:true");
    if (ret != 0) {
        goto END;
    }

    char *dnsInfo = (char*)malloc(MAX_DNS_NAME_LEN);
    if (dnsInfo == NULL){
        goto END;
    }
    memset_s(dnsInfo, MAX_DNS_NAME_LEN, 0, MAX_DNS_NAME_LEN);
    int usedLen = sprintf_s(dnsInfo, 256, "DNS:%s", certInfo->commonName);
    if (usedLen <= 0) {
        printf("the sprintf_s failed\n");
        goto END;
    }

    ret = SetExtension(x509, x509, NID_subject_alt_name, dnsInfo);
    if (ret != 0) {
        goto END;
    }


    // for EVP_sm3(), the signature algorithm is : 1.2.156.10197.1.501
    // for EVP_sha256(), the signature algorithm is :sha256ECDSA
    X509_sign(x509, pkey, EVP_sm3());

END:
    if (dnsInfo != NULL) {
        free(dnsInfo);
        dnsInfo = NULL;
    }
    if (ret != 0) {
        if (x509 != NULL) {
            X509_free(x509);
            x509 = NULL;
        }
    }
    return x509;
}


void X509ToPEM(X509 *cert, char *pemBuf, uint32_t *pemBufLen)
{
    BIO *bio = NULL;

    if ((cert == NULL) || (pemBuf == NULL) || (pemBufLen == NULL)) {
        return;
    }
    if (*pemBufLen == 0) {
        return;
    }

    bio = BIO_new(BIO_s_mem());
    if (bio == NULL) {
        return;
    }

    if (PEM_write_bio_X509(bio, cert) == 0) {
        BIO_set_close(bio, BIO_CLOSE);
        BIO_free(bio);
        return;
    }

    BUF_MEM *mem = NULL;
    BIO_get_mem_ptr(bio, &mem);

    errno_t rc = memcpy_s(pemBuf, *pemBufLen, mem->data, mem->length);
    if (rc != EOK) {
        BIO_set_close(bio, BIO_CLOSE);
        BIO_free(bio);
        printf("cp pem x509 failed %d\n", rc);
        return;
    }

    *pemBufLen = mem->length;
    BIO_set_close(bio, BIO_CLOSE);
    BIO_free(bio);
    return;
}

CertInfo *NewCertInfo(void)
{
    CertInfo *certInfo = (CertInfo *)malloc(sizeof(CertInfo));
    if (certInfo == NULL) {
        return NULL;
    }
    certInfo->country = (char *)malloc(sizeof(char) * CERT_INFO_LENGTH);
    if (certInfo->country == NULL) {
        CertInfoFree(certInfo);
        return NULL;
    }
    certInfo->province = (char *)malloc(sizeof(char) * CERT_INFO_LENGTH);
    if (certInfo->province == NULL) {
        CertInfoFree(certInfo);
        return NULL;
    }
    certInfo->locality = (char *)malloc(sizeof(char) * CERT_INFO_LENGTH);
    if (certInfo->locality == NULL) {
        CertInfoFree(certInfo);
        return NULL;
    }
    certInfo->org = (char *)malloc(sizeof(char) * CERT_INFO_LENGTH);
    if (certInfo->org == NULL) {
        CertInfoFree(certInfo);
        return NULL;
    }
    certInfo->ou = (char *)malloc(sizeof(char) * CERT_INFO_LENGTH);
    if (certInfo->ou == NULL) {
        CertInfoFree(certInfo);
        return NULL;
    }
    certInfo->commonName = (char *)malloc(sizeof(char) * CERT_INFO_LENGTH);
    if (certInfo->commonName == NULL) {
        CertInfoFree(certInfo);
        return NULL;
    }
    return certInfo;
}


_Bool SetCountryProvinceLocality(CertInfo *certInfo, const char *country, const char *province, const char *locality)
{
    if ((certInfo == NULL) || (country == NULL) || (province == NULL) || (locality == NULL)) {
        return false;
    }
    errno_t rc = memset_s(certInfo->country, CERT_INFO_LENGTH, 0, CERT_INFO_LENGTH);
    if (rc != EOK) {
        return false;
    }
    rc = memset_s(certInfo->province, CERT_INFO_LENGTH, 0, CERT_INFO_LENGTH);
    if (rc != EOK) {
        return false;
    }
    rc = memset_s(certInfo->locality, CERT_INFO_LENGTH, 0, CERT_INFO_LENGTH);
    if (rc != EOK) {
        return false;
    }
    rc = memcpy_s(certInfo->country, CERT_INFO_LENGTH, country, strlen(country));
    if (rc != EOK) {
        return false;
    }
    rc = memcpy_s(certInfo->province, CERT_INFO_LENGTH, province, strlen(province));
    if (rc != EOK) {
        return false;
    }
    rc = memcpy_s(certInfo->locality, CERT_INFO_LENGTH, locality, strlen(locality));
    if (rc != EOK) {
        return false;
    }
    return true;
}

_Bool SetOrgOuCommonNameYears(CertInfo *certInfo, const char *org, const char *ou, const char *commonName,
    const uint32_t years)
{
    if ((certInfo == NULL) || (org == NULL) || (ou == NULL) || (commonName == NULL)) {
        return false;
    }

    errno_t rc = memset_s(certInfo->org, CERT_INFO_LENGTH, 0, CERT_INFO_LENGTH);
    if (rc != EOK) {
        return false;
    }
    rc = memset_s(certInfo->ou, CERT_INFO_LENGTH, 0, CERT_INFO_LENGTH);
    if (rc != EOK) {
        return false;
    }
    rc = memset_s(certInfo->commonName, CERT_INFO_LENGTH, 0, CERT_INFO_LENGTH);
    if (rc != EOK) {
        return false;
    }
    rc = memcpy_s(certInfo->org, CERT_INFO_LENGTH, org, strlen(org));
    if (rc != EOK) {
        return false;
    }
    rc = memcpy_s(certInfo->ou, CERT_INFO_LENGTH, ou, strlen(ou));
    if (rc != EOK) {
        return false;
    }
    rc = memcpy_s(certInfo->commonName, CERT_INFO_LENGTH, commonName, strlen(commonName));
    if (rc != EOK) {
        return false;
    }
    certInfo->years = years;
    return true;
}

void CertInfoFree(CertInfo *certInfo)
{
    if (certInfo == NULL) {
        return;
    }
    if (certInfo->country != NULL) {
        free(certInfo->country);
    }
    if (certInfo->province != NULL) {
        free(certInfo->province);
    }
    if (certInfo->locality != NULL) {
        free(certInfo->locality);
    }
    if (certInfo->org != NULL) {
        free(certInfo->org);
    }
    if (certInfo->ou != NULL) {
        free(certInfo->ou);
    }
    if (certInfo->commonName != NULL) {
        free(certInfo->commonName);
    }

    free(certInfo);
}

int isCaCert(X509* certX509)
{
    if (certX509 == NULL) {
        return -1;
    }
    BASIC_CONSTRAINTS *constraint = NULL;
    constraint = (BASIC_CONSTRAINTS*)X509_get_ext_d2i(certX509, NID_basic_constraints, NULL, NULL);
    if (constraint == NULL){
        return -1;
    }
    if (!constraint->ca) {
        BASIC_CONSTRAINTS_free(constraint);
        return 0;
    }
    BASIC_CONSTRAINTS_free(constraint);
    return 1;
}
