/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2022. All rights reserved.
 */

// Package wrapper encapsulating go interface for the gmssl C library
package wrapper

/*
#cgo CFLAGS: -I/usr/local/include/openssl -I/usr/local/include/securec
#cgo LDFLAGS: -L/usr/local/include/openssl  -lgmcrypto -L/usr/local/include/securec -lsecurec
#include <stdlib.h>
#include "gmcert.h"
*/
import "C"
import (
	"fmt"
	"time"
	"unsafe"

	"github.com/pkg/errors"
)

// GmCertInfo gm certificate info struct
type GmCertInfo struct {
	Country         string
	Province        string
	Locality        string
	Org             string
	OU              string
	CN              string
	ValidationYears uint
}

const maxPemCertSize = 8192

var monthMap = map[string]time.Month{
	"Jan": 1,
	"Feb": 2,
	"Mar": 3,
	"Apr": 4,
	"May": 5,
	"Jun": 6,
	"Jul": 7,
	"Aug": 8,
	"Sep": 9,
	"Oct": 10,
	"Nov": 11,
	"Dec": 12}

// GetSignAlgorithm get sign algorithm of certificate
func (cert *Certificate) GetSignAlgorithm() string {
	const maxSigAlgLen = 128
	var sigAlgBuf [maxSigAlgLen]byte
	var bufLen C.uint = maxSigAlgLen

	ret := C.GetSignAlgorithm(cert.x509, (*C.char)(unsafe.Pointer(&sigAlgBuf)), (*C.uint)(&bufLen))
	if ret < 0 {
		return ""
	}
	return string(sigAlgBuf[:bufLen])
}

// GetExpireTime get the certificate expire time
func (cert *Certificate) GetExpireTime() time.Time {
	var bufLen C.uint = 128
	var timeBuf [128]byte
	res := C.GetExipreTime(cert.x509, (*C.char)(unsafe.Pointer(&timeBuf)), bufLen)
	if res == 0 {
		return time.Now()
	}

	var month string
	var day int
	var hour int
	var min int
	var sec int
	var year int
	var timeZone string

	_, err := fmt.Sscanf(string(timeBuf[:]), "%s %d %d:%d:%d %d %s", &month, &day, &hour, &min, &sec, &year, &timeZone)
	if err != nil {
		return time.Now()
	}
	mon, ok := monthMap[month]
	if !ok {
		return time.Now()
	}

	location := time.FixedZone(timeZone, 0)
	return time.Date(year, mon, day, hour, min, sec, sec, location)
}

// GetFingerPrint get finger print of certificate
func (cert *Certificate) GetFingerPrint() string {
	var fingerPrintBuf [int(C.EVP_MAX_MD_SIZE)]byte
	var bufLen C.uint = C.EVP_MAX_MD_SIZE
	var fingerPrint string

	bufLenRet := C.GetFingerPrintf(cert.x509, (*C.char)(unsafe.Pointer(&fingerPrintBuf)), C.uint(bufLen))

	length := int(bufLenRet)
	if length == 0 {
		return ""
	}
	if length >= len(fingerPrintBuf) {
		return ""
	}
	for i := 0; i <= length-1; i++ {
		item := fmt.Sprintf("%02x", fingerPrintBuf[i])
		fingerPrint = fingerPrint + item
	}

	return fingerPrint
}

// VerifyCert verify cert with CA cert
func (cert *Certificate) VerifyCert(caCert *Certificate) bool {
	var res C.int
	res = C.VerifyCert(cert.x509, caCert.x509)

	if res == 1 {
		return true
	}

	return false
}

func freeCString(cStrs ...*C.char) {
	for _, cStr := range cStrs {
		C.free(unsafe.Pointer(cStr))
	}
}

func newCertInfo(gmCertInfo *GmCertInfo) (*C.CertInfo, error) {
	cCountry := C.CString(gmCertInfo.Country)
	cProvince := C.CString(gmCertInfo.Province)
	cLocality := C.CString(gmCertInfo.Locality)
	cOrg := C.CString(gmCertInfo.Org)
	cOU := C.CString(gmCertInfo.OU)
	cCommonName := C.CString(gmCertInfo.CN)
	defer freeCString(cCountry, cProvince, cLocality, cOrg, cOU, cCommonName)
	certInfo := C.NewCertInfo()
	ret := C.SetCountryProvinceLocality(certInfo, cCountry, cProvince, cLocality)
	if !ret {
		return nil, errors.New("creat cert info failed")
	}
	ret = C.SetOrgOuCommonNameYears(certInfo, cOrg, cOU, cCommonName, C.uint(gmCertInfo.ValidationYears))
	if !ret {
		return nil, errors.New("creat cert info failed")
	}
	return certInfo, nil
}

// GenerateMiddleCACert ge middle ca cert
func GenerateMiddleCACert(gmCertInfo *GmCertInfo, pubKey *PublicKey, caPriKey *PrivateKey,
	caCert *Certificate) (*Certificate, error) {
	certInfo, err := newCertInfo(gmCertInfo)

	defer C.CertInfoFree(certInfo)
	if err != nil {
		return nil, err
	}

	x509 := C.GenerateMiddleCACert(certInfo, pubKey.pkey, caPriKey.pkey, caCert.x509)
	if x509 == nil {
		return nil, errors.New("create middle ca cert failed")
	}

	cert := &Certificate{x509: x509}

	return cert, nil
}

// GenerateCert generate self sign certificate
func GenerateCert(gmCertInfo *GmCertInfo, pubKey *PublicKey, caPriKey *PrivateKey,
	caCert *Certificate) (*Certificate, error) {
	certInfo, err := newCertInfo(gmCertInfo)
	defer C.CertInfoFree(certInfo)
	if err != nil {
		return nil, err
	}
	x509 := C.GenerateCert(certInfo, pubKey.pkey, caPriKey.pkey, caCert.x509)
	if x509 == nil {
		return nil, errors.New("create cert failed")
	}

	cert := &Certificate{x509: x509}
	return cert, nil
}

// GenerateSelfSignCert generate self sign certificate
func GenerateSelfSignCert(gmCertInfo *GmCertInfo, key *PrivateKey) (*Certificate, error) {
	certInfo, err := newCertInfo(gmCertInfo)
	defer C.CertInfoFree(certInfo)
	if err != nil {
		return nil, err
	}

	sm2Key := key
	x509 := C.GenerateSelfSignCert(certInfo, sm2Key.pkey)
	if x509 == nil {
		return nil, errors.New("create self sign cert failed")
	}

	cert := &Certificate{x509: x509}

	return cert, nil
}

// GetPemFromCert get pem from cert
func GetPemFromCert(certificate *Certificate) []byte {
	var pemBuf [maxPemCertSize]byte
	var pemBufLen C.uint = maxPemCertSize
	C.X509ToPEM(certificate.x509, (*C.char)(unsafe.Pointer(&pemBuf)), (*C.uint)(&pemBufLen))

	pemCert := pemBuf[:pemBufLen]
	return pemCert
}

// GetPemWithoutEnc get pem private key without encryption
func (sk *PrivateKey) GetPemWithoutEnc() (string, error) {
	bio := C.BIO_new(C.BIO_s_mem())
	if bio == nil {
		return "", GetErrors()
	}
	defer C.BIO_free(bio)

	if C.PEM_write_bio_PKCS8PrivateKey(bio, sk.pkey, nil,
		nil, C.int(0), nil, nil) != 1 {
		return "", GetErrors()
	}

	var p *C.char
	priKeyLen := C.BIO_get_mem_data_func(bio, &p)
	if priKeyLen <= 0 {
		return "", GetErrors()
	}

	return C.GoString(p)[:priKeyLen], nil
}
