/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2022. All rights reserved.
 */

package wrapper

/*
#include <stdlib.h>
#include <openssl/bio.h>
#include "gmcert.h"
#include "gmkey.h"
*/
import "C"

import (
	"errors"
	"reflect"
	"runtime"
	"unsafe"
)

const sm2DigestLen = 32

// PublicKey encapsulate public key as type EVP_PKEY
type PublicKey struct {
	pkey *C.EVP_PKEY
	ctx  *C.EVP_PKEY_CTX
}

// PrivateKey encapsulate private key as type EVP_PKEY
type PrivateKey struct {
	pkey *C.EVP_PKEY
}

// GeneratePrivateKey Generate private key by algorith and args
func GeneratePrivateKey(alg string, args [][2]string) (*PrivateKey, error) {
	calg := C.CString(alg)
	defer C.free(unsafe.Pointer(calg))

	ctx := C.NewPkeyKeygenCtx(calg, nil)
	defer C.EVP_PKEY_CTX_free(ctx)

	if ctx == nil {
		return nil, GetErrors()
	}
	var pkey *C.EVP_PKEY

	if alg == "DH" || alg == "DSA" {

		if 1 != C.EVP_PKEY_paramgen_init(ctx) {
			return nil, GetErrors()
		}
		err := procKeyGenerationArgs(args, ctx)
		if err != nil {
			return nil, err
		}
		if 1 != C.EVP_PKEY_paramgen(ctx, &pkey) {
			return nil, GetErrors()
		}
		if 1 != C.EVP_PKEY_keygen_init(ctx) {
			return nil, GetErrors()
		}
		if 1 != C.EVP_PKEY_keygen(ctx, &pkey) {
			return nil, GetErrors()
		}

	} else {
		if 1 != C.EVP_PKEY_keygen_init(ctx) {
			return nil, GetErrors()
		}

		err := procKeyGenerationArgs(args, ctx)
		if err != nil {
			return nil, err
		}

		if 1 != C.EVP_PKEY_keygen(ctx, &pkey) {
			return nil, GetErrors()
		}
	}

	sk := &PrivateKey{pkey: pkey}
	runtime.SetFinalizer(sk, func(sk *PrivateKey) {
		C.EVP_PKEY_free(sk.pkey)
	})

	return sk, nil
}

func procKeyGenerationArgs(args [][2]string, ctx *C.EVP_PKEY_CTX) error {
	for _, arg := range args {
		name := arg[0]
		value := arg[1]
		cName := C.CString(name)
		defer C.free(unsafe.Pointer(cName))
		cValue := C.CString(value)
		defer C.free(unsafe.Pointer(cValue))
		if C.EVP_PKEY_CTX_ctrl_str(ctx, cName, cValue) <= 0 {
			return GetErrors()
		}
	}
	return nil
}

// NewPrivateKeyFromPEM new private key from pem
func NewPrivateKeyFromPEM(pem string, pass string) (*PrivateKey, error) {
	cpem := C.CString(pem)
	defer C.free(unsafe.Pointer(cpem))
	cpass := C.CString(pass)
	defer C.free(unsafe.Pointer(cpass))
	bio := C.BIO_new_mem_buf(unsafe.Pointer(cpem), -1)
	if bio == nil {
		return nil, GetErrors()
	}
	defer C.BIO_free(bio)
	pkey := C.PEM_read_bio_PrivateKey(bio, nil, nil, unsafe.Pointer(cpass))
	if pkey == nil {
		return nil, GetErrors()
	}
	sk := &PrivateKey{pkey: pkey}
	runtime.SetFinalizer(sk, func(sk *PrivateKey) {
		C.EVP_PKEY_free(sk.pkey)
	})
	return sk, nil
}

// GetPublicKeyPEM get public key pem format value
func (sk *PrivateKey) GetPublicKeyPEM() (string, error) {
	bio := C.BIO_new(C.BIO_s_mem())
	if bio == nil {
		return "", GetErrors()
	}
	defer C.BIO_free(bio)
	if C.PemWriteBioPubKey(bio, sk.pkey) != 1 {
		return "", GetErrors()
	}
	var p *C.char
	pubKeyLen := C.BIO_get_mem_data_func(bio, &p)
	if pubKeyLen <= 0 {
		return "", GetErrors()
	}
	return C.GoString(p)[:pubKeyLen], nil
}

// NewPublicKeyFromPEM new public key struct from pem format
func NewPublicKeyFromPEM(pem string) (*PublicKey, error) {
	cpem := C.CString(pem)
	defer C.free(unsafe.Pointer(cpem))
	bio := C.BIO_new_mem_buf(unsafe.Pointer(cpem), -1)
	if bio == nil {
		return nil, GetErrors()
	}
	defer C.BIO_free(bio)
	pkey := C.PemReadBioPubKey(bio)
	if pkey == nil {
		return nil, GetErrors()
	}
	ctx := C.GetPubKeyCtxFromPk(pkey)
	if ctx == nil {
		return nil, errors.New("the public key ctx is nil")
	}
	pk := &PublicKey{pkey: pkey, ctx: ctx}
	runtime.SetFinalizer(pk, func(pk *PublicKey) {
		C.EVP_PKEY_free(pk.pkey)
	})
	return pk, nil
}

// GetDer get der format public key
func (pk *PublicKey) GetDer() ([]byte, error) {
	var der *C.uchar
	var length C.int
	defer func() {
		if der != nil {
			C.free(unsafe.Pointer(der))
		}
	}()
	length = C.i2d_PublicKey(pk.pkey, &der)
	if length < 1 {
		return nil, errors.New("get der failed for public key")
	}
	goDer := C.GoBytes(unsafe.Pointer(der), length)
	return goDer, nil
}

// GetPEM get pem format value of public key
func (pk *PublicKey) GetPEM() (string, error) {
	bio := C.BIO_new(C.BIO_s_mem())
	if bio == nil {
		return "", GetErrors()
	}
	defer C.BIO_free(bio)
	if C.PemWriteBioPubKey(bio, pk.pkey) != 1 {
		return "", GetErrors()
	}
	var p *C.char
	pubKeyLen := C.BIO_get_mem_data_func(bio, &p)
	if pubKeyLen <= 0 {
		return "", GetErrors()
	}
	return C.GoString(p)[:pubKeyLen], nil
}

// Sign sign with private key
func (sk *PrivateKey) Sign(digest []byte) ([]byte, error) {
	var sigLen C.size_t
	digestSliceHeader := (*reflect.SliceHeader)(unsafe.Pointer(&digest))
	sig := C.SkSign(sk.pkey, (*C.uchar)(unsafe.Pointer(digestSliceHeader.Data)),
		C.size_t(len(digest)), &sigLen, nil)
	if sig == nil {
		C.ERR_print_errors_fp(C.stderr)
		return nil, GetErrors()
	}
	defer C.OPENSSL_free_func(unsafe.Pointer(sig))
	return C.GoBytes(unsafe.Pointer(sig), C.int(sigLen)), nil
}

// Verify verify with public key
func (pk *PublicKey) Verify(digest, sig []byte) error {
	if C.PkVerify(pk.pkey, pk.ctx, (*C.uchar)(&digest[0]), C.size_t(len(digest)),
		(*C.uchar)(&sig[0]), C.size_t(len(sig)), nil) != 1 {
		C.ERR_print_errors_fp(C.stderr)
		return GetErrors()
	}
	return nil
}

// ComputeSM2IDDigest compute id digest for SM2 by public key
func (pk *PublicKey) ComputeSM2IDDigest(id string) ([]byte, error) {
	if C.EVP_PKEY_EC != C.EVP_PKEY_id(pk.pkey) {
		return nil, errors.New("invalid public key type")
	}
	outBuf := make([]byte, sm2DigestLen)
	outLen := C.size_t(len(outBuf))
	if C.ComputeIDDigest(pk.pkey, (*C.uchar)(&outBuf[0]), &outLen) != 1 {
		return nil, GetErrors()
	}
	return outBuf[:sm2DigestLen], nil
}

// ComputeSM2IDDigest compute id digest for SM2 by private key
func (sk *PrivateKey) ComputeSM2IDDigest(id string) ([]byte, error) {
	if C.EVP_PKEY_EC != C.EVP_PKEY_id(sk.pkey) {
		return nil, errors.New("invalid private key type")
	}
	outBuf := make([]byte, sm2DigestLen)
	outLen := C.size_t(len(outBuf))
	if C.ComputeIDDigest(sk.pkey, (*C.uchar)(&outBuf[0]), &outLen) != 1 {
		return nil, GetErrors()
	}
	return outBuf[:sm2DigestLen], nil
}

// GetDer get der format private key
func (sk *PrivateKey) GetDer() ([]byte, error) {
	var der *C.uchar
	var length C.int
	defer func() {
		if der != nil {
			C.free(unsafe.Pointer(der))
		}
	}()
	length = C.i2d_PrivateKey(sk.pkey, &der)
	if length < 1 {
		return nil, errors.New("get der failed for public key")
	}
	goDer := C.GoBytes(unsafe.Pointer(der), length)
	return goDer, nil
}
