/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2020. All rights reserved.
 */

package action

import (
	"context"
	"sync"
	"time"

	"github.com/pkg/errors"
	"google.golang.org/grpc/grpclog"

	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/utils"
)

const maxStreamConnTryTimes = 3

// ContractAction is the action for contract operations.
type ContractAction struct {
	action
	contractClient    nodeservice.ContractClient
	transactionClient nodeservice.TransactionSenderClient
	clientLck         sync.Mutex
}

// NewContractAction is used to create contract action instance with action config.
func NewContractAction(config *Config) *ContractAction {
	action := &ContractAction{action: action{config: config}}
	return action
}

// Invoke is used to send invoke request raw message to contract by grpc.
func (action *ContractAction) Invoke(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	if rawMsg == nil {
		return nil, errors.New("raw message is nil")
	}
	client, err := action.getContractClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get contract client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.Invoke(ctx, rawMsg)
}

// QueryStream is used to send QueryStream request raw message to contract by grpc.
func (action *ContractAction) QueryStream(rawMsg *common.RawMessage) (*StreamResultIterator, error) {
	if rawMsg == nil {
		return nil, errors.New("raw message is nil")
	}
	client, err := action.getContractClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get contract client error")
	}
	for i := 1; i <= maxStreamConnTryTimes; i++ {
		streamResultIterator, err := createStream(client, rawMsg)
		if err != nil {
			grpclog.Errorf("create stream err:%s, the try time is %d", err.Error(), i)
			time.Sleep(time.Second * time.Duration(i))
			continue
		}
		return streamResultIterator, nil
	}
	return nil, errors.New("connect error")
}

// StreamResultIterator is the definition of block result iterator.
type StreamResultIterator struct {
	client nodeservice.Contract_QueryStreamClient
	ctx    context.Context
	cancel context.CancelFunc
}

// Next is used to get the next element of iterator.
func (itr *StreamResultIterator) Next() ([]byte, error) {
	streamMsg, err := itr.client.Recv()
	if err != nil {
		return nil, errors.WithMessage(err, "recv data result error")
	}
	sm := common.StreamMessage{}
	err = proto.Unmarshal(streamMsg.Payload, &sm)
	if err != nil {
		return nil, err
	}
	if sm.Type == common.STREAM_ERROR {
		return nil, errors.New("parse result error: " + string(sm.Payload))
	}
	select {
	case <-itr.ctx.Done():
		return nil, errors.New("grpc stream has closed")
	default:
		return sm.Payload, nil
	}
}

// Cancel is used to cancel this data chan.
func (itr *StreamResultIterator) Cancel() {
	err := itr.client.CloseSend()
	if err != nil {
		itr.cancel()
		return
	}
}

// Query is used to send query request raw message to contract by grpc.
func (action *ContractAction) Query(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	if rawMsg == nil {
		return nil, errors.New("raw message is nil")
	}
	client, err := action.getContractClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get contract client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.Query(ctx, rawMsg)
}

// ContractImport is used to send import contract request raw message by grpc.
func (action *ContractAction) ContractImport(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	client, err := action.getContractClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get contract client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.Import(ctx, rawMsg)
}

// ContractUnImport is used to send unimport contract request raw message by grpc.
func (action *ContractAction) ContractUnImport(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	client, err := action.getContractClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get contract client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.UnImport(ctx, rawMsg)
}

// Transaction is used to send transaction request raw message by grpc.
func (action *ContractAction) Transaction(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	if rawMsg == nil {
		return nil, errors.New("raw message is nil")
	}
	client, err := action.getTransactionClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get transaction client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.SendTransaction(ctx, rawMsg)
}

func (action *ContractAction) getContractClient() (nodeservice.ContractClient, error) {
	if err := action.resetConn(); err != nil {
		return nil, errors.WithMessage(err, "reset clients error")
	}
	if action.contractClient == nil {
		action.newClients()
	}
	return action.contractClient, nil
}

func (action *ContractAction) getTransactionClient() (nodeservice.TransactionSenderClient, error) {
	if err := action.resetConn(); err != nil {
		return nil, errors.WithMessage(err, "reset clients error")
	}
	if action.transactionClient == nil {
		action.newClients()
	}
	return action.transactionClient, nil
}

func (action *ContractAction) newClients() {
	action.clientLck.Lock()
	defer action.clientLck.Unlock()
	if action.contractClient == nil || action.transactionClient == nil {
		action.contractClient = nodeservice.NewContractClient(action.conn)
		action.transactionClient = nodeservice.NewTransactionSenderClient(action.conn)
	}
}

// QueryState is used to send import contract request raw message by grpc.
func (action *ContractAction) QueryState(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	client, err := action.getContractClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get contract client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.QueryState(ctx, rawMsg)
}

func createStream(client nodeservice.ContractClient, rawMsg *common.RawMessage) (*StreamResultIterator, error) {
	stream, err := client.QueryStream(context.Background())
	if err != nil {
		return nil, errors.WithMessage(err, "QueryStream error")
	}
	ctx, cancel := context.WithCancel(context.Background())
	streamResultIterator := &StreamResultIterator{
		client: stream,
		ctx:    ctx,
		cancel: cancel,
	}
	err = stream.Send(rawMsg)
	if err != nil {
		cancel()
		grpclog.Errorf("Stream send err:%s", err.Error())
	}
	return streamResultIterator, nil
}
