/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package event

import (
	"context"
	"encoding/hex"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/rawmessage"
)

type blockListener struct {
	chainID     string
	msgBuilder  rawmessage.MsgBuilder
	ch          chan msg
	client      nodeservice.EventServiceClient
	eventClient nodeservice.EventService_RegisterResultEventClient
	ctx         context.Context
	cancel      context.CancelFunc
}

func newBlockListener(params *listenerParam) *blockListener {
	return &blockListener{
		chainID:    params.chainID,
		msgBuilder: params.msgBuilder,
		ch:         params.ch,
		client:     params.client,
		ctx:        params.ctx,
		cancel:     params.cancel,
	}
}

func (l *blockListener) init() error {
	if err := l.registerEvent(); err != nil {
		return errors.WithMessage(err, "register result event error")
	}
	go l.recv()
	return nil
}

func (l *blockListener) reInit() error {
	if err := reInit(l.registerEvent); err != nil {
		return errors.WithMessage(err, "re init block listener error")
	}
	go l.recv()
	return nil
}

func (l *blockListener) registerEvent() error {
	startPoint := &nodeservice.EventStartPoint{}
	startPoint.ChainId = l.chainID
	startPoint.Type = nodeservice.LATEST
	payload, err := startPoint.Marshal()
	if err != nil {
		return errors.WithMessage(err, "marshal EventStartPoint error")
	}
	rawMsg, err := l.msgBuilder.GetRawMessage(payload)
	if err != nil {
		return errors.WithMessage(err, "message builder get raw message error")
	}
	l.eventClient, err = l.client.RegisterResultEvent(l.ctx, rawMsg)
	if err != nil {
		// return original grpc error to trigger reconnect or re init.
		return err
	}
	rawMsg, err = l.eventClient.Recv()
	if err != nil {
		// return original grpc error to trigger reconnect or re init.
		return err
	}
	return handleErr(rawMsg.Payload)
}

func (l *blockListener) registerTx(txHash []byte, opts ...ShardIDOption) (chan *common.TxResult, error) {
	return registerSyncSend(txHash, l.ch)
}

func registerSyncSend(txHash []byte, msgCh chan<- msg, opts ...ShardIDOption) (chan *common.TxResult, error) {
	if msgCh == nil {
		return nil, errors.New("message channel is nil")
	}
	ch := make(chan *common.TxResult, 1)
	statusCh := make(chan status, 1)

	defaultShardIDOpt := DefaultShardIDOptions()
	for _, opt := range opts {
		opt.Apply(&defaultShardIDOpt)
	}

	m := &registerMsg{
		id:       hex.EncodeToString(txHash),
		statusCh: statusCh,
		resultCh: ch,
	}
	msgCh <- m
	s := <-statusCh
	if s != success {
		return nil, errors.New("register tx hash to dispatch map error")
	}
	return ch, nil
}

func (l *blockListener) recv() {
	processor := func(rawMsg *common.RawMessage) {
		br := &common.BlockResult{}
		err := proto.Unmarshal(rawMsg.Payload, br)
		if err != nil {
			res := &common.Response{}
			err := proto.Unmarshal(rawMsg.Payload, res)
			if err != nil {
				log.Errorf("unmarshal response error: %v", err)
			}
			log.Errorf("response status: %v, info: %v", res.Status, res.StatusInfo)
			return
		}
		for _, txResult := range br.TxResults {
			msg := &txResultMsg{txResult: txResult}
			l.ch <- msg
		}
	}
	listen(l.eventClient.Recv, processor, l.reInit)
}
