/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

// Package event provide the implementation of event service.
package event

import (
	"context"

	"github.com/pkg/errors"
	"google.golang.org/grpc/codes"
	rpcStatus "google.golang.org/grpc/status"

	"huawei.com/huaweichain/common/logger"
	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/rawmessage"
)

var log = logger.GetModuleLogger("go sdk", "event")

// BlockEventService is the definition of block event service.
type BlockEventService struct {
	chainID    string
	msgBuilder rawmessage.MsgBuilder
	client     nodeservice.EventServiceClient
	ctx        context.Context
	cancel     context.CancelFunc
	holdCount  int32
	clean      func()
}

// NewBlockEventService is used to create an instance of block event service.
func NewBlockEventService(ctx context.Context, client nodeservice.EventServiceClient, chainID string,
	msgBuilder rawmessage.MsgBuilder, clean func()) *BlockEventService {
	childCtx, cancel := context.WithCancel(ctx)
	return &BlockEventService{
		chainID:    chainID,
		client:     client,
		msgBuilder: msgBuilder,
		ctx:        childCtx,
		cancel:     cancel,
		holdCount:  1,
		clean:      clean,
	}
}

// TryHold is used to hold block event service.
func (s *BlockEventService) TryHold() bool {
	return tryHold(&s.holdCount)
}

// Close is used to close block event service.
func (s *BlockEventService) Close() {
	count := close(&s.holdCount)
	if count == 0 {
		s.clean()
		s.cancel()
	}
}

// RegisterBlockEvent is used to register block event from latest block.
func (s *BlockEventService) RegisterBlockEvent() (*BlockIterator, error) {
	return s.registerBlockEventFrom(nodeservice.LATEST, uint64(0))
}

// RegisterBlockEventFrom is used to register block event from the specified block number.
func (s *BlockEventService) RegisterBlockEventFrom(startNum uint64) (*BlockIterator, error) {
	return s.registerBlockEventFrom(nodeservice.SPECIFIC, startNum)
}

func (s *BlockEventService) registerBlockEventFrom(startPointType nodeservice.StartPointType,
	startNum uint64) (*BlockIterator, error) {
	itr := &BlockIterator{
		iterator: s.newIter(startNum),
	}
	if err := itr.init(startPointType, startNum); err != nil {
		return nil, errors.WithMessage(err, "block iterator error")
	}
	return itr, nil
}

// RegisterBlockNumEvent is used to register block number event from latest block.
func (s *BlockEventService) RegisterBlockNumEvent() (*BlockNumIterator, error) {
	return s.registerBlockNumEventFrom(nodeservice.LATEST, uint64(0))
}

func (s *BlockEventService) registerBlockNumEventFrom(startPointType nodeservice.StartPointType,
	startNum uint64) (*BlockNumIterator, error) {
	itr := &BlockNumIterator{
		iterator: s.newIter(startNum),
	}
	if err := itr.init(startPointType, startNum); err != nil {
		return nil, errors.WithMessage(err, "block number iterator error")
	}
	return itr, nil
}

func (s *BlockEventService) newIter(startNum uint64) iterator {
	return iterator{
		chainID:    s.chainID,
		blockNum:   startNum,
		msgBuilder: s.msgBuilder,
		client:     s.client,
		ctx:        s.ctx,
		cancel:     s.cancel,
	}
}

// RegisterBlockResultEvent is used to register block result event from latest block.
func (s *BlockEventService) RegisterBlockResultEvent() (*BlockResultIterator, error) {
	return s.registerBlockResultEventFrom(nodeservice.LATEST, uint64(0))
}

// RegisterBlockResultEventFrom is used to register block result event from the specified block number.
func (s *BlockEventService) RegisterBlockResultEventFrom(startNum uint64) (*BlockResultIterator, error) {
	return s.registerBlockResultEventFrom(nodeservice.SPECIFIC, startNum)
}

func (s *BlockEventService) registerBlockResultEventFrom(startPointType nodeservice.StartPointType,
	startNum uint64) (*BlockResultIterator, error) {
	itr := &BlockResultIterator{
		iterator: s.newIter(startNum),
	}
	if err := itr.init(startPointType, startNum); err != nil {
		return nil, errors.WithMessage(err, "block result iterator error")
	}
	return itr, nil
}

// RegisterBlockAndResultEvent is used to register block and result event from latest block.
func (s *BlockEventService) RegisterBlockAndResultEvent() (*BlockAndResultIterator, error) {
	return s.registerBlockAndResultEventFrom(nodeservice.LATEST, 0)
}

// RegisterBlockAndResultEventFrom is used to register block and result event from the specified block number.
func (s *BlockEventService) RegisterBlockAndResultEventFrom(startNum uint64) (*BlockAndResultIterator, error) {
	return s.registerBlockAndResultEventFrom(nodeservice.SPECIFIC, startNum)
}

func (s *BlockEventService) registerBlockAndResultEventFrom(startPointType nodeservice.StartPointType,
	startNum uint64) (*BlockAndResultIterator, error) {
	itr := &BlockAndResultIterator{
		iterator: s.newIter(startNum),
	}
	if err := itr.init(startPointType, startNum); err != nil {
		return nil, errors.WithMessage(err, "block and result iterator error")
	}
	return itr, nil
}

type iterator struct {
	chainID    string
	blockNum   uint64
	msgBuilder rawmessage.MsgBuilder
	client     nodeservice.EventServiceClient
	ctx        context.Context
	cancel     context.CancelFunc
}

func (itr *iterator) buildRawMsg(startPointType nodeservice.StartPointType,
	startNum uint64) (*common.RawMessage, error) {
	startPoint := &nodeservice.EventStartPoint{
		ChainId:  itr.chainID,
		Type:     startPointType,
		BlockNum: startNum,
	}
	payload, err := startPoint.Marshal()
	if err != nil {
		return nil, errors.WithMessage(err, "marshal EventStartPoint error")
	}
	return itr.msgBuilder.GetRawMessage(payload)
}

func (itr *iterator) registerEvent(startPointType nodeservice.StartPointType,
	startNum uint64, register func(rawMsg *common.RawMessage) (func() (*common.RawMessage, error), error)) error {
	rawMsg, err := itr.buildRawMsg(startPointType, startNum)
	if err != nil {
		return errors.WithMessage(err, "build raw message error")
	}
	recv, err := register(rawMsg)
	if err != nil {
		// return original grpc error to trigger reconnect or re init.
		return err
	}
	rawMsg, err = recv()
	if err != nil {
		// return original grpc error to trigger reconnect or re init.
		return err
	}
	return handleErr(rawMsg.Payload)
}

// BlockIterator is the definition of block iterator.
type BlockIterator struct {
	iterator
	eventClient nodeservice.EventService_RegisterBlockEventClient
}

func (itr *BlockIterator) init(startPointType nodeservice.StartPointType,
	startNum uint64) error {
	return itr.registerEvent(startPointType, startNum, itr.registerEventClient)
}

func (itr *BlockIterator) reInit() (func() (*common.RawMessage, error), error) {
	registerEvent := func() error {
		return itr.registerEvent(nodeservice.SPECIFIC, itr.blockNum+1, itr.registerEventClient)
	}

	if err := reInit(registerEvent); err != nil {
		return nil, errors.WithMessage(err, "re init block iterator error")
	}

	log.Info("re init block iterator success")
	return itr.eventClient.Recv, nil
}

func (itr *BlockIterator) registerEventClient(rawMsg *common.RawMessage) (func() (*common.RawMessage, error), error) {
	var err error
	itr.eventClient, err = itr.client.RegisterBlockEvent(itr.ctx, rawMsg)
	if err != nil {
		// return original grpc error to trigger reconnect or re init.
		return nil, err
	}
	return itr.eventClient.Recv, nil
}

// Next is used to get the next element of iterator.
func (itr *BlockIterator) Next() (*common.Block, error) {
	payload, err := next(itr.eventClient.Recv, itr.reInit)
	if err != nil {
		return nil, errors.WithMessage(err, "recv next element error")
	}
	block := &common.Block{}
	err = unmarshal(payload, block)
	if err != nil {
		return nil, err
	}
	itr.blockNum = block.Header.Number
	return block, nil
}

// BlockNumIterator is the definition of block number iterator.
type BlockNumIterator struct {
	iterator
	eventClient nodeservice.EventService_RegisterBlockNumEventClient
}

func (itr *BlockNumIterator) init(startPointType nodeservice.StartPointType,
	startNum uint64) error {
	return itr.registerEvent(startPointType, startNum, itr.registerEventClient)
}

func (itr *BlockNumIterator) reInit() (func() (*common.RawMessage, error), error) {
	registerEvent := func() error {
		return itr.registerEvent(nodeservice.SPECIFIC, itr.blockNum+1, itr.registerEventClient)
	}

	if err := reInit(registerEvent); err != nil {
		return nil, errors.WithMessage(err, "re init block number iterator error")
	}

	log.Info("re init block number iterator success")
	return itr.eventClient.Recv, nil
}

func (itr *BlockNumIterator) registerEventClient(
	rawMsg *common.RawMessage) (func() (*common.RawMessage, error), error) {
	var err error
	itr.eventClient, err = itr.client.RegisterBlockNumEvent(itr.ctx, rawMsg)
	if err != nil {
		// return original grpc error to trigger reconnect or re init.
		return nil, err
	}
	return itr.eventClient.Recv, nil
}

// Next is used to get the next element of iterator.
func (itr *BlockNumIterator) Next() (uint64, error) {
	payload, err := next(itr.eventClient.Recv, itr.reInit)
	if err != nil {
		return uint64(0), errors.WithMessage(err, "recv next element error")
	}
	blockNumEvent := &nodeservice.BlockNumEvent{}
	err = unmarshal(payload, blockNumEvent)
	if err != nil {
		return uint64(0), err
	}
	if !blockNumEvent.Success {
		return uint64(0), errors.Errorf("get block number error: %v", blockNumEvent.Message)
	}
	itr.blockNum = blockNumEvent.Number
	return blockNumEvent.Number, nil
}

// BlockResultIterator is the definition of block result iterator.
type BlockResultIterator struct {
	iterator
	eventClient nodeservice.EventService_RegisterResultEventClient
}

func (itr *BlockResultIterator) init(startPointType nodeservice.StartPointType,
	startNum uint64) error {
	return itr.registerEvent(startPointType, startNum, itr.registerEventClient)
}

func (itr *BlockResultIterator) reInit() (func() (*common.RawMessage, error), error) {
	registerEvent := func() error {
		return itr.registerEvent(nodeservice.SPECIFIC, itr.blockNum+1, itr.registerEventClient)
	}

	if err := reInit(registerEvent); err != nil {
		return nil, errors.WithMessage(err, "re init block result iterator error")
	}
	log.Info("re init block result iterator success")
	return itr.eventClient.Recv, nil
}

func (itr *BlockResultIterator) registerEventClient(
	rawMsg *common.RawMessage) (func() (*common.RawMessage, error), error) {
	var err error
	itr.eventClient, err = itr.client.RegisterResultEvent(itr.ctx, rawMsg)
	if err != nil {
		// return original grpc error to trigger reconnect or re init.
		return nil, err
	}
	return itr.eventClient.Recv, nil
}

// Next is used to get the next element of iterator.
func (itr *BlockResultIterator) Next() (*common.BlockResult, error) {
	payload, err := next(itr.eventClient.Recv, itr.reInit)
	if err != nil {
		return nil, errors.WithMessage(err, "recv next element error")
	}
	blockResult := &common.BlockResult{}
	err = unmarshal(payload, blockResult)
	if err != nil {
		return nil, err
	}
	itr.blockNum = blockResult.BlockNum
	return blockResult, nil
}

// BlockAndResultIterator is the definition of block and result iterator.
type BlockAndResultIterator struct {
	iterator
	eventClient nodeservice.EventService_RegisterBlockAndResultEventClient
}

func (itr *BlockAndResultIterator) init(startPointType nodeservice.StartPointType,
	startNum uint64) error {
	return itr.registerEvent(startPointType, startNum, itr.registerEventClient)
}

func (itr *BlockAndResultIterator) reInit() (func() (*common.RawMessage, error), error) {
	registerEvent := func() error {
		return itr.registerEvent(nodeservice.SPECIFIC, itr.blockNum+1, itr.registerEventClient)
	}

	if err := reInit(registerEvent); err != nil {
		return nil, errors.WithMessage(err, "re init block and result iterator error")
	}
	log.Info("re init block and result iterator success")
	return itr.eventClient.Recv, nil
}

func (itr *BlockAndResultIterator) registerEventClient(
	rawMsg *common.RawMessage) (func() (*common.RawMessage, error), error) {
	var err error
	itr.eventClient, err = itr.client.RegisterBlockAndResultEvent(itr.ctx, rawMsg)
	if err != nil {
		// return original grpc error to trigger reconnect or re init.
		return nil, err
	}
	return itr.eventClient.Recv, nil
}

// Next is used to get the next element of iterator.
func (itr *BlockAndResultIterator) Next() (*common.BlockAndResult, error) {
	payload, err := next(itr.eventClient.Recv, itr.reInit)
	if err != nil {
		return nil, errors.WithMessage(err, "recv next element error")
	}
	br := &common.BlockAndResult{}
	err = unmarshal(payload, br)
	if err != nil {
		return nil, err
	}
	itr.blockNum = br.Result.BlockNum
	return br, nil
}

func unmarshal(payload []byte, pb proto.Unmarshaler) error {
	err := proto.Unmarshal(payload, pb)
	if err != nil {
		responseErr := handleErr(payload)
		if responseErr != nil {
			return errors.WithMessagef(responseErr, "after unmarshal error: %v, "+
				"unmarshal response error", err.Error())
		}
	}
	return nil
}

type recvFunc = func() (*common.RawMessage, error)

func next(recv recvFunc, reInit func() (recvFunc, error)) ([]byte, error) {
	for {
		rawMsg, err := recv()
		if err == nil && rawMsg != nil {
			if rawMsg.Payload == nil {
				return nil, errors.Errorf("recv empty message, maybe chain has quited")
			}
			return rawMsg.Payload, nil
		}

		rpcErr, ok := rpcStatus.FromError(err)
		if !ok {
			return nil, errors.WithMessage(err, "failed to recv message, please recover service manually")
		}
		// client close stream actively
		if rpcErr.Code() == codes.Canceled {
			return nil, errors.Errorf("recv error: %v, may client close stream actively please recover service "+
				"manually", rpcErr)
		}
		if rpcErr.Code() != codes.Unavailable {
			if rpcErr.Code() != codes.Unknown || rpcErr.Message() != rpcErrDesc {
				return nil, errors.Errorf("recv error: %v, please recover service manually", rpcErr)
			}
		}
		log.Errorf("recv error: %+v", err)
		recv, err = reInit()
		if err != nil {
			return nil, errors.WithMessage(err, "re init failed, please recover service manually")
		}
	}
}

func handleErr(payload []byte) error {
	res := &common.Response{}
	err := proto.Unmarshal(payload, res)
	if err != nil {
		return errors.WithMessage(err, "unmarshal response error")
	}
	if res.Status != common.SUCCESS {
		return errors.Errorf("response status: %v, info: %v", res.Status, res.StatusInfo)
	}
	return nil
}
