/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

package event

import (
	"context"
	"encoding/hex"
	"time"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/rawmessage"
)

type status int32

const (
	timeout        = 30
	success status = 0
	failed  status = 1
)

type txListener struct {
	chainID     string
	ch          chan msg
	sendCh      chan *sendMsg
	dispatchCh  chan msg
	errCh       chan struct{}
	registerMap map[string]chan *nodeservice.TxEventRes
	msgBuilder  rawmessage.MsgBuilder
	client      nodeservice.EventServiceClient
	eventClient nodeservice.EventService_RegisterTxEventClient
	ctx         context.Context
	cancel      context.CancelFunc
	shardID     uint32
}

func newTxListener(params *listenerParam) *txListener {
	return &txListener{
		chainID:     params.chainID,
		ch:          params.ch,
		sendCh:      make(chan *sendMsg, buffSize),
		dispatchCh:  make(chan msg, buffSize),
		registerMap: make(map[string]chan *nodeservice.TxEventRes),
		errCh:       make(chan struct{}),
		msgBuilder:  params.msgBuilder,
		client:      params.client,
		ctx:         params.ctx,
		cancel:      params.cancel,
		shardID:     params.shardID,
	}
}

func (l *txListener) init() error {
	if err := l.registerEvent(); err != nil {
		return errors.WithMessage(err, "register event error")
	}
	go l.send()
	go l.recv()
	go l.dispatch()
	return nil
}

func (l *txListener) reInit() error {
	if err := reInit(l.registerEvent); err != nil {
		return errors.WithMessage(err, "re init tx listener error")
	}
	go l.recv()
	return nil
}

func (l *txListener) registerEvent() error {
	var err error
	l.eventClient, err = l.client.RegisterTxEvent(l.ctx)
	if err != nil {
		// return original grpc error to trigger reconnect or re init.
		return err
	}
	txEvent := &nodeservice.TxEvent{ChainId: l.chainID,
		ShardID: l.shardID, Type: nodeservice.REGISTER_CLIENT}
	payload, err := txEvent.Marshal()
	if err != nil {
		return errors.WithMessage(err, "marshal tx event error")
	}
	rawMsg, err := l.msgBuilder.GetRawMessage(payload)
	if err != nil {
		return errors.WithMessage(err, "message builder get raw message error")
	}
	err = l.eventClient.Send(rawMsg)
	if err != nil {
		// return original grpc error to trigger reconnect or re init.
		return err
	}

	rawMsg, err = l.eventClient.Recv()
	if err != nil {
		// return original grpc error to trigger reconnect or re init.
		return err
	}

	resp := &nodeservice.TxEventRes{}
	err = proto.Unmarshal(rawMsg.Payload, resp)
	if err != nil {
		return errors.WithMessage(err, "unmarshal tx event response result error")
	}
	if resp.Type != nodeservice.REGISTER_CLIENT {
		return errors.Errorf("incorrect tx event response type: %v", resp.Type)
	}
	if resp.Status != nodeservice.SUCCESS {
		return errors.Errorf("register tx event client failed, register tx event client "+
			"response status: %v", resp.Status)
	}
	return nil
}

func (l *txListener) registerTx(txHash []byte, opts ...ShardIDOption) (chan *common.TxResult, error) {
	defaultShardIDOpt := DefaultShardIDOptions()
	for _, opt := range opts {
		opt.Apply(&defaultShardIDOpt)
	}
	txEvent := &nodeservice.TxEvent{ChainId: l.chainID, TxHash: txHash, Type: nodeservice.REGISTER_TX_HASH,
		ShardID: defaultShardIDOpt.ShardID}
	txEventResMsg, err := l.register(txEvent)
	if err != nil {
		return nil, errors.WithMessage(err, "register tx hash event error")
	}
	if txEventResMsg.txEventRes.Status != nodeservice.SUCCESS {
		return nil, errors.WithMessagef(err, "register tx event client failed, register tx hash event "+
			"response status: %v", txEventResMsg.txEventRes.Status)
	}

	return registerSyncSend(txHash, l.ch)
}

func (l *txListener) register(txEvent *nodeservice.TxEvent) (*txEventResMsg, error) {
	statusCh := make(chan status, 1)
	dispatchCh := make(chan *nodeservice.TxEventRes, 1)

	var id string
	if txEvent.Type == nodeservice.REGISTER_CLIENT {
		id = txEvent.ChainId
	} else {
		id = hex.EncodeToString(txEvent.TxHash)
	}

	l.dispatchCh <- &registerMsg{
		id:         id,
		statusCh:   statusCh,
		dispatchCh: dispatchCh,
	}
	s := <-statusCh
	if s != success {
		return nil, errors.New("register tx event to dispatch map error")
	}

	sCh := make(chan status, 1)
	l.sendCh <- &sendMsg{
		ch:      sCh,
		txEvent: txEvent,
	}
	s = <-sCh
	if s != success {
		if err := l.unregister(id); err != nil {
			return nil, errors.WithMessagef(err, "unregister tx event error: id: [%v]", id)
		}
		return nil, errors.New("send register tx event error")
	}
	select {
	case txEventRes := <-dispatchCh:
		return &txEventResMsg{txEventRes: txEventRes}, nil
	case <-time.After(time.Duration(timeout) * time.Second):
		if err := l.unregister(id); err != nil {
			return nil, errors.WithMessagef(err, "unregister tx event error: id: [%v]", id)
		}
		return nil, errors.Errorf("register tx event time out for wait [%d] second, id: [%v]", timeout, id)
	case <-l.ctx.Done():
		return nil, errors.New("listener closed")
	}
}

func (l *txListener) unregister(id string) error {
	statusCh := make(chan status, 1)
	l.dispatchCh <- &unregisterMsg{
		id:       id,
		statusCh: statusCh,
	}
	s := <-statusCh
	if s != success {
		return errors.New("unregister tx event from dispatch map error")
	}
	return nil
}

func (l *txListener) send() {
	if l.ctx.Done() == nil {
		log.Error(" context done chan is nil")
		return
	}
	for {
		select {
		case sMsg := <-l.sendCh:
			payload, err := sMsg.txEvent.Marshal()
			if err != nil {
				log.Errorf("marshal tx event error: %v", err)
				sMsg.ch <- failed
			}
			rawMsg, err := l.msgBuilder.GetRawMessage(payload)
			if err != nil {
				log.Errorf("message builder get raw message error: %v", err)
				sMsg.ch <- failed
			}
			if err := l.eventClient.Send(rawMsg); err != nil {
				log.Errorf("send register tx id info error: %v", err)
				sMsg.ch <- failed
			}
			sMsg.ch <- success
		case <-l.ctx.Done():
			log.Debug("send done.")
			return
		}
	}
}

func (l *txListener) recv() {
	processor := func(rawMsg *common.RawMessage) {
		resp := &nodeservice.TxEventRes{}
		err := proto.Unmarshal(rawMsg.Payload, resp)
		if err != nil {
			log.Errorf("unmarshal tx event response result error: %v", err)
			return
		}
		switch resp.Type {
		case nodeservice.REGISTER_CLIENT:
			fallthrough
		case nodeservice.REGISTER_TX_HASH:
			msg := &txEventResMsg{txEventRes: resp}
			l.dispatchCh <- msg
		case nodeservice.TX_RESULT:
			txResult := &common.TxResult{}
			err = proto.Unmarshal(resp.Payload, txResult)
			if err != nil {
				log.Errorf("unmarshal tx result error: %v", err)
				return
			}
			msg := &txResultMsg{txResult: txResult}
			l.ch <- msg
		default:
			log.Panicf("invalid service type %s", resp.Type)
		}
	}
	listen(l.eventClient.Recv, processor, l.reInit)
}

func (l *txListener) dispatch() {
	for {
		select {
		case m := <-l.dispatchCh:
			l.processMsg(m)
		case <-l.ctx.Done():
			log.Debug("dispatch done.")
			return
		}
	}
}

func (l *txListener) processMsg(m msg) {
	switch u := m.(type) {
	case *unregisterMsg:
		delete(l.registerMap, u.id)
		u.statusCh <- success
	case *registerMsg:
		l.registerMap[u.id] = u.dispatchCh
		u.statusCh <- success
	case *txEventResMsg:
		ch, ok := l.registerMap[u.txEventRes.Id]
		if !ok {
			return
		}
		if ch == nil {
			log.Error("tx result chan is nil")
			delete(l.registerMap, u.txEventRes.Id)
			return
		}
		ch <- u.txEventRes
		delete(l.registerMap, u.txEventRes.Id)
	default:
	}
}
