/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

package action

import (
	"context"
	"sync"

	"github.com/pkg/errors"

	pb "huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/action/event"
	"huawei.com/huaweichain/sdk/rawmessage"
)

// EventAction is the action for event operations.
type EventAction struct {
	action
	msgBuilder rawmessage.MsgBuilder
	client     nodeservice.EventServiceClient
	clientLck  sync.Mutex
	blkSrvMgr  *event.Mgr
	txSrvMgr   *event.Mgr
}

// NewEventAction is used to create event action instance with a grpc client connection.
func NewEventAction(config *Config) *EventAction {
	action := &EventAction{
		action:     action{config: config},
		msgBuilder: rawmessage.NewMsgBuilderImpl(config.Crypto),
		blkSrvMgr:  &event.Mgr{},
		txSrvMgr:   &event.Mgr{},
	}
	return action
}

// GetBlockEventService is used to get block event service.
func (action *EventAction) GetBlockEventService(chainID string) (*event.BlockEventService, error) {
	return action.GetBlockEventServiceWithCtx(context.Background(), chainID)
}

// GetBlockEventServiceWithCtx creates new block event service with context control.
func (action *EventAction) GetBlockEventServiceWithCtx(ctx context.Context, chainID string) (
	*event.BlockEventService, error) {
	supply := func() (event.Service, error) {
		client, err := action.getClient()
		if err != nil {
			return nil, errors.WithMessage(err, "get client error")
		}
		clean := func() {
			action.blkSrvMgr.Remove(chainID)
		}
		return event.NewBlockEventService(ctx, client, chainID, action.msgBuilder, clean), nil
	}
	service, err := action.blkSrvMgr.Get(chainID, supply)
	if err != nil {
		return nil, errors.WithMessage(err, "tx event service manager get tx event service error")
	}
	blockEventService, ok := service.(*event.BlockEventService)
	if !ok {
		return nil, errors.New("type assert error: *event.TxEventService")
	}
	return blockEventService, nil
}

// GetTxEventService is used to get tx event service by default event source type. default tx event.
func (action *EventAction) GetTxEventService(chainID string) (*event.TxEventService, error) {
	return action.GetTxEventServiceWithSourceType(chainID, event.Block)
}

// GetTxEventServiceWithSourceType is used to get tx event service with specified event source type.
func (action *EventAction) GetTxEventServiceWithSourceType(chainID string,
	source event.SourceType, opts ...event.ShardIDOption) (*event.TxEventService, error) {
	return action.GetTxListener(context.Background(), chainID, source, opts...)
}

// GetTxListener new a TxEventService with context and source type.
func (action *EventAction) GetTxListener(ctx context.Context, chainID string, source event.SourceType,
	opts ...event.ShardIDOption) (
	*event.TxEventService, error) {
	supply := func() (event.Service, error) {
		client, err := action.getClient()
		if err != nil {
			return nil, errors.WithMessage(err, "get client error")
		}
		clean := func() {
			action.txSrvMgr.Remove(chainID)
		}
		defaultShardIDOpt := event.DefaultShardIDOptions()
		for _, opt := range opts {
			opt.Apply(&defaultShardIDOpt)
		}
		params := event.Params{
			ChainID:    chainID,
			Source:     source,
			MsgBuilder: action.msgBuilder,
			Client:     client,
			Clean:      clean,
			ShardID:    defaultShardIDOpt.ShardID,
		}
		return event.NewTxSrvWithSrc(ctx, &params), nil
	}
	service, err := action.txSrvMgr.Get(chainID, supply)
	if err != nil {
		return nil, errors.WithMessage(err, "tx event service manager get tx event service error")
	}
	txEventService, ok := service.(*event.TxEventService)
	if !ok {
		return nil, errors.New("type assert error: *event.TxEventService")
	}
	return txEventService, nil
}

// Listen is used to register result event to server and get the register result client.
//
// Deprecated: Use BlockEventService.RegisterBlockResultEvent instead.
func (action *EventAction) Listen(chainID string) (nodeservice.EventService_RegisterResultEventClient, error) {
	startPoint := &nodeservice.EventStartPoint{}
	startPoint.ChainId = chainID
	startPoint.Type = nodeservice.LATEST
	rawMsg, err := action.getRawMessage(startPoint)
	if err != nil {
		return nil, errors.WithMessage(err, "get raw message error")
	}
	client, err := action.getClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get client error")
	}
	event, err := client.RegisterResultEvent(context.Background(), rawMsg)
	if err != nil {
		return nil, errors.WithMessage(err, "event action RegisterResultEvent failed")
	}
	return event, nil
}

// ListenBlockAndResult is used to register block and result event to server and get the register result client.
//
// Deprecated: Use BlockEventService.RegisterBlockAndResultEvent instead.
func (action *EventAction) ListenBlockAndResult(chainID string,
	number uint64) (nodeservice.EventService_RegisterBlockAndResultEventClient, error) {
	startPoint := &nodeservice.EventStartPoint{
		ChainId:  chainID,
		Type:     nodeservice.SPECIFIC,
		BlockNum: number,
	}
	rawMsg, err := action.getRawMessage(startPoint)
	if err != nil {
		return nil, errors.WithMessage(err, "get raw message error")
	}
	client, err := action.getClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get client error")
	}
	event, err := client.RegisterBlockAndResultEvent(context.Background(), rawMsg)
	if err != nil {
		return nil, errors.WithMessage(err, "event action RegisterBlockAndResultEvent failed")
	}
	return event, nil
}

// RegisterTxEvent is used to register tx event to server and get the register tx event client.
//
// Deprecated: Use GetTxEventService instead.
func (action *EventAction) RegisterTxEvent(chainID string) (*TxEvent, error) {
	txEventService, err := action.GetTxEventService(chainID)
	if err != nil {
		return nil, errors.WithMessage(err, "get tx event service error")
	}
	return &TxEvent{txEventService: txEventService}, nil
}

func (action *EventAction) getRawMessage(msg pb.Message) (*common.RawMessage, error) {
	payload, err := msg.Marshal()
	if err != nil {
		return nil, errors.WithMessage(err, "marshal tx event error: %v")
	}
	return action.msgBuilder.GetRawMessage(payload)
}

func (action *EventAction) getClient() (nodeservice.EventServiceClient, error) {
	if err := action.resetConn(); err != nil {
		return nil, errors.WithMessage(err, "reset clients error: %v")
	}
	if action.client == nil {
		action.newClient()
	}
	return action.client, nil
}

func (action *EventAction) newClient() {
	action.clientLck.Lock()
	defer action.clientLck.Unlock()
	if action.client == nil {
		action.client = nodeservice.NewEventServiceClient(action.conn)
	}
}

// TxEvent is the definition of TxEvent.
type TxEvent struct {
	txEventService *event.TxEventService
}

// RegisterTx is used to register tx id to server.
func (e *TxEvent) RegisterTx(txHash []byte) (chan *common.TxResult, error) {
	return e.txEventService.RegisterTx(txHash)
}

// Close is used to close tx event.
func (e *TxEvent) Close() {
	e.txEventService.Close()
}
