/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2020. All rights reserved.
 */

// Package client provide the implementation of client module, which is the entrance of go sdk.
package client

import (
	"regexp"
	"strings"
	"time"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/cryptomgr"
	_ "huawei.com/huaweichain/common/cryptomgr/crypto" // initialize crypto provider register
	"huawei.com/huaweichain/common/cryptomgr/cryptoimpl"
	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/config"
	"huawei.com/huaweichain/sdk/crypto"
	"huawei.com/huaweichain/sdk/node"
	"huawei.com/huaweichain/sdk/rawmessage"
	"huawei.com/huaweichain/sdk/utils"
)

// GatewayClient is the definition of gateway client.
type GatewayClient struct {
	tls                 *crypto.TLS
	cpt                 crypto.Crypto
	ChainRawMessage     *rawmessage.ChainRawMessage
	ContractRawMessage  *rawmessage.ContractRawMessage
	LifecycleRawMessage *rawmessage.LifecycleRawMessage
	QueryRawMessage     *rawmessage.QueryRawMessage
	ConfigRawMessage    *rawmessage.ConfigRawMessage
	Nodes               map[string]*node.WNode
	Conf                *config.ClientConfig
	Decrypt             func(bytes []byte) ([]byte, error)
}

// NodeNameReg is the node name regex format.
const (
	nodeNameReg = "^[A-Za-z0-9_][A-Za-z0-9_\\-\\.]{0,99}$"
)

// NewEmptyGatewayClient is used to create an instance of gateway client which is empty.
func NewEmptyGatewayClient() *GatewayClient {
	return &GatewayClient{Nodes: make(map[string]*node.WNode)}
}

// NewGatewayClient is used to create an instance of gateway client by specifying the config file path.
func NewGatewayClient(configPath string, decrypts ...func(bytes []byte) ([]byte, error)) (*GatewayClient, error) {
	cfg, err := config.NewClientConfig(configPath)
	if err != nil {
		return nil, errors.WithMessage(err, "load config file error")
	}

	return NewGatewayClientWithCfg(cfg, decrypts...)
}

// NewGatewayClientWithCfg is used to create an instance of gateway client by specifying the config data.
func NewGatewayClientWithCfg(cfg *config.ClientConfig, decrypts ...func(bytes []byte) ([]byte,
	error)) (*GatewayClient, error) {
	client := &GatewayClient{
		Conf:  cfg,
		Nodes: make(map[string]*node.WNode),
	}
	decrypt := func(bytes []byte) ([]byte, error) {
		return bytes, nil
	}
	if len(decrypts) > 0 {
		decrypt = decrypts[0]
	}
	client.Decrypt = decrypt
	var err error
	client.cpt, err = getCrypto(cfg.Client, decrypt)
	if err != nil {
		return nil, errors.WithMessage(err, "get crypto error")
	}
	err = client.setRawMsg(client.cpt)
	if err != nil {
		client.CleanSenInfo()
		return nil, errors.WithMessage(err, " set raw message error")
	}

	client.Nodes, err = getNodes(cfg, client.cpt, decrypt)
	if err != nil {
		client.CleanSenInfo()
		return nil, errors.WithMessage(err, "get nodes error")
	}
	return client, nil
}

// CleanSenInfo clean sensitive info in client object
// Call this function when this client object is no longer used
func (client *GatewayClient) CleanSenInfo() {
	if client.cpt != nil {
		client.cpt.CleanSenInfo()
		client.cpt = nil
	}
	if client.tls != nil {
		client.tls.CleanSenInfo()
		client.tls = nil
	}
	if client.Nodes != nil {
		for _, n := range client.Nodes {
			n.CleanSenInfo()
		}
		client.Nodes = nil
	}
}

// SetIdentity is used to set identity for sdk.
func (client *GatewayClient) SetIdentity(algorithm string, key []byte, cert []byte) error {
	c, err := crypto.NewCryptoWithIdentity(algorithm, cert, key)
	if err != nil {
		return errors.WithMessage(err, "new crypto with identity error")
	}
	client.cpt = c
	return client.setRawMsg(c)
}

// SetTLS is used to set TLS for sdk.
func (client *GatewayClient) SetTLS(certPEMBlock []byte, keyPEMBlock []byte, roots [][]byte) {
	gmCertFactory := cryptoimpl.GetCertFactory(cryptomgr.Sm2WithSm3)
	if gmCertFactory != nil {
		_, err := gmCertFactory.GetCertFromPem(certPEMBlock)
		if err == nil {
			client.tls = crypto.NewTLS(certPEMBlock, keyPEMBlock, roots, cryptomgr.Sm2WithSm3)
			return
		}
	}
	client.tls = crypto.NewTLS(certPEMBlock, keyPEMBlock, roots, cryptomgr.EcdsaWithSha256)
}

// AddNode is used to add node proxy for sdk.
func (client *GatewayClient) AddNode(name string, hostOverride string, host string, port int) error {
	n := config.Node{
		ID:           name,
		HostOverride: hostOverride,
		Host:         host,
		Port:         port,
	}
	return client.addNode(&n)
}

// Add configured node.
func (client *GatewayClient) Add(node *config.Node) error {
	return client.addNode(node)
}

func (client *GatewayClient) addNode(n *config.Node) error {
	if err := checkNodeName(n.ID); err != nil {
		return errors.WithMessage(err, "check node name error")
	}
	if err := checkHost(n.Host); err != nil {
		return errors.WithMessage(err, "failed to check host")
	}
	if err := checkHost(n.HostOverride); err != nil {
		return errors.WithMessage(err, "failed to check host override")
	}
	if err := checkPort(n.Port); err != nil {
		return errors.WithMessage(err, "failed to check port")
	}
	if n.RestPort != 0 {
		if err := checkPort(n.RestPort); err != nil {
			return errors.WithMessage(err, "failed to check port")
		}
	}
	n.ID = strings.ToLower(n.ID)

	tlsEnable := false
	if client.tls != nil {
		tlsEnable = true
	}
	if client.cpt == nil {
		return errors.New("add node need to set crypto first")
	}
	var err error
	client.Nodes[n.ID], err = node.NewNode(n, client.cpt, tlsEnable, client.tls)
	if err != nil {
		return errors.WithMessage(err, "new node error")
	}
	return nil
}

func checkPort(port int) error {
	const minPort = 1024
	const maxPort = 65535
	if port < minPort || port > maxPort {
		return errors.Errorf("port is not in range[%v:%v]", minPort, maxPort)
	}
	return nil
}

// GetNode get node to send message, 'name' is case-insensitive.
func (client *GatewayClient) GetNode(name string) (*node.WNode, bool) {
	wnode, ok := client.Nodes[strings.ToLower(name)]
	return wnode, ok
}

// SetTimeout set client timeout (/s)
func (client *GatewayClient) SetTimeout(seconds int64) {
	utils.SetTimeout(time.Duration(seconds))
}

// GetSingleApplier provides func to get latest block number when build txHeader.
func (client *GatewayClient) GetSingleApplier(addr, chainID string) func() (uint64, error) {
	return func() (uint64, error) {
		var wnode *node.WNode
		for _, n := range client.Nodes {
			if n.GetNodeAddr() == addr {
				wnode = n
				break
			}
		}
		if wnode == nil {
			return 0, errors.Errorf("node addr %s is unknown", addr)
		}
		reqRaw, err := client.QueryRawMessage.BuildLatestChainStateRawMessage(chainID)
		if err != nil {
			return 0, err
		}
		resRaw, err := wnode.QueryAction.GetLatestChainState(reqRaw)
		if err != nil {
			return 0, err
		}
		response := &common.Response{}
		if err = proto.Unmarshal(resRaw.Payload, response); err != nil {
			return 0, errors.WithMessage(err, "unmarshal response error")
		}
		if response.Status != common.SUCCESS {
			return 0, errors.Errorf("response error: status: %v, info: %v",
				response.Status.String(), response.StatusInfo)
		}
		latestChainState := &nodeservice.LatestChainState{}
		if err = proto.Unmarshal(response.Payload, latestChainState); err != nil {
			return 0, errors.WithMessage(err, "unmarshal latest chain state error")
		}
		return latestChainState.Height + 1, nil
	}
}

func (client *GatewayClient) setRawMsg(crypto crypto.Crypto) error {
	msgBuilder := rawmessage.NewMsgBuilderImpl(crypto)
	client.ChainRawMessage = rawmessage.NewChainRawMessage(msgBuilder, crypto)
	client.ContractRawMessage = rawmessage.NewContractRawMessage(msgBuilder)
	client.LifecycleRawMessage = rawmessage.NewLifecycleRawMessage(msgBuilder, crypto)
	client.QueryRawMessage = rawmessage.NewQueryRawMessage(msgBuilder)
	client.ConfigRawMessage = rawmessage.NewConfigRawMessage(msgBuilder)
	return nil
}

func getNodes(c *config.ClientConfig, cpt crypto.Crypto, decrypt config.DecryptFunc) (map[string]*node.WNode, error) {
	nodes := make(map[string]*node.WNode)
	for _, n := range c.Nodes {
		if err := checkNodeName(n.ID); err != nil {
			return nil, errors.WithMessage(err, "node name error in config")
		}
		var wnode *node.WNode
		var tls *crypto.TLS
		var err error
		if c.Client.TLS.Enable {
			tls, err = node.ConvertTLS(c.Client.TLS, c.Client.Type, decrypt)
			if err != nil {
				return nil, errors.WithMessage(err, "convert node tls from config tls error")
			}
		}
		wnode, err = node.NewNode(n, cpt, c.Client.TLS.Enable, tls)
		if err != nil {
			return nil, errors.WithMessage(err, "new node error")
		}
		nodes[n.ID] = wnode
	}
	return nodes, nil
}

func getCrypto(c *config.Client, decrypt config.DecryptFunc) (crypto.Crypto, error) {
	if c == nil {
		return nil, errors.New("the client config is nil")
	}
	return crypto.NewCrypto(c.Type, c.Identity.CertPath, c.Identity.KeyPath, decrypt)
}

func checkNodeName(nodeName string) error {
	ok := regexp.MustCompile(nodeNameReg).MatchString(nodeName)
	if !ok {
		return errors.Errorf("node name not match format: %v", nodeNameReg)
	}
	return nil
}

func checkHost(host string) error {
	const hostLen = 8192
	if len(host) > hostLen {
		return errors.Errorf("host len out of range %v", hostLen)
	}
	const hostReg = "^[A-Za-z0-9_\\-\\.]+$"
	if !regexp.MustCompile(hostReg).MatchString(host) {
		return errors.Errorf("host name not match format: %v", hostReg)
	}
	return nil
}

// Crypto returns client's identity crypto utilities.
func (client *GatewayClient) Crypto() crypto.Crypto {
	return client.cpt
}
