/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2020. All rights reserved.
 */

package config

import (
	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto/common"
	consensus "huawei.com/huaweichain/proto/schema"
	"huawei.com/huaweichain/proto/shard"
)

// GenesisConfig is the definition of GenesisConfig.
type GenesisConfig struct {
	Consenters   []*consensus.Consenter
	Organs       []*common.Organization
	GenesisBlock GenesisBlock
	Decrypt      func([]byte) ([]byte, error)
}

// NewGenesisConfig is used to new an instance of genesis config
func NewGenesisConfig(genesisConfigPath string,
	decrypts ...func(bytes []byte) ([]byte, error)) (*GenesisConfig, error) {
	v, errMsg := parseCfg(genesisConfigPath)
	if errMsg != nil {
		return nil, errors.WithMessage(errMsg, "parse config file error")
	}

	decrypt := func(bytes []byte) ([]byte, error) {
		return bytes, nil
	}
	if len(decrypts) > 0 {
		decrypt = decrypts[0]
	}

	genesisConfig := &GenesisConfig{Decrypt: decrypt}
	errMsg = unmarshalGenesisConfig(v, &genesisConfig)
	if errMsg != nil {
		return nil, errors.WithMessage(errMsg, "error unmarshaling config into struct")
	}
	return genesisConfig, nil
}

// AddConsenter is used to add consenter.
func (gc *GenesisConfig) AddConsenter(consenter *consensus.Consenter) {
	gc.Consenters = append(gc.Consenters, consenter)
}

// AddOrganization is used to add organization.
func (gc *GenesisConfig) AddOrganization(name string, admin []byte, root []byte, tls []byte) {
	organ := &common.Organization{
		Name:        name,
		RootCert:    root,
		AdminCert:   admin,
		TLSRootCert: tls,
	}
	gc.Organs = append(gc.Organs, organ)
}

// GenesisBlock is the definition of GenesisBlock.
type GenesisBlock struct {
	DBType             string
	OrgLimit           bool
	Consensus          Consensus
	Raft               Raft
	Flic               Flic
	Hotstuff           Hotstuff
	Dumbo              Dumbo
	Solo               Solo
	Policy             Policy
	Organizations      []Organization
	Consenters         []Consenter
	Sharding           Sharding
	Net                *Net
	BlockLimit         int64
	MinPlatformVersion string
	ApprovalNoCert     bool
}

// Consensus is the definition of Consensus common cfg.
type Consensus struct {
	Type                  string
	BlockPackage          *BlockPackage `yaml:"blockPackage"`
	Transport             *Transport
	InitBlockSeqCap       uint64 `yaml:"initBlockSeqCap"`
	InitPendingEntriesCap uint64 `yaml:"initPendingEntriesCap"`
	BlockLimitRate        uint64 `yaml:"blockLimitRate"`
	TxLimitRate           uint64 `yaml:"txLimitRate"`
	TickBufSize           uint64 `yaml:"tickBufSize"`
	Pipeline              *Pipeline
	MaxConsenterNum       uint64 `mapstructure:"max_consenter_num" yaml:"max_consenter_num"`
}

// BlockPackage is the definition of Consensus module cfg.
type BlockPackage struct {
	MaxTxCount   uint64 `yaml:"maxTxCount"`
	MaxBlockSize uint64 `yaml:"maxBlockSize"`
	TickInterval uint64 `yaml:"tickInterval"`
	TimeoutTick  uint64 `yaml:"timeoutTick"`
	TxBufSize    uint64 `yaml:"txBufSize"`
	MaxRoutines  uint64 `yaml:"maxRoutines"`
}

// Pipeline is the definition of pipeline.
type Pipeline struct {
	ProposedBlockBufSize  uint64 `yaml:"proposedBlockBufSize"`
	CommittedBlockBufSize uint64 `yaml:"committedBlockBufSize"`
	PersistedBlockBufSize uint64 `yaml:"persistedBlockBufSize"`
}

// Transport is the definition of consensus node send and receive msg buf cfg.
type Transport struct {
	SendBufSize uint64 `yaml:"sendBufSize"`
	RecvBufSize uint64 `yaml:"recvBufSize"`
}

// Hotstuff is the definition of Hotstuff consensus.
type Hotstuff struct {
	Nodes           []string
	Learners        []string `yaml:",omitempty"`
	ProposerType    string   `mapstructure:"proposer_type" yaml:"proposer_type"`
	TickInterval    uint64   `mapstructure:"tick_interval" yaml:"tick_interval"`
	ViewTimeoutTick uint32   `mapstructure:"view_timeout_tick" yaml:"view_timeout_tick"`
	WaitTxDuration  uint64   `mapstructure:"wait_tx_duration" yaml:"wait_tx_duration"`
	MaxPrunedBlocks uint64   `mapstructure:"max_pruned_blocks" yaml:"max_pruned_blocks"`
	MaxBlockBatch   uint64   `mapstructure:"max_block_batch" yaml:"max_block_batch"`
	ForwardProp     bool     `mapstructure:"forward_prop" yaml:"forward_prop"`
	MemoryStore     bool     `mapstructure:"memory_store" yaml:"memory_store"`
	VerifyTimestamp bool     `mapstructure:"verify_timestamp" yaml:"verify_timestamp"`
}

// Raft is the definition of Raft consensus.
type Raft struct {
	Nodes                       []string
	Learners                    []string `yaml:",omitempty"`
	TickInterval                uint64   `yaml:"tickInterval"`
	HeartbeatTick               uint64   `yaml:"heartbeatTick"`
	ElectionTick                uint64   `yaml:"electionTick"`
	SafeTicker                  bool     `yaml:"safeTicker,omitempty"`
	UnreachableMissingHeartbeat uint64   `yaml:"unreachableMissingHeartbeat"`
	CheckQuorum                 bool     `yaml:"checkQuorum,omitempty"`
	PreVote                     bool     `yaml:"preVote,omitempty"`
	DisableForward              bool     `yaml:"disableForward,omitempty"`
	SnapshotThreshold           uint64   `yaml:"snapshotThreshold"`
	MaxBatchSize                uint64   `yaml:"maxBatchSize"`
	TermRoundBound              uint32   `yaml:"term_round_bound,omitempty"`
}

// Dumbo is the definition of Dumbo consensus.
type Dumbo struct {
	Nodes          []string
	Learners       []string `yaml:",omitempty"`
	ForwardProp    bool     `mapstructure:"forward_prop" yaml:"forward_prop,omitempty"`
	TxBatchSize    uint64   `mapstructure:"tx_batch_size" yaml:"tx_batch_size"`
	MaxBlockBatch  uint64   `mapstructure:"max_block_batch" yaml:"max_block_batch"`
	MaxFutureEpoch uint64   `mapstructure:"max_future_epoch" yaml:"max_future_epoch"`
}

// Flic is the definition of Flic consensus.
type Flic struct {
	MaxFaultNodes uint64 `yaml:"maxFaultNodes,omitempty"`
	ReqTimeout    uint64 `yaml:"reqTimeout,omitempty"`
}

// Solo is the definition of Solo consensus.
type Solo struct {
}

// Policy is the definition of proposal policy.
type Policy struct {
	Config    string
	Lifecycle string
	Archive   string
}

// Organization is the definition of Organization.
type Organization struct {
	ID          string `yaml:"id,omitempty"` // defined to store key when convert map to struct array
	AdminCert   string `yaml:"adminCert"`
	RootCert    string `yaml:"rootCert"`
	TLSRootCert string `yaml:"tlsRootCert"`
}

// Consenter is the definition of consenter node.
type Consenter struct {
	ID       string `yaml:"id,omitempty"` // defined to store key when convert map to struct array
	Org      string
	Host     string
	Port     uint64
	ReqPort  uint32 `mapstructure:"req_port" yaml:"req_port"`
	RestPort uint32 `mapstructure:"rest_port" yaml:"rest_port"`
	ReeCert  string `yaml:"reeCert"`
	TeeCert  string `yaml:"teeCert,omitempty"`
}

// ShardNode is the definition of a node config for sharding
type ShardNode struct {
	ID   string // defined to store key when convert map to struct array
	Host string
	Port uint64
}

// Sharding is the definition of shard option.
type Sharding struct {
	Enabled    bool
	ShardNum   uint32                `mapstructure:"shard_num" yaml:"shard_num"`
	ShardID    uint32                `mapstructure:"shard_id" yaml:"shard_id"`
	PolicyType shard.ShardPolicyType `mapstructure:"policy_type" yaml:"policy_type"`
	PolicyPath string                `mapstructure:"policy_path" yaml:"policy_path"`
}

// Net is the definition of node network.
type Net struct {
	Sync         Sync
	ZoneTemplate ZoneTemplate           `mapstructure:"zone_template" yaml:"zone_template"`
	ConsZone     map[string]interface{} `mapstructure:"cons_zone" yaml:"cons_zone"`
	Zones        []map[string]interface{}
	Domains      []Domain
}

// Sync is the definition of sync.
type Sync struct {
	TickInterval uint32 `mapstructure:"tick_interval" yaml:"tick_interval"`
	Heartbeat    uint32
	Timeout      uint32
	TimeoutCount uint32 `mapstructure:"timeout_count" yaml:"timeout_count"`
}

// ZoneTemplate  is the definition of zone template.
type ZoneTemplate struct {
	Attachable bool       `yaml:",omitempty"`
	BlockBatch uint32     `mapstructure:"block_batch" yaml:"block_batch"`
	Alloc      Allocator  `yaml:",omitempty"`
	LinkPolicy LinkPolicy `mapstructure:"link_policy" yaml:"link_policy"`
}

// Allocator is the definition of allocator.
type Allocator struct {
	Type           string                 `yaml:",omitempty"`
	CoordinatorNum uint32                 `mapstructure:"coordinator_num" yaml:"coordinator_num,omitempty"`
	PeerNum        uint32                 `mapstructure:"peer_num" yaml:"peer_num,omitempty"`
	Conf           map[string]interface{} `yaml:",omitempty"`
}

// LinkPolicy is the definition of link policy.
type LinkPolicy struct {
	Limit  uint32 `yaml:",omitempty"`
	Period uint32 `yaml:",omitempty"`
}

// Zone is the definition of zone.
type Zone struct {
	ID         string
	BlockBatch uint32 `mapstructure:"block_batch" yaml:"block_batch"`
	Alloc      map[string]interface{}
	Domain     []string
	Attachable bool
}

// Domain is the definition of domain.
type Domain struct {
	Path string
}

// DecryptFunc is the type for Decrypt Func
type DecryptFunc func(bytes []byte) ([]byte, error)
