/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

package config

import (
	"io/ioutil"
	"path/filepath"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto/common"
)

type updateConfig struct {
	Orgs []*oneOrg
}

type oneOrg struct {
	Name      string
	Option    string
	RootPath  string
	AdminPath string
	TlsPath   string // nolint:golint,stylecheck // conflict with java.
}

// GetOrgUpdates get org updates form yaml
func GetOrgUpdates(path string, decrypt DecryptFunc) (*common.ConfigSet_OrgUpdates, error) {
	v, errMsg := parseCfg(path)
	if errMsg != nil {
		return nil, errors.WithMessage(errMsg, "parse config file error")
	}
	config := &updateConfig{}
	errMsg = unmarshal(v, &config)
	if errMsg != nil {
		return nil, errors.WithMessage(errMsg, "error unmarshal config into struct")
	}

	updates := make([]*common.ConfigSet_OrgUpdate, 0, len(config.Orgs))
	for _, org := range config.Orgs {
		update, err := appendUpdate(org, decrypt)
		if err != nil {
			return nil, err
		}
		updates = append(updates, update)
	}
	return &common.ConfigSet_OrgUpdates{OrgUpdate: updates}, nil
}

func appendUpdate(param *oneOrg, decrypt DecryptFunc) (*common.ConfigSet_OrgUpdate, error) {
	if param.Name == "" {
		return nil, errors.New("invalid org, org name should not be empty")
	}
	org := &common.Organization{Name: param.Name}
	var updates *common.ConfigSet_OrgUpdate
	switch param.Option {
	case "remove":
		updates = &common.ConfigSet_OrgUpdate{
			Operation:    common.OP_REMOVE,
			Organization: org,
		}
	case "append":
		err := loadOrgCerts(param, org, decrypt)
		if err != nil {
			return nil, err
		}
		updates = &common.ConfigSet_OrgUpdate{
			Operation:    common.OP_APPEND,
			Organization: org,
		}
	case "replace":
		err := replaceOrgCerts(param, org, decrypt)
		if err != nil {
			return nil, err
		}
		updates = &common.ConfigSet_OrgUpdate{
			Operation:    common.OP_REPLACE,
			Organization: org,
		}
	case "freeze":
		updates = &common.ConfigSet_OrgUpdate{
			Operation:    common.OP_FREEZE,
			Organization: org,
		}
	case "unfreeze":
		updates = &common.ConfigSet_OrgUpdate{
			Operation:    common.OP_UNFREEZE,
			Organization: org,
		}
	default:
		return nil, errors.New("unknown operation")
	}
	return updates, nil
}

func loadOrgCerts(org *oneOrg, o *common.Organization, decrypt DecryptFunc) error {
	opts := []certOption{
		withRootCert(org.RootPath),
		withAdminCert(org.AdminPath),
		withTLSRootCert(org.TlsPath),
	}
	for _, opt := range opts {
		if err := opt(o, decrypt); err != nil {
			return err
		}
	}
	return nil
}

func replaceOrgCerts(org *oneOrg, o *common.Organization, decrypt DecryptFunc) error {
	var opts []certOption
	if org.RootPath != "" {
		opts = append(opts, withRootCert(org.RootPath))
	}
	if org.AdminPath != "" {
		opts = append(opts, withAdminCert(org.AdminPath))
	}
	if org.TlsPath != "" {
		opts = append(opts, withTLSRootCert(org.TlsPath))
	}
	if len(opts) == 0 {
		return errors.New("replace must contain at least one of root, admin or tls")
	}
	for _, opt := range opts {
		if err := opt(o, decrypt); err != nil {
			return err
		}
	}
	return nil
}

type certOption func(org *common.Organization, decrypt DecryptFunc) error

func withRootCert(path string) certOption {
	return func(org *common.Organization, decrypt DecryptFunc) error {
		cert, err := readFile(path, decrypt)
		if err != nil {
			return errors.WithMessage(err, "read root cert error")
		}
		org.RootCert = cert
		return nil
	}
}

func withAdminCert(path string) certOption {
	return func(org *common.Organization, decrypt DecryptFunc) error {
		cert, err := readFile(path, decrypt)
		if err != nil {
			return errors.WithMessage(err, "read admin cert error")
		}
		org.AdminCert = cert
		return nil
	}
}

func withTLSRootCert(path string) certOption {
	return func(org *common.Organization, decrypt DecryptFunc) error {
		cert, err := readFile(path, decrypt)
		if err != nil {
			return errors.WithMessage(err, "read tls cert error")
		}
		org.TLSRootCert = cert
		return nil
	}
}

func readFile(path string, decrypt DecryptFunc) ([]byte, error) {
	cert, err := ioutil.ReadFile(filepath.Clean(filepath.FromSlash(path)))
	if err != nil {
		return nil, errors.Wrap(err, "read file error")
	}
	return decrypt(cert)
}
