/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2020. All rights reserved.
 */

// Package crypto provide the implementation of crypto.
package crypto

import (
	"crypto/tls"
	"crypto/x509"
	"io/ioutil"
	"path/filepath"

	"github.com/pkg/errors"
	"google.golang.org/grpc/credentials"

	"huawei.com/huaweichain/common/cryptomgr"
	"huawei.com/huaweichain/common/cryptomgr/cryptoimpl"
)

var registry = make(map[string]func(string, *TLS) (credentials.TransportCredentials, error))

// nolint:gochecknoinits // register ecdsa crypto for tls configuration.
func init() {
	registry[cryptomgr.EcdsaWithSha256] = func(serverName string, t *TLS) (credentials.TransportCredentials, error) {
		c, err := t.initTLSConfig(serverName)
		if err != nil {
			return nil, err
		}
		return credentials.NewTLS(c), nil
	}
}

// Info is the definition of Info
type Info struct {
	alg    string
	cert   cryptomgr.Cert
	key    cryptomgr.Key
	hashFn func([]byte) []byte
}

// CleanSenInfo clean sensitive info in Crypto object
func (c *Info) CleanSenInfo() {
	if c.key != nil {
		c.key.CleanKeySensitiveInfo()
		c.key = nil
	}
}

// Sign is used to sign for a message.
func (c *Info) Sign(message []byte) ([]byte, error) {
	return c.key.Sign(message, c.alg)
}

// GetCertificate is used to get certificate.
func (c *Info) GetCertificate() ([]byte, error) {
	return c.cert.GetPemCertBytes(), nil
}

// GetCommonName is used to get common name from certificate.
func (c *Info) GetCommonName() string {
	return c.cert.GetCommonName()
}

// GetOrg is used to get organization from certificate.
func (c *Info) GetOrg() string {
	if len(c.cert.GetOrganization()) == 0 {
		return ""
	}
	return c.cert.GetOrganization()[0]
}

// Hash is the function to compute hash. It could be sha256 or sm3 depends on the identity
// algorithm config.
func (c *Info) Hash(data []byte) []byte {
	return c.hashFn(data)
}

// Crypto is the definition of crypto interface.
type Crypto interface {
	Sign(message []byte) ([]byte, error)
	GetCertificate() ([]byte, error)
	GetCommonName() string
	GetOrg() string
	Hash(data []byte) []byte
	CleanSenInfo()
}

// NewCrypto is used to new an instance of crypto.
func NewCrypto(alg string, certPath string, keyPath string,
	decrypt func(bytes []byte) ([]byte, error)) (Crypto, error) {
	certPem, err := getPemInfoFromPath(certPath, decrypt)
	if err != nil {
		return nil, errors.WithMessage(err, "getPemInfoFromPath error")
	}
	keyPem, err := getPemInfoFromPath(keyPath, decrypt)
	if err != nil {
		return nil, errors.WithMessage(err, "getPemInfoFromPath error")
	}
	c, err := NewCryptoWithIdentity(alg, certPem, keyPem)
	if err != nil {
		cryptoimpl.CleanBytes(keyPem)
		return nil, err
	}
	return c, nil
}

// NewCryptoWithIdentity is used to create an instance by specify certificate and key byte array.
func NewCryptoWithIdentity(alg string, certPem []byte, keyPem []byte) (Crypto, error) {
	certFactory := cryptoimpl.GetCertFactory(alg)
	if certFactory == nil {
		return nil, errors.New("not supported crypto algorithm for cert factory")
	}
	keyFactory := cryptoimpl.GetAsKeyFactory(alg)
	if keyFactory == nil {
		return nil, errors.New("not supported crypto algorithm for key factory")
	}
	hashFn := cryptoimpl.GetHashFunc(cryptoimpl.GetCfgHashAlg(alg))
	if hashFn == nil {
		return nil, errors.New("not supported crypto algorithm for hash function")
	}

	cert, err := certFactory.GetCertFromPem(certPem)
	if err != nil {
		return nil, errors.WithMessage(err, "GetCertFromPem error")
	}
	key, err := keyFactory.GetKeyFromPem(keyPem)
	if err != nil {
		return nil, errors.WithMessage(err, "GetKeyFromPem error")
	}
	return &Info{
		alg:    alg,
		cert:   cert,
		key:    key,
		hashFn: hashFn,
	}, nil
}

func getPemInfoFromPath(path string, decrypt func(bytes []byte) ([]byte, error)) ([]byte, error) {
	info, err := ioutil.ReadFile(filepath.Clean(filepath.FromSlash(path)))
	if err != nil {
		return nil, errors.WithMessage(err, "read file error")
	}
	info, err = decrypt(info)
	if err != nil {
		return nil, errors.WithMessage(err, "decrypt message error")
	}
	return info, nil
}

// TLS is the definition of TLS.
type TLS struct {
	Algorithm    string
	certPEMBlock []byte
	keyPEMBlock  []byte
	roots        [][]byte
}

// NewTLS is use to new an instance of TLS.
func NewTLS(certPEMBlock []byte, keyPEMBlock []byte, roots [][]byte, algorithm string) *TLS {
	return &TLS{
		Algorithm:    algorithm,
		certPEMBlock: certPEMBlock,
		keyPEMBlock:  keyPEMBlock,
		roots:        roots,
	}
}

// GetTransportCredentials from TLS configuration for grpc tls.
func (t *TLS) GetTransportCredentials(serverName string) (credentials.TransportCredentials, error) {
	tc, ok := registry[t.Algorithm]
	if !ok {
		return nil, errors.Errorf("unsupported algorithm %s", t.Algorithm)
	}
	return tc(serverName, t)
}

// EcTLSClientConf creates tls configuration for ECDSA.
func EcTLSClientConf(serverName string, t *TLS) (*tls.Config, error) {
	return t.initTLSConfig(serverName)
}

func (t *TLS) initTLSConfig(serverName string) (*tls.Config, error) {
	cert, err := tls.X509KeyPair(t.certPEMBlock, t.keyPEMBlock)
	if err != nil {
		return nil, errors.WithMessage(err, "x509 key pair error")
	}
	certPool := x509.NewCertPool()
	for _, root := range t.roots {
		if !certPool.AppendCertsFromPEM(root) {
			return nil, errors.New("fail to append root ca")
		}
	}

	return &tls.Config{
		ServerName:               serverName,
		Certificates:             []tls.Certificate{cert},
		RootCAs:                  certPool,
		MinVersion:               tls.VersionTLS12,
		PreferServerCipherSuites: true,
		CipherSuites: []uint16{
			tls.TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,
			tls.TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,
			tls.TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256,
			tls.TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384,
		},
	}, nil
}

// CleanSenInfo clean sensitive info in TLS
func (t *TLS) CleanSenInfo() {
	cryptoimpl.CleanBytes(t.keyPEMBlock)
	t.certPEMBlock = nil
}
