/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package flow

import (
	"context"

	"github.com/gogo/protobuf/proto"
	"github.com/pkg/errors"
	"google.golang.org/grpc"

	"huawei.com/huaweichain/proto/archive"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/flow/internal"
)

// ArchiveFlow provides some functions for archive snapshot.
type ArchiveFlow struct {
	stub  internal.Stub
	chain string
	node  string
}

func newArchiveFlow(stub internal.Stub, chain string, node string) *ArchiveFlow {
	return &ArchiveFlow{
		stub:  stub,
		chain: chain,
		node:  node,
	}
}

func (a *ArchiveFlow) getClient() (nodeservice.ArchiveServiceClient, error) {
	node, err := a.stub.GetNode(a.node)
	if err != nil {
		return nil, errors.WithMessagef(err, "failed to fetch client for remote node %s", a.node)
	}
	conn, err := node.ChainAction.GetReadyConn()
	if err != nil {
		return nil, errors.WithMessage(err, "failed to get ready connection for grpc")
	}
	client := nodeservice.NewArchiveServiceClient(conn)
	return client, nil
}

type archiveClientFunc = func(client nodeservice.ArchiveServiceClient) archiveFunc
type archiveFunc = func(ctx context.Context, in *common.RawMessage, opts ...grpc.CallOption) (*common.RawMessage, error)

func (a *ArchiveFlow) invokeTask(ctx context.Context, method archiveFunc, msg *common.RawMessage) (string, error) {
	res, err := a.invoke(ctx, method, msg)
	if err != nil {
		return "", err
	}
	taskID := res.StatusInfo
	return taskID, nil
}

func (a *ArchiveFlow) queryCheckpoint(ctx context.Context, method archiveFunc, msg *common.RawMessage) (
	[]uint64, error) {
	res, err := a.invoke(ctx, method, msg)
	if err != nil {
		return nil, err
	}
	var cps nodeservice.ArchiveCheckPoints
	err = proto.Unmarshal(res.Payload, &cps)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to parse checkpoints")
	}
	return cps.Checkpoints, nil
}

func (a *ArchiveFlow) invoke(parent context.Context, method archiveFunc, msg *common.RawMessage) (
	*common.Response, error) {
	ctx, cancel := context.WithCancel(parent)
	defer cancel()
	resMsg, err := method(ctx, msg)
	if err != nil {
		return nil, errors.Wrapf(err, "failed to invoke archive request")
	}
	res, err := internal.HandleResult(resMsg)
	if err != nil {
		return nil, err
	}
	return res, nil
}

// Snapshot calls specific node to start a snapshot task with block number.
// If snapshot task is accepted, returns task id. Use compact with true, the snapshot will be compacted.
func (a *ArchiveFlow) Snapshot(ctx context.Context, checkpoint uint64, compact bool) (string, error) {
	req := nodeservice.ArchiveRequest{
		ChainId:    a.chain,
		CheckPoint: checkpoint,
		IsCompact:  compact,
	}
	msg, err := a.stub.Seal(&req)
	if err != nil {
		return "", err
	}
	client, err := a.getClient()
	if err != nil {
		return "", err
	}
	return a.invokeTask(ctx, client.Snapshot, msg)
}

// Delete calls specific node to start a task for deleting a snapshot with checkpoint.
// If snapshot task is accepted, returns task id.
func (a *ArchiveFlow) Delete(ctx context.Context, checkpoint uint64) (string, error) {
	f := func(client nodeservice.ArchiveServiceClient) archiveFunc {
		return client.Delete
	}
	return a.sendTask(ctx, checkpoint, f)
}

// Recover calls specific node to start a task for recovering state by a snapshot with checkpoint.
// If snapshot task is accepted, returns task id.
func (a *ArchiveFlow) Recover(ctx context.Context, checkpoint uint64) (string, error) {
	f := func(client nodeservice.ArchiveServiceClient) archiveFunc {
		return client.Recover
	}
	return a.sendTask(ctx, checkpoint, f)
}

func (a *ArchiveFlow) sendTask(ctx context.Context, checkpoint uint64, f archiveClientFunc) (string, error) {
	req := nodeservice.ArchiveRequest{
		ChainId:    a.chain,
		CheckPoint: checkpoint,
	}
	msg, err := a.stub.Seal(&req)
	if err != nil {
		return "", err
	}
	client, err := a.getClient()
	if err != nil {
		return "", err
	}
	return a.invokeTask(ctx, f(client), msg)
}

// Store calls specific node to start a task for storing a snapshot with checkpoint.
// If snapshot task is accepted, returns task id.
func (a *ArchiveFlow) Store(ctx context.Context, checkpoint uint64) (string, error) {
	f := func(client nodeservice.ArchiveServiceClient) archiveFunc {
		return client.Store
	}
	return a.sendTask(ctx, checkpoint, f)
}

// TaskStatus calls specific node to fetch a task status with task id.
func (a *ArchiveFlow) TaskStatus(ctx context.Context, taskID string) (*archive.Task, error) {
	req := nodeservice.TaskRequest{
		ChainId: a.chain,
		TaskId:  taskID,
	}
	msg, err := a.stub.Seal(&req)
	if err != nil {
		return nil, err
	}
	client, err := a.getClient()
	if err != nil {
		return nil, err
	}
	res, err := a.invoke(ctx, client.GetTask, msg)
	if err != nil {
		return nil, err
	}
	var task archive.Task
	err = proto.Unmarshal(res.Payload, &task)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to parse task status info")
	}
	return &task, nil
}

// CurrentTasks calls specific node to fetch all uncompleted task status.
func (a *ArchiveFlow) CurrentTasks(ctx context.Context) ([]*archive.Task, error) {
	msg, client, err := a.prepareRequest()
	if err != nil {
		return nil, err
	}
	res, err := a.invoke(ctx, client.GetCurTasks, msg)
	if err != nil {
		return nil, err
	}
	var tasks nodeservice.ArchiveTasks
	err = proto.Unmarshal(res.Payload, &tasks)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to parse task list")
	}
	return tasks.Tasks, nil
}

// OnChainCheckpoint calls specific node to fetch all checkpoint block number on chain.
func (a *ArchiveFlow) OnChainCheckpoint(ctx context.Context) ([]uint64, error) {
	msg, client, err := a.prepareRequest()
	if err != nil {
		return nil, err
	}
	return a.queryCheckpoint(ctx, client.GetOnChainCheckpoints, msg)
}

func (a *ArchiveFlow) prepareRequest() (*common.RawMessage, nodeservice.ArchiveServiceClient, error) {
	req := nodeservice.ArchiveRequest{
		ChainId: a.chain,
	}
	msg, err := a.stub.Seal(&req)
	if err != nil {
		return nil, nil, err
	}
	client, err := a.getClient()
	if err != nil {
		return nil, nil, err
	}
	return msg, client, nil
}

// OffChainCheckpoint calls specific node to fetch all checkpoint block number off chain.
func (a *ArchiveFlow) OffChainCheckpoint(ctx context.Context) ([]uint64, error) {
	msg, client, err := a.prepareRequest()
	if err != nil {
		return nil, err
	}
	return a.queryCheckpoint(ctx, client.GetOffChainCheckpoints, msg)
}
