/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

// Package flow is the implementation of flow sdk for go client.
package flow

import (
	"context"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/logger"
	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/proto/shard"
	"huawei.com/huaweichain/sdk/config"
	"huawei.com/huaweichain/sdk/flow/internal"
	"huawei.com/huaweichain/sdk/flow/status"
	"huawei.com/huaweichain/sdk/node"
	"huawei.com/huaweichain/sdk/utils"
)

// ChainFlow is used to do the chain related operation
// which includes create, delete and query
type ChainFlow struct {
	stub       internal.Stub
	node       string
	genesis    string // chain config path
	decrypt    func(bytes []byte) ([]byte, error)
	entrypoint *common.Entrypoint
	shardID    string
	p2pSeeds   []*P2PSeed

	lg *logger.WienerLogger
}

// newChainFlow is used to return a new ChainFlow
func newChainFlow(stub internal.Stub, node string) *ChainFlow {
	return &ChainFlow{
		stub: stub,
		node: node,
		decrypt: func(bytes []byte) ([]byte, error) {
			return bytes, nil
		},
		lg: logger.GetModuleLogger("sdk.flow", "chain"),
	}
}

// Decrypt is used to set decrypt for sdk
func (c *ChainFlow) Decrypt(decrypt func(bytes []byte) ([]byte, error)) *ChainFlow {
	c.decrypt = decrypt
	return c
}

// Entrypoint is used to set EntryPoint for sdk
func (c *ChainFlow) Entrypoint(zone string, master bool, seeds []string) *ChainFlow {
	e := &common.Entrypoint{
		ZoneId:        zone,
		Coordinator:   false, // auto assign, no need to set
		InitialMaster: master,
	}
	if master {
		e.Linkers = seeds
	} else {
		e.Seeds = seeds
	}
	c.entrypoint = e
	return c
}

// GenesisFilePath is used to set genesis path for sdk
func (c *ChainFlow) GenesisFilePath(genesis string) *ChainFlow {
	c.genesis = genesis
	return c
}

// P2PSeeds is used to set p2p seed nodes for sdk
func (c *ChainFlow) P2PSeeds(p2pSeeds ...*P2PSeed) *ChainFlow {
	c.p2pSeeds = p2pSeeds
	return c
}

// ShardID is used to set shard id for sdk
func (c *ChainFlow) ShardID(shardID string) *ChainFlow {
	c.shardID = shardID
	return c
}

// Create is used to create chain for sdk
func (c *ChainFlow) Create(ctx context.Context, chain string) error {
	// read chain config from file
	genesisConfig, err := config.NewGenesisConfig(c.genesis, c.decrypt)
	if err != nil {
		return errors.WithMessage(err, "new genesis config error")
	}

	genesisBlock, err := c.stub.ChainBuilder().BuildGenesisBlockWithGenesisConfig(chain, genesisConfig)
	if err != nil {
		return errors.WithMessage(err, "build genesis block error")
	}

	genesisBlockBytes, err := genesisBlock.Marshal()
	if err != nil {
		return errors.WithMessage(err, "marshal genesis block error")
	}
	// construct raw message
	rawMsg, err := c.constructRawMessage(genesisBlockBytes)
	if err != nil {
		return errors.WithMessage(err, "construct raw message error")
	}
	// send message
	rpc, err := c.client()
	if err != nil {
		return errors.WithMessage(err, "client error")
	}
	responseMsg, err := rpc.CreateChain(ctx, rawMsg)
	if err != nil {
		return errors.WithMessage(err, "chain action join chain error")
	}
	var res common.Response
	if err := proto.Unmarshal(responseMsg.Payload, &res); err != nil {
		return errors.WithMessage(err, "unmarshal response error")
	}
	return status.New(res.Status, res.StatusInfo).Err()
}

func (c *ChainFlow) constructRawMessage(genesisBlockBytes []byte) (*common.RawMessage, error) {
	var rawMsg *common.RawMessage
	var err error
	if len(c.p2pSeeds) == 0 {
		rawMsg, err = c.stub.ChainBuilder().BuildJoinMsgWithEntrypoint(genesisBlockBytes, c.entrypoint)
		if err != nil {
			return nil, errors.WithMessage(err, "build join msg with entrypoint error")
		}
	} else {
		p2pSeedsByte, err := c.shardInfoBytes()
		if err != nil {
			return nil, errors.WithMessage(err, "shard info bytes error")
		}
		rawMsg, err = c.stub.ChainBuilder().BuildJoinMsgWithShardInfo(genesisBlockBytes, p2pSeedsByte)
		if err != nil {
			return nil, errors.WithMessage(err, "build join msg with shard info error")
		}
	}

	return rawMsg, nil
}

func (c *ChainFlow) shardInfoBytes() ([]byte, error) {
	shardNodes := make([]*shard.ShardNode, 0, len(c.p2pSeeds))
	for _, p2pSeedNode := range c.p2pSeeds {
		shardNodes = append(shardNodes, &shard.ShardNode{
			Host: p2pSeedNode.Host,
			Port: p2pSeedNode.Port,
		})
	}
	shardInfo := &shard.ShardInfo{
		ShardNodes: shardNodes,
	}
	shardInfoBytes, err := proto.Marshal(shardInfo)
	if err != nil {
		return nil, errors.WithMessage(err, "marshal shard info error")
	}

	return shardInfoBytes, nil
}

// Join is used to join chain for sdk
func (c *ChainFlow) Join(ctx context.Context, chain, existedNode string) error {
	if c.shardID != "" {
		shardChainID, err := utils.ShardChainID(chain, c.shardID)
		if err != nil {
			return errors.WithMessage(err, "shard chain id error")
		}
		chain = shardChainID
	}
	genesis, err := c.genesisBlock(ctx, chain, existedNode)
	if err != nil {
		return errors.WithMessage(err, "get genesis block response error")
	}
	chainInfo, infoErr := c.chainConf(ctx, chain, existedNode)
	if infoErr != nil {
		return errors.WithMessage(infoErr, "get config info error")
	}

	if c.entrypoint == nil {
		return errors.WithMessage(err, "need to set entryPoint config")
	}
	rawReq, err := c.stub.ChainBuilder().BuildJoinMsgWithLatestConf(genesis, chainInfo, c.entrypoint)
	if err != nil {
		return errors.WithMessage(err, "build join chain raw message error")
	}
	rpc, err := c.client()
	if err != nil {
		return err
	}
	rawRes, err := rpc.CreateChain(ctx, rawReq)
	if err != nil {
		return errors.WithMessage(err, "chain action join chain error")
	}
	var res common.Response
	if err = proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return errors.WithMessage(err, "unmarshal response error")
	}
	return status.New(res.Status, res.StatusInfo).Err()
}

func (c *ChainFlow) check() error {
	if c.entrypoint == nil {
		return errors.New("entrypoint info not found")
	}
	if c.node == "" {
		return errors.New("target node info not found")
	}
	return nil
}

// JoinWithSnap join chain with snapshot.
func (c *ChainFlow) JoinWithSnap(ctx context.Context, chain, existNode string) (Future, error) {
	if err := c.check(); err != nil {
		return nil, err
	}
	var err error
	ctx, cancel := context.WithCancel(ctx)
	defer func() {
		if err != nil {
			cancel()
		}
	}()

	c.lg.Debugf("begin to fetch genesis block from node %s", existNode)
	genesis, err := c.genesisBlock(ctx, chain, existNode)
	if err != nil {
		return nil, errors.WithMessage(err, "get genesis block response error")
	}
	c.lg.Debugf("begin to fetch latest chain configuration from node %s", existNode)
	chainInfo, err := c.chainConf(ctx, chain, existNode)
	if err != nil {
		return nil, errors.WithMessagef(err, "failed to fetch latest chain configuration from node %s", existNode)
	}

	source, err := c.newHostInfo(existNode)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to construct host info")
	}
	sourceBytes, err := proto.Marshal(source)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to marshal host info")
	}

	req, err := c.stub.ChainBuilder().BuildJoinMsgWithLatestConf(genesis, chainInfo, c.entrypoint, sourceBytes)
	if err != nil {
		return nil, errors.WithMessage(err,
			"failed to build request raw message to join chain with latest chain configuration")
	}
	return c.joinWithSnap(ctx, cancel, chain, req, existNode)
}

func (c *ChainFlow) joinWithSnap(ctx context.Context, cancel context.CancelFunc, chain string, req *common.RawMessage,
	existNode string) (Future, error) {
	sc, err := c.client()
	if err != nil {
		return nil, err
	}
	stream, err := sc.JoinChainWithSnapshot(ctx, req)
	if err != nil {
		return nil, errors.WithMessage(err, "chain action join chain error")
	}
	ps := joinParams{
		chain:     chain,
		node:      c.node,
		existNode: existNode,
		stub:      c.stub,
		stream:    stream,
	}
	return joinWithSnap(ctx, cancel, &ps), nil
}

func (c *ChainFlow) client() (nodeservice.ChainManagerClient, error) {
	n, err := c.stub.GetNode(c.node)
	if err != nil {
		return nil, err
	}
	return c.nodeClient(n)
}

func (c *ChainFlow) nodeClient(n *node.WNode) (nodeservice.ChainManagerClient, error) {
	cc, err := n.ChainAction.GetReadyConn()
	if err != nil {
		return nil, err
	}
	return nodeservice.NewChainManagerClient(cc), nil
}

func (c *ChainFlow) newHostInfo(existedNode string) (*nodeservice.HostInfo, error) {
	eNode, err := c.stub.GetNode(existedNode)
	if err != nil {
		return nil, errors.WithMessagef(err, "exist node not exist： %v", existedNode)
	}
	return &nodeservice.HostInfo{
		Host: eNode.Host,
		Port: uint32(eNode.Port),
	}, nil
}

func (c *ChainFlow) genesisBlock(ctx context.Context, chain, existedNode string) ([]byte, error) {
	genesis, err := loadStub(c.stub).User().Observer().Target(chain, existedNode).Genesis(ctx)
	if err != nil {
		return nil, err
	}
	return proto.Marshal(genesis)
}

// chainConf is used to query chain for sdk
func (c *ChainFlow) chainConf(ctx context.Context, chain, node string) (*common.ConfigInfo, error) {
	info, err := c.chainInfo(ctx, chain, node)
	if err != nil {
		return nil, err
	}
	if info == nil {
		return nil, errors.Errorf("empty chain information from %s for chain %s", node, chain)
	}
	return info.Config, nil
}

func (c *ChainFlow) chainInfo(ctx context.Context, chain string, node string) (*nodeservice.ChainInfo, error) {
	wNode, err := c.stub.GetNode(node)
	if err != nil {
		return nil, err
	}
	req, err := c.stub.ChainBuilder().BuildQueryChainRawMessage(chain)
	if err != nil {
		return nil, errors.WithMessage(err, "build query single chain raw message error")
	}
	rpc, err := c.nodeClient(wNode)
	if err != nil {
		return nil, err
	}
	rawRes, err := rpc.QueryChainInfo(ctx, req)
	if err != nil {
		return nil, errors.WithMessage(err, "chain action query single chain error")
	}
	var res common.Response
	if err := proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return nil, errors.WithMessage(err, "unmarshal response error")
	}
	if res.Status != common.SUCCESS {
		return nil, status.New(res.Status, res.StatusInfo).Err()
	}
	var queryInfoResponse nodeservice.QueryInfoResponse
	if err := proto.Unmarshal(res.Payload, &queryInfoResponse); err != nil {
		return nil, errors.WithMessage(err, "unmarshal response error")
	}
	return queryInfoResponse.ChainInfo, nil
}

// Exist returns true if the node has been joined chain.
func (c *ChainFlow) Exist(ctx context.Context, chain string) (bool, error) {
	info, err := c.chainInfo(ctx, chain, c.node)
	if err != nil {
		return false, errors.WithMessagef(err, "failed to find chain in node %s", c.node)
	}
	return info != nil, nil
}

// ChainConf is used to query chain for sdk
func (c *ChainFlow) ChainConf(ctx context.Context, chain string) (*common.ConfigInfo, error) {
	return c.chainConf(ctx, chain, c.node)
}

// Quit is used to delete chain for sdk
func (c *ChainFlow) Quit(ctx context.Context, chain string) error {
	rawReq, err := c.stub.ChainBuilder().BuildQuitChainRawMessage(chain)
	if err != nil {
		return errors.WithMessage(err, "build quit chain raw message error")
	}
	client, err := c.client()
	if err != nil {
		return err
	}
	rawRes, err := client.DeleteChain(ctx, rawReq)
	if err != nil {
		return errors.WithMessage(err, "chain action quit chain error")
	}
	var res common.Response
	if err := proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return err
	}
	if res.Status != common.SUCCESS {
		return status.ResponseToErr(&res)
	}
	return nil
}

// AllChainConf is used to query all chains for sdk
func (c *ChainFlow) AllChainConf(ctx context.Context) ([]*common.ConfigInfo, error) {
	rawMsg, err := c.stub.ChainBuilder().BuildQueryAllChainRawMessage()
	if err != nil {
		return nil, errors.WithMessage(err, "build query all chains raw message error")
	}
	client, err := c.client()
	if err != nil {
		return nil, err
	}
	rawRes, err := client.QueryAllChainInfos(ctx, rawMsg)
	if err != nil {
		return nil, errors.WithMessage(err, "chain action query all chains error")
	}
	var res common.Response
	if err := proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return nil, errors.WithMessage(err, "unmarshal response error")
	}
	if res.Status != common.SUCCESS {
		return nil, status.ResponseToErr(&res)
	}
	var info nodeservice.QueryAllInfoResponse
	if err := proto.Unmarshal(res.Payload, &info); err != nil {
		return nil, errors.WithMessage(err, "unmarshal query all info error")
	}
	var infos []*common.ConfigInfo
	for _, ci := range info.ChainInfos {
		infos = append(infos, ci.Config)
	}
	return infos, nil
}
