/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package flow

import (
	"regexp"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/util"
	"huawei.com/huaweichain/proto/nodeservice"
)

const (
	// CircleNameReg is the circle name regex format.
	CircleNameReg = "^[A-Za-z0-9][A-Za-z0-9_-]{0,99}$"
)

// checkCircleVoteParams check the format of circle vote parameters
func checkCircleVoteParams(circleConfig *nodeservice.CircleConfig) error {
	if err := util.CheckChainID(circleConfig.ChainId); err != nil {
		return errors.WithMessage(err, "check chain id error")
	}
	if err := checkCircleName(circleConfig.Name); err != nil {
		return errors.WithMessage(err, "check circle name error")
	}
	if err := util.CheckPolicy(circleConfig.Policy); err != nil {
		return errors.WithMessage(err, "check policy error")
	}
	if err := checkPeer(circleConfig.RequiredPeer, circleConfig.MaximumPeer); err != nil {
		return errors.WithMessage(err, "check peer error")
	}
	if err := checkCircleDBType(circleConfig.DbType); err != nil {
		return errors.WithMessage(err, "check db type error")
	}
	return nil
}

// checkCircleName check the circle name format
func checkCircleName(name string) error {
	r, err := regexp.Compile(CircleNameReg)
	if err != nil {
		return errors.WithMessage(err, "regexp compile error")
	}
	if !r.MatchString(name) {
		return errors.Errorf("check circle name not match format: %s", CircleNameReg)
	}
	return nil
}

// checkCircleName check the circle name format
func checkPeer(requiredPeer uint32, maxPeer uint32) error {
	if maxPeer < requiredPeer {
		return errors.New("circleInfo maximum peer count less than required")
	}
	return nil
}

func checkCircleDBType(dbType nodeservice.DBType) error {
	if nodeservice.DBType_name[int32(dbType)] == "" {
		return errors.Errorf("the circle dbType is %s", dbType)
	}
	return nil
}
