/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package flow

import (
	"context"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/proto/schema"
	"huawei.com/huaweichain/sdk/flow/internal"
	"huawei.com/huaweichain/sdk/flow/status"
	"huawei.com/huaweichain/sdk/flow/types"
	"huawei.com/huaweichain/sdk/rawmessage"
)

// ConfigFlow is used to do the config related operation
type ConfigFlow struct {
	stub  internal.Stub
	cons  *internal.Consenter
	chain string
}

func newConfigFlow(chain string, stub internal.Stub) *ConfigFlow {
	configFlowInstance := ConfigFlow{
		chain: chain,
		stub:  stub,
		cons:  internal.NewConsenterFlow(chain, stub),
	}
	return &configFlowInstance
}

// Reader assigns specific remote node for request.
func (c *ConfigFlow) Reader(node string) *ConfigReaderFlow {
	return &ConfigReaderFlow{
		chain: c.chain,
		stub:  c.stub,
		node:  node,
	}
}

// ConfigReaderFlow represents only read flow instance for chain configuration.
type ConfigReaderFlow struct {
	stub  internal.Stub
	chain string
	node  string
}

// CertStatus query all invalid certificates with serial number.
func (c *ConfigReaderFlow) CertStatus(ctx context.Context, org string) (frozen, revocation []string, err error) {
	rawReq, err := c.stub.QueryBuilder().BuildQueryCertStatus(c.chain, org)
	if err != nil {
		return
	}
	client, err := c.client()
	if err != nil {
		return
	}
	rawRes, err := client.Query(ctx, rawReq)
	if err != nil {
		return
	}
	var res common.Response
	if err = proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return nil, nil, errors.WithMessage(err, "failed to unmarshal response")
	}
	if res.Status != common.SUCCESS {
		return nil, nil, status.ResponseToErr(&res)
	}

	fs, err := c.extractList(&res, rawmessage.GetFrozenSerial)
	if err != nil {
		return nil, nil, err
	}
	rs, err := c.extractList(&res, rawmessage.GetRevocationSerial)
	if err != nil {
		return nil, nil, err
	}
	return fs, rs, nil
}

func (c *ConfigReaderFlow) client() (nodeservice.VoteManagerClient, error) {
	wNode, err := c.stub.GetNode(c.node)
	if err != nil {
		return nil, err
	}
	conn, err := wNode.QueryAction.GetReadyConn()
	if err != nil {
		return nil, err
	}
	return nodeservice.NewVoteManagerClient(conn), nil
}

func (c *ConfigReaderFlow) shardClient() (nodeservice.ShardServiceClient, error) {
	wNode, err := c.stub.GetNode(c.node)
	if err != nil {
		return nil, err
	}
	conn, err := wNode.QueryAction.GetReadyConn()
	if err != nil {
		return nil, err
	}
	client := nodeservice.NewShardServiceClient(conn)
	return client, nil
}

func (c *ConfigReaderFlow) circleClient() (nodeservice.CircleClient, error) {
	wNode, err := c.stub.GetNode(c.node)
	if err != nil {
		return nil, err
	}
	conn, err := wNode.QueryAction.GetReadyConn()
	if err != nil {
		return nil, err
	}
	client := nodeservice.NewCircleClient(conn)
	return client, nil
}

type serialFunc = func(payload []byte) (map[string]interface{}, error)

func (c *ConfigReaderFlow) extractList(res *common.Response, serial serialFunc) ([]string, error) {
	set, err := serial(res.Payload)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to get serial number")
	}
	list := make([]string, 0, len(set))
	for k := range set {
		list = append(list, k)
	}
	return list, nil
}

// ChainVoteStatus is used to query chain vote process
func (c *ConfigReaderFlow) ChainVoteStatus(ctx context.Context, subject types.Subject) ([]*nodeservice.Vote, error) {
	rawReq, err := c.stub.QueryBuilder().BuildQueryChainUpdateVoteRawMessage(c.chain, string(subject))
	if err != nil {
		return nil, errors.WithMessage(err, "build query chain update vote raw message error")
	}
	client, err := c.client()
	if err != nil {
		return nil, err
	}
	rawRes, err := client.Query(ctx, rawReq)
	if err != nil {
		return nil, errors.WithMessage(err, "chain action query get vote error")
	}
	var res common.Response
	if err = proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return nil, errors.WithMessage(err, "unmarshal response raw error: %v")
	}
	if res.Status != common.SUCCESS {
		return nil, status.ResponseToErr(&res)
	}
	var voteRes nodeservice.VoteQueryResponse
	if err = proto.Unmarshal(res.Payload, &voteRes); err != nil {
		return nil, errors.WithMessage(err, "unmarshal vote query response error")
	}
	return voteRes.Votes, nil
}

// ShardPolicyType is used to query shard policy type.
func (c *ConfigReaderFlow) ShardPolicyType(ctx context.Context) (string, error) {
	rawReq, err := c.stub.QueryBuilder().BuildShardPolicyRawMessage(c.chain)
	if err != nil {
		return "", errors.WithMessage(err, "errors occur in building query contract state raw message")
	}
	client, err := c.shardClient()
	if err != nil {
		return "", err
	}
	rawRes, err := client.GetShardPolicy(ctx, rawReq)
	if err != nil {
		return "", err
	}
	var res common.Response
	if err := proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return "", errors.WithMessage(err, "unmarshal response error")
	}
	if res.Status != common.SUCCESS {
		return "", status.ResponseToErr(&res)
	}
	shardPolicy := &nodeservice.ShardPolicy{}
	if err = proto.Unmarshal(res.Payload, shardPolicy); err != nil {
		return "", errors.WithMessage(err, "unmarshal shard policy error")
	}
	return shardPolicy.Result, nil
}

// Circles is used to query circle config list.
func (c *ConfigReaderFlow) Circles(ctx context.Context) (*nodeservice.CircleConfigs, error) {
	rawReq, err := c.stub.Seal(&nodeservice.CircleRequest{ChainId: c.chain})
	if err != nil {
		return nil, errors.WithMessage(err, "errors occur in building query circle list raw message")
	}
	client, err := c.circleClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get circle service client error")
	}
	rawRes, err := client.QueryChainCircles(ctx, rawReq)
	if err != nil {
		return nil, err
	}
	var res common.Response
	if err = proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return nil, err
	}
	if res.Status != common.SUCCESS {
		return nil, status.ResponseToErr(&res)
	}
	var circles nodeservice.CircleConfigs
	if err = proto.Unmarshal(res.Payload, &circles); err != nil {
		return nil, errors.WithMessage(err, "unmarshal circle error")
	}
	return &circles, nil
}

// Circle is used to query circle config.
func (c *ConfigReaderFlow) Circle(ctx context.Context, circleName string) (*nodeservice.CircleConfig, error) {
	if len(circleName) == 0 {
		return nil, errors.New("null circle name in query param")
	}
	rawReq, err := c.stub.Seal(&nodeservice.CircleRequest{ChainId: c.chain, CircleName: circleName})
	if err != nil {
		return nil, errors.WithMessage(err, "errors occur in building query circle list raw message")
	}
	client, err := c.circleClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get circle service client error")
	}
	rawRes, err := client.QueryCircleConfig(ctx, rawReq)
	if err != nil {
		return nil, err
	}
	var res common.Response
	if err = proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return nil, err
	}
	if res.Status != common.SUCCESS {
		return nil, status.ResponseToErr(&res)
	}
	var circle nodeservice.CircleConfig
	if err = proto.Unmarshal(res.Payload, &circle); err != nil {
		return nil, errors.WithMessage(err, "unmarshal circle error")
	}
	return &circle, nil
}

// UpdatePolicy is used to update policy
func (c *ConfigFlow) UpdatePolicy(ctx context.Context, policy string) (*types.Receipt, error) {
	rawMsg, err := c.stub.ConfigBuilder().UpdateConfig.BuildUpdateConfPolicyRawMessage(c.chain, policy)
	if err != nil {
		return nil, errors.WithMessage(err, "build update config policy raw message error")
	}
	return c.cons.SendTx(ctx, rawMsg)
}

// UpdateLifecyclePolicy is used to update lifecycle
func (c *ConfigFlow) UpdateLifecyclePolicy(ctx context.Context, policy string) (*types.Receipt, error) {
	rawMsg, err := c.stub.ConfigBuilder().UpdateConfig.BuildUpdateLifecycleRawMessage(c.chain, policy)
	if err != nil {
		return nil, errors.WithMessage(err, "build update lifecycle raw message error")
	}
	return c.cons.SendTx(ctx, rawMsg)
}

// MinPlatformVersion sets minimum platform version for compatibility.
func (c *ConfigFlow) MinPlatformVersion(ctx context.Context, version string) (*types.Receipt, error) {
	rawReq, err := c.stub.ConfigBuilder().UpdateConfig.BuildUpdateMinPltfVersionRawMessage(c.chain, version)
	if err != nil {
		return nil, err
	}
	return c.cons.SendTx(ctx, rawReq)
}

// VoteType represents vote transaction type.
type VoteType int

const (
	// ConfigVote represents the vote tx is vote command.
	ConfigVote VoteType = iota
	// ConfigComplete represents the vote tx is complete command.
	ConfigComplete
)

// AddConsenter sends vote transaction to change consensus configuration to add consensus node.
func (c *ConfigFlow) AddConsenter(ctx context.Context, vote VoteType, consenter ...Consenter) (*types.Receipt, error) {
	ccs := make([]*schema.Consenter, len(consenter))
	for i := 0; i < len(consenter); i++ {
		cons := &consenter[i]
		ccs[i] = &schema.Consenter{
			Name:     cons.Name,
			Org:      cons.Org,
			Host:     cons.Addr,
			Port:     uint64(cons.RpcPort),
			ReqPort:  cons.ReqPort,
			RestPort: cons.RestPort,
			ReeCert:  cons.PemCert,
			TeeCert:  cons.PemCert,
		}
	}
	updates := schema.ConsenterUpdates{
		Op:         schema.AddVoter,
		Consenters: ccs,
	}
	return c.consenterUpdate(ctx, &updates, convertConsenterVoteHandler(vote))
}

func convertConsenterVoteHandler(vote VoteType) string {
	var handler string
	switch vote {
	case ConfigVote:
		handler = rawmessage.VoteHandler
	case ConfigComplete:
		handler = rawmessage.CompleteHandler
	default:
		log.Panicf("invalid vote type %d", vote)
	}
	return handler
}

func (c *ConfigFlow) getConsensus(ctx context.Context) (schema.ConsType, error) {
	n, err := c.cons.Consenter()
	if err != nil {
		return 0, err
	}
	config, err := loadStub(c.stub).Admin().Chain(n.ID).ChainConf(ctx, c.chain)
	if err != nil {
		return 0, err
	}
	var cons schema.SysCons
	if err := proto.Unmarshal(config.Config.Consensus, &cons); err != nil {
		return 0, err
	}
	return cons.Type, nil
}

func (c *ConfigFlow) consenterUpdate(ctx context.Context, updates *schema.ConsenterUpdates, handler string) (
	*types.Receipt, error) {
	consType, err := c.getConsensus(ctx)
	if err != nil {
		return nil, errors.Wrap(err, "failed to get consensus")
	}
	var update schema.ConsensusUpdate
	err = c.checkCons(&update, consType, updates)
	if err != nil {
		return nil, err
	}

	updateBytes, err := proto.Marshal(&update)
	if err != nil {
		return nil, errors.Wrapf(err, "failed to marshal %T", updates)
	}
	set := common.ConfigSet{
		Value: &common.ConfigSet_ConsenterUpdate{
			ConsenterUpdate: updateBytes,
		},
	}
	setBytes, err := proto.Marshal(&set)
	if err != nil {
		return nil, errors.Wrapf(err, "failed to marshal %T", set)
	}
	builder := rawmessage.NewMsgBuilderImpl(c.stub.Crypto())
	data := common.VoteTxData{
		Handler: handler,
		Payload: setBytes,
	}
	tx, err := builder.BuildVoteTx(c.chain, &data)
	if err != nil {
		return nil, errors.WithMessage(err, "build add tx error")
	}
	msg, err := builder.GetTxRawMsg(tx)
	if err != nil {
		return nil, err
	}
	return c.cons.SendTx(ctx, msg)
}

func (c *ConfigFlow) checkCons(update *schema.ConsensusUpdate,
	consType schema.ConsType, updates *schema.ConsenterUpdates) error {
	switch consType {
	case schema.Raft:
		raftConfChange := make([]*schema.ConsenterUpdates, 0)
		raftConfChange = append(raftConfChange, updates)
		cc := schema.RaftConsenterUpdates{
			RaftConfChange: raftConfChange,
		}
		*update = schema.ConsensusUpdate{
			Update: &schema.ConsensusUpdate_RaftConfChange{
				RaftConfChange: &cc,
			},
		}
	case schema.Hotstuff:
		*update = schema.ConsensusUpdate{
			Update: &schema.ConsensusUpdate_ConfChange{
				ConfChange: updates,
			},
		}
	default:
		return errors.New("config change can only handle raft and hotstuff")
	}
	return nil
}

// RemoveConsenter sends vote transaction to change consensus configuration to remove consensus node.
func (c *ConfigFlow) RemoveConsenter(ctx context.Context, vote VoteType, consenter ...string) (*types.Receipt, error) {
	ccs := make([]*schema.Consenter, len(consenter))
	for i, cons := range consenter {
		ccs[i] = &schema.Consenter{Name: cons}
	}
	updates := schema.ConsenterUpdates{
		Op:         schema.Remove,
		Consenters: ccs,
	}
	return c.consenterUpdate(ctx, &updates, convertConsenterVoteHandler(vote))
}

// Consenter is used to set consenter list
func (c *ConfigFlow) Consenter(consenter ...string) *ConfigFlow {
	c.cons.AvailableConsenter(consenter)
	return c
}

// Org returns specific flow for organization configuration.
func (c *ConfigFlow) Org() *OrgUpdateFlow {
	return &OrgUpdateFlow{flow: c}
}

// Zone returns specific flow for zone management.
func (c *ConfigFlow) Zone() *ZoneUpdateFlow {
	return &ZoneUpdateFlow{flow: c}
}

// Cert returns specific flow for certificate management.
func (c *ConfigFlow) Cert() *CertUpdateFlow {
	return &CertUpdateFlow{flow: c}
}

// Shard is used to return ShardFlow
func (c *ConfigFlow) Shard() *ShardFlow {
	return &ShardFlow{flow: c}
}

// Circle returns specific flow for circle management.
func (c *ConfigFlow) Circle() *CircleUpdateFlow {
	return &CircleUpdateFlow{flow: c}
}

// OrgUpdateFlow represents a flow for organization updating.
type OrgUpdateFlow struct {
	flow    *ConfigFlow
	updates []*common.ConfigSet_OrgUpdate
}

// Add new organizations on chain.
func (o *OrgUpdateFlow) Add(orgs ...*common.Organization) *OrgUpdateFlow {
	for _, org := range orgs {
		o.updates = append(o.updates, &common.ConfigSet_OrgUpdate{
			Operation:    common.OP_APPEND,
			Organization: org,
		})
	}
	return o
}

// Remove some organization by org name on chain.
func (o *OrgUpdateFlow) Remove(orgs ...string) *OrgUpdateFlow {
	for _, org := range orgs {
		o.updates = append(o.updates, &common.ConfigSet_OrgUpdate{
			Operation:    common.OP_REMOVE,
			Organization: &common.Organization{Name: org},
		})
	}
	return o
}

// Replace organization certificate on chain.
func (o *OrgUpdateFlow) Replace(orgs ...*common.Organization) *OrgUpdateFlow {
	for _, org := range orgs {
		o.updates = append(o.updates, &common.ConfigSet_OrgUpdate{
			Operation:    common.OP_REPLACE,
			Organization: org,
		})
	}
	return o
}

// Freeze some organization member on chain.
func (o *OrgUpdateFlow) Freeze(orgs ...string) *OrgUpdateFlow {
	for _, org := range orgs {
		o.updates = append(o.updates, &common.ConfigSet_OrgUpdate{
			Operation:    common.OP_FREEZE,
			Organization: &common.Organization{Name: org},
		})
	}
	return o
}

// Unfreeze some frozen organization by org name on chain.
func (o *OrgUpdateFlow) Unfreeze(orgs ...string) *OrgUpdateFlow {
	for _, org := range orgs {
		o.updates = append(o.updates, &common.ConfigSet_OrgUpdate{
			Operation:    common.OP_UNFREEZE,
			Organization: &common.Organization{Name: org},
		})
	}
	return o
}

// Update sends vote request for updating organization.
func (o *OrgUpdateFlow) Update(ctx context.Context) (*types.Receipt, error) {
	orgMap := make(map[string]struct{})
	us := make([]*common.ConfigSet_OrgUpdate, 0, len(o.updates))
	for _, u := range o.updates {
		if _, ok := orgMap[u.Organization.Name]; ok {
			return nil, errors.Errorf("there are two organization %s duplicate each other", u.Organization.Name)
		}
		orgMap[u.Organization.Name] = struct{}{}
		us = append(us, u)
	}
	flow := o.flow
	rawReq, err := flow.stub.ConfigBuilder().UpdateConfig.
		BuildUpdateOrgRawMessage(flow.chain, &common.ConfigSet_OrgUpdates{OrgUpdate: us})
	if err != nil {
		return nil, err
	}
	return flow.cons.SendTx(ctx, rawReq)
}

// ZoneUpdateFlow represents specific flow for zone configuration updating.
type ZoneUpdateFlow struct {
	flow     *ConfigFlow
	toAdd    []*common.ZoneProperties
	toRemove []string
}

// Add new zone with properties.
func (z *ZoneUpdateFlow) Add(zoneProps ...*common.ZoneProperties) *ZoneUpdateFlow {
	z.toAdd = zoneProps
	return z
}

// Remove existed zone by zone id.
func (z *ZoneUpdateFlow) Remove(zone ...string) *ZoneUpdateFlow {
	z.toRemove = zone
	return z
}

// Update send vote request for updating zone configuration.
func (z *ZoneUpdateFlow) Update(ctx context.Context) (*types.Receipt, error) {
	if len(z.toAdd) != 0 {
		rawReq, err := z.flow.stub.ConfigBuilder().ZoneConfig.BuildAddRawMsg(z.flow.chain, z.toAdd)
		if err != nil {
			return nil, err
		}
		return z.flow.cons.SendTx(ctx, rawReq)
	}
	if len(z.toRemove) != 0 {
		rawReq, err := z.flow.stub.ConfigBuilder().ZoneConfig.BuildRemoveRawMsg(z.flow.chain, z.toRemove)
		if err != nil {
			return nil, err
		}
		return z.flow.cons.SendTx(ctx, rawReq)
	}
	return nil, errors.New("no valid zone updates for adding or removing")
}

// CertUpdateFlow represents specific flow for certificate updating.
type CertUpdateFlow struct {
	flow *ConfigFlow
	crls []actions
}

type actions struct {
	sn     string
	action string
}

// Freeze certificate by serial number.
func (c *CertUpdateFlow) Freeze(serialNum ...string) *CertUpdateFlow {
	for _, sn := range serialNum {
		c.crls = append(c.crls, actions{sn: sn, action: "freeze"})
	}
	return c
}

// Unfreeze frozen certificate by serial number.
func (c *CertUpdateFlow) Unfreeze(serialNum ...string) *CertUpdateFlow {
	for _, sn := range serialNum {
		c.crls = append(c.crls, actions{sn: sn, action: "unfreeze"})
	}
	return c
}

// Revocation existed certificate by serial number.
func (c *CertUpdateFlow) Revocation(serialNum ...string) *CertUpdateFlow {
	for _, sn := range serialNum {
		c.crls = append(c.crls, actions{sn: sn, action: "revocation"})
	}
	return c
}

// Update sends vote request for certificate updating.
func (c *CertUpdateFlow) Update(ctx context.Context) (*types.Receipt, error) {
	snMap := make(map[string]string, len(c.crls))
	for i := range c.crls {
		crl := &c.crls[i]
		_, ok := snMap[crl.sn]
		if ok {
			return nil, errors.Errorf("there are two same certs with different updates, serial number = %s", crl.sn)
		}
		snMap[crl.sn] = crl.action
	}
	entries := make([]*common.CRLEntry, 0, len(snMap))
	for sn, action := range snMap {
		entry, err := rawmessage.ToCRLEntry(sn, action)
		if err != nil {
			return nil, err
		}
		entries = append(entries, entry)
	}
	org := c.flow.stub.Crypto().GetOrg()
	payload, err := rawmessage.GetChainCRLBytes(org, entries)
	if err != nil {
		return nil, err
	}

	voteTxData := common.VoteTxData{
		Handler: common.CERT_STATUS_CHANGE.String(),
		Payload: payload,
	}
	vote, err := c.flow.stub.SealVote(c.flow.chain, &voteTxData)
	if err != nil {
		return nil, err
	}
	return c.flow.cons.SendTx(ctx, vote)
}

const (
	circleHandler = "circle_config"
)

// CircleUpdateFlow represents specific flow for circle configuration updating.
type CircleUpdateFlow struct {
	flow         *ConfigFlow
	circleConfig *nodeservice.CircleConfig
}

// Update send vote request for updating zone configuration.
func (c *CircleUpdateFlow) Update(ctx context.Context) (*types.Receipt, error) {
	if c.circleConfig == nil {
		return nil, errors.New("nil circle config")
	}
	if err := checkCircleVoteParams(c.circleConfig); err != nil {
		return nil, errors.WithMessage(err, "check circle vote params error")
	}
	payload, err := proto.Marshal(c.circleConfig)
	if err != nil {
		return nil, errors.WithMessage(err, "marshal circle definition error")
	}
	voteTxData := common.VoteTxData{
		Handler: circleHandler,
		Payload: payload,
	}
	vote, err := c.flow.stub.SealVote(c.flow.chain, &voteTxData)
	if err != nil {
		return nil, err
	}
	return c.flow.cons.SendTx(ctx, vote)

}

// Config set circle config with property.
func (c *CircleUpdateFlow) Config(circleConfig *nodeservice.CircleConfig) *CircleUpdateFlow {
	c.circleConfig = circleConfig
	return c
}
