/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package flow

import (
	"bytes"
	"context"
	"fmt"
	"sort"
	"strconv"
	"time"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/logger"
	"huawei.com/huaweichain/common/util"
	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/schema"
	"huawei.com/huaweichain/sdk/client"
	"huawei.com/huaweichain/sdk/flow/rest"
	"huawei.com/huaweichain/sdk/flow/types"
)

// Consenter represents consensus node information.
type Consenter struct {
	Name     string
	Org      string
	Addr     string
	RpcPort  uint32 // nolint:golint,stylecheck // ignore
	ReqPort  uint32
	RestPort uint32
	PemCert  []byte
}

// AddConsenter adds multiple nodes to consensus cluster in batch. If any node has not been joined chain,
// try to add it to zone cluster as peer first, then synchronize block progress and transfer to consensus node.
func AddConsenter(client *client.GatewayClient, chain string, consensusNode string, seeds []string,
	consenter ...Consenter) (*Result, error) {
	if len(chain) == 0 {
		return nil, errors.Wrap(types.ErrEmptyChainID, "failed to add consenter")
	}
	if len(consensusNode) == 0 {
		return nil, errors.New("consensus node should not be empty")
	}
	if _, ok := client.Nodes[consensusNode]; !ok {
		return nil, errors.Errorf("consensus node %s doesn't exist", consensusNode)
	}
	if len(consenter) == 0 {
		return nil, errors.New("empty consenter list which add to consensus cluster")
	}
	cons := convPointerSlice(consenter)
	for _, n := range cons {
		if len(n.Name) == 0 {
			return nil, errors.New("consenter name is empty")
		}
		if _, ok := client.Nodes[n.Name]; !ok {
			return nil, errors.Errorf("pending consensus node %s doesn't exist in local network configuration", n.Name)
		}
	}
	sort.Slice(cons, func(i, j int) bool {
		return cons[i].Name < cons[j].Name
	})
	lg := logger.GetModuleLogger("sdk.flow.consensus", "config_change")
	result, childCtx := newResult(context.Background(), lg)
	p := params{
		client:     client,
		chain:      chain,
		existNode:  consensusNode,
		consenters: cons,
		res:        result,
	}
	a := addProcessor{
		params: &p,
		seeds:  seeds,
	}
	if err := a.validate(); err != nil {
		return nil, err
	}
	go a.add(childCtx)
	return result, nil
}

func (p *addProcessor) validate() error {
	ns := make([]string, len(p.consenters))
	for i := range p.consenters {
		ns[i] = p.consenters[i].Name
	}
	if err := validateName(ns); err != nil {
		return err
	}
	nodes := make([]*Consenter, len(p.consenters))
	copy(nodes, p.consenters)
	sort.Slice(nodes, func(i, j int) bool {
		return bytes.Compare(nodes[i].PemCert, nodes[j].PemCert) <= 0
	})

	for i, n := range nodes {
		if len(n.Addr) == 0 {
			return errors.Errorf("node %s address is empty", n.Name)
		}
		if len(n.PemCert) == 0 {
			return errors.Errorf("node %s pem certificate is empty", n.Name)
		}
		if i == len(nodes)-1 {
			break
		}
		if bytes.Equal(n.PemCert, nodes[i+1].PemCert) {
			return errors.Errorf("duplicate node %s and node %s cert in consenter change list", n.Name, nodes[i+1].Name)
		}
	}
	return nil
}

func validateName(consenters []string) error {
	nodes := make([]string, len(consenters))
	copy(nodes, consenters)
	sort.Slice(nodes, func(i, j int) bool {
		return nodes[i] < nodes[j]
	})
	for i, n := range nodes {
		if i == len(nodes)-1 {
			break
		}
		if n == nodes[i+1] {
			return errors.Errorf("duplicate node %s name in consenter change list", n)
		}
	}
	return nil
}

// RemoveConsenter removes multiple nodes from consensus cluster in batch.
// The removed nodes will create new root zone instance as single node.
func RemoveConsenter(client *client.GatewayClient, chain string, consensusNode string,
	consenter ...Consenter) (*Result, error) {
	if len(chain) == 0 {
		return nil, errors.Wrap(types.ErrEmptyChainID, "failed to remove consenter")
	}
	if len(consensusNode) == 0 {
		return nil, errors.New("consensus node should not be empty")
	}
	consenters := convPointerSlice(consenter)
	sort.Slice(consenters, func(i, j int) bool {
		return consenters[i].Name < consenters[j].Name
	})
	var consNames []string
	for _, c := range consenters {
		consNames = append(consNames, c.Name)
	}
	if _, ok := client.Nodes[consensusNode]; !ok {
		return nil, errors.Errorf("consensus node %s doesn't exist in local network configuration", consensusNode)
	}
	if len(consNames) == 0 {
		return nil, errors.New("empty consenter list which remove from consensus cluster")
	}
	for _, n := range consNames {
		if len(n) == 0 {
			return nil, errors.New("consenter name is empty")
		}
		if _, ok := client.Nodes[n]; !ok {
			return nil, errors.Errorf("consensus node %s doesn't exist in local network configuration", n)
		}
	}
	result, childCtx := newResult(context.Background(), nil)
	p := params{
		client:     client,
		chain:      chain,
		existNode:  consensusNode,
		consenters: consenters,
		res:        result,
	}
	a := removeProcessor{
		params:    &p,
		consNames: consNames,
	}

	if err := validateName(consNames); err != nil {
		return nil, err
	}
	go a.remove(childCtx)
	return result, nil
}

type removeProcessor struct {
	*params
	consNames []string
	quitChain bool
}

const (
	retryDuration = 5 * time.Second
)

func (p *removeProcessor) remove(ctx context.Context) {
	flow := LoadFlow(p.client)

	ok, err := p.check(ctx, flow)
	if err != nil {
		p.res.sendNodeError(p.existNode, err)
	}
	if ok {
		if err = p.vote(ctx, flow); err != nil {
			p.res.sendNodeError(p.existNode, err)
		}
	}

	if p.checkFinish(ctx) {
		return
	}

	if p.quitChain {
		p.removeChain(ctx, flow)
	}
	p.res.finish()
}

func (p *removeProcessor) vote(ctx context.Context, flow *Flow) error {
	configFlow := flow.Operator().Config(p.chain).Consenter(p.existNode)
	r, err := configFlow.RemoveConsenter(ctx, ConfigVote, p.consNames...)
	if err = handleResponse(r, err); err != nil {
		return err
	}
	p.res.sendChangePending()
	r, err = configFlow.RemoveConsenter(ctx, ConfigComplete, p.consNames...)
	if err = handleResponse(r, err); err != nil {
		return err
	}
	p.res.sendChangeComplete()
	return nil
}

func handleResponse(receipt *types.Receipt, err error) error {
	if err != nil {
		return err
	}
	if !receipt.Ok() {
		return errors.Errorf("tx is invalid with code %s", receipt.Status)
	}
	return nil
}

func (p *removeProcessor) checkFinish(parent context.Context) bool {
	errs := p.finalCheck(parent, false)
	if len(errs) != 0 {
		p.res.sendAbort(errs, "there are some consenters which failed to convert to peer")
		return true
	}
	return false
}

func (p *addProcessor) checkFinish(parent context.Context) bool {
	errs := p.finalCheck(parent, true)
	if len(errs) != 0 {
		p.res.sendAbort(errs, "there are some peers which failed to convert to consenter")
		return true
	}
	return false
}

func (p *params) finalCheck(parent context.Context, expect bool) map[string]error {
	group, ctx := util.WithContext(parent)
	for i := range p.consenters {
		cons := p.consenters[i]
		group.Go(cons.Name, func() error {
			node, ok := p.client.GetNode(cons.Name)
			if !ok {
				return errors.Errorf("failed to get node %s config from local network config", cons.Name)
			}
			node.RestPort = int(cons.RestPort)
			restc, err := rest.NewRestClient(node)
			if err != nil {
				return errors.WithMessagef(err, "failed to create rest client for node %s", cons.Name)
			}
			for {
				time.Sleep(retryDuration)
				isConsenter, err := rest.IsConsenter(ctx, restc, p.chain)
				if err != nil && errors.Is(err, rest.ErrCheckConsenterTimeout) {
					continue
				}
				if err != nil {
					return errors.WithMessagef(err, "failed to get node %s role", cons.Name)
				}
				if isConsenter == expect {
					break
				}
			}
			return nil
		})
	}
	return group.Wait()
}

func (p *removeProcessor) removeChain(parent context.Context, flow *Flow) {
	removeGroup, ctx := util.WithContext(parent)
	for _, c := range p.consNames {
		cons := c
		removeGroup.Go(cons, func() error {
			node := flow.Admin().Chain(cons)
			ok, err := node.Exist(ctx, p.chain)
			if err != nil {
				return errors.WithMessagef(err, "there is not the chain %s in node %s", p.chain, cons)
			}
			if !ok {
				return errors.Errorf("there is not the chain %s in node %s", p.chain, cons)
			}
			if err := node.Quit(ctx, p.chain); err != nil {
				return errors.WithMessagef(err, "failed to remove chain for %s with chain id %s", cons, p.chain)
			}
			return nil
		})
	}
	if errs := removeGroup.Wait(); len(errs) != 0 {
		p.res.sendAbort(errs, "there are some consensus nodes which failed to remove chain")
	}
	p.res.sendRemove()
}

type params struct {
	client     *client.GatewayClient
	chain      string
	existNode  string
	consenters []*Consenter
	res        *Result
}

type addProcessor struct {
	*params
	seeds     []string
	joinChain bool
}

var stepNum uint64 = 10

func (p *addProcessor) add(ctx context.Context) {
	flow := LoadFlow(p.client)

	ok, err := p.check(ctx, flow)
	if err != nil {
		p.res.sendNodeError(p.existNode, err)
		return
	}
	if ok {
		target, ok := p.join(ctx, flow)
		if !ok {
			return
		}

		if ok := p.sync(ctx, flow, target); !ok {
			return
		}

		if err = p.vote(ctx, flow); err != nil {
			p.res.sendNodeError(p.existNode, err)
		}
	}

	if p.checkFinish(ctx) {
		return
	}

	p.res.finish()
}

func convPointerSlice(consenters []Consenter) []*Consenter {
	cons := make([]*Consenter, 0, len(consenters))
	for i := range consenters {
		cons = append(cons, &consenters[i])
	}
	return cons
}

func convSlice(consenters []*Consenter) []Consenter {
	cons := make([]Consenter, 0, len(consenters))
	for _, c := range consenters {
		cons = append(cons, *c)
	}
	return cons
}

func (p *addProcessor) check(ctx context.Context, flow *Flow) (bool, error) {
	consMap, err := p.consenterSet(ctx, flow)
	if err != nil {
		return false, err
	}
	if ok := p.allNotExist(consMap); ok {
		return true, nil
	}
	if ok := p.allExist(consMap); ok {
		return false, nil
	}
	return false, errors.New("there are some consensus nodes in the chain consensus node list")
}

func (p *removeProcessor) check(ctx context.Context, flow *Flow) (bool, error) {
	consMap, err := p.consenterSet(ctx, flow)
	if err != nil {
		return false, err
	}
	if ok := p.allExist(consMap); ok {
		return true, nil
	}
	if ok := p.allNotExist(consMap); ok {
		return false, nil
	}
	return false, errors.New("there are some consensus nodes not in the chain consensus node list")
}

func (p *params) all(consMap map[string]struct{}, predicate func(bool) bool) bool {
	for _, c := range p.consenters {
		if _, ok := consMap[c.Name]; predicate(ok) {
			return false
		}
	}
	return true
}

func (p *params) allExist(consMap map[string]struct{}) bool {
	return p.all(consMap, func(ok bool) bool {
		return !ok
	})
}

func (p *params) allNotExist(consMap map[string]struct{}) bool {
	return p.all(consMap, func(ok bool) bool {
		return ok
	})
}

func (p *params) consenterSet(ctx context.Context, flow *Flow) (map[string]struct{}, error) {
	chainConf, err := flow.Admin().Chain(p.existNode).ChainConf(ctx, p.chain)
	if err != nil {
		return nil, errors.WithMessagef(err, "failed to fetch latest chain config from exist node %s", p.existNode)
	}
	var cons schema.SysCons
	if err := proto.Unmarshal(chainConf.Config.Consensus, &cons); err != nil {
		return nil, errors.WithMessagef(err, "failed to parse consensus configuration "+
			"which extracted from chain configuration fetched from node %s", p.existNode)
	}
	consMap := make(map[string]struct{}, len(cons.Consenter))
	for _, c := range cons.Consenter {
		consMap[c.Name] = struct{}{}
	}
	return consMap, nil
}

func (p *addProcessor) join(parent context.Context, flow *Flow) (uint64, bool) {
	joinGroup, ctx := util.WithContext(parent)
	for i := range p.consenters {
		cons := p.consenters[i]
		joinGroup.Go(cons.Name, func() error {
			ok, err := flow.Admin().Chain(cons.Name).Exist(ctx, p.chain)
			if ok {
				return nil
			}
			if err != nil {
				return err
			}
			if !p.joinChain {
				return errors.WithMessagef(err, "node %s should join chain %s first", cons.Name, p.chain)
			}

			if err := flow.Admin().Chain(cons.Name).
				Entrypoint("", false, p.seeds).
				Join(ctx, p.chain, p.existNode); err != nil {
				return errors.WithMessagef(err, "failed to join chain for %s with chain id = %s", cons.Name, p.chain)
			}
			return nil
		})
	}
	if errs := joinGroup.Wait(); len(errs) != 0 {
		p.res.sendAbort(errs, "there are some pending consensus nodes which failed to join chain")
		return 0, false
	}

	target, err := flow.User().Observer().Target(p.chain, p.existNode).Height(parent)
	if err != nil {
		p.res.sendNodeError(p.existNode, err)
		return 0, false
	}
	p.res.sendJoin(target)
	return target, true
}

func (p *addProcessor) sync(parent context.Context, flow *Flow, target uint64) bool {
	group, ctx := util.WithContext(parent)
	for i := range p.consenters {
		cons := p.consenters[i]
		group.Go(cons.Name, func() error {
			return p.listenBlockHeight(ctx, flow, cons.Name, target)
		})
	}
	if errs := group.Wait(); len(errs) != 0 {
		p.res.sendAbort(errs, "there are some pending consensus nodes which failed to synchronized")
		return false
	}
	p.res.sendSynchronized()
	return true
}

func (p *addProcessor) listenBlockHeight(ctx context.Context, flow *Flow, cons string, target uint64) error {
	ch, err := flow.User().Event().Target(p.chain, cons).BlockNumEvent(ctx)
	if err != nil {
		return err
	}
	for {
		select {
		case res, ok := <-ch:
			if !ok {
				return ctx.Err()
			}
			h, err := res.Get()
			if err != nil {
				return err
			}
			h++
			if h >= target {
				p.res.sendSyncProgress(cons, h, target)
				return nil
			} else if h%stepNum == 0 {
				p.res.sendSyncProgress(cons, h, target)
			}
		case <-ctx.Done():
			return ctx.Err()
		}
	}
}

func (p *addProcessor) vote(ctx context.Context, flow *Flow) error {
	configFlow := flow.Operator().Config(p.chain).Consenter(p.existNode)
	r, err := configFlow.AddConsenter(ctx, ConfigVote, convSlice(p.consenters)...)
	if err = handleResponse(r, err); err != nil {
		return err
	}
	p.res.sendChangePending()
	r, err = configFlow.AddConsenter(ctx, ConfigComplete, convSlice(p.consenters)...)
	if err = handleResponse(r, err); err != nil {
		return err
	}
	p.res.sendChangeComplete()
	return nil
}

// ConsensusStep represents consensus configuration change operation steps.
type ConsensusStep int32

const (
	// Success represents the operation is executed successfully.
	Success ConsensusStep = 1000 + iota
	// Abort represents the operation has been aborted with some error.
	Abort
	// JoinSuccess represents node has joined the chain.
	JoinSuccess
	// BlockSyncProgress represents progress of block synchronizing.
	BlockSyncProgress
	// Synchronized represents block has been synchronized.
	Synchronized
	// Voting represents configuration change in voting state.
	Voting // single policy
	// Pending represents configuration change in pending state.
	Pending
	// Completed represents configuration change in completed state.
	Completed
	// Removed represents node has been removed from chain.
	Removed
)

var stepName = map[int32]string{
	1000: "Success",
	1001: "Abort",
	1002: "JoinSuccess",
	1003: "BlockSyncProgress",
	1004: "Synchronized",
	1005: "Voting",
	1006: "Pending",
	1007: "Completed",
	1008: "Removed",
}

func (s ConsensusStep) String() string {
	v, ok := stepName[int32(s)]
	if ok {
		return v
	}
	return strconv.Itoa(int(s))
}

// Val returns ConsensusStep integer value.
func (s ConsensusStep) Val() int32 {
	return int32(s)
}

// Progress represents progress result of async multiple operations.
type Progress struct {
	Step        Step               `json:"step_code,omitempty"`
	StepName    string             `json:"step,omitempty"`
	BlockGap    uint64             `json:"block_gap,omitempty"`
	SyncPercent map[string]float32 `json:"sync_percent,omitempty"`
	Message     string             `json:"message,omitempty"`
	NodeErrs    map[string]error   `json:"node_errs,omitempty"`
}

// StepCode returns enum step code.
func (p *Progress) StepCode() Step {
	return p.Step
}

// Msg returns error message.
func (p *Progress) Msg() string {
	return p.Message
}

// Errs returns errors with each node.
func (p *Progress) Errs() map[string]error {
	return p.NodeErrs
}

// Result represents result and context for async operation request.
type Result struct {
	*result
	ch chan *Progress
}

func newResult(parent context.Context, lg *logger.WienerLogger) (*Result, context.Context) {
	ctx, cancelFunc := context.WithCancel(parent)
	r := result{
		res:    make(chan Response, 1),
		cancel: cancelFunc,
		lg:     lg,
	}
	ch := make(chan *Progress, 1)
	go func() {
		for val := range r.Get() {
			v, ok := val.(*Progress)
			if ok {
				ch <- v
			}
		}
		close(ch)
	}()
	return &Result{
		ch:     ch,
		result: &r,
	}, ctx
}

// Response returns channel for progress result.
func (r *Result) Response() <-chan *Progress {
	return r.ch
}

func (r *Result) sendJoin(height uint64) {
	p := Progress{
		Step:     JoinSuccess,
		StepName: JoinSuccess.String(),
		BlockGap: height,
		Message:  "all pending consenters joined into chain",
	}
	r.send(&p)
}

func (r *Result) sendSyncProgress(node string, height uint64, target uint64) {
	var p Progress
	if height >= target {
		p = Progress{
			Step:     BlockSyncProgress,
			StepName: BlockSyncProgress.String(),
			BlockGap: target,
			SyncPercent: map[string]float32{
				node: 1,
			},
			Message: fmt.Sprintf("%s's block height is synchronized", node),
		}
	} else {
		pre := float32(height) / float32(target)
		p = Progress{
			Step:     BlockSyncProgress,
			StepName: BlockSyncProgress.String(),
			BlockGap: target,
			SyncPercent: map[string]float32{
				node: pre,
			},
			Message: fmt.Sprintf("%s's block height has been synchronized %.2f%%", node, pre),
		}
	}
	r.send(&p)
}

func (r *Result) sendSynchronized() {
	p := Progress{
		Step:     Synchronized,
		StepName: Synchronized.String(),
		Message:  "all consenters have been synchronized",
	}
	r.send(&p)
}

func (r *Result) sendChangeComplete() {
	p := Progress{
		Step:     Completed,
		StepName: Completed.String(),
		Message:  "consensus configuration change has been completed, vote state converts to completed",
	}
	r.send(&p)
}

func (r *Result) sendChangePending() {
	p := Progress{
		Step:     Pending,
		StepName: Pending.String(),
		Message:  "consensus configuration change has been voted, vote state converts to pending",
	}
	r.send(&p)
}

func (r *Result) sendRemove() {
	p := Progress{
		Step:     Removed,
		StepName: Removed.String(),
		Message:  "all consenters haven been removed from chain",
	}
	r.send(&p)
}

func (r *Result) sendNodeError(node string, err error) {
	r.sendAbort(map[string]error{node: err}, err.Error())
}

func (r *Result) sendAbort(errs map[string]error, msg string) {
	p := Progress{
		Step:     Abort,
		StepName: Abort.String(),
		NodeErrs: errs,
		Message:  msg,
	}
	r.send(&p)
}

func (r *Result) finish() {
	p := Progress{Step: Success, StepName: Success.String()}
	r.send(&p)
}
