/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package flow

import (
	"context"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/sdk/flow/internal"
	"huawei.com/huaweichain/sdk/flow/types"
	rm "huawei.com/huaweichain/sdk/rawmessage"
)

// ContractFlow is used to handle contract related process
// which includes query and invoke
type ContractFlow struct {
	e        *internal.Endorser
	c        *internal.Consenter
	domains  []string
	chain    string
	contract string
	stub     internal.Stub
	shardID  string

	et []*common.EndorseTarget
}

func newContractFlow(stub internal.Stub, chain string, contract string) *ContractFlow {
	return &ContractFlow{
		stub:     stub,
		chain:    chain,
		contract: contract,
		e:        internal.NewEndorserFlow(stub),
		c:        internal.NewConsenterFlow(chain, stub),
		domains:  []string{"/**"},
	}
}

// EventSource sets event source type for transaction result.
func (c *ContractFlow) EventSource(source types.EventSource) *ContractFlow {
	c.c.Event(source)
	return c
}

// Domains sets domain list for data visibility.
func (c *ContractFlow) Domains(domains []string) *ContractFlow {
	c.domains = domains
	return c
}

// ShardID sets shardID for shard event listen function.
func (c *ContractFlow) ShardID(shardID string) *ContractFlow {
	c.shardID = shardID
	return c
}

// EndorseTargets sets target node for endorsement in agent mode.
func (c *ContractFlow) EndorseTargets(targets []*common.EndorseTarget) *ContractFlow {
	c.et = targets
	return c
}

// ProposeTarget sets target node for consensus in agent mode.
func (c *ContractFlow) ProposeTarget(target *common.ProposeTarget) *ContractFlow {
	c.e.ProposeTarget(target)
	return c
}

// Consenter is used to set consenter list
func (c *ContractFlow) Consenter(consenter ...string) *ContractFlow {
	c.c.AvailableConsenter(consenter)
	return c
}

// Endorser sets available node list for endorsement.
func (c *ContractFlow) Endorser(endorser ...string) *ContractFlow {
	c.e.Endorser(endorser)
	return c
}

// EnableEndorseByOrg enables choosing random endorser by organization in endorsement step.
func (c *ContractFlow) EnableEndorseByOrg(org ...string) *ContractFlow {
	c.e.EnableEndorseByOrg(org)
	return c
}

// Invoke is used to send invoke message
func (c *ContractFlow) Invoke(ctx context.Context, method string, args ...string) (*types.Receipt, error) {
	// send invoke message
	helper := c.stub.ContractBuilder()
	invocation := helper.BuildContractInvocation(c.contract, method, args)
	tx, val, err := c.endorse(ctx, helper, invocation)
	if err != nil {
		return nil, err
	}
	receipt, err := c.c.SendTx(ctx, tx)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to send transaction to consensus")
	}
	receipt.ReturnValue = val
	return receipt, nil
}

func (c *ContractFlow) endorse(ctx context.Context, helper *rm.ContractRawMessage,
	invocation *common.ContractInvocation) (*rm.TxRawMsg, []byte, error) {
	if len(c.et) == 0 {
		req, err := helper.BuildInvokeRawMsg(c.chain, invocation, c.domains)
		if err != nil {
			return nil, nil, errors.WithMessage(err, "BuildInvokeRawMsg error")
		}
		return c.e.Endorse(ctx, req)
	}

	msgs, err := helper.BuildInvokeRawMsgWithTargets(c.chain, invocation, c.domains, c.et)
	if err != nil {
		return nil, nil, err
	}
	return c.e.EndorseWithProxy(ctx, msgs)
}

// Query is used to query invoke message
func (c *ContractFlow) Query(ctx context.Context, method string, args ...string) ([]byte, error) {
	rawReq, err := c.queryReq(method, args)
	if err != nil {
		return nil, err
	}
	return c.e.Call(ctx, rawReq)
}

func (c *ContractFlow) queryReq(method string, args []string) (*common.RawMessage, error) {
	invocation := c.stub.ContractBuilder().BuildContractInvocation(c.contract, method, args)
	rawReq, err := c.stub.ContractBuilder().BuildInvokeRawMsg(c.chain, invocation, c.domains)
	if err != nil {
		return nil, errors.WithMessage(err, "BuildInvokeRawMsg error")
	}
	return rawReq, nil
}

// StreamQuery returns an iterator for stream query by specific contract method.
func (c *ContractFlow) StreamQuery(method string, args ...string) (*types.StreamIter, error) {
	rawReq, err := c.queryReq(method, args)
	if err != nil {
		return nil, err
	}
	return c.e.Stream(rawReq)
}

// Listener sets listener for transaction result event.
func (c *ContractFlow) Listener(listener string) *ContractFlow {
	c.c.SetListener(listener)
	return c
}
