/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

// Package event provides event api implementation in flow pattern.
package event

import (
	"context"
	"github.com/pkg/errors"
	"huawei.com/huaweichain/common/logger"
	"huawei.com/huaweichain/proto"
	commPb "huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/action/event"
	"huawei.com/huaweichain/sdk/flow/internal"
	"huawei.com/huaweichain/sdk/flow/types"
)

var lg = logger.GetModuleLogger("sdk.flow.event", "event-flow")

// Flow defined event operations by flow style api.
type Flow struct {
	stub internal.Stub

	chain    string
	node     string
	startNum uint64
}

// NewEventFlow creates a new event flow with existed types.Stub instance.
func NewEventFlow(stub internal.Stub) *Flow {
	return &Flow{
		stub: stub,
	}
}

// Target set listen target with chain id and node name.
func (f *Flow) Target(chain, node string) *Flow {
	f.chain = chain
	f.node = node
	return f
}

func (f *Flow) check() error {
	if f.chain == "" {
		return types.ErrEmptyChainID
	}
	if !f.stub.RouterContains(f.node) {
		return types.ErrNoSuchNode
	}
	return nil
}

// Get events with event filter constructed by some filter options.
func (f *Flow) Get(ctx context.Context, opts ...FilterOption) ([]*commPb.Event, error) {
	client, err := f.getClient()
	if err != nil {
		return nil, err
	}
	msg, err := f.constructReq(opts)
	if err != nil {
		return nil, err
	}

	resMsg, err := client.GetEvent(ctx, msg)
	if err != nil {
		return nil, err
	}
	res, err := internal.HandleResult(resMsg)
	if err != nil {
		return nil, err
	}
	var es nodeservice.EventList
	if err := proto.Unmarshal(res.Payload, &es); err != nil {
		return nil, err
	}
	return es.Events, nil
}

func (f *Flow) getClient() (nodeservice.EventServiceClient, error) {
	if err := f.check(); err != nil {
		return nil, errors.Wrap(err, "failed to check event flow parameters")
	}
	node, err := f.stub.GetNode(f.node)
	if err != nil {
		return nil, err
	}
	conn, err := node.EventAction.GetReadyConn()
	if err != nil {
		return nil, err
	}
	return nodeservice.NewEventServiceClient(conn), nil
}

func (f *Flow) constructReq(opts []FilterOption) (*commPb.RawMessage, error) {
	req := newFilter(f.chain)
	var err error
	for _, opt := range opts {
		if err = opt(req); err != nil {
			return nil, err
		}
	}
	msg, err := f.stub.Seal(req.proto())
	if err != nil {
		return nil, err
	}
	return msg, nil
}

// Iter events with event filter constructed by some filter options. Use iterator to read event data.
func (f *Flow) Iter(ctx context.Context, opts ...FilterOption) (Iterator, error) {
	client, err := f.getClient()
	if err != nil {
		return nil, err
	}
	msg, err := f.constructReq(opts)
	if err != nil {
		return nil, err
	}

	iter, err := client.IterEvent(ctx, msg)
	if err != nil {
		return nil, err
	}
	return &iterator{client: iter}, nil
}

// BlockIter represents an iterator for block.
type BlockIter struct {
	*event.BlockIterator
	cancel context.CancelFunc
}

// Close iterator and terminate rpc stream.
func (b *BlockIter) Close() {
	b.cancel()
}

// BlockNumIter represents an iterator for block number.
type BlockNumIter struct {
	*event.BlockNumIterator
	cancel context.CancelFunc
}

// Close iterator and terminate rpc stream.
func (b *BlockNumIter) Close() {
	b.cancel()
}

// BlockResIter represents an iterator for block result.
type BlockResIter struct {
	*event.BlockResultIterator
	cancel context.CancelFunc
}

// Close iterator and terminate rpc stream.
func (b *BlockResIter) Close() {
	b.cancel()
}

// BlockWithResIter represents an iterator for block and result.
type BlockWithResIter struct {
	*event.BlockAndResultIterator
	cancel context.CancelFunc
}

// Close iterator and terminate rpc stream.
func (b *BlockWithResIter) Close() {
	b.cancel()
}

// BlockIter returns an iterator for new block.
func (f *Flow) BlockIter(ctx context.Context) (*BlockIter, error) {
	service, err := f.blockEventService(ctx)
	if err != nil {
		return nil, err
	}
	var iter *event.BlockIterator
	if f.startNum == 0 {
		iter, err = service.RegisterBlockEvent()
		if err != nil {
			return nil, err
		}
	} else {
		iter, err = service.RegisterBlockEventFrom(f.startNum)
		if err != nil {
			return nil, err
		}
	}
	return &BlockIter{
		BlockIterator: iter,
		cancel:        service.Close,
	}, nil
}

// StartNum set optional start number for iterator.
func (f *Flow) StartNum(num uint64) *Flow {
	f.startNum = num
	return f
}

func (f *Flow) blockEventService(ctx context.Context) (*event.BlockEventService, error) {
	node, err := f.stub.GetNode(f.node)
	if err != nil {
		return nil, err
	}
	service, err := node.EventAction.GetBlockEventServiceWithCtx(ctx, f.chain)
	if err != nil {
		return nil, err
	}
	return service, nil
}

// BlockResIter returns an iterator for block result.
func (f *Flow) BlockResIter(ctx context.Context) (*BlockResIter, error) {
	service, err := f.blockEventService(ctx)
	if err != nil {
		return nil, err
	}
	var iter *event.BlockResultIterator
	if f.startNum == 0 {
		iter, err = service.RegisterBlockResultEvent()
		if err != nil {
			return nil, err
		}
	} else {
		iter, err = service.RegisterBlockResultEventFrom(f.startNum)
		if err != nil {
			return nil, err
		}
	}
	return &BlockResIter{
		BlockResultIterator: iter,
		cancel:              service.Close,
	}, nil
}

// BlockWithResIter returns an iterator for block raw data and result.
func (f *Flow) BlockWithResIter(ctx context.Context) (*BlockWithResIter, error) {
	service, err := f.blockEventService(ctx)
	if err != nil {
		return nil, err
	}
	var iter *event.BlockAndResultIterator
	if f.startNum == 0 {
		iter, err = service.RegisterBlockAndResultEvent()
		if err != nil {
			return nil, err
		}
	} else {
		iter, err = service.RegisterBlockAndResultEventFrom(f.startNum)
		if err != nil {
			return nil, err
		}
	}
	return &BlockWithResIter{
		BlockAndResultIterator: iter,
		cancel:                 service.Close,
	}, nil
}

// BlockNumIter returns an iterator for block number.
func (f *Flow) BlockNumIter(ctx context.Context) (*BlockNumIter, error) {
	service, err := f.blockEventService(ctx)
	if err != nil {
		return nil, err
	}
	iter, err := service.RegisterBlockNumEvent()
	if err != nil {
		return nil, err
	}
	return &BlockNumIter{
		BlockNumIterator: iter,
		cancel:           service.Close,
	}, nil
}

// Listen events from result channel with event filter constructed by some filter options.
func (f *Flow) Listen(ctx context.Context, ch chan<- *Result, opts ...FilterOption) error {
	if ch == nil {
		return errors.New("result channel should not be nil")
	}
	if err := f.check(); err != nil {
		return errors.Wrap(err, "failed to check event flow parameters")
	}
	node, err := f.stub.GetNode(f.node)
	if err != nil {
		return err
	}
	conn, err := node.EventAction.GetReadyConn()
	if err != nil {
		return err
	}
	service := nodeservice.NewEventServiceClient(conn)
	msg, err := f.constructReq(opts)
	if err != nil {
		return err
	}
	client, err := service.RegisterEvent(ctx, msg)
	if err != nil {
		return err
	}
	res, err := client.Recv()
	if err != nil {
		return err
	}
	_, err = internal.HandleResult(res)
	if err != nil {
		return err
	}
	go handle(client, ch)
	return nil
}

func handle(client nodeservice.EventService_RegisterEventClient, ch chan<- *Result) {
	recv := func() (*nodeservice.EventList, error) {
		msg, err := client.Recv()
		if err != nil {
			return nil, err
		}
		lg.Debug("received message")
		res, err := internal.HandleResult(msg)
		if err != nil {
			return nil, err
		}
		var es nodeservice.EventList
		if err := proto.Unmarshal(res.Payload, &es); err != nil {
			return nil, err
		}
		return &es, nil
	}

	for {
		events, err := recv()
		if err != nil {
			select {
			case ch <- &Result{err: err}:
				lg.Debugf("receive error message: %+v", err)
				return
			case <-client.Context().Done():
				ch <- &Result{err: errors.WithMessagef(err, "context error: %s", client.Context().Err())}
				return
			}
		}
		lg.Debugf("handle message, event len = %d", len(events.Events))
		for _, e := range events.Events {
			select {
			case ch <- &Result{val: e}:
			case <-client.Context().Done():
				return
			}
		}
	}
}

// Result returns event or error in result pattern.
type Result struct {
	val *commPb.Event
	err error
}

// Err returns error, otherwise nil.
func (r *Result) Err() error {
	return r.err
}

// Value returns result value, returns nil if error exists.
func (r *Result) Value() *commPb.Event {
	return r.val
}

// Iterator is the definition of listen events result iterator.
type Iterator interface {
	Next() bool
	Value() *commPb.Event
	Err() error
	Cancel()
}

type iterator struct {
	current *commPb.Event
	client  nodeservice.EventService_IterEventClient
	err     error
}

// Next is used to get the next element of iterator.
func (i *iterator) Next() bool {
	next, err := i.next()
	if err != nil {
		i.err = err
		return false
	}
	i.current = next
	return true
}

func (i *iterator) next() (*commPb.Event, error) {
	msg, err := i.client.Recv()
	if err != nil {
		return nil, err
	}
	res, err := internal.HandleResult(msg)
	if err != nil {
		return nil, err
	}
	var es nodeservice.EventList
	if err := proto.Unmarshal(res.Payload, &es); err != nil {
		return nil, err
	}
	if len(es.Events) == 0 || len(es.Events) != 1 {
		return nil, errors.New("event list is empty or not only one")
	}
	return es.Events[0], nil
}

// Value is used to get the next element's value of iterator.
func (i *iterator) Value() *commPb.Event {
	return i.current
}

// Err is used to get error.
func (i *iterator) Err() error {
	return i.err
}

// Cancel is used to cancel this data chan.
func (i *iterator) Cancel() {
	if err := i.client.CloseSend(); err != nil {
		i.err = err
	}
}

// NumberResult contains block number or error.
type NumberResult struct {
	height uint64
	err    error
}

// Get current block number.
func (r *NumberResult) Get() (uint64, error) {
	return r.height, r.err
}

// BlockNumEvent listens block number event and return result channel immediately.
func (f *Flow) BlockNumEvent(ctx context.Context) (<-chan NumberResult, error) {
	iter, err := f.BlockNumIter(ctx)
	if err != nil {
		return nil, err
	}
	ch := make(chan NumberResult)
	go listenBlockEvent(ctx, ch, iter)
	return ch, nil
}

func listenBlockEvent(ctx context.Context, ch chan NumberResult, iter *BlockNumIter) {
	defer func() {
		close(ch)
		iter.Close()
	}()
	for {
		next, err := iter.Next()
		if err != nil {
			select {
			case ch <- NumberResult{err: err}:
			case <-ctx.Done():
			}
			return
		}
		select {
		case ch <- NumberResult{height: next}:
		case <-ctx.Done():
			return
		}
	}
}
