/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package event

import (
	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/flow/types"
)

type filter struct {
	blockHash    []byte
	from         uint64
	to           uint64
	chain        string
	contractName string
	eventName    string
	address      []byte
	topics       [][]byte
}

func newFilter(chain string) *filter {
	return &filter{chain: chain}
}

func (f *filter) proto() *nodeservice.EventFilter {
	return &nodeservice.EventFilter{
		ChainId:      f.chain,
		ContractName: f.contractName,
		EventName:    f.eventName,
		Address:      f.address,
		Topics:       f.topics,
		FromHeight:   f.from,
		ToHeight:     f.to,
		BlockHash:    f.blockHash,
	}
}

// FilterOption configures how we filter event from receipt.
type FilterOption func(f *filter) error

// WithCustomTopics returns a FilterOption which search from bloom by custom topic index in container contract.
func WithCustomTopics(topics ...string) FilterOption {
	return func(f *filter) error {
		if len(topics) == 0 {
			return errors.New("topic list is empty")
		}
		ts := make([][]byte, 0, len(topics))
		for _, topic := range topics {
			ts = append(ts, []byte(topic))
		}
		f.topics = ts
		return nil
	}
}

// WithEVMTopics returns a FilterOption which search from bloom by evm log topic.
func WithEVMTopics(topics ...string) FilterOption {
	return func(f *filter) error {
		if len(topics) == 0 {
			return errors.New("topic list is empty")
		}
		ts := make([][]byte, len(topics))
		for i, t := range topics {
			if t == "" {
				ts[i] = []byte{}
				continue
			}
			topic, err := types.StringToHash(t)
			if err != nil {
				return err
			}
			ts[i] = topic.Bytes()
		}
		f.topics = ts
		return nil
	}
}

// WithVMTopics returns a FilterOption which search from bloom by vm contract log topic.
func WithVMTopics(topics ...string) FilterOption {
	return WithEVMTopics(topics...)
}

// WithEVMAddress returns a FilterOption which search from bloom by evm contract address.
func WithEVMAddress(address string) FilterOption {
	return func(f *filter) error {
		addr, err := types.StringToAddress(address)
		if err != nil {
			return err
		}
		f.address = addr.Bytes()
		return nil
	}
}

// WithXmartAddress returns a FilterOption which search from bloom by xmart contract address.
func WithXmartAddress(address string) FilterOption {
	return func(f *filter) error {
		addr, err := types.StringToHash(address)
		if err != nil {
			return err
		}
		f.address = addr.Bytes()
		return nil
	}
}

// WithEventName returns a FilterOption which search from bloom by event name.
func WithEventName(name string) FilterOption {
	return func(f *filter) error {
		if name == "" {
			return errors.New("event name is empty")
		}
		f.eventName = name
		return nil
	}
}

// WithContractName returns a FilterOption which search from bloom by contract name.
func WithContractName(name string) FilterOption {
	return func(f *filter) error {
		if name == "" {
			return errors.New("contract name is empty")
		}
		f.contractName = name
		return nil
	}
}

// WithBlockRange returns a FilterOption which search events from ledger by block height range.
func WithBlockRange(from, to uint64) FilterOption {
	return func(f *filter) error {
		if len(f.blockHash) != 0 {
			return errors.New("block hash has already been configured")
		}
		if from == 0 {
			f.from = 1
		} else {
			f.from = from
		}
		f.to = to
		return nil
	}
}

// withBlockHash returns a FilterOption which search events from ledger by specific block hash.
func withBlockHash(hash string) FilterOption {
	return func(f *filter) error {
		if f.from != 0 || f.to != 0 {
			return errors.New("block range has already been configured")
		}
		h, err := types.StringToHash(hash)
		if err != nil {
			return err
		}
		f.blockHash = h.Bytes()
		return nil
	}
}

// WithEVMContract returns a FilterOption which search only evm event.
func WithEVMContract() FilterOption {
	return func(f *filter) error {
		f.contractName = types.NativeContractEVM
		return nil
	}
}

// WithXmartContract returns a FilterOption which search only xmart event.
func WithXmartContract() FilterOption {
	return func(f *filter) error {
		f.contractName = types.NativeContractXmart
		return nil
	}
}
