/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package flow

import (
	"context"
	"encoding/hex"
	"encoding/json"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/sdk/flow/types"
)

// CallParams contains all parameters for evm calling.
type CallParams struct {
	Gas uint64
}

// EVMContractFlow is used to handle contract related process
// specific for EVM contract.
type EVMContractFlow struct {
	params CallParams
	flow   *ContractFlow
}

func newEVMContractFlow(flow *ContractFlow) *EVMContractFlow {
	return &EVMContractFlow{
		flow: flow,
	}
}

// Endorser sets available node list for endorsement.
func (e *EVMContractFlow) Endorser(endorser string) *EVMContractFlow {
	e.flow.Endorser(endorser)
	return e
}

// Listener sets listener for transaction result event.
func (e *EVMContractFlow) Listener(listener string) *EVMContractFlow {
	e.flow.Listener(listener)
	return e
}

// Consenter is used to set consenter list
func (e *EVMContractFlow) Consenter(consenter ...string) *EVMContractFlow {
	e.flow.Consenter(consenter...)
	return e
}

// Transact represents contract function name for executing encoded command for evm transaction.
func (e *EVMContractFlow) Transact(ctx context.Context, address string, encodedInput string) (*types.Receipt, error) {
	p, err := json.Marshal(&e.params)
	if err != nil {
		return nil, err
	}
	return e.flow.Invoke(ctx, "call", address, encodedInput, string(p))
}

// Call represents contract function name for executing encoded command for evm.
func (e *EVMContractFlow) Call(ctx context.Context, address string, encodedInput string) (string, error) {
	p, err := json.Marshal(&e.params)
	if err != nil {
		return "", err
	}
	res, err := e.flow.Query(ctx, "call", address, encodedInput, string(p))
	if err != nil {
		return "", err
	}
	return hex.EncodeToString(res), nil
}

// Gas set gas limit for contract invocation.
func (e *EVMContractFlow) Gas(gas uint64) *EVMContractFlow {
	e.params.Gas = gas
	return e
}

// Deploy represents contract function name for deploying compiled contract bytecode.
func (e *EVMContractFlow) Deploy(ctx context.Context, bytecode string) (*types.Receipt, error) {
	p, err := json.Marshal(&e.params)
	if err != nil {
		return nil, err
	}
	return e.flow.Invoke(ctx, "deploy", bytecode, string(p))
}

// Bytecode represents contract function name for querying deployed executable bytecode of specific contract address.
func (e *EVMContractFlow) Bytecode(ctx context.Context, address string) (string, error) {
	res, err := e.flow.Query(ctx, "bytecode", address)
	if err != nil {
		return "", err
	}
	return string(res), nil
}

// Account represents contract function name for querying account address of caller.
func (e *EVMContractFlow) Account(ctx context.Context) (string, error) {
	res, err := e.flow.Query(ctx, "account")
	if err != nil {
		return "", err
	}
	if len(res) != 40 {
		return "", errors.Errorf("valid account address should with 20 bytes length, "+
			"current is %s", string(res))
	}
	return string(res), nil
}
