/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package internal

import (
	"context"
	"crypto/rand"
	"encoding/hex"
	"math/big"
	"strconv"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/action/event"
	"huawei.com/huaweichain/sdk/flow/status"
	"huawei.com/huaweichain/sdk/flow/types"
	"huawei.com/huaweichain/sdk/node"
	"huawei.com/huaweichain/sdk/rawmessage"
	"huawei.com/huaweichain/sdk/utils"
)

// Consenter provides consensus process for transaction.
type Consenter struct {
	stub       Stub
	chain      string
	listener   string
	consenters []string
	shardID    string
	source     types.EventSource
}

// SetListener for transaction result event listening.
func (c *Consenter) SetListener(listener string) {
	c.listener = listener
}

// ShardID for shard send.
func (c *Consenter) ShardID(shardID string) *Consenter {
	c.shardID = shardID
	return c
}

func (c *Consenter) setShardID(txRawMsg *rawmessage.TxRawMsg) error {
	var tx common.Transaction
	if err := proto.Unmarshal(txRawMsg.Msg.Payload, &tx); err != nil {
		return errors.WithMessage(err, "tx unmarshal error")
	}
	shardIDs, err := utils.GetShards(&tx)
	if err != nil {
		return errors.WithMessagef(err, "failed to get shards from tx byte "+
			"which tx id is %s", hex.EncodeToString(txRawMsg.Hash))
	}
	if c.shardID == "" && len(shardIDs) > 0 {
		c.shardID = strconv.Itoa(int(shardIDs[0]))
	}

	return nil
}

// AvailableConsenter is used to set consenter node list
func (c *Consenter) AvailableConsenter(consenters []string) {
	c.consenters = consenters
}

// Consenter is used to get a usable Consenter node
func (c *Consenter) Consenter() (*node.WNode, error) {
	i, err := rand.Int(rand.Reader, big.NewInt(int64(len(c.consenters))))
	if err != nil {
		return nil, err
	}
	return c.stub.GetNode(c.consenters[i.Int64()])
}

// SendTx is used to send tx
func (c *Consenter) SendTx(ctx context.Context, tx *rawmessage.TxRawMsg) (*types.Receipt, error) {
	if ctx == nil {
		return nil, errors.New("nil ctx")
	}
	if tx == nil {
		return nil, errors.New("nil tx")
	}
	txID := hex.EncodeToString(tx.Hash)
	if len(c.consenters) == 0 {
		return nil, errors.New("no available consenter")
	}
	if c.listener == "" {
		c.listener = c.consenters[0]
	}
	l, err := c.stub.GetNode(c.listener)
	if err != nil {
		return nil, err
	}
	if err := c.setShardID(tx); err != nil {
		return nil, errors.WithMessagef(err, "failed to set shardID which tx id is %s", txID)
	}
	listener, err := c.eventListener(l)
	if err != nil {
		return nil, errors.WithMessagef(err, "failed to start event listener for chain %s", c.chain)
	}
	defer listener.Close()
	es, err := c.txResultChan(listener, tx)
	if err != nil {
		return nil, errors.WithMessagef(err, "failed to get tx result chan which tx id is %s", txID)
	}

	consenter, err := c.Consenter()
	if err != nil {
		return nil, err
	}
	conn, err := consenter.ContractAction.GetReadyConn()
	if err != nil {
		return nil, err
	}
	client := nodeservice.NewTransactionSenderClient(conn)
	rawResp, err := client.SendTransaction(ctx, tx.Msg)
	if err != nil {
		return nil, errors.WithMessagef(err, "failed to send transaction to consensus service which tx id is %s", txID)
	}
	var resp common.Response
	if err = proto.Unmarshal(rawResp.Payload, &resp); err != nil {
		return nil, errors.WithMessagef(err, "failed to unmarshal response from consensus service which tx id is %s", txID)
	}
	if resp.Status != common.SUCCESS {
		return nil, status.ResponseToErr(&resp)
	}
	return c.waitForEvent(ctx, tx, es)
}

func (c *Consenter) eventListener(l *node.WNode) (*event.TxEventService, error) {
	var listener *event.TxEventService
	var err error
	if c.shardID != "" {
		shardID, err := strconv.Atoi(c.shardID)
		if err != nil {
			return nil, errors.WithMessage(err, "failed to transform shardID from string to uint32")
		}
		if listener, err = l.EventAction.GetTxEventServiceWithSourceType(c.chain, event.SourceType(c.source),
			event.WithShardID(uint32(shardID))); err != nil {
			return nil, errors.WithMessage(err, "failed to get tx event service with source type")
		}
	} else {
		if listener, err = l.EventAction.GetTxEventServiceWithSourceType(c.chain,
			event.SourceType(c.source)); err != nil {
			return nil, errors.WithMessage(err, "failed to get tx event service with source type")
		}
	}

	return listener, nil
}

func (c *Consenter) txResultChan(listener *event.TxEventService,
	tx *rawmessage.TxRawMsg) (chan *common.TxResult, error) {
	var es chan *common.TxResult
	var err error
	txID := hex.EncodeToString(tx.Hash)
	if c.shardID != "" {
		var shardID int
		if shardID, err = strconv.Atoi(c.shardID); err != nil {
			return nil, errors.WithMessagef(err, "failed to transfer shardID from "+
				"string to int which tx id is %s", txID)
		}
		if es, err = listener.RegisterTx(tx.Hash, event.WithShardID(uint32(shardID))); err != nil {
			return nil, errors.WithMessagef(err, "failed to register transaction event which tx id is %s", txID)
		}
	} else {
		if es, err = listener.RegisterTx(tx.Hash); err != nil {
			return nil, errors.WithMessagef(err, "failed to register transaction event which tx id is %s", txID)
		}
	}

	return es, nil
}

func (c *Consenter) waitForEvent(ctx context.Context, tx *rawmessage.TxRawMsg,
	event chan *common.TxResult) (*types.Receipt, error) {
	select {
	case e := <-event:
		if e == nil {
			return nil, errors.New("transaction event is empty")
		}
		if e.Status != common.VALID {
			return &types.Receipt{
				TxHash: hex.EncodeToString(tx.Hash),
				Status: e.Status,
				Msg:    "failed to validate transaction",
			}, nil
		}
		return &types.Receipt{
			TxHash: hex.EncodeToString(tx.Hash),
			Status: common.VALID,
			Msg:    "transaction has been completed",
		}, nil
	case <-ctx.Done():
		return nil, status.FromContextError(ctx.Err()).Err()
	}
}

// Event sets event source type for listening.
func (c *Consenter) Event(source types.EventSource) {
	c.source = source
}

// NewConsenterFlow creates consensus implementation for ContractFlow.
func NewConsenterFlow(chain string, stub Stub) *Consenter {
	return &Consenter{
		stub:   stub,
		chain:  chain,
		source: types.TxEventSource,
	}
}
