/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package internal

import (
	"context"
	"math/rand"
	"sync"
	"time"

	"github.com/pkg/errors"
	"golang.org/x/sync/errgroup"

	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/flow/status"
	"huawei.com/huaweichain/sdk/flow/types"
	"huawei.com/huaweichain/sdk/node"
	rm "huawei.com/huaweichain/sdk/rawmessage"
)

// Endorser provides endorsement implementation for transaction process.
type Endorser struct {
	stub      Stub
	endorsers []string
	orgs      []string
	pt        *common.ProposeTarget
}

// NewEndorserFlow creates an endorsement handler for transaction.
func NewEndorserFlow(stub Stub) *Endorser {
	return &Endorser{stub: stub}
}

// ProposeTarget sets target node for proposing.
func (e *Endorser) ProposeTarget(target *common.ProposeTarget) {
	e.pt = target
}

// Endorser sets endorser list for endorsement.
func (e *Endorser) Endorser(endorser []string) {
	e.endorsers = endorser
}

// EndorseWithProxy executes endorsement for transaction by proxy.
func (e *Endorser) EndorseWithProxy(ctx context.Context, msgs []*common.RawMessage) (*rm.TxRawMsg, []byte, error) {
	eg, err := e.endorserGroup()
	if err != nil {
		return nil, nil, err
	}
	return e.endorse(ctx, eg, msgs)
}

// Endorse executes endorsement for transaction.
func (e *Endorser) Endorse(ctx context.Context, msg *common.RawMessage) (*rm.TxRawMsg, []byte, error) {
	eg, err := e.endorserGroup()
	if err != nil {
		return nil, nil, err
	}
	msgs := make([]*common.RawMessage, eg.num())
	for i := 0; i < len(msgs); i++ {
		msgs[i] = msg
	}
	return e.endorse(ctx, eg, msgs)
}

func (e *Endorser) endorse(ctx context.Context, eg *endorserGroup, msgs []*common.RawMessage) (
	*rm.TxRawMsg, []byte, error) {
	helper := e.stub.ContractBuilder()
	approvals, err := eg.endorse(ctx, msgs)
	if err != nil {
		return nil, nil, errors.WithMessage(err, "failed to endorse for transaction")
	}
	return e.validateApprovals(helper, approvals)
}

func (e *Endorser) validateApprovals(helper *rm.ContractRawMessage, approvals []*common.RawMessage) (
	*rm.TxRawMsg, []byte, error) {
	// send transaction message
	tx, err := e.constructTx(helper, approvals)
	if err != nil {
		return nil, nil, err
	}
	var returnValue []byte
	for _, res := range approvals {
		val, err := helper.ParseQueryResult(res)
		if err != nil {
			if s, ok := status.FromError(err); ok {
				if common.Status(s.Code()) == common.HASH {
					continue
				}
			}
			return nil, nil, err
		}
		returnValue = val
		break
	}
	return tx, returnValue, nil
}

func (e *Endorser) constructTx(helper *rm.ContractRawMessage, approvals []*common.RawMessage) (
	*rm.TxRawMsg, error) {
	if e.pt != nil {
		return helper.BuildTxRawMsgWithProxy(approvals, e.pt)
	}
	return helper.BuildTxRawMsg(approvals)
}

// Call for querying contract world state by request.
func (e *Endorser) Call(ctx context.Context, msg *common.RawMessage) ([]byte, error) {
	client, err := e.client()
	if err != nil {
		return nil, err
	}
	rawRes, err := client.Query(ctx, msg)
	if err != nil {
		return nil, err
	}
	return e.stub.ContractBuilder().ParseQueryResult(rawRes)
}

// Stream for querying contract world state by request stream.
func (e *Endorser) Stream(msg *common.RawMessage) (*types.StreamIter, error) {
	client, err := e.pickup()
	if err != nil {
		return nil, err
	}
	iter, err := client.ContractAction.QueryStream(msg)
	if err != nil {
		return nil, err
	}
	return &types.StreamIter{
		StreamResultIterator: iter,
	}, nil
}

func (e *Endorser) client() (nodeservice.ContractClient, error) {
	stub, err := e.pickup()
	if err != nil {
		return nil, err
	}
	conn, err := stub.ContractAction.GetReadyConn()
	if err != nil {
		return nil, errors.WithMessage(err, "failed to get connection for contract endorsement")
	}
	client := nodeservice.NewContractClient(conn)
	return client, nil
}

func (e *Endorser) pickup() (*node.WNode, error) {
	if len(e.endorsers) == 0 {
		return nil, errors.New("no available endorser")
	}
	stub, err := e.stub.GetNode(e.endorsers[0])
	if err != nil {
		return nil, err
	}
	return stub, nil
}

// EnableEndorseByOrg for choosing endorser node by organization.
func (e *Endorser) EnableEndorseByOrg(org []string) {
	e.orgs = org
}

func (e *Endorser) endorserGroup() (*endorserGroup, error) {
	if len(e.orgs) != 0 {
		return e.endorserGroupByOrg()
	}
	clients := make([]nodeservice.ContractClient, 0, len(e.endorsers))
	for _, n := range e.endorsers {
		stub, err := e.stub.GetNode(n)
		if err != nil {
			return nil, err
		}
		conn, err := stub.ContractAction.GetReadyConn()
		if err != nil {
			return nil, errors.WithMessage(err, "failed to get connection for contract endorsement")
		}
		clients = append(clients, nodeservice.NewContractClient(conn))
	}
	return &endorserGroup{clients: clients}, nil
}

func (e *Endorser) endorserGroupByOrg() (*endorserGroup, error) {
	var cs []nodeservice.ContractClient
	for _, org := range e.orgs {
		nodes, err := e.stub.NodeByOrg(e.endorsers, org)
		if err != nil {
			return nil, err
		}
		n, err := randomNode(nodes)
		if err != nil {
			return nil, err
		}
		conn, err := n.ContractAction.GetReadyConn()
		if err != nil {
			return nil, err
		}
		cs = append(cs, nodeservice.NewContractClient(conn))
	}
	return &endorserGroup{clients: cs}, nil
}

type endorserGroup struct {
	clients []nodeservice.ContractClient
}

func (e *endorserGroup) endorse(ctx context.Context, msgs []*common.RawMessage) ([]*common.RawMessage, error) {
	if len(msgs) != len(e.clients) {
		return nil, errors.New("message count should equal to client count")
	}
	var approvals []*common.RawMessage
	ch := make(chan *common.RawMessage)
	wg := sync.WaitGroup{}
	wg.Add(1)
	go func() {
		for r := range ch {
			approvals = append(approvals, r)
		}
		wg.Done()
	}()

	group, ctx := errgroup.WithContext(ctx)
	for i, c := range e.clients {
		lc := c
		li := i
		group.Go(func() error {
			res, err := lc.Invoke(ctx, msgs[li])
			if err != nil {
				return err
			}
			ch <- res
			return nil
		})
	}
	err := group.Wait()
	close(ch)
	if err != nil {
		return nil, err
	}
	wg.Wait()
	return approvals, nil
}

func (e *endorserGroup) num() int {
	return len(e.clients)
}

func randomNode(nodeList []*node.WNode) (*node.WNode, error) {
	length := len(nodeList)
	if length == 0 {
		return nil, errors.New("nodeList is null")
	}
	rand.Seed(time.Now().UnixNano())
	randomIndex := rand.Intn(length) // nolint:gosec // not security case
	if check := checkNodeConnected(nodeList[randomIndex]); check {
		return nodeList[randomIndex], nil
	}
	if length <= 1 {
		return nil, errors.New("fail to select nodes in method randomNode")
	}
	nodeList = append(nodeList[:randomIndex], nodeList[randomIndex+1:]...)
	return randomNode(nodeList)
}

// checkNodeConnected is used to check node connection state by send a invoke message
func checkNodeConnected(node *node.WNode) bool {
	rawMes := &common.RawMessage{}
	_, err := node.ContractAction.Invoke(rawMes)
	return err == nil
}
