/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

// Package internal provides internal impplementation for flow sdk.
package internal

import (
	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/logger"
	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/sdk/client"
	"huawei.com/huaweichain/sdk/config"
	"huawei.com/huaweichain/sdk/crypto"
	"huawei.com/huaweichain/sdk/flow/types"
	"huawei.com/huaweichain/sdk/node"
	"huawei.com/huaweichain/sdk/rawmessage"
)

var log = logger.GetModuleLogger("flow", "internal")

// Stub is used to cover gatewayClient
type stub struct {
	sdk *client.GatewayClient
	lg  *logger.WienerLogger

	orgIdx map[string]map[string]struct{}
	err    error
}

// NewStub with empty client.
func NewStub() Stub {
	return LoadStub(client.NewEmptyGatewayClient())
}

// LoadStub from configured client instance.
func LoadStub(sdk *client.GatewayClient) Stub {
	return &stub{
		sdk:    sdk,
		orgIdx: make(map[string]map[string]struct{}),
		lg:     log,
	}
}

func (s *stub) Destroy() {
	s.sdk.CleanSenInfo()
}

// SetIdentity is used to call the same method in gatewayClient
func (s *stub) SetIdentity(algorithm string, key []byte, cert []byte) {
	err := s.sdk.SetIdentity(algorithm, key, cert)
	if err != nil {
		s.ReturnErr(err)
	}
}

// SetTLS is used to call the same method in gatewayClient
func (s *stub) SetTLS(certPEMBlock []byte, keyPEMBlock []byte, roots [][]byte) {
	s.sdk.SetTLS(certPEMBlock, keyPEMBlock, roots)
}

// AddNode is used to call the same method in gatewayClient
func (s *stub) AddNode(node *types.Node) {
	n := config.Node{
		ID:           node.Name,
		Org:          node.Org,
		HostOverride: node.IP,
		Host:         node.Host,
		Port:         node.RPCPort,
		RestPort:     node.RestPort,
		RPCCompress:  node.RPCCompress,
	}
	err := s.sdk.Add(&n)
	if err != nil {
		s.ReturnErr(err)
		return
	}
	s.addOrgIdx(node.Name, node.Org)
}

func (s *stub) addOrgIdx(name string, org string) {
	m, ok := s.orgIdx[org]
	if ok {
		m[name] = struct{}{}
	} else {
		s.orgIdx[org] = map[string]struct{}{name: {}}
	}
}

// Node is used to transform the struct node
func (s *stub) Node(name string) (*types.Node, error) {
	nodeMap := s.sdk.Nodes
	n, ok := nodeMap[name]
	if !ok {
		return nil, errors.Errorf("node %s not found", name)
	}
	return &types.Node{
		Name:        n.ID,
		Org:         n.Org,
		Host:        n.Host,
		IP:          n.HostOverride,
		RPCPort:     n.Port,
		RestPort:    n.RestPort,
		RPCCompress: n.RPCCompress,
	}, nil
}

// GetNode returns remote node stub for interaction with node service.
func (s *stub) GetNode(name string) (*node.WNode, error) {
	n, ok := s.sdk.GetNode(name)
	if ok {
		return n, nil
	}
	return nil, errors.Errorf("node %s not found", name)
}

// NodeByOrg returns node list filtered by organization.
func (s *stub) NodeByOrg(nodes []string, org string) ([]*node.WNode, error) {
	var ns []*node.WNode
	for _, name := range nodes {
		_, ok := s.orgIdx[org][name]
		if ok {
			wNode, err := s.GetNode(name)
			if err != nil {
				return nil, err
			}
			ns = append(ns, wNode)
		}
	}
	return ns, nil
}

// ChainBuilder is used to get chain raw message
func (s *stub) ChainBuilder() *rawmessage.ChainRawMessage {
	return s.sdk.ChainRawMessage
}

// QueryBuilder is used to get query raw message
func (s *stub) QueryBuilder() *rawmessage.QueryRawMessage {
	return s.sdk.QueryRawMessage
}

// ContractBuilder is used to get contract raw message
func (s *stub) ContractBuilder() *rawmessage.ContractRawMessage {
	return s.sdk.ContractRawMessage
}

// LifecycleBuilder is used to get lifecycle raw message
func (s *stub) LifecycleBuilder() *rawmessage.LifecycleRawMessage {
	return s.sdk.LifecycleRawMessage
}

// ConfigBuilder is used to get config raw message
func (s *stub) ConfigBuilder() *rawmessage.ConfigRawMessage {
	return s.sdk.ConfigRawMessage
}

// Crypto returns crypto utility.
func (s *stub) Crypto() crypto.Crypto {
	return s.sdk.Crypto()
}

// RouterContains returns true if node has been configured in router.
func (s *stub) RouterContains(node string) bool {
	_, ok := s.sdk.Nodes[node]
	return ok
}

// Seal payload for request common.RawMessage.
func (s *stub) Seal(req proto.Marshaler) (*common.RawMessage, error) {
	payload, err := proto.Marshal(req)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to build archive request payload")
	}
	builder := rawmessage.NewMsgBuilderImpl(s.Crypto())
	msg, err := builder.GetRawMessage(payload)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to build raw message for request")
	}
	return msg, nil
}

// SealVote payload to transaction for vote proposal.
func (s *stub) SealVote(chain string, msg proto.Marshaler) (*rawmessage.TxRawMsg, error) {
	builder := rawmessage.NewMsgBuilderImpl(s.Crypto())
	tx, err := builder.BuildVoteTx(chain, msg)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to build raw message for transaction request")
	}
	return builder.GetTxRawMsg(tx)
}

// ReturnErr handles error
func (s *stub) ReturnErr(err error) {
	s.err = err
}

// Err returns error from previous process.
func (s *stub) Err() error {
	err := s.err
	if err != nil {
		return errors.WithMessage(err, "flow error")
	}
	return nil
}

// RemoveNode from router table and close all underlying connection.
func (s *stub) RemoveNode(nodeID string) error {
	wn, err := s.GetNode(nodeID)
	if err != nil {
		return err
	}
	if err := wn.ChainAction.Close(); err != nil {
		return err
	}
	if err := wn.ContractAction.Close(); err != nil {
		return err
	}
	if err := wn.EventAction.Close(); err != nil {
		return err
	}
	if err := wn.QueryAction.Close(); err != nil {
		return err
	}
	delete(s.sdk.Nodes, nodeID)
	return nil
}

// Stub provides fundamental functions for client.
type Stub interface {
	SetIdentity(algorithm string, key []byte, cert []byte)
	SetTLS(certPEMBlock []byte, keyPEMBlock []byte, roots [][]byte)
	AddNode(node *types.Node)
	Node(name string) (*types.Node, error)
	GetNode(name string) (*node.WNode, error)
	NodeByOrg(nodes []string, org string) ([]*node.WNode, error)
	ChainBuilder() *rawmessage.ChainRawMessage
	QueryBuilder() *rawmessage.QueryRawMessage
	ContractBuilder() *rawmessage.ContractRawMessage
	LifecycleBuilder() *rawmessage.LifecycleRawMessage
	ConfigBuilder() *rawmessage.ConfigRawMessage
	Crypto() crypto.Crypto
	RouterContains(node string) bool
	Seal(req proto.Marshaler) (*common.RawMessage, error)
	SealVote(chain string, msg proto.Marshaler) (*rawmessage.TxRawMsg, error)

	ReturnErr(err error)
	Err() error
	RemoveNode(n string) error
	Destroy()
}
