/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package flow

import (
	"context"

	"github.com/pkg/errors"
	"google.golang.org/grpc"

	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/flow/internal"
	"huawei.com/huaweichain/sdk/flow/status"
	"huawei.com/huaweichain/sdk/flow/types"
	"huawei.com/huaweichain/sdk/rawmessage"
)

// LifecycleFlow is used to handle the contract lifecycle operation
// which includes import, vote, freeze
type LifecycleFlow struct {
	cons  *internal.Consenter
	chain string

	contractName string
	version      string
	betaVersion  uint32
	packagePath  string
	platform     ContractPlatform

	historySupport bool
	initRequired   bool
	shardID        string
	stub           internal.Stub
}

// ContractPlatform represents supported contract platform.
type ContractPlatform string

// Optional ContractPlatform contains go, fabric go, assembly script, java and binary.
const (
	Go             ContractPlatform = "Go"
	FabricGo       ContractPlatform = "FabricGo"
	AssemblyScript ContractPlatform = "AssemblyScript"
	Java           ContractPlatform = "Java"
	Binary         ContractPlatform = "Binary"
)

func (cp ContractPlatform) String() string {
	return string(cp)
}

var platformToType = map[ContractPlatform]contractType{
	Go: {
		sandbox: "docker",
		format:  "go",
	},
	FabricGo: {
		sandbox: "docker",
		format:  "go",
	},
	Binary: {
		sandbox: "docker",
		format:  "binary",
	},
	Java: {
		sandbox: "docker",
		format:  "java",
	},
	AssemblyScript: {
		sandbox: "nativewasm",
		format:  "wasm",
	},
}

// contractType is used to pre-define the contract
type contractType struct {
	sandbox string
	format  string
}

// fromPlatform is used to get contract type
func fromPlatform(platform ContractPlatform) (*contractType, error) {
	t, ok := platformToType[platform]
	if ok {
		return &t, nil
	}
	return nil, errors.Errorf("unsupported contract platform %s", platform)
}

// LifecycleReaderFlow represents only read flow instance for contract lifecycle.
type LifecycleReaderFlow struct {
	stub  internal.Stub
	chain string
	node  string
	flow  *LifecycleFlow
}

func newLifecycleFlow(chain string, stub internal.Stub) *LifecycleFlow {
	lifecycleFlowInstance := LifecycleFlow{
		stub:  stub,
		cons:  internal.NewConsenterFlow(chain, stub),
		chain: chain,
	}
	return &lifecycleFlowInstance
}

// Reader assigns specific remote node for request.
func (l *LifecycleFlow) Reader(node string) *LifecycleReaderFlow {
	return &LifecycleReaderFlow{
		stub:  l.stub,
		chain: l.chain,
		node:  node,
		flow:  l,
	}
}

// Define is used to set parameters for LifecycleFlow
func (l *LifecycleFlow) Define(name, version, pkgPath string, platform ContractPlatform) *LifecycleFlow {
	l.contractName = name
	l.version = version
	l.packagePath = pkgPath
	l.platform = platform
	return l
}

// Contract sets contract basic info contains name and version.
func (l *LifecycleFlow) Contract(name, version string) *LifecycleFlow {
	l.contractName = name
	l.version = version
	return l
}

// Beta sets beta version for contract.
func (l *LifecycleFlow) Beta(version uint32) *LifecycleFlow {
	l.betaVersion = version
	return l
}

// EnableHistory is used to enable historySupport or not
func (l *LifecycleFlow) EnableHistory() *LifecycleFlow {
	l.historySupport = true
	return l
}

// RequireInit is used to enable initRequired
func (l *LifecycleFlow) RequireInit() *LifecycleFlow {
	l.initRequired = true
	return l
}

// Install is used to import contract
func (l *LifecycleFlow) Install(ctx context.Context, node string) error {
	t, err := fromPlatform(l.platform)
	if err != nil {
		return err
	}

	contract, err := l.contractDefinition()
	if err != nil {
		return err
	}
	var rawReq *common.RawMessage
	if l.platform == FabricGo {
		rawReq, err = l.stub.LifecycleBuilder().BuildImportFabricRawMessage(contract,
			l.packagePath, t.sandbox, t.format)
	} else {
		rawReq, err = l.stub.LifecycleBuilder().BuildImportRawMessage(contract,
			l.packagePath, t.sandbox, t.format)
	}
	if err != nil {
		return errors.WithMessage(err, "build lifecycle import raw message error")
	}

	client, err := l.client(node)
	if err != nil {
		return err
	}
	return l.instantiate(ctx, client.Import, rawReq)
}

type generalRpcFunc = func(ctx context.Context, in *common.RawMessage,
	opts ...grpc.CallOption) (*common.RawMessage, error)

func (l *LifecycleFlow) instantiate(ctx context.Context, f generalRpcFunc, rawReq *common.RawMessage) error {
	rawRes, err := f(ctx, rawReq)
	if err != nil {
		return err
	}
	var res common.Response
	if err = proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return errors.WithMessage(err, "errors occur in proto unmarshal")
	}
	if res.Status != common.SUCCESS {
		return status.ResponseToErr(&res)
	}
	return nil
}

func (l *LifecycleFlow) contractDefinition() (*rawmessage.Contract, error) {
	if l.chain == "" {
		return nil, errors.New("chain id is empty")
	}
	if l.contractName == "" {
		return nil, errors.New("contract name is empty")
	}
	if l.version == "" {
		return nil, errors.New("version is empty")
	}
	var contract *rawmessage.Contract
	if l.betaVersion == 0 {
		contract = rawmessage.NewContract(l.chain, l.contractName, l.version)
	} else {
		contract = rawmessage.NewBetaContract(l.chain, l.contractName, l.version, l.betaVersion)
	}
	return contract, nil
}

func (l *LifecycleReaderFlow) client() (nodeservice.ContractClient, error) {
	return l.flow.client(l.node)
}

func (l *LifecycleFlow) client(node string) (nodeservice.ContractClient, error) {
	stub, err := l.stub.GetNode(node)
	if err != nil {
		return nil, err
	}
	conn, err := stub.ContractAction.GetReadyConn()
	if err != nil {
		return nil, errors.WithMessage(err, "contract action import error")
	}
	client := nodeservice.NewContractClient(conn)
	return client, nil
}

// Uninstall is used to  unImport contract
func (l *LifecycleFlow) Uninstall(ctx context.Context, node string) error {
	definition, err := l.contractDefinition()
	if err != nil {
		return err
	}
	rawReq, err := l.stub.LifecycleBuilder().BuildUnImportRawMessage(definition)
	if err != nil {
		return errors.WithMessage(err, "errors occur in building unimport raw message")
	}
	client, err := l.client(node)
	if err != nil {
		return err
	}
	return l.instantiate(ctx, client.UnImport, rawReq)
}

// Freeze is used to freeze contract
func (l *LifecycleFlow) Freeze(ctx context.Context, contract string) (*types.Receipt, error) {
	return l.manage(ctx, contract, "freeze")
}

// Unfreeze is used to unfreeze contract
func (l *LifecycleFlow) Unfreeze(ctx context.Context, contract string) (*types.Receipt, error) {
	return l.manage(ctx, contract, "unfreeze")
}

// Remove is used to remove contract
func (l *LifecycleFlow) Remove(ctx context.Context, contract string) (*types.Receipt, error) {
	return l.manage(ctx, contract, "destroy")
}

func (l *LifecycleFlow) manage(ctx context.Context, contract string, option string) (*types.Receipt, error) {
	rawMessage, err := l.stub.LifecycleBuilder().BuildManageRawMessage(l.chain, contract, option)
	if err != nil {
		return nil, errors.WithMessage(err, "errors occur in building manage raw message")
	}
	r, err := l.cons.SendTx(ctx, rawMessage)
	if err != nil {
		return nil, errors.WithMessage(err, "errors occur in send tx")
	}
	return r, nil
}

// Vote is used to vote
func (l *LifecycleFlow) Vote(ctx context.Context, desc string, policy string) (*types.Receipt, error) {
	rawMessage, err := l.constructRawMessage(desc, policy)
	if err != nil {
		return nil, errors.WithMessage(err, "errors occur in constructRawMessage")
	}
	r, err := l.cons.SendTx(ctx, rawMessage)
	if err != nil {
		return nil, errors.WithMessage(err, "errors occur in sendTx")
	}
	return r, nil
}

func (l *LifecycleFlow) constructRawMessage(desc string, policy string) (*rawmessage.TxRawMsg, error) {
	definition, err := l.contractDefinition()
	if err != nil {
		return nil, err
	}

	var rawMessage *rawmessage.TxRawMsg
	if l.shardID == "" {
		if rawMessage, err = l.stub.LifecycleBuilder().BuildVoteRawMessage(definition,
			desc, policy, l.historySupport, l.initRequired); err != nil {
			return nil, errors.WithMessage(err, "errors occur in building vote raw message")
		}
	} else {
		voteParam := &rawmessage.VoteParams{
			C:              definition,
			ShardID:        l.shardID,
			Desc:           desc,
			Policy:         policy,
			HistorySupport: l.historySupport,
			InitRequired:   l.initRequired,
		}
		if rawMessage, err = l.stub.LifecycleBuilder().BuildShardVoteRawMessage(voteParam); err != nil {
			return nil, errors.WithMessage(err, "errors occur in building shard vote raw message")
		}
	}

	return rawMessage, nil
}

// Consenter is used to set consenterList
func (l *LifecycleFlow) Consenter(consenter ...string) *LifecycleFlow {
	l.cons.AvailableConsenter(consenter)
	return l
}

// Listener for transaction result event listening.
func (l *LifecycleFlow) Listener(listener string) *LifecycleFlow {
	l.cons.SetListener(listener)
	return l
}

// ShardID is used to set shardID
func (l *LifecycleFlow) ShardID(shardID string) *LifecycleFlow {
	l.shardID = shardID
	l.cons.ShardID(shardID)
	return l
}

// ContractState is used to query contract state
func (l *LifecycleReaderFlow) ContractState(ctx context.Context, contract string) (string, error) {
	rawReq, err := l.stub.LifecycleBuilder().BuildQueryStateRawMessage(l.chain, contract)
	if err != nil {
		return "", errors.WithMessage(err, "errors occur in building query contract state raw message")
	}
	client, err := l.client()
	if err != nil {
		return "", err
	}
	rawRes, err := client.QueryState(ctx, rawReq)
	if err != nil {
		return "", err
	}
	var res common.Response
	if err := proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return "", errors.WithMessage(err, "unmarshal response error")
	}
	if res.Status != common.SUCCESS {
		return "", status.ResponseToErr(&res)
	}
	return res.StatusInfo, nil
}

// Lifecycle represents contract lifecycle.
type Lifecycle string

// Optional lifecycle contains starting, freezing, unfreezing, destroying.
const (
	Start    Lifecycle = "start"
	Freeze   Lifecycle = "freeze"
	Unfreeze Lifecycle = "unfreeze"
	Destroy  Lifecycle = "destroy"
)

// VoteState query lifecycle vote state for specific contract.
func (l *LifecycleReaderFlow) VoteState(ctx context.Context, contract string, lifecycle Lifecycle) (
	[]*nodeservice.Vote, error) {
	rawReq, err := l.stub.QueryBuilder().BuildQueryLifecycleVoteRawMessage(l.chain, contract, string(lifecycle))
	if err != nil {
		return nil, err
	}
	stub, err := l.stub.GetNode(l.node)
	if err != nil {
		return nil, err
	}
	conn, err := stub.QueryAction.GetReadyConn()
	if err != nil {
		return nil, errors.WithMessage(err, "contract action import error")
	}
	client := nodeservice.NewVoteManagerClient(conn)
	rawRes, err := client.Query(ctx, rawReq)
	if err != nil {
		return nil, err
	}
	var res common.Response
	if err := proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return nil, errors.WithMessage(err, "")
	}
	if res.Status != common.SUCCESS {
		return nil, status.ResponseToErr(&res)
	}
	var vote nodeservice.VoteQueryResponse
	if err := proto.Unmarshal(res.Payload, &vote); err != nil {
		return nil, errors.WithMessage(err, "unmarshal vote query response error")
	}
	return vote.Votes, nil
}
