/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package flow

import (
	"context"
	"encoding/hex"
	"encoding/json"
	"fmt"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/action"
	"huawei.com/huaweichain/sdk/flow/internal"
	"huawei.com/huaweichain/sdk/flow/rest"
	"huawei.com/huaweichain/sdk/flow/status"
)

// ObserverFlow is the role of observer
type ObserverFlow struct {
	stub  internal.Stub
	chain string
	node  string
}

func newObserverFlow(stub internal.Stub) *ObserverFlow {
	return &ObserverFlow{
		stub: stub,
	}
}

// Genesis is used to query genesis block
func (o *ObserverFlow) Genesis(ctx context.Context) (*common.Block, error) {
	return o.Block(ctx, 0)
}

func (o *ObserverFlow) client() (nodeservice.ChainServiceClient, error) {
	node, err := o.stub.GetNode(o.node)
	if err != nil {
		return nil, err
	}
	conn, err := node.QueryAction.GetReadyConn()
	if err != nil {
		return nil, err
	}
	return nodeservice.NewChainServiceClient(conn), nil
}
func (o *ObserverFlow) restClient() (*rest.Client, error) {
	node, err := o.stub.GetNode(o.node)
	if err != nil {
		return nil, err
	}
	return rest.NewRestClient(node)
}

func (o *ObserverFlow) shardClient() (nodeservice.ShardServiceClient, error) {
	node, err := o.stub.GetNode(o.node)
	if err != nil {
		return nil, err
	}
	conn, err := node.QueryAction.GetReadyConn()
	if err != nil {
		return nil, err
	}
	return nodeservice.NewShardServiceClient(conn), nil
}

// ShardNodeRouter is used to query node router in shard scene.
func (o *ObserverFlow) ShardNodeRouter(ctx context.Context) (string, error) {
	msg, err := o.stub.QueryBuilder().BuildShardRouterInfoRawMessage(o.node)
	if err != nil {
		return "", errors.WithMessage(err, "failed to build shard router info raw message")
	}
	client, err := o.shardClient()
	if err != nil {
		return "", errors.WithMessage(err, "failed to build shard client")
	}
	rawRes, err := client.GetShardRouterInfo(ctx, msg)
	if err != nil {
		return "", errors.WithMessage(err, "failed to query shard router info")
	}
	var res common.Response
	if err = proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return "", errors.WithMessage(err, "failed to unmarshal response")
	}
	if res.Status != common.SUCCESS {
		return "", status.ResponseToErr(&res)
	}
	var info nodeservice.ShardRouterInfo
	if err = proto.Unmarshal(res.Payload, &info); err != nil {
		return "", errors.WithMessagef(err,
			"failed to unmarshal shard router information with payload %s", string(rawRes.Payload))
	}
	return info.Result, nil
}

// Block is used to query block through block num
func (o *ObserverFlow) Block(ctx context.Context, num uint64) (*common.Block, error) {
	rawMsg, err := o.stub.QueryBuilder().BuildBlockRawMessage(o.chain, num)
	if err != nil {
		return nil, errors.WithMessage(err, "build create chain raw message error")
	}
	client, err := o.client()
	if err != nil {
		return nil, err
	}
	responseRaw, err := client.GetBlockByNum(ctx, rawMsg)
	if err != nil {
		return nil, errors.WithMessage(err, "query action query block by blockheight error")
	}
	return o.parseBlock(responseRaw)
}

func (o *ObserverFlow) parseBlock(responseRaw *common.RawMessage) (*common.Block, error) {
	response := &common.Response{}
	if err := proto.Unmarshal(responseRaw.Payload, response); err != nil {
		return nil, errors.WithMessage(err, "unmarshal responseraw error: %v")
	}
	if response.Status != common.SUCCESS {
		return nil, status.ResponseToErr(response)
	}
	block := &common.Block{}
	if err := proto.Unmarshal(response.Payload, block); err != nil {
		return nil, errors.WithMessage(err, "unmarshal block error")
	}
	return block, nil
}

// Height is used to query chain state
func (o *ObserverFlow) Height(ctx context.Context) (uint64, error) {
	req, err := o.stub.QueryBuilder().BuildLatestChainStateRawMessage(o.chain)
	if err != nil {
		return 0, errors.WithMessage(err, "build query chain state raw message error")
	}
	client, err := o.client()
	if err != nil {
		return 0, err
	}
	rawRes, err := client.GetLatestChainState(ctx, req)
	if err != nil {
		return 0, errors.WithMessage(err, "query action get latest chain state error")
	}
	var res common.Response
	if err = proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return 0, errors.WithMessage(err, "unmarshal responseraw error: %v")
	}
	if res.Status != common.SUCCESS {
		return 0, status.ResponseToErr(&res)
	}
	var state nodeservice.LatestChainState
	if err := proto.Unmarshal(res.Payload, &state); err != nil {
		return 0, errors.Wrap(err, "failed to parse chain configuration data")
	}

	return state.Height, nil
}

// Tx is used to query tx
func (o *ObserverFlow) Tx(ctx context.Context, txID string) (*common.Tx, error) {
	txHash, err := hex.DecodeString(txID)
	if err != nil {
		return nil, errors.Wrap(err, "failed to parse tx id")
	}
	rawReq, err := o.stub.QueryBuilder().BuildTxRawMessage(o.chain, txHash)
	if err != nil {
		return nil, errors.WithMessage(err, "build query tx raw message error")
	}
	client, err := o.client()
	if err != nil {
		return nil, err
	}
	rawRes, err := client.GetTxByHash(ctx, rawReq)
	if err != nil {
		return nil, errors.WithMessage(err, "query action gettxbyhash error")
	}
	var res common.Response
	if err := proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return nil, errors.WithMessage(err, "unmarshal responseraw error: %v")
	}
	if res.Status != common.SUCCESS {
		return nil, status.ResponseToErr(&res)
	}
	var tx common.Tx
	if err := proto.Unmarshal(res.Payload, &tx); err != nil {
		return nil, errors.WithMessage(err, "unmarshal tx error")
	}
	return &tx, nil
}

// TxResult is used to query tx result
func (o *ObserverFlow) TxResult(ctx context.Context, txID string) (common.TxStatus, error) {
	txHash, err := hex.DecodeString(txID)
	if err != nil {
		return 0, err
	}
	rawReq, err := o.stub.QueryBuilder().BuildTxRawMessage(o.chain, txHash)
	if err != nil {
		return 0, errors.WithMessage(err, "build query tx result raw message error")
	}
	client, err := o.client()
	if err != nil {
		return 0, err
	}
	rawRes, err := client.GetTxResultByTxHash(ctx, rawReq)
	if err != nil {
		return 0, errors.WithMessage(err, "query action gettxresultbytxhash error")
	}
	var res common.Response
	if err := proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return 0, errors.WithMessage(err, "unmarshal responseraw error: %v")
	}
	if res.Status != common.SUCCESS {
		return 0, status.ResponseToErr(&res)
	}
	var txResult common.TxResult
	if err := proto.Unmarshal(res.Payload, &txResult); err != nil {
		return 0, errors.WithMessage(err, "unmarshal transaction error")
	}
	return txResult.Status, nil
}

// BlockByTxHash query raw block data by tx hash which contains.
func (o *ObserverFlow) BlockByTxHash(ctx context.Context, txID string) (*common.Block, error) {
	txHash, err := hex.DecodeString(txID)
	if err != nil {
		return nil, errors.WithMessage(err, "decode tx id")
	}
	rawMsg, err := o.stub.QueryBuilder().BuildTxRawMessage(o.chain, txHash)
	if err != nil {
		return nil, errors.WithMessage(err, "build tx raw message error")
	}
	client, err := o.client()
	if err != nil {
		return nil, err
	}
	rawRes, err := client.GetBlockByTxHash(ctx, rawMsg)
	if err != nil {
		return nil, errors.WithMessage(err, "")
	}
	return o.parseBlock(rawRes)
}

// Contract query specific contract definition detail info.
func (o *ObserverFlow) Contract(ctx context.Context, name string) (*common.ContractDefinition, error) {
	rawReq, err := o.stub.QueryBuilder().BuildContractRawMessage(o.chain, name)
	if err != nil {
		return nil, errors.WithMessage(err, "build contract raw message error")
	}
	client, err := o.client()
	if err != nil {
		return nil, err
	}
	rawRes, err := client.GetContractInfo(ctx, rawReq)
	if err != nil {
		return nil, errors.WithMessage(err, "query action get contract info error")
	}
	var res common.Response
	if err = proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return nil, errors.WithMessage(err, "unmarshal response error")
	}
	if res.Status != common.SUCCESS {
		return nil, status.ResponseToErr(&res)
	}
	var definition common.ContractDefinition
	if err = proto.Unmarshal(res.Payload, &definition); err != nil {
		return nil, errors.WithMessage(err, "unmarshal block error")
	}
	return &definition, nil
}

// ContractInfo represents contract detail information.
type ContractInfo struct {
	Status  string `json:"status"`
	Message string `json:"message"`
	*common.ContractDefinition
	Policy string `json:"policy,omitempty"`
}

// Contracts queries contract list from specific node by chain.
func (o *ObserverFlow) Contracts(ctx context.Context) ([]*ContractInfo, error) {
	cli, err := o.restClient()
	if err != nil {
		return nil, err
	}
	if o.chain == "" {
		return nil, errors.New("chain id is empty")
	}
	path := fmt.Sprintf("/node/stat/chain/contracts?chain=%s", o.chain)
	body, err := cli.Send(ctx, path)
	if err != nil {
		return nil, err
	}
	res := struct {
		Contracts []*common.Response `json:"contracts"`
	}{}
	if err = json.Unmarshal(body, &res); err != nil {
		return nil, errors.Wrapf(err, "failed to unmarshal json payload with %s", body)
	}
	cs := make([]*ContractInfo, 0, len(res.Contracts))
	for _, c := range res.Contracts {
		ci, err := parseContractInfo(c)
		if err != nil {
			return nil, err
		}
		cs = append(cs, ci)
	}
	return cs, nil
}

func parseContractInfo(c *common.Response) (*ContractInfo, error) {
	var cd common.ContractDefinition
	err := proto.Unmarshal(c.Payload, &cd)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to parse contract list from response")
	}
	extensions := cd.ValidatorExtensions
	cd.ValidatorExtensions = nil
	ci := &ContractInfo{
		Status:             c.Status.String(),
		Message:            c.StatusInfo,
		ContractDefinition: &cd,
	}
	for k, v := range extensions {
		switch k {
		case "policy":
			ci.Policy = string(v)
		default:
		}
	}
	return ci, nil
}

// StateByPrefix query world state for specific contract by key prefix.
func (o *ObserverFlow) StateByPrefix(ctx context.Context, contract, prefix string) (*StateIter, error) {
	rawReq, err := o.stub.QueryBuilder().BuildGetStateRawMessage(o.chain, contract, prefix)
	if err != nil {
		return nil, err
	}
	node, err := o.stub.GetNode(o.node)
	if err != nil {
		return nil, err
	}
	return node.QueryAction.StateOfChain(ctx, rawReq)
}

// StateByRange query world state for specific contract by range from begin to end.
func (o *ObserverFlow) StateByRange(ctx context.Context, contract, begin, end string) (*StateIter, error) {
	rawReq, err := o.stub.QueryBuilder().BuildGetStateByRangeRawMessage(o.chain, contract, begin, end)
	if err != nil {
		return nil, err
	}
	node, err := o.stub.GetNode(o.node)
	if err != nil {
		return nil, err
	}
	return node.QueryAction.StateOfChain(ctx, rawReq)
}

// StateIter represents iterator for world state stream querying.
type StateIter = action.StreamChainStateIterator

// Target sets chain id and node id for construct specific query request.
func (o *ObserverFlow) Target(chain, node string) *ObserverFlow {
	o.chain = chain
	o.node = node
	return o
}
