/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

// Package rest provides restful api client for querying information about chain and node.
package rest

import (
	"context"
	"fmt"
	"net/http"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/cryptomgr"
	"huawei.com/huaweichain/common/logger"
	"huawei.com/huaweichain/sdk/client"
	"huawei.com/huaweichain/sdk/crypto"
	"huawei.com/huaweichain/sdk/node"
)

var lg = logger.GetModuleLogger("sdk.rest", "")

var (
	registry    = make(map[string]func(*node.WNode) (*Client, error))
	rawRegistry = make(map[string]func(string, *crypto.TLS) (httpClient, error))
)

// nolint:gochecknoinits // register ecdsa crypto for tls configuration.
func init() {
	registry[cryptomgr.EcdsaWithSha256] = func(node *node.WNode) (*Client, error) {
		c, err := crypto.EcTLSClientConf(node.HostOverride, node.TLS)
		if err != nil {
			return nil, errors.WithMessagef(err, "failed to export http tls configuration from node %s", node.ID)
		}
		return &Client{
			httpClient: &ecClient{
				conf: c,
			},
			protocol: "https",
			host:     node.Host,
			port:     node.RestPort,
		}, nil
	}

	rawRegistry[cryptomgr.EcdsaWithSha256] = func(hostname string, tls *crypto.TLS) (httpClient, error) {
		conf, err := crypto.EcTLSClientConf(hostname, tls)
		if err != nil {
			return nil, err
		}
		return &ecClient{conf: conf}, nil
	}
}

// Client wrappers http.Client for rest api client.
type Client struct {
	httpClient
	protocol string
	host     string
	port     int
}

// NewClient creates a rest client instance by GatewayClient and specific node name.
func NewClient(sdk *client.GatewayClient, nodeName string) (*Client, error) {
	n, ok := sdk.GetNode(nodeName)
	if !ok {
		return nil, errors.Errorf("node %s doesn't exsit in local network configuration", nodeName)
	}
	return NewRestClient(n)
}

// NewRestClient creates a new rest http client for node rest api.
func NewRestClient(node *node.WNode) (*Client, error) {
	if node.RestPort == 0 {
		return nil, errors.Errorf("failed to initialize rest client for node %s "+
			"without rest port in local configuration", node.ID)
	}
	if !node.TLSEnable() {
		return &Client{
			httpClient: newInsecureClient(),
			protocol:   "http",
			host:       node.Host,
			port:       node.RestPort,
		}, nil
	}
	f, ok := registry[node.TLS.Algorithm]
	if !ok {
		return nil, errors.Errorf("unsupported algorithm %s", node.TLS.Algorithm)
	}
	return f(node)
}

// SendWithCode http GET request by rest api path.
func (c *Client) SendWithCode(ctx context.Context, path string) (int, []byte, error) {
	url := fmt.Sprintf("%s://%s:%d%s", c.protocol, c.host, c.port, path)
	lg.Debugf("rest api url = %s", url)
	return c.sendWithCode(ctx, http.MethodGet, url, nil)
}

// Send http GET request by rest api path.
func (c *Client) Send(ctx context.Context, path string) ([]byte, error) {
	code, body, err := c.SendWithCode(ctx, path)
	if err != nil {
		return nil, err
	}
	if code != http.StatusOK {
		return nil, errors.Errorf("failed to handle rest http request with code: %d and message: %s", code, body)
	}
	return body, nil
}
