//go:build gm && linux
// +build gm,linux

/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package rest

import (
	"bufio"
	"bytes"
	"context"
	"io"
	"net/http"
	"net/url"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/cryptomgr"
	"huawei.com/huaweichain/gmssl/tls"
	"huawei.com/huaweichain/sdk/crypto"
	"huawei.com/huaweichain/sdk/node"
)

// nolint:gochecknoinits // register GM crypto for tls configuration.
func init() {
	registry[cryptomgr.Sm2WithSm3] = func(node *node.WNode) (*Client, error) {
		return &Client{
			httpClient: &gmClient{
				conf: crypto.GmTLSClientConf(node.HostOverride, node.TLS),
			},
			protocol: "https",
			host:     node.Host,
			port:     node.RestPort,
		}, nil
	}

	rawRegistry[cryptomgr.Sm2WithSm3] = func(hostname string, tls *crypto.TLS) (httpClient, error) {
		conf := crypto.GmTLSClientConf(hostname, tls)
		return &gmClient{
			conf: conf,
		}, nil
	}
}

type gmClient struct {
	conf *tls.Config
}

func (g *gmClient) sendWithCode(ctx context.Context, method string, rawURL string,
	payload []byte) (int, []byte, error) {
	uri, err := url.ParseRequestURI(rawURL)
	if err != nil {
		return 0, nil, err
	}
	conn, err := tls.Dial("tcp", uri.Host, g.conf)
	if err != nil {
		return 0, nil, errors.WithMessage(err, "dial error")
	}
	defer func() {
		if err = conn.Close(); err != nil {
			lg.Errorf("%+v", err)
		}
	}()

	res, err := g.doSend(ctx, method, rawURL, conn, payload)
	if err != nil {
		return 0, nil, err
	}
	return handleRes(res)
}

func (g *gmClient) send(ctx context.Context, method string, rawURL string, payload []byte) ([]byte, error) {
	_, body, err := g.sendWithCode(ctx, method, rawURL, payload)
	if err != nil {
		return nil, err
	}
	return body, nil
}

func (g *gmClient) doSend(ctx context.Context, method string, rawURL string, conn io.ReadWriter,
	payload []byte) (*http.Response, error) {
	lg.Debugf("rest api url = %s", rawURL)
	req, err := http.NewRequestWithContext(ctx, method, rawURL, bytes.NewBuffer(payload))
	if err != nil {
		return nil, err
	}
	req.Header.Set("Content-Type", "application/json")
	if err := req.Write(conn); err != nil {
		return nil, errors.WithMessage(err, "write msg error")
	}
	reader := bufio.NewReader(conn)
	if _, err := reader.Peek(1); err != nil {
		return nil, errors.WithMessage(err, "read response error")
	}
	return http.ReadResponse(reader, req)
}
