/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package rest

import (
	"context"
	"encoding/hex"
	"encoding/json"
	goerrors "errors"
	"fmt"
	"net/http"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/sdk/client"
)

// ErrCheckConsenterTimeout returns timeout error when check node whether consensus node.
var ErrCheckConsenterTimeout = goerrors.New("node is busy, check consenter timeout")

// Node represents rest client for fetching information about node.
type Node struct {
	sdk *client.GatewayClient
}

// NewNode creates an instance of rest client about node info.
func NewNode(sdk *client.GatewayClient) *Node {
	return &Node{sdk: sdk}
}

// Payload represents response json payload data structure.
type Payload struct {
	IsConsenter bool `json:"is_consenter"`
}

// IsConsenter checks whether the node is consenter for specific chain.
func (n *Node) IsConsenter(ctx context.Context, node string, chain string) (bool, error) {
	c, err := NewClient(n.sdk, node)
	if err != nil {
		return false, err
	}
	return IsConsenter(ctx, c, chain)
}

// IsConsenter checks whether the node is consenter for specific chain.
func IsConsenter(ctx context.Context, c *Client, chain string) (bool, error) {
	path := fmt.Sprintf("/node/stat?chain=%s", chain)
	code, body, err := c.SendWithCode(ctx, path)
	if err != nil {
		return false, err
	}
	if code == http.StatusInternalServerError {
		return false, errors.Wrapf(ErrCheckConsenterTimeout, "failed to fetch node role from rest api with message %s", body)
	}
	if code != http.StatusOK {
		return false, errors.Errorf("failed to fetch node role from rest api "+
			"with code %d and message %s", code, body)
	}
	var p Payload
	if err := json.Unmarshal(body, &p); err != nil {
		return false, errors.Wrapf(err, "failed to unmarshal json payload with %s", body)
	}
	return p.IsConsenter, nil
}

const hashBytesLen = 32

// GetCertByHash query PEM format certificate data by hash of certificate in hex string.
func GetCertByHash(ctx context.Context, c *Client, chain string, hash string) ([]byte, error) {
	if chain == "" {
		return nil, errors.New("chain id is empty")
	}
	hashBytes, err := hex.DecodeString(hash)
	if err != nil {
		return nil, errors.Wrapf(err, "hash hex string %s is invalid", hash)
	}
	if len(hashBytes) != hashBytesLen {
		return nil, errors.Errorf("cert hash should be 32 bytes, but current is %s", hash)
	}

	path := fmt.Sprintf("/node/stat/chain/cert?chain=%s&cert_hash=%s", chain, hash)
	body, err := c.Send(ctx, path)
	if err != nil {
		return nil, err
	}
	res := struct {
		Success bool   `json:"success"`
		PemCert []byte `json:"pem_cert"`
		Message string `json:"message"`
	}{}
	if err = json.Unmarshal(body, &res); err != nil {
		return nil, errors.Wrapf(err, "failed to unmarshal json payload with %s", body)
	}
	if !res.Success {
		return nil, errors.New(res.Message)
	}
	if len(res.PemCert) == 0 {
		return nil, errors.New("certificate is empty")
	}
	return res.PemCert, nil
}

// GetCertMap query all PEM format certificates with hash of certificate in hex string.
func GetCertMap(ctx context.Context, c *Client, chain string) (map[string][]byte, error) {
	if chain == "" {
		return nil, errors.New("chain id is empty")
	}

	path := fmt.Sprintf("/node/stat/chain/certs?chain=%s", chain)
	body, err := c.Send(ctx, path)
	if err != nil {
		return nil, err
	}
	res := struct {
		Success bool              `json:"success"`
		CertMap map[string][]byte `json:"cert_map"`
		Message string            `json:"message"`
	}{}
	if err = json.Unmarshal(body, &res); err != nil {
		return nil, errors.Wrapf(err, "failed to unmarshal json payload with %s", body)
	}
	if !res.Success {
		return nil, errors.New(res.Message)
	}
	return res.CertMap, nil
}
