/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package rest

import (
	"bytes"
	"context"
	"crypto/tls"
	"io/ioutil"
	"net/http"
	"net/url"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/sdk/client"
	"huawei.com/huaweichain/sdk/crypto"
	"huawei.com/huaweichain/sdk/node"
)

// Curl executes raw http method with url.
func Curl(ctx context.Context, sdk *client.GatewayClient, method, rawURL string, payload []byte) (int, []byte, error) {
	uri, err := url.ParseRequestURI(rawURL)
	if err != nil {
		return 0, nil, err
	}
	if uri.Scheme != "https" {
		c := newInsecureClient()
		return c.sendWithCode(ctx, method, rawURL, payload)
	}

	c := sdk.Conf.Client
	conf, err := node.ConvertTLS(c.TLS, c.Type, sdk.Decrypt)
	if err != nil {
		return 0, nil, err
	}
	cli, err := newHTTPClient(c.Type, conf, uri)
	if err != nil {
		return 0, nil, err
	}
	return cli.sendWithCode(ctx, method, rawURL, payload)
}

type ecClient struct {
	conf *tls.Config
}

func (e *ecClient) send(ctx context.Context, method string, rawURL string, payload []byte) ([]byte, error) {
	_, body, err := e.sendWithCode(ctx, method, rawURL, payload)
	if err != nil {
		return nil, err
	}
	return body, nil
}

func (e *ecClient) sendWithCode(ctx context.Context, method string, rawURL string,
	payload []byte) (int, []byte, error) {
	c := &http.Client{
		Transport: &http.Transport{
			TLSClientConfig:   e.conf,
			DisableKeepAlives: true,
		},
	}
	return sendWithCode(ctx, method, rawURL, payload, c)
}

type httpClient interface {
	send(ctx context.Context, method string, rawURL string, payload []byte) ([]byte, error)
	sendWithCode(ctx context.Context, method string, rawURL string, payload []byte) (int, []byte, error)
}

type insecureClient struct {
	*http.Client
}

func (c insecureClient) send(ctx context.Context, method string, rawURL string, payload []byte) ([]byte, error) {
	_, body, err := c.sendWithCode(ctx, method, rawURL, payload)
	if err != nil {
		return nil, err
	}
	return body, nil
}

func (c insecureClient) sendWithCode(ctx context.Context, method string, rawURL string,
	payload []byte) (int, []byte, error) {
	return sendWithCode(ctx, method, rawURL, payload, c.Client)
}

func sendWithCode(ctx context.Context, method string, rawURL string, payload []byte, c *http.Client) (
	int, []byte, error) {
	defer c.CloseIdleConnections()

	req, err := http.NewRequestWithContext(ctx, method, rawURL, bytes.NewBuffer(payload))
	if err != nil {
		return 0, nil, err
	}
	req.Header.Set("Content-Type", "application/json")
	res, err := c.Do(req)
	if err != nil {
		return 0, nil, err
	}
	return handleRes(res)
}

func newHTTPClient(alg string, tls *crypto.TLS, url *url.URL) (httpClient, error) {
	hostname := url.Hostname()
	f, ok := rawRegistry[alg]
	if !ok {
		return nil, errors.Errorf("unsupported algorithm %s", alg)
	}
	return f(hostname, tls)
}

func newInsecureClient() *insecureClient {
	return &insecureClient{Client: &http.Client{Transport: &http.Transport{DisableKeepAlives: true}}}
}

func handleRes(res *http.Response) (int, []byte, error) {
	defer func() {
		if err := res.Body.Close(); err != nil {
			lg.Error("failed to close request body")
		}
	}()
	body, err := ioutil.ReadAll(res.Body)
	if err != nil {
		return res.StatusCode, nil, errors.Wrapf(err, "failed to read response body with %s", res.Body)
	}
	return res.StatusCode, body, nil
}
