/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package flow

import (
	"context"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/util"
	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/shard"
	"huawei.com/huaweichain/sdk/flow/types"
)

// ShardPolicy is alias of proto shard.ShardPolicyType
type ShardPolicy = shard.ShardPolicyType

// ShardPolicy type definitions.
const (
	ShardPolicyHashMod = shard.HashMod
	ShardPolicyRange   = shard.Range
	ShardPolicyConHash = shard.ConsHash
	ShardPolicyCustom  = shard.Custom
)

const (
	// shardHandler is the shard handler string
	shardHandler = "shard"
	// shardUpdate is a subject of hard policy handler
	shardUpdate = "update"
)

// ShardFlow represents specific flow for shard config.
type ShardFlow struct {
	flow        *ConfigFlow
	shardID     string
	policyBytes []byte
}

// ShardID is used to set shard id.
func (s *ShardFlow) ShardID(shardID string) *ShardFlow {
	s.shardID = shardID
	return s
}

// CustomPolicyBytecode sets compiled wasm bytecode for custom sharding policy.
func (s *ShardFlow) CustomPolicyBytecode(bytecode []byte) *ShardFlow {
	s.policyBytes = bytecode
	return s
}

// UpdateShardPolicy sets shard chain policy.
func (s *ShardFlow) UpdateShardPolicy(ctx context.Context, policy ShardPolicy) (*types.Receipt, error) {
	chainID := s.flow.chain
	if err := util.CheckChainID(chainID); err != nil {
		return nil, errors.WithMessagef(err, "failed to validate chain id %s", chainID)
	}
	content, err := s.policyContent(policy)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to get policy content")
	}
	def := shard.ShardPolicyDefinition{
		Type:    policy,
		Content: content,
	}
	defBytes, err := proto.Marshal(&def)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to marshal shard policy definition")
	}
	voteTxData := common.VoteTxData{
		Handler: shardHandler,
		Subject: shardUpdate,
		Payload: defBytes,
	}

	msg, err := s.flow.stub.SealVote(chainID, &voteTxData)
	if err != nil {
		return nil, errors.WithMessage(err, "errors occur in building manage raw message")
	}

	return s.flow.cons.ShardID(s.shardID).SendTx(ctx, msg)
}

func (s *ShardFlow) policyContent(policy ShardPolicy) ([]byte, error) {
	if policy == ShardPolicyCustom {
		if len(s.policyBytes) == 0 {
			return nil, errors.New("custom sharding policy, but compiled policy bytecode not found")
		}
		return s.policyBytes, nil
	}
	return []byte(policy.String()), nil
}

// P2PSeed is the definition of p2p Node.
type P2PSeed struct {
	Host string
	Port uint64
}
