/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package flow

import (
	"context"
	"fmt"
	"io"
	"strconv"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/logger"
	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/flow/internal"
)

type joinParams struct {
	chain     string
	node      string
	existNode string

	stub   internal.Stub
	stream nodeservice.ChainManager_JoinChainWithSnapshotClient
}

// joinWithSnap join chain with snapshot.
func joinWithSnap(ctx context.Context, cancel context.CancelFunc, p *joinParams) Future {
	lg := logger.GetModuleLogger("sdk.flow.snap", "join_with_snap")
	r := &result{
		res:    make(chan Response, 1),
		cancel: cancel,
		lg:     lg,
	}
	a := joinProcessor{
		joinParams: p,
		res:        r,
		lg:         lg,
	}
	go a.join(ctx)
	return r
}

type joinProcessor struct {
	*joinParams
	res *result
	lg  *logger.WienerLogger
}

func (p *joinProcessor) join(ctx context.Context) {
	if err := p.handleSnap(); err != nil {
		p.sendNodeError(p.node, err)
		return
	}
	target, err := loadStub(p.stub).User().Observer().
		Target(p.chain, p.existNode).Height(ctx)
	if err != nil {
		p.sendNodeError(p.existNode, err)
		return
	}
	p.sendJoin(target)
	if err := p.sync(ctx, target); err != nil {
		p.sendNodeError(p.node, err)
		return
	}
	p.sendSynchronized()
	p.finish()
}

func (p *joinProcessor) sync(ctx context.Context, target uint64) error {
	resCh, err := loadStub(p.stub).User().Event().Target(p.chain, p.node).BlockNumEvent(ctx)
	if err != nil {
		return err
	}
	for {
		select {
		case res, ok := <-resCh:
			if !ok {
				return errors.New("listener channel is closed")
			}
			h, err := res.Get()
			if err != nil {
				return err
			}
			h++
			if h >= target {
				p.sendSyncProgress(p.node, h, target)
				return nil
			} else if h%stepNum == 0 {
				p.sendSyncProgress(p.node, h, target)
			}
		case <-ctx.Done():
			return ctx.Err()
		}
	}
}

func (p *joinProcessor) handleSnap() error {
	for {
		msg, err := p.stream.Recv()
		if err != nil {
			if errors.Is(err, io.EOF) {
				return nil
			}
			return errors.Wrap(err, "receive error from rpc stream")
		}
		var resp common.Response
		if err := proto.Unmarshal(msg.Payload, &resp); err != nil {
			return errors.WithMessage(err, "failed to unmarshal response")
		}
		if resp.Status != common.SUCCESS {
			return errors.Errorf("receive response with error code %s and message: %s", resp.Status, resp.StatusInfo)
		}

		var joinResp common.JoinResponse
		if err := proto.Unmarshal(resp.Payload, &joinResp); err != nil {
			return errors.WithMessage(err, "failed to unmarshal join response")
		}
		progress, err := handleJoinSnapResp(&joinResp)
		if err != nil {
			return err
		}
		p.res.send(progress)
		if joinResp.Status == common.JOIN_CHAIN_FAILED {
			return nil
		}
	}
}

func handleJoinSnapResp(joinResp *common.JoinResponse) (*snapProgress, error) {
	var progress snapProgress
	switch joinResp.Status {
	case common.FETCH_SNAPSHOT_PROCESSING:
		progress = snapProgress{
			SnapStep: FetchProcessing,
			Step:     FetchProcessing.String(),
			Message:  joinResp.StatusInfo,
		}
	case common.FETCH_SNAPSHOT_SUCCESS:
		progress = snapProgress{
			SnapStep: FetchSuccess,
			Step:     FetchSuccess.String(),
			Message:  joinResp.StatusInfo,
		}
	case common.FETCH_SNAPSHOT_FAILED:
		progress = snapProgress{
			SnapStep: FetchFailed,
			Step:     FetchFailed.String(),
			Message:  joinResp.StatusInfo,
		}
	case common.JOIN_CHAIN_PENDING:
		progress = snapProgress{
			SnapStep: JoinPending,
			Step:     JoinPending.String(),
			Message:  joinResp.StatusInfo,
		}
	case common.JOIN_CHAIN_FAILED:
		progress = snapProgress{
			SnapStep: JoinFailed,
			Step:     JoinFailed.String(),
			Message:  joinResp.StatusInfo,
		}
	case common.JOIN_CHAIN_SUCCESS:
		progress = snapProgress{
			SnapStep: JoinWithSnapSuccess,
			Step:     JoinWithSnapSuccess.String(),
			Message:  joinResp.StatusInfo,
		}
	default:
		return nil, errors.Errorf("receive unknown status, response status: %s, status info: %v",
			joinResp.Status, joinResp.StatusInfo)
	}
	return &progress, nil
}

func (p *joinProcessor) sendJoin(height uint64) {
	progress := snapProgress{
		SnapStep: JoinWithSnapSuccess,
		Step:     JoinWithSnapSuccess.String(),
		BlockGap: height,
		Message:  "node joined into chain",
	}
	p.res.send(&progress)
}

func (p *joinProcessor) sendNodeError(node string, err error) {
	progress := snapProgress{
		SnapStep: Abort,
		Step:     Abort.String(),
		Message:  err.Error(),
		NodeErrs: map[string]error{node: err},
	}
	p.res.send(&progress)
}

func (p *joinProcessor) sendSyncProgress(node string, height uint64, target uint64) {
	var progress snapProgress
	if height >= target {
		progress = snapProgress{
			SnapStep: BlockSyncProgress,
			Step:     BlockSyncProgress.String(),
			BlockGap: target,
			SyncPercent: map[string]float32{
				node: 1,
			},
			Message: fmt.Sprintf("%s's block height is synchronized", node),
		}
	} else {
		pre := float32(height*100) / float32(target)
		progress = snapProgress{
			SnapStep: BlockSyncProgress,
			Step:     BlockSyncProgress.String(),
			BlockGap: target,
			SyncPercent: map[string]float32{
				node: pre,
			},
			Message: fmt.Sprintf("%s's block height has been synchronized %.2f%%", node, pre),
		}
	}
	p.res.send(&progress)
}

func (p *joinProcessor) sendSynchronized() {
	progress := snapProgress{
		SnapStep: Synchronized,
		Step:     Synchronized.String(),
		Message:  "new node has been synchronized",
	}
	p.res.send(&progress)
}

func (p *joinProcessor) finish() {
	progress := snapProgress{SnapStep: Success, Step: Success.String()}
	p.res.send(&progress)
}

// SnapStep represents consensus configuration change operation steps.
type SnapStep int32

const (
	// FetchProcessing represents progress of node is fetching snapshot from remote.
	FetchProcessing SnapStep = 1100 + iota
	// FetchSuccess represents node has fetched snapshot from remote.
	FetchSuccess
	// FetchFailed represents node fetch snapshot from remote failed.
	FetchFailed
	// JoinPending represents progress of node is joining chaining .
	JoinPending
	// JoinWithSnapSuccess represents node has joined the chain.
	JoinWithSnapSuccess
	// JoinFailed represents node joined the chain failed.
	JoinFailed
)

var snapStepName = map[int32]string{
	1100: "FetchProcessing",
	1101: "FetchSuccess",
	1102: "FetchFailed",
	1103: "JoinPending",
	1104: "JoinWithSnapSuccess",
	1105: "JoinFailed",
}

// Val returns SnapStep integer value.
func (s SnapStep) Val() int32 {
	return int32(s)
}

// String returns SnapStep string name.
func (s SnapStep) String() string {
	v, ok := snapStepName[int32(s)]
	if ok {
		return v
	}
	return strconv.Itoa(int(s))
}

// snapProgress represents progress result of async multiple operations.
type snapProgress struct {
	SnapStep    Step               `json:"step_code,omitempty"`
	Step        string             `json:"step,omitempty"`
	BlockGap    uint64             `json:"block_gap,omitempty"`
	SyncPercent map[string]float32 `json:"sync_percent,omitempty"`
	Message     string             `json:"message,omitempty"`
	NodeErrs    map[string]error   `json:"node_errs,omitempty"`
}

func (p *snapProgress) StepCode() Step {
	return p.SnapStep
}

func (p *snapProgress) Msg() string {
	return p.Message
}

func (p *snapProgress) Errs() map[string]error {
	return p.NodeErrs
}
