/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

// Package status provides error with code handler utility.
package status

import (
	"context"
	"encoding/hex"
	"fmt"

	"github.com/pkg/errors"
	"google.golang.org/grpc/status"

	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/sdk/codes"
)

// Status represents status with code.
type Status interface {
	Code() codes.Code
	Err() error
	Message() string
	fmt.Stringer
}

type response struct {
	res *common.Response
}

func (s *response) Err() error {
	if s.Code() == codes.OK {
		return nil
	}
	return &Error{s: s}
}

func (s *response) Code() codes.Code {
	if s == nil || s.res == nil {
		return codes.OK
	}
	if s.res.Status == common.SUCCESS {
		return codes.OK
	}
	return codes.Code(s.res.Status)
}

func (s *response) String() string {
	return s.res.Status.String()
}

func (s *response) Message() string {
	return s.res.StatusInfo
}

type txResult struct {
	res *common.TxResult
}

func (s *txResult) Err() error {
	if s.Code() == codes.OK {
		return nil
	}
	return &Error{s: s}
}

func (s *txResult) Code() codes.Code {
	if s == nil || s.res == nil {
		return codes.OK
	}
	if s.res.Status == common.VALID {
		return codes.OK
	}
	return codes.Code(s.res.Status)
}

func (s *txResult) String() string {
	return s.res.Status.String()
}

func (s *txResult) Message() string {
	return hex.EncodeToString(s.res.TxHash)
}

// New Status with proto common.Status.
func New(c common.Status, msg string) Status {
	return &response{res: &common.Response{Status: c, StatusInfo: msg}}
}

// Newf Status with proto common.Status and format message.
func Newf(c common.Status, format string, args ...interface{}) Status {
	return New(c, fmt.Sprintf(format, args...))
}

// NewStatus constructs Status with internal status code.
func NewStatus(c codes.Code, msg string) Status {
	return &response{res: &common.Response{Status: common.Status(c), StatusInfo: msg}}
}

// NewStatusf constructs Status with internal status code and format message.
func NewStatusf(c codes.Code, format string, args ...interface{}) Status {
	return NewStatus(c, fmt.Sprintf(format, args...))
}

// FromProto converts common.Response to Status.
func FromProto(s *common.Response) Status {
	return New(s.Status, s.StatusInfo)
}

// FromTxProto converts common.TxResult to Status.
func FromTxProto(s *common.TxResult) Status {
	return &txResult{res: s}
}

// Err returns error from proto common.Status.
func Err(c common.Status, msg string) error {
	return New(c, msg).Err()
}

// IsProposed checks Status is TxResult or not.
func IsProposed(s Status) bool {
	_, ok := s.(*txResult)
	return ok
}

// Error represents custom error for status code.
type Error struct {
	s Status
}

// ChainStatus returns Status from error.
func (e *Error) ChainStatus() Status {
	return e.s
}

// Error returns error message.
func (e *Error) Error() string {
	return fmt.Sprintf("chain error: code = %s decs = %s", e.s.String(), e.s.Message())
}

// FromError converts error to Status, if error is original status error returns true.
func FromError(err error) (Status, bool) {
	if err == nil {
		return nil, true
	}
	if se, ok := err.(interface { // nolint:errorlint // need to convert error to Status
		ChainStatus() Status
	}); ok {
		return se.ChainStatus(), true
	}
	return NewStatus(codes.Unknown, err.Error()), false
}

// FromRPCError converts error to grpc status, returns true if error returned from grpc.
func FromRPCError(err error) (*status.Status, bool) {
	return status.FromError(err)
}

// FromContextError converts error to Status which occurs by context.
func FromContextError(err error) Status {
	if err == nil {
		return nil
	}
	if errors.Is(err, context.DeadlineExceeded) {
		return NewStatus(codes.DeadlineExceeded, err.Error())
	}
	if errors.Is(err, context.Canceled) {
		return NewStatus(codes.Canceled, err.Error())
	}
	return NewStatus(codes.Unknown, err.Error())
}

// ResponseToErr converts proto common.Response to error.
func ResponseToErr(response *common.Response) error {
	return FromProto(response).Err()
}
