/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

// Package test is the implementation of flow sdk mock server for go client.
package test

import (
	"context"
	"crypto/tls"
	"crypto/x509"
	"net"
	"strconv"

	"github.com/pkg/errors"
	"google.golang.org/grpc"
	"google.golang.org/grpc/credentials"

	"huawei.com/huaweichain/common/logger"
	"huawei.com/huaweichain/common/version"
	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
)

var cacert = []byte("-----BEGIN CERTIFICATE-----\nMIICQTCCAeigAwIBAgIRAOnLWWnBXGtly8VR+u1eRAAwCgYIKoZIzj0EAwIwWTEJ\nMAcGA1UEBhMAMQkwBwYDVQQIEwAxCTAHBgNVBAcTADENMAsGA1UEChMEb3JnMTEJ\nMAcGA1UECxMAMRwwGgYDVQQDExNjYS5vcmcxLmV4YW1wbGUuY29tMCAXDTIyMDUx\nNjEyNTcxM1oYDzIwNTIwNTA4MTI1NzEzWjBZMQkwBwYDVQQGEwAxCTAHBgNVBAgT\nADEJMAcGA1UEBxMAMQ0wCwYDVQQKEwRvcmcxMQkwBwYDVQQLEwAxHDAaBgNVBAMT\nE2NhLm9yZzEuZXhhbXBsZS5jb20wWTATBgcqhkjOPQIBBggqhkjOPQMBBwNCAARi\nYhaWVqpbEMWh/n9R9CQBzDiNf+6yPrtG3TNtss1kdMh8bgB399DKFQj0qtnfZ28M\nvdfMuF5fwztUOuR+BsRjo4GOMIGLMA4GA1UdDwEB/wQEAwIBpjAdBgNVHSUEFjAU\nBggrBgEFBQcDAQYIKwYBBQUHAwIwDwYDVR0TAQH/BAUwAwEB/zApBgNVHQ4EIgQg\nJ+GfTSZKvS5ITEg9iXwrEMqce/vZKWx8Yx85/YZSLzswHgYDVR0RBBcwFYITY2Eu\nb3JnMS5leGFtcGxlLmNvbTAKBggqhkjOPQQDAgNHADBEAiBPdjfl30+86BOV/E+I\nGzwoJA5B0YWSUOsYavs3zdYdxgIgIfveyTdy2kv6IzM0QzIxt++Pzgw/JHFFJoEu\ngG50mVw=\n-----END CERTIFICATE-----")
var serverkey = []byte("-----BEGIN PRIVATE KEY-----\nMIGHAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBG0wawIBAQQgiz6Of/dGRGynSBWL\nVMjl9pXI1tx0cU9Sf4HVn21ePIqhRANCAATE09sP+hBko2umBX0V8rDDie/B6fFi\nI2PP3RTT52BTBHsZKDAFwF/4JXSfjPXhI5IBQVxj7/8DCNJW1xjzLSaS\n-----END PRIVATE KEY-----\n")
var servercert = []byte("-----BEGIN CERTIFICATE-----\nMIICJTCCAcugAwIBAgIQX4caHhz5oO81O63a+fAwkjAKBggqhkjOPQQDAjBZMQkw\nBwYDVQQGEwAxCTAHBgNVBAgTADEJMAcGA1UEBxMAMQ0wCwYDVQQKEwRvcmcxMQkw\nBwYDVQQLEwAxHDAaBgNVBAMTE2NhLm9yZzEuZXhhbXBsZS5jb20wIBcNMjIwNTE2\nMTI1NzEzWhgPMjA1MjA1MDgxMjU3MTNaMFwxCTAHBgNVBAYTADEJMAcGA1UECBMA\nMQkwBwYDVQQHEwAxDTALBgNVBAoTBG9yZzExCTAHBgNVBAsTADEfMB0GA1UEAxMW\ncGVlcjAub3JnMS5leGFtcGxlLmNvbTBZMBMGByqGSM49AgEGCCqGSM49AwEHA0IA\nBMTT2w/6EGSja6YFfRXysMOJ78Hp8WIjY8/dFNPnYFMEexkoMAXAX/gldJ+M9eEj\nkgFBXGPv/wMI0lbXGPMtJpKjcDBuMA4GA1UdDwEB/wQEAwIFoDAMBgNVHRMBAf8E\nAjAAMCsGA1UdIwQkMCKAICfhn00mSr0uSExIPYl8KxDKnHv72SlsfGMfOf2GUi87\nMCEGA1UdEQQaMBiCFnBlZXIwLm9yZzEuZXhhbXBsZS5jb20wCgYIKoZIzj0EAwID\nSAAwRQIhAJ//vj9W88m3ZfueRFvw8XImAcwIyt6wSz/WFJANZaQfAiAiJpMV3Lo6\nf9haXGsd3tW1ztJUxu3HpUeLBrP+cy29DQ==\n-----END CERTIFICATE-----\n")
var log = logger.GetModuleLogger("go flow sdk", "test")

// sglkChainService is used to mock chain service in server
type sglkChainService struct {
	nodeservice.UnimplementedChainManagerServer
	c *ChainMgr
}

type sglkContractService struct {
	nodeservice.UnimplementedContractServer
}

type sglkTransactionService struct {
	nodeservice.UnimplementedTransactionSenderServer
}

// SendTransaction is used to mock sendTransaction
func (*sglkTransactionService) SendTransaction(ctx context.Context, req *common.RawMessage) (*common.RawMessage, error) {
	log.Info("send transaction successfully")
	return ConstructResponse(common.SUCCESS, "", nil), nil
}

// Invoke is used to mock Invoke
func (*sglkContractService) Invoke(ctx context.Context, req *common.RawMessage) (*common.RawMessage, error) {
	log.Info("invoke successfully")
	return ConstructResponse(common.SUCCESS, "", nil), nil
}

// Import is used to mock Import
func (*sglkContractService) Import(ctx context.Context, req *common.RawMessage) (*common.RawMessage, error) {
	log.Info("import contract successfully")
	return ConstructResponse(common.SUCCESS, "", nil), nil
}

// ConstructResponse is used to construct rawMessage
func ConstructResponse(status common.Status, statusInfo string, payload []byte) *common.RawMessage {
	var response *common.Response
	if payload != nil {
		response = &common.Response{Status: status, StatusInfo: statusInfo, Payload: payload}
	} else {
		response = &common.Response{Status: status, StatusInfo: statusInfo}
	}

	responseBytes, err := proto.Marshal(response)
	if err != nil {
		return nil
	}

	rawMessage := &common.RawMessage{Payload: responseBytes, PlatformVersion: version.PlatformVersion}

	return rawMessage
}

// ParseGenesisBlock is used to parse chain config from genesis block.
func ParseGenesisBlock(genesisBlock *common.Block) (*common.ChainConfig, error) {
	var err error
	blockBody := &common.BlockBody{}
	if err = proto.Unmarshal(genesisBlock.Body, blockBody); err != nil {
		return nil, errors.WithMessage(err, "unmarshal block body error")
	}
	if len(blockBody.TxList) < 1 {
		return nil, errors.Errorf("config tx not exist")
	}
	txPayload := &common.TxPayload{}
	if err = proto.Unmarshal(blockBody.TxList[0].GetFull().Payload, txPayload); err != nil {
		return nil, errors.WithMessage(err, "unmarshal txPayload error")
	}
	voteTxData := &common.VoteTxData{}
	if err = proto.Unmarshal(txPayload.Data, voteTxData); err != nil {
		return nil, errors.WithMessage(err, "unmarshal chain config error")
	}
	chainConfig := &common.ChainConfig{}
	if err = proto.Unmarshal(voteTxData.Payload, chainConfig); err != nil {
		return nil, errors.WithMessage(err, "unmarshal chain config error")
	}
	return chainConfig, nil
}

// CreateChain is used to mock createChain
func (s *sglkChainService) CreateChain(ctx context.Context, rawMsg *common.RawMessage) (*common.RawMessage, error) {
	info := &nodeservice.CreateInfo{}
	if err := proto.Unmarshal(rawMsg.Payload, info); err != nil {
		return ConstructResponse(common.BAD_REQUEST,
			"The payload of RawMessage is not CreateInfo.", nil), nil
	}
	genesisBlock := &common.Block{}
	if err := proto.Unmarshal(info.GenesisBlock, genesisBlock); err != nil {
		return ConstructResponse(common.BAD_REQUEST,
			"unmarshal GenesisBlock error", nil), nil
	}
	chainConfig, _ := ParseGenesisBlock(genesisBlock)
	log.Info("create chain successfully")
	s.c, _ = NewChainMgr()
	s.c.chains[chainConfig.ChainId] = "sglkMock"
	return ConstructResponse(common.SUCCESS, "", nil), nil
}

// DeleteChain is used to mock deleteChain
func (s *sglkChainService) DeleteChain(ctx context.Context, rawMsg *common.RawMessage) (*common.RawMessage, error) {
	info := &nodeservice.DeleteInfo{}
	if err := proto.Unmarshal(rawMsg.Payload, info); err != nil {
		rawMessage := ConstructResponse(common.BAD_REQUEST,
			"The RawMessage payload is wrong.", nil)
		return rawMessage, nil
	}
	chainID := info.ChainId
	if s.c == nil || s.c.chains == nil {
		log.Info("chain is empty")
		return nil, errors.New("chain is empty")
	}
	_, ok := s.c.chains[chainID]
	if !ok {
		log.Infof("chain [%s] not found", chainID)
		return nil, errors.Errorf("chain [%s] not found", chainID)
	}
	delete(s.c.chains, chainID)
	log.Infof("delete chain id [%s] successfully", info.ChainId)
	rawMessage := ConstructResponse(common.SUCCESS, "", nil)
	return rawMessage, nil
}

// QueryChainInfo is used to queryChainInfo
func (s *sglkChainService) QueryChainInfo(ctx context.Context, rawMsg *common.RawMessage) (*common.RawMessage, error) {
	if s.c == nil || s.c.chains == nil {
		log.Infof("chain is empty")
		return nil, errors.New("chain is empty")
	}
	info := &nodeservice.QueryInfo{}
	if err := proto.Unmarshal(rawMsg.Payload, info); err != nil {
		rawMessage := ConstructResponse(common.BAD_REQUEST,
			"The RawMessage payload is not query chain info request.", nil)
		return rawMessage, nil
	}
	_, ok := s.c.chains[info.ChainId]
	if !ok {
		log.Infof("chain [%s] not found", info.ChainId)
		return nil, errors.Errorf("chain [%s] not found", info.ChainId)
	}
	log.Infof("query action - chain [%s] info is [%s]", info.ChainId, s.c.chains[info.ChainId])
	rawMessage := ConstructResponse(common.SUCCESS, "", nil)
	return rawMessage, nil
}

// QueryAllChainInfos is used to query all chain info
func (s *sglkChainService) QueryAllChainInfos(ctx context.Context, req *common.RawMessage) (*common.RawMessage, error) {
	if s.c == nil {
		log.Info("chain is empty")
		return nil, errors.New("chain is empty")
	}
	for key, value := range s.c.chains {
		log.Infof("chain id = [%s],chain message is [%s]", key, value)
	}
	rawMessage := ConstructResponse(common.SUCCESS, "", nil)
	return rawMessage, nil
}

// GetFreePort is used to get free local port
func GetFreePort() (int, error) {
	addr, err := net.ResolveTCPAddr("tcp", "localhost:0")
	if err != nil {
		return 0, err
	}
	l, err := net.ListenTCP("tcp", addr)
	if err != nil {
		return 0, err
	}
	defer func() {
		err := l.Close()
		if err != nil {
			log.Fatal("GetFreePort - close error")
		}
	}()
	return l.Addr().(*net.TCPAddr).Port, nil
}

func startServe(freePort int) {
	certificate, err := tls.X509KeyPair(servercert, serverkey)
	if err != nil {
		log.Fatal(err.Error())
	}
	certPool := x509.NewCertPool()
	if ok := certPool.AppendCertsFromPEM(cacert); !ok {
		log.Fatal("failed to append certs")
	}
	_ = credentials.NewTLS(&tls.Config{
		MinVersion:   tls.VersionTLS12,
		Certificates: []tls.Certificate{certificate},
		ClientAuth:   tls.RequireAndVerifyClientCert, // NOTE: this is optional!
		ClientCAs:    certPool,
	})
	rpcServe := grpc.NewServer()
	lis, err := net.Listen("tcp", ":"+strconv.Itoa(freePort))
	if err != nil {
		log.Fatal(err.Error())
	}
	nodeservice.RegisterChainManagerServer(rpcServe, &sglkChainService{})
	nodeservice.RegisterContractServer(rpcServe, &sglkContractService{})
	nodeservice.RegisterTransactionSenderServer(rpcServe, &sglkTransactionService{})
	go func() {
		err := rpcServe.Serve(lis)
		if err != nil {
			log.Fatal(err.Error())
		}
	}()
}

// ChainMgr is used to mock chain mgt storage
type ChainMgr struct {
	chains map[string]string
}

// NewChainMgr is used to create a chainMgr instance
func NewChainMgr() (*ChainMgr, error) {
	return &ChainMgr{
		chains: make(map[string]string),
	}, nil
}
