/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

// Package types provides some basic data structure definitions for ease of use.
package types

import (
	"encoding/hex"
	"strings"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/sdk/action"
	"huawei.com/huaweichain/sdk/action/event"
)

type (
	// Hash defined 32-bytes hash byte array.
	Hash [hashLen]byte
	// Address defined 20-bytes address byte array.
	Address [addrLen]byte
)

// Bytes returns byte slice from hash.
func (h Hash) Bytes() []byte {
	return h[:]
}

// Hex returns hex string from hash.
func (h Hash) Hex() string {
	return hex.EncodeToString(h[:])
}

// StringToHash converts hex string format to Hash.
func StringToHash(hash string) (Hash, error) {
	var h Hash
	if hash == "" {
		return h, errors.New("hash is empty")
	}
	hashBytes, err := hex.DecodeString(strings.TrimPrefix(hash, "0x"))
	if err != nil {
		return h, errors.Errorf("hash %s should be hex string", hash)
	}
	if len(hashBytes) != hashLen {
		return h, errors.Errorf("hash %s length should be 32 bytes", hash)
	}
	return BytesToHash(hashBytes)
}

const (
	hashLen = 32
	addrLen = 20
)

// BytesToHash converts byte slice to Hash, returns error if bytes length is not equal to 32.
func BytesToHash(b []byte) (Hash, error) {
	var hash Hash
	if len(b) != hashLen {
		return hash, errors.New("hash length should be 32 bytes")
	}
	copy(hash[:], b)
	return hash, nil
}

// Bytes returns byte slice from Address.
func (a Address) Bytes() []byte {
	return a[:]
}

// Hex returns address' hex string format.
func (a Address) Hex() string {
	return hex.EncodeToString(a[:])
}

// StringToAddress converts hex string to Address.
func StringToAddress(address string) (Address, error) {
	var addr Address
	if address == "" {
		return addr, errors.New("address is empty")
	}
	addrBytes, err := hex.DecodeString(strings.TrimPrefix(address, "0x"))
	if err != nil {
		return addr, errors.Wrapf(err, "address %s should be hex string format", address)
	}
	if len(addrBytes) != addrLen {
		return addr, errors.New("address length must be 20 bytes")
	}
	return BytesToAddress(addrBytes)
}

// BytesToAddress converts byte slice to Address.
func BytesToAddress(b []byte) (Address, error) {
	var addr Address
	if len(b) != addrLen {
		return addr, errors.New("address length should be 20 bytes")
	}
	copy(addr[:], b)
	return addr, nil
}

// EventSource represents the way for listening transaction result from event.
type EventSource int32

// event source type.
const (
	BlockEventSource = EventSource(event.Block)
	TxEventSource    = EventSource(event.Tx)
)

// Receipt represents transaction result.
type Receipt struct {
	TxHash      string
	Status      common.TxStatus
	Msg         string
	ReturnValue []byte
}

// Ok checks transaction is valid.
func (r *Receipt) Ok() bool {
	return r.Status == common.VALID
}

// StreamIter represents an iterator for querying a large number of data from contract.
type StreamIter struct {
	*action.StreamResultIterator
}

// Close iterator and terminates underlying rpc stream for resource release.
func (s *StreamIter) Close() {
	s.StreamResultIterator.Cancel()
}

// Subject represents vote handler subject.
type Subject string

const (
	// SubjectPolicy is subject of configuration voting policy.
	SubjectPolicy Subject = "policy"
	// SubjectLifecycle is subject of lifecycle voting.
	SubjectLifecycle Subject = "lifecycle"
	// SubjectShardPolicy is subject of shard policy voting.
	SubjectShardPolicy Subject = "shard_policy"
	// SubjectOrg is subject of organization configuration.
	SubjectOrg Subject = "org"
	// SubjectConsensus is subject of consensus configuration change voting.
	SubjectConsensus Subject = "consensus"
	// SubjectDomain is subject of domain configuration.
	SubjectDomain Subject = "domain"
	// SubjectZone is subject of zone configuration.
	SubjectZone Subject = "zone"
)
