/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package flow

import (
	"context"
	"sync"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/logger"
)

// Future represents a future result for async task.
type Future interface {
	Get() <-chan Response
	Cancel()
	Err() error
}

// Response represents s single result for async task.
type Response interface {
	StepCode() Step
	Msg() string
	Errs() map[string]error
}

// Step represents enum of multiple steps for async task result.
type Step interface {
	String() string
	Val() int32
}

// result represents result and context for async operation request.
type result struct {
	mu     sync.Mutex
	res    chan Response
	done   bool
	err    error
	cancel context.CancelFunc
	lg     *logger.WienerLogger
}

// Cancel operation context.
func (r *result) Cancel() {
	r.cancel()
}

// Get returns channel for progress result.
func (r *result) Get() <-chan Response {
	return r.res
}

// Err returns abort error.
func (r *result) Err() error {
	r.mu.Lock()
	err := r.err
	r.mu.Unlock()
	return err
}

// send process to process channel.
func (r *result) send(progress Response) {
	r.mu.Lock()
	defer r.mu.Unlock()

	if r.done || r.err != nil {
		return
	}

	defer func() {
		if err := recover(); err != nil {
			r.lg.Errorf("Error snapProgress = %+v, %+v", progress, err)
		}
	}()

	select {
	case r.res <- progress:
	case <-r.res:
		select {
		case r.res <- progress:
			r.lg.Debugf("replace previous progress from result channel with new %+v", progress)
		default:
			r.lg.Panicf("invalid branch "+
				"when replace previous progress from result channel with new %+v", progress)
		}
	}
	if progress.StepCode() == Success {
		r.done = true
		close(r.res)
	}
	if progress.StepCode() == Abort {
		r.err = errors.New(progress.Msg())
		close(r.res)
	}
}
