//go:build xmart && linux
// +build xmart,linux

/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 */

package flow

import (
	"context"
	"encoding/hex"
	"strings"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/xmart"
	"huawei.com/huaweichain/sdk/flow/internal/cgo"
	"huawei.com/huaweichain/sdk/flow/status"
	"huawei.com/huaweichain/sdk/flow/types"
)

// XmartContract returns contract flow for specific xmart platform.
func (u *UserFlow) XmartContract(chain string) *XmartFlow {
	return newXmartFlow(u.Contract(chain, types.NativeContractXmart))
}

// XmartFlow is used to handle xmart contract related process
// which includes deploy, transact and call
type XmartFlow struct {
	flow     *ContractFlow
	metadata []byte
	address  string
}

const (
	deployXmart = "deploy"
	callXmart   = "call"
	getAccount  = "account"
	getMetadata = "metadata"

	hexAddressLen = 64
)

func newXmartFlow(flow *ContractFlow) *XmartFlow {
	return &XmartFlow{
		flow: flow,
	}
}

// Endorser sets available node list for endorsement.
func (x *XmartFlow) Endorser(endorser string) *XmartFlow {
	x.flow.Endorser(endorser)
	return x
}

// Listener sets listener for transaction result event.
func (x *XmartFlow) Listener(listener string) *XmartFlow {
	x.flow.Listener(listener)
	return x
}

// Consenter is used to set consenter list
func (x *XmartFlow) Consenter(consenter ...string) *XmartFlow {
	x.flow.Consenter(consenter...)
	return x
}

// Metadata sets xmart contract metadata for xmart deploy/transact/call.
func (x *XmartFlow) Metadata(metadata []byte) *XmartFlow {
	x.metadata = metadata
	return x
}

// Address sets xmart contract address for xmart call/transact.
func (x *XmartFlow) Address(address string) *XmartFlow {
	x.address = address
	return x
}

// EventSource sets event source type for transaction result.
func (x *XmartFlow) EventSource(source types.EventSource) *XmartFlow {
	x.flow.c.Event(source)
	return x
}

// Deploy represents contract function name for deploying compiled contract bytecode.
func (x *XmartFlow) Deploy(ctx context.Context, wasm []byte, method string,
	args ...string) (*types.Receipt, error) {
	if len(x.metadata) == 0 {
		return nil, errors.New("xmart metadata not set, please set metadata first")
	}
	inputStr, err := x.generateEncodedInput(x.metadata, method, args...)
	if err != nil {
		return nil, err
	}
	return x.flow.Invoke(ctx, deployXmart, hex.EncodeToString(wasm), hex.EncodeToString(x.metadata), inputStr)
}

// Transact represents contract function name for executing encoded command for xmart transaction.
func (x *XmartFlow) Transact(ctx context.Context, method string, args ...string) (*types.Receipt, error) {
	if err := x.checkMetaAndAddr(); err != nil {
		return nil, err
	}
	inputStr, err := x.generateEncodedInput(x.metadata, method, args...)
	if err != nil {
		return nil, err
	}
	return x.invoke(ctx, method, x.address, inputStr) // method is defined in contract
}

func (x *XmartFlow) invoke(ctx context.Context, method string, args ...string) (*types.Receipt, error) {
	helper := x.flow.stub.ContractBuilder()
	invocation := helper.BuildContractInvocation(x.flow.contract, callXmart, args) // method in invocation is "call"
	tx, val, err := x.flow.endorse(ctx, helper, invocation)
	if err != nil {
		return nil, err
	}
	finalVal, err := x.validateXmartRes(val, method)
	if err != nil {
		return nil, err // return this err directly, not carry message
	}
	receipt, err := x.flow.c.SendTx(ctx, tx)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to send transaction to consensus")
	}
	receipt.ReturnValue = finalVal
	return receipt, nil
}

func (x *XmartFlow) validateXmartRes(encodedOut []byte, method string) ([]byte, error) {
	res := cgo.DecodeXmartRes(x.metadata, encodedOut, method)
	if strings.HasPrefix(res, "Err") && strings.HasSuffix(res, ")") {
		return nil, status.Newf(common.BAD_REQUEST, "xmart contract execute result is error: %s", res).Err()
	}
	return []byte(res), nil
}

// Call represents contract function name for executing encoded command for xmart.
func (x *XmartFlow) Call(ctx context.Context, method string, args ...string) (string, error) {
	if err := x.checkMetaAndAddr(); err != nil {
		return "", err
	}
	inputStr, err := x.generateEncodedInput(x.metadata, method, args...)
	if err != nil {
		return "", err
	}
	res, err := x.flow.Query(ctx, callXmart, x.address, inputStr)
	if err != nil {
		return "", err
	}
	finalVal, err := x.validateXmartRes(res, method)
	if err != nil {
		return "", err // return this err directly, not carry message
	}
	return string(finalVal), nil
}

// Account represents contract function name for querying account address of caller.
func (x *XmartFlow) Account(ctx context.Context) (string, error) {
	res, err := x.flow.Query(ctx, getAccount)
	if err != nil {
		return "", err
	}
	if len(res) != hexAddressLen {
		return "", errors.Errorf("valid account address length %d not equal 32", len(res))
	}
	return string(res), nil
}

// GetMetadata return metadata, if metadata in cache is not available, query from blockchain.
func (x *XmartFlow) GetMetadata(ctx context.Context) ([]byte, error) {
	if len(x.address) == 0 {
		return nil, errors.New("xmart address not set, please set address first")
	}
	if len(x.metadata) > 0 {
		return x.metadata, nil
	}
	return x.flow.Query(ctx, getMetadata, x.address)
}

func (x *XmartFlow) generateEncodedInput(meta []byte, method string, args ...string) (string, error) {
	inputArgs := &xmart.XmartContractArgs{
		Metadata: meta,
		FuncName: method,
		Args:     args,
	}
	encodedInput, err := cgo.EncodeXmartInput(inputArgs)
	if err != nil {
		return "", err
	}
	return hex.EncodeToString(encodedInput), nil
}

func (x *XmartFlow) checkMetaAndAddr() error {
	if len(x.metadata) == 0 {
		return errors.New("xmart metadata not set, please set metadata first")
	}
	if len(x.address) == 0 {
		return errors.New("xmart address not set, please set address first")
	}
	return nil
}
