/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2020. All rights reserved.
 */

// Package node is the definition of node.
package node

import (
	"fmt"
	"io/ioutil"
	"path/filepath"

	"github.com/pkg/errors"
	"google.golang.org/grpc/credentials"

	"huawei.com/huaweichain/common/cryptomgr/cryptoimpl"
	"huawei.com/huaweichain/sdk/action"
	"huawei.com/huaweichain/sdk/config"
	"huawei.com/huaweichain/sdk/crypto"
)

// WNode is the definition of WNode.
type WNode struct {
	*config.Node
	TLS            *crypto.TLS
	cpt            crypto.Crypto
	ChainAction    *action.ChainAction
	ContractAction *action.ContractAction
	EventAction    *action.EventAction
	QueryAction    *action.QueryAction
	tlsEnable      bool
}

// NewNode is used to create a wnode proxy.
func NewNode(node *config.Node, cpt crypto.Crypto, tlsEnable bool, tls *crypto.TLS) (*WNode, error) {
	config, err := getConfig(node, cpt, tlsEnable, tls)
	if err != nil {
		return nil, errors.WithMessage(err, "get config error")
	}
	return &WNode{
		Node:           node,
		tlsEnable:      tlsEnable,
		TLS:            tls,
		cpt:            cpt,
		ChainAction:    action.NewChainAction(config),
		ContractAction: action.NewContractAction(config),
		EventAction:    action.NewEventAction(config),
		QueryAction:    action.NewQueryAction(config),
	}, nil
}

// GetNodeAddr is used to get node address.
func (n *WNode) GetNodeAddr() string {
	return fmt.Sprintf("%s:%d", n.Host, n.Port)
}

// ConvertTLS is used to convert config TLS to node TLS.
func ConvertTLS(tls config.TLS, algorithm string, decrypt func(bytes []byte) ([]byte, error)) (*crypto.TLS, error) {
	certPEMBlock, err := ioutil.ReadFile(filepath.Clean(filepath.FromSlash(tls.CertPath)))
	if err != nil {
		return nil, errors.WithMessagef(err, "read file failed, file: %v", tls.CertPath)
	}
	certPEMBlock, err = decrypt(certPEMBlock)
	if err != nil {
		return nil, errors.WithMessagef(err, "decrypt cert failed, file: %v", tls.CertPath)
	}

	roots := make([][]byte, len(tls.RootPath))
	for i, rootCertPath := range tls.RootPath {
		root, err := ioutil.ReadFile(filepath.Clean(filepath.FromSlash(rootCertPath)))
		if err != nil {
			return nil, errors.WithMessagef(err, "read file failed, file: %v", rootCertPath)
		}
		root, err = decrypt(root)
		if err != nil {
			return nil, errors.WithMessagef(err, "decrypt root cert failed, file: %v", rootCertPath)
		}
		roots[i] = root
	}
	keyPEMBlock, err := ioutil.ReadFile(filepath.Clean(filepath.FromSlash(tls.KeyPath)))
	if err != nil {
		return nil, errors.WithMessagef(err, "read file failed, file: %v", tls.KeyPath)
	}
	keyPEMBlock, err = decrypt(keyPEMBlock)
	if err != nil {
		if keyPEMBlock != nil {
			cryptoimpl.CleanBytes(keyPEMBlock)
		}
		return nil, errors.WithMessagef(err, "decrypt key failed, file: %v", tls.KeyPath)
	}
	return crypto.NewTLS(certPEMBlock, keyPEMBlock, roots, algorithm), nil
}

func getConfig(node *config.Node, cpt crypto.Crypto, tlsEnable bool, tls *crypto.TLS) (*action.Config, error) {
	if node.HostOverride == "" {
		node.HostOverride = node.Host
	}

	hostInfo := action.HostInfo{
		HostOverride: node.HostOverride,
		Host:         node.Host,
		Port:         node.Port,
	}

	var tc credentials.TransportCredentials
	if tlsEnable {
		c, err := tls.GetTransportCredentials(node.HostOverride)
		if err != nil {
			return nil, errors.WithMessage(err, "get ssl context error")
		}
		tc = c
	}
	return &action.Config{
		Addr:        hostInfo,
		Crypto:      cpt,
		TLSEnable:   tlsEnable,
		Credentials: tc,
		Compress:    node.RPCCompress,
	}, nil
}

// TLSEnable checks whether tls config is enabled.
func (n *WNode) TLSEnable() bool {
	return n.tlsEnable
}

// CleanSenInfo clean sensitive info in WNode
func (n *WNode) CleanSenInfo() {
	if n.cpt != nil {
		n.cpt.CleanSenInfo()
		n.cpt = nil
	}
	if n.TLS != nil {
		n.TLS.CleanSenInfo()
		n.TLS = nil
	}
}
