/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

package rawmessage

import (
	"github.com/gogo/protobuf/proto"
	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/util"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/proto/shard"
	"huawei.com/huaweichain/sdk/config"
	"huawei.com/huaweichain/sdk/crypto"
	"huawei.com/huaweichain/sdk/genesisblock"
	"huawei.com/huaweichain/sdk/utils"
)

const (
	// HANDLER is the key value of handler.
	HANDLER = "config"
	// BlockNumber is the specified block number.
	BlockNumber = 0
)

// ChainRawMessage is the definition of ChainRawMessage.
type ChainRawMessage struct {
	builder MsgBuilder
	crypto  crypto.Crypto
}

// NewChainRawMessage is used to create an instance of chain raw message by specifying an message builder.
func NewChainRawMessage(builder MsgBuilder, crypto crypto.Crypto) *ChainRawMessage {
	return &ChainRawMessage{builder: builder, crypto: crypto}
}

// BuildGenesisBlock is used to build genesis block for create chain by specifying genesis block config file path.
func (msg *ChainRawMessage) BuildGenesisBlock(chainID string,
	genesisConfigPath string, decrypts ...func(bytes []byte) ([]byte, error)) (*common.Block, error) {
	if err := util.CheckChainID(chainID); err != nil {
		return nil, errors.WithMessage(err, "check chain id error")
	}
	genesisConfig, err := config.NewGenesisConfig(genesisConfigPath, decrypts...)
	if err != nil {
		return nil, errors.WithMessage(err, "new genesis config error")
	}

	return msg.BuildGenesisBlockWithGenesisConfig(chainID, genesisConfig)
}

// BuildGenesisBlockWithGenesisConfig is used to build genesis block with genesis config.
func (msg *ChainRawMessage) BuildGenesisBlockWithGenesisConfig(chainID string,
	genesisConfig *config.GenesisConfig) (*common.Block, error) {
	if err := util.CheckChainID(chainID); err != nil {
		return nil, errors.WithMessage(err, "check chain id error")
	}
	chainConfig, err := genesisblock.GetChainConfig(genesisConfig, chainID)
	if err != nil {
		return nil, errors.WithMessage(err, "get chain config error")
	}
	tx, err := msg.getVoteTransaction(chainID, chainConfig)
	if err != nil {
		return nil, errors.WithMessage(err, "get transaction error")
	}

	txHash := utils.HashSha256(tx.Payload)
	t := &common.Tx{
		Hash: txHash,
		Data: &common.Tx_Full{Full: tx},
	}
	blockBody := &common.BlockBody{TxList: []*common.Tx{t}}
	bodyBytes, err := proto.Marshal(blockBody)
	if err != nil {
		return nil, errors.WithMessage(err, "block body marshal error")
	}

	header := &common.BlockHeader{
		Number:    BlockNumber,
		BodyHash:  txHash,
		Timestamp: int64(utils.GenerateTimestamp()),
		Version:   chainConfig.MinPlatformVersion,
	}
	genesisBlock := &common.Block{Header: header, Body: bodyBytes}
	return genesisBlock, nil
}

// BuildQuitChainRawMessage is used to build quit chain raw message for quit chain by specifying chain id.
func (msg *ChainRawMessage) BuildQuitChainRawMessage(chainID string) (*common.RawMessage, error) {
	deleteInfo := &nodeservice.DeleteInfo{ChainId: chainID}
	payload, err := proto.Marshal(deleteInfo)
	if err != nil {
		return nil, errors.WithMessage(err, "marshal delete info error")
	}
	rawMsg, err := msg.builder.GetRawMessage(payload)
	if err != nil {
		return nil, errors.WithMessage(err, "get raw message error")
	}
	return rawMsg, nil
}

// BuildJoinChainRawMessage is used to build join chain raw message for join chain by specifying genesis block bytes.
func (msg *ChainRawMessage) BuildJoinChainRawMessage(genesisBlockBytes []byte) (*common.RawMessage, error) {
	return msg.BuildJoinMsgWithLatestConf(genesisBlockBytes, nil, getDefaultInitialEntrypoint())
}

// BuildJoinMsgWithShardInfo is used to build join chain raw message with shard info.
func (msg *ChainRawMessage) BuildJoinMsgWithShardInfo(genesisBlockBytes []byte,
	shardInfo []byte) (*common.RawMessage, error) {
	return msg.BuildJoinMsgWithLatestConf(genesisBlockBytes, nil, getDefaultInitialEntrypoint(),
		nil, shardInfo)
}

// BuildJoinMsgWithEntrypoint is used to build join chain raw message with config info.
func (msg *ChainRawMessage) BuildJoinMsgWithEntrypoint(genesisBlockBytes []byte,
	entrypoint *common.Entrypoint) (*common.RawMessage, error) {
	return msg.BuildJoinMsgWithLatestConf(genesisBlockBytes, nil, entrypoint)
}

// BuildJoinMsgWithLatestConf is used to build join chain raw message with config info and entrypoint.
func (msg *ChainRawMessage) BuildJoinMsgWithLatestConf(genesisBlockBytes []byte,
	latestConf *common.ConfigInfo, entrypoint *common.Entrypoint, configs ...[]byte) (*common.RawMessage, error) {
	if entrypoint == nil {
		entrypoint = getDefaultInitialEntrypoint()
		log.Debugf("default initial entrypoint setted, default zone id: %v", entrypoint.ZoneId)
	}
	if err := config.CheckEntryPoint(entrypoint); err != nil {
		return nil, err
	}
	entrypointBytes, err := proto.Marshal(entrypoint)
	if err != nil {
		return nil, errors.WithMessage(err, "marshal entrypoint error")
	}
	createInfo := &nodeservice.CreateInfo{
		GenesisBlock: genesisBlockBytes,
		Config:       latestConf,
		Entrypoint:   entrypointBytes,
	}
	applyConfig(createInfo, configs...)
	return msg.buildJoinChainRawMessage(createInfo)
}

const (
	hostInfoIdx  = 0
	shardInfoIdx = 1
)

func applyConfig(createInfo *nodeservice.CreateInfo, configs ...[]byte) {
	if len(configs) > hostInfoIdx {
		createInfo.HostInfo = configs[hostInfoIdx]
	}
	if len(configs) > shardInfoIdx {
		createInfo.ShardInfo = configs[shardInfoIdx]
	}
}

// BuildShardInfo is used to build shard info.
func (msg *ChainRawMessage) BuildShardInfo(configPath string) (*shard.ShardInfo, error) {
	return nil, errors.New("failed to get shard info because unsupported shard feature yet")
}

// BuildQueryChainRawMessage is used to build query chain raw message for query chain by specifying chain id.
func (msg *ChainRawMessage) BuildQueryChainRawMessage(chainID string) (*common.RawMessage, error) {
	return msg.buildQueryChainRawMessage(chainID)
}

// BuildQueryAllChainRawMessage is used to build query all chains raw message for query all chains.
func (msg *ChainRawMessage) BuildQueryAllChainRawMessage() (*common.RawMessage, error) {
	return msg.buildQueryChainRawMessage()
}

func (msg *ChainRawMessage) buildJoinChainRawMessage(createInfo *nodeservice.CreateInfo) (*common.RawMessage, error) {
	if createInfo.GenesisBlock == nil {
		return nil, errors.New("genesis block is nil")
	}
	payload, err := proto.Marshal(createInfo)
	if err != nil {
		return nil, errors.WithMessage(err, "marshal create info error")
	}
	rawMsg, err := msg.builder.GetRawMessage(payload)
	if err != nil {
		return nil, errors.WithMessage(err, "get raw message error")
	}
	return rawMsg, nil
}

func (msg *ChainRawMessage) buildQueryChainRawMessage(chainIDs ...string) (*common.RawMessage, error) {
	queryInfo := &nodeservice.QueryInfo{}
	if len(chainIDs) > 0 {
		queryInfo.ChainId = chainIDs[0]
	}
	payload, err := proto.Marshal(queryInfo)
	if err != nil {
		return nil, errors.WithMessage(err, "marshal query info error")
	}
	rawMsg, err := msg.builder.GetRawMessage(payload)
	if err != nil {
		return nil, errors.WithMessage(err, "get raw message error")
	}
	return rawMsg, nil
}

func (msg *ChainRawMessage) getVoteTransaction(chainID string,
	chainConfig *common.ChainConfig) (*common.Transaction, error) {
	if chainConfig.ChainId == "" {
		return nil, errors.New("chain id is empty")
	}
	chainConfigBytes, err := proto.Marshal(chainConfig)
	if err != nil {
		return nil, errors.WithMessage(err, "marshal chain config error")
	}
	voteTxData := &common.VoteTxData{
		Handler: HANDLER,
		Payload: chainConfigBytes,
	}
	return msg.builder.BuildVoteTx(chainID, voteTxData)
}

func getDefaultInitialEntrypoint() *common.Entrypoint {
	return &common.Entrypoint{
		ZoneId:        "",
		Coordinator:   true,
		InitialMaster: true,
	}
}
