/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

package rawmessage

import (
	"io/ioutil"
	"path/filepath"
	"strconv"
	"strings"

	"github.com/gogo/protobuf/proto"
	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/util"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/crypto"
	"huawei.com/huaweichain/sdk/utils"
)

// LifecycleRawMessage is the definition of LifecycleRawMessage.
type LifecycleRawMessage struct {
	builder MsgBuilder
	crypto  crypto.Crypto
}

// Contract is the definition of contract.
type Contract struct {
	chainID     string
	name        string
	version     string
	betaVersion uint32
}

// VoteParams is the definition of vote params.
type VoteParams struct {
	C              *Contract
	ShardID        string
	Desc           string
	Policy         string
	Schema         string
	HistorySupport bool
	InitRequired   bool
}

// NewLifecycleRawMessage is used to create an instance of lifecycle raw message by specifying an message builder.
func NewLifecycleRawMessage(builder MsgBuilder, crypto crypto.Crypto) *LifecycleRawMessage {
	return &LifecycleRawMessage{builder: builder, crypto: crypto}
}

// NewContract is used to create an instance of contract by specifying chain id, contract name and contract version.
func NewContract(chainID string, name string, version string) *Contract {
	return &Contract{chainID: chainID, name: name, version: version}
}

// NewBetaContract is used to create an instance of contract by specifying chain id, contract name and contract version.
func NewBetaContract(chainID string, name string, version string, betaVersion uint32) *Contract {
	return &Contract{
		chainID:     chainID,
		name:        name,
		version:     version,
		betaVersion: betaVersion,
	}
}

func checkContractParams(c *Contract) error {
	if c == nil {
		return errors.New("contract is nil")
	}
	if err := util.CheckParams(c.chainID, c.name, c.version); err != nil {
		return errors.WithMessage(err, "check params error")
	}
	return nil
}

// BuildImportRawMessage is used to build import raw message for import contract.
func (msg *LifecycleRawMessage) BuildImportRawMessage(c *Contract, path string,
	sandbox string, format string) (*common.RawMessage, error) {
	if err := checkContractParams(c); err != nil {
		return nil, err
	}
	if err := util.CheckParams(c.chainID, c.name, c.version); err != nil {
		return nil, errors.WithMessage(err, "check params error")
	}
	lowerSandbox := strings.ToLower(sandbox)
	var sandboxType common.ContractRunEnv
	switch lowerSandbox {
	case "docker":
		sandboxType = common.Docker
	case "native":
		sandboxType = common.Native
	case "nativewasm":
		sandboxType = common.NativeWasm
	case "teewasm":
		sandboxType = common.TEEWasm
	default:
		return nil, errors.New("unsupported sandbox type")
	}

	codes, err := ioutil.ReadFile(filepath.Clean(filepath.FromSlash(path)))
	if err != nil {
		return nil, errors.WithMessage(err, "read contract code error")
	}
	externalContractInfo := &common.ExternalContractInfo{
		SandboxType:   sandboxType,
		Code:          codes,
		SchemaVersion: c.version,
		Format:        format,
		BetaVersion:   c.betaVersion,
	}
	contractInfoType := &nodeservice.ContractInfo_ExternalContractInfo{ExternalContractInfo: externalContractInfo}
	contractInfo := &nodeservice.ContractInfo{ChainId: c.chainID, ContractName: c.name, Type: contractInfoType}
	return msg.buildImportRawMsg(contractInfo)
}

// BuildImportFabricRawMessage is used to build import raw message for import fabric contract.
func (msg *LifecycleRawMessage) BuildImportFabricRawMessage(c *Contract, path string,
	sandbox string, format string) (*common.RawMessage, error) {
	if err := checkContractParams(c); err != nil {
		return nil, err
	}
	lowerSandbox := strings.ToLower(sandbox)
	var sandboxType common.ContractRunEnv
	if lowerSandbox != "docker" {
		return nil, errors.New("unsupported sandbox type")
	}
	sandboxType = common.Docker
	codes, err := ioutil.ReadFile(filepath.Clean(filepath.FromSlash(path)))
	if err != nil {
		return nil, errors.WithMessage(err, "read contract code error")
	}
	ext := map[string][]byte{
		"contractType": []byte("fabric"),
	}

	externalContractInfo := &common.ExternalContractInfo{
		SandboxType:   sandboxType,
		Code:          codes,
		SchemaVersion: c.version,
		Format:        format,
		Extensions:    ext,
		BetaVersion:   c.betaVersion,
	}
	contractInfoType := &nodeservice.ContractInfo_ExternalContractInfo{ExternalContractInfo: externalContractInfo}
	contractInfo := &nodeservice.ContractInfo{ChainId: c.chainID,
		ContractName: c.name, Type: contractInfoType}
	return msg.buildImportRawMsg(contractInfo)
}

// checkUnImportParams check the format of contract parameters
func checkUnImportParams(c *Contract) error {
	if c == nil {
		return errors.New("contract is nil")
	}
	if err := util.CheckChainID(c.chainID); err != nil {
		return errors.WithMessage(err, "check chain id error")
	}
	if err := util.CheckContractName(c.name); err != nil {
		return errors.WithMessage(err, "check contract name error")
	}
	if c.version == "" {
		return nil
	}
	if err := util.CheckContractVersion(c.version); err != nil {
		return errors.WithMessage(err, "check contract version error")
	}
	return nil
}

// BuildUnImportRawMessage is used to build unimport raw message for unimport contract.
func (msg *LifecycleRawMessage) BuildUnImportRawMessage(c *Contract) (*common.RawMessage, error) {
	if err := checkUnImportParams(c); err != nil {
		return nil, err
	}
	externalContractInfo := &common.ExternalContractInfo{
		SchemaVersion: c.version,
	}
	contractInfoType := &nodeservice.ContractInfo_ExternalContractInfo{
		ExternalContractInfo: externalContractInfo,
	}
	contractInfo := &nodeservice.ContractInfo{
		ChainId:      c.chainID,
		ContractName: c.name,
		Type:         contractInfoType,
	}
	return msg.buildImportRawMsg(contractInfo)
}

func checkVoteParams(c *Contract, desc string, policy string) error {
	if c == nil {
		return errors.New("contract is nil")
	}
	if err := util.CheckVoteParams(c.chainID, c.name, c.version, policy, desc); err != nil {
		return errors.WithMessage(err, "check vote params error")
	}
	return nil
}

// BuildVoteRawMessage is used to build vote raw message for vote contract.
func (msg *LifecycleRawMessage) BuildVoteRawMessage(c *Contract, desc string, policy string,
	historySupport bool, initRequired bool) (*TxRawMsg, error) {
	if err := checkVoteParams(c, desc, policy); err != nil {
		return nil, err
	}
	vp := &VoteParams{
		C:              c,
		ShardID:        "",
		Desc:           desc,
		Policy:         policy,
		Schema:         "",
		HistorySupport: historySupport,
		InitRequired:   initRequired,
	}
	rawMsg, err := msg.buildVoteRawMessage(vp)
	if err != nil {
		return nil, errors.WithMessage(err, "build vote raw message error")
	}
	return rawMsg, nil
}

// BuildShardVoteRawMessage is used to build vote raw message for vote contract.
func (msg *LifecycleRawMessage) BuildShardVoteRawMessage(vp *VoteParams) (*TxRawMsg, error) {
	if err := utils.CheckShardID(vp.ShardID); err != nil {
		return nil, errors.WithMessage(err, "check shard id error")
	}
	if err := checkVoteParams(vp.C, vp.Desc, vp.Policy); err != nil {
		return nil, errors.WithMessage(err, "check vote params error")
	}
	rawMsg, err := msg.buildVoteRawMessage(vp)
	if err != nil {
		return nil, errors.WithMessage(err, "build vote raw message error")
	}
	return rawMsg, nil
}

// BuildSQLVoteRawMessage is used to build SQL vote raw message for vote contract.
// If no table to create when update contract, schema can be "".
func (msg *LifecycleRawMessage) BuildSQLVoteRawMessage(c *Contract, desc string, policy string,
	schema string, initRequired bool) (*TxRawMsg, error) {
	if err := checkVoteParams(c, desc, policy); err != nil {
		return nil, err
	}
	vp := &VoteParams{
		C:            c,
		ShardID:      "",
		Desc:         desc,
		Policy:       policy,
		Schema:       schema,
		InitRequired: initRequired,
	}
	rawMsg, err := msg.buildVoteRawMessage(vp)
	if err != nil {
		return nil, errors.WithMessage(err, "build sql vote raw message error")
	}
	return rawMsg, nil
}

func (msg *LifecycleRawMessage) buildVoteRawMessage(vp *VoteParams) (*TxRawMsg, error) {
	validatorExtensions := make(map[string][]byte)
	validatorExtensions["policy"] = []byte(vp.Policy)
	contractDefinition := &common.ContractDefinition{
		ContractName:        vp.C.name,
		SchemaVersion:       vp.C.version,
		SequenceNumber:      0,
		Description:         vp.Desc,
		HistorySupport:      vp.HistorySupport,
		RequireInit:         vp.InitRequired,
		ApprovalValidator:   "default",
		ValidatorExtensions: validatorExtensions,
		BetaVersion:         vp.C.betaVersion,
	}
	if vp.Schema != "" {
		contractDefinition.SqlDbSchema = vp.Schema
	}
	contractDefinitionBytes, err := proto.Marshal(contractDefinition)
	if err != nil {
		return nil, errors.WithMessage(err, "marshal contract definition error")
	}

	voteTxData := &common.VoteTxData{
		Handler: LifecycleHandler,
		Subject: start,
		Payload: contractDefinitionBytes,
	}

	tx, err := msg.builder.BuildVoteTx(vp.C.chainID, voteTxData)
	if err != nil {
		return nil, errors.WithMessage(err, "build vote tx error")
	}
	if vp.ShardID != "" {
		extensions := make(map[string][]byte)
		num, err := strconv.ParseUint(vp.ShardID, parseUintBase, parseUintBitSize)
		if err != nil {
			return nil, errors.WithMessage(err, "shardID is error")
		}
		extensions[TargetShard] = proto.EncodeVarint(num)
		tx.Extensions = extensions
	}

	return msg.builder.GetTxRawMsg(tx)
}

// BuildConfirmBetaRawMsg is to build contract vote raw message for confirm beta contract
func (msg *LifecycleRawMessage) BuildConfirmBetaRawMsg(chain, contract, version string,
	betaVersion uint32) (*TxRawMsg, error) {
	contractDefinition := &common.ContractDefinition{
		ContractName:  contract,
		SchemaVersion: version,
		BetaVersion:   betaVersion,
	}
	contractDefinitionBytes, err := proto.Marshal(contractDefinition)
	if err != nil {
		return nil, errors.WithMessage(err, "marshal contract definition error")
	}
	voteTxData := &common.VoteTxData{
		Handler: "lifecycle",
		Subject: confirmBetaVersion,
		Payload: contractDefinitionBytes,
	}
	tx, err := msg.builder.BuildVoteTx(chain, voteTxData)
	if err != nil {
		return nil, errors.WithMessage(err, "build vote tx error")
	}
	return msg.builder.GetTxRawMsg(tx)
}

// BuildManageRawMessage is used to build contract manage vote raw message for vote contract.
func (msg *LifecycleRawMessage) BuildManageRawMessage(chain string, contract string,
	option string) (*TxRawMsg, error) {
	if option != freeze && option != unfreeze && option != destroy {
		return nil, errors.Errorf("support option:[%s,%s,%s]", freeze, unfreeze, destroy)
	}
	voteTxData := &common.VoteTxData{
		Handler: "lifecycle",
		Subject: option,
		Payload: []byte(contract),
	}
	tx, err := msg.builder.BuildVoteTx(chain, voteTxData)
	if err != nil {
		return nil, errors.WithMessage(err, "build vote tx error")
	}
	return msg.builder.GetTxRawMsg(tx)
}

// BuildQueryStateRawMessage is used to build query contract state raw message for import contract.
func (msg *LifecycleRawMessage) BuildQueryStateRawMessage(chain string, contract string) (*common.RawMessage, error) {
	contractInfo := &nodeservice.ContractInfo{
		ChainId:      chain,
		ContractName: contract,
	}
	payload, err := proto.Marshal(contractInfo)
	if err != nil {
		return nil, errors.WithMessage(err, "marshal contract info error")
	}
	return msg.builder.GetRawMessage(payload)
}

func (msg *LifecycleRawMessage) buildImportRawMsg(
	contractInfo *nodeservice.ContractInfo) (*common.RawMessage, error) {
	if contractInfo.ChainId == "" {
		return nil, errors.New("contract info field chain id is empty")
	}
	payload, err := proto.Marshal(contractInfo)
	if err != nil {
		return nil, errors.WithMessage(err, "marshal contract info error")
	}
	rawMsg, err := msg.builder.GetRawMessage(payload)
	if err != nil {
		return nil, errors.WithMessage(err, "get raw message error")
	}
	return rawMsg, nil
}
