/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

package rawmessage

import (
	"github.com/gogo/protobuf/proto"
	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
)

const (
	// UpdateHandler is config update vote handler name
	UpdateHandler = "update"
	// LifecycleHandler is the lifecycle handler string
	LifecycleHandler = "lifecycle"

	start              = "start"
	freeze             = "freeze"
	unfreeze           = "unfreeze"
	destroy            = "destroy"
	revocation         = "revocation"
	confirmBetaVersion = "confirmBetaVersion"
	betaID             = "beta"
)

// QueryRawMessage is the definition of QueryRawMessage.
type QueryRawMessage struct {
	builder MsgBuilder
}

// NewQueryRawMessage is used to create an instance of query raw message by specifying an message builder.
func NewQueryRawMessage(builder MsgBuilder) *QueryRawMessage {
	return &QueryRawMessage{builder: builder}
}

// BuildTxRawMessage is used to build tx raw message for query transaction.
func (msg *QueryRawMessage) BuildTxRawMessage(chainID string, txHash []byte) (*common.RawMessage, error) {
	request := &nodeservice.ChainServiceRequest{ChainId: chainID,
		Type: &nodeservice.ChainServiceRequest_TxHash{TxHash: txHash}}
	return msg.buildQueryRawMessage(request)
}

// BuildLatestChainStateRawMessage is used to build latest chain state raw message for query latest chain state.
func (msg *QueryRawMessage) BuildLatestChainStateRawMessage(chainID string) (*common.RawMessage, error) {
	request := &nodeservice.ChainServiceRequest{ChainId: chainID}
	return msg.buildQueryRawMessage(request)
}

// BuildShardRouterInfoRawMessage is used to build raw message for query  shard router information.
func (msg *QueryRawMessage) BuildShardRouterInfoRawMessage(nodeID string) (*common.RawMessage, error) {
	request := &nodeservice.ShardRouterInfoRequest{NodeId: nodeID}
	return msg.buildQueryRawMessage(request)
}

// BuildShardPolicyRawMessage is used to build raw message for query  shard policy information.
func (msg *QueryRawMessage) BuildShardPolicyRawMessage(chainID string) (*common.RawMessage, error) {
	request := &nodeservice.ShardPolicyRequest{ChainId: chainID}
	return msg.buildQueryRawMessage(request)
}

// BuildGetStateByRangeRawMessage is used to build query chain state raw message by range.
// chainID: the chain name, which can not be nil
// contractName: the contract name, if nil, data of all contracts on the chain should be queried
// begin, end: the begin and end key of this query, the data range is [begin, end), for example:
// keys in state DB are "11","12","13", if begin="11", end="13", you will get "11", "12", not including "13"
// if end key is empty, the data range is [begin, utf8.MaxRune)
func (msg *QueryRawMessage) BuildGetStateByRangeRawMessage(chainID string, contractName string,
	begin string, end string) (*common.RawMessage, error) {
	request := &nodeservice.ChainStateRequest{ChainId: chainID, ContractName: contractName, Begin: begin, End: end}
	return msg.buildQueryRawMessage(request)
}

// BuildGetStateRawMessage is used to build query chain state raw message.
func (msg *QueryRawMessage) BuildGetStateRawMessage(chainID string, contractName string,
	prefix string) (*common.RawMessage, error) {
	request := &nodeservice.ChainStateRequest{ChainId: chainID, ContractName: contractName, Begin: prefix}
	return msg.buildQueryRawMessage(request)
}

// BuildBlockRawMessage is used to build block raw message for query block.
func (msg *QueryRawMessage) BuildBlockRawMessage(chainID string, blockNum uint64) (*common.RawMessage, error) {
	request := &nodeservice.ChainServiceRequest{ChainId: chainID,
		Type: &nodeservice.ChainServiceRequest_BlockNum{BlockNum: blockNum}}
	return msg.buildQueryRawMessage(request)
}

// BuildBlockAndResultRawMessage is used to build block and result raw message for query block.
func (msg *QueryRawMessage) BuildBlockAndResultRawMessage(chainID string, blockNum uint64) (*common.RawMessage,
	error) {
	return msg.BuildBlockRawMessage(chainID, blockNum)
}

func newQueryRequest(chainID, contract string) *nodeservice.ChainServiceRequest {
	return &nodeservice.ChainServiceRequest{
		ChainId:    chainID,
		Extensions: make(map[string][]byte),
		Type: &nodeservice.ChainServiceRequest_Contract{
			Contract: contract,
		},
	}
}

// BuildBetaContractRawMessage is used to build contract raw message for query contract.
func (msg *QueryRawMessage) BuildBetaContractRawMessage(chainID string, contract string) (*common.RawMessage, error) {
	request := newQueryRequest(chainID, contract)
	request.Extensions["versionType"] = []byte(betaID)
	return msg.buildQueryRawMessage(request)
}

// BuildContractRawMessage is used to build contract raw message for query contract.
func (msg *QueryRawMessage) BuildContractRawMessage(chainID string, contract string) (*common.RawMessage, error) {
	request := newQueryRequest(chainID, contract)
	return msg.buildQueryRawMessage(request)
}

// BuildQueryChainUpdateVoteRawMessage is used to build query chain update vote raw message for query chain update vote.
func (msg *QueryRawMessage) BuildQueryChainUpdateVoteRawMessage(chainID string,
	subject string) (*common.RawMessage, error) {
	s, err := Subject(subject).conv()
	if err != nil {
		return nil, err
	}
	request := &nodeservice.VoteQuery{ChainId: chainID, Handler: UpdateHandler, Subject: s}
	return msg.buildQueryRawMessage(request)
}

// BuildQueryLifecycleVoteRawMessage is used to build query lifecycle vote raw message for query lifecycle vote.
func (msg *QueryRawMessage) BuildQueryLifecycleVoteRawMessage(chainID string, contract string,
	option string) (*common.RawMessage, error) {
	if option != start && option != freeze && option != unfreeze && option != destroy {
		return nil, errors.Errorf("support option:[%s,%s,%s,%s]", start, freeze, unfreeze, destroy)
	}
	request := &nodeservice.VoteQuery{
		ChainId: chainID,
		Handler: LifecycleHandler,
		Subject: option + "_" + contract,
	}
	return msg.buildQueryRawMessage(request)
}

// BuildQueryCertStatus is used to build query cert status
func (msg *QueryRawMessage) BuildQueryCertStatus(chainID string, org string) (*common.RawMessage, error) {
	request := &nodeservice.VoteQuery{
		ChainId: chainID,
		Handler: common.CERT_STATUS_CHANGE.String(),
		Subject: org,
	}
	return msg.buildQueryRawMessage(request)
}

// GetFrozenSerial get freeze serial number from query response
func GetFrozenSerial(payload []byte) (map[string]interface{}, error) {
	updateFrozen := func(crlEntry *common.CRLEntry, frozenSet map[string]interface{}) map[string]interface{} {
		if crlEntry.CrlEntryExt == nil {
			return frozenSet
		}
		if crlEntry.CrlEntryExt.ReasonCode == common.CertificateHold {
			frozenSet[crlEntry.SerialNumber] = nil
		}
		if crlEntry.CrlEntryExt.ReasonCode == common.RemoveFromCRL {
			delete(frozenSet, crlEntry.SerialNumber)
		}
		return frozenSet
	}

	return getSerialSet(payload, updateFrozen)
}

// GetRevocationSerial get revocation serial number from query response
func GetRevocationSerial(payload []byte) (map[string]interface{}, error) {
	updateRevocation := func(crlEntry *common.CRLEntry, revocationSet map[string]interface{}) map[string]interface{} {
		if crlEntry.CrlEntryExt == nil {
			revocationSet[crlEntry.SerialNumber] = nil
		}
		return revocationSet
	}

	return getSerialSet(payload, updateRevocation)
}

func getSerialSet(payload []byte, update func(crlEntry *common.CRLEntry,
	revocationSet map[string]interface{}) map[string]interface{}) (map[string]interface{}, error) {
	revocationSet := make(map[string]interface{})
	voteQueryResponse := &nodeservice.VoteQueryResponse{}
	if err := proto.Unmarshal(payload, voteQueryResponse); err != nil {
		return nil, err
	}
	for _, vote := range voteQueryResponse.Votes {
		orgCRLInfo := &common.OrgCRLInfo{}
		if err := proto.Unmarshal(vote.Payload, orgCRLInfo); err != nil {
			return nil, err
		}
		for _, chainCRL := range orgCRLInfo.ChainCRLs {
			crlInfo := &common.CRLInfo{}
			if err := proto.Unmarshal(chainCRL.Payload, crlInfo); err != nil {
				return nil, err
			}
			for _, crlEntry := range crlInfo.Entries {
				revocationSet = update(crlEntry, revocationSet)
			}
		}
	}
	return revocationSet, nil
}

func (msg *QueryRawMessage) buildQueryRawMessage(pb proto.Message) (*common.RawMessage, error) {
	payload, err := proto.Marshal(pb)
	if err != nil {
		return nil, errors.WithMessage(err, "marshal vote query error")
	}
	return msg.builder.GetRawMessage(payload)
}

// Subject of vote configuration.
type Subject string

type innerSubject string

const (
	// SubjectPolicy is subject of configuration voting policy.
	SubjectPolicy Subject = "policy"
	// SubjectLifecycle is subject of lifecycle voting.
	SubjectLifecycle Subject = "lifecycle"
	// SubjectOrg is subject of organization configuration.
	SubjectOrg Subject = "org"
	// SubjectConsensus is subject of consensus configuration change voting.
	SubjectConsensus Subject = "consensus"
	// SubjectDomain is subject of domain configuration.
	SubjectDomain Subject = "domain"
	// SubjectZone is subject of zone configuration.
	SubjectZone Subject = "zone"

	subjectPolicy    innerSubject = "cfgPolicy"
	subjectLifecycle innerSubject = "lifecycle"
	subjectOrg       innerSubject = "organization"
	subjectConsensus innerSubject = "consensus_configuration_change_pending"
	subjectDomain    innerSubject = "DOMAIN"
	subjectZone      innerSubject = "ZONE"
)

func (s Subject) conv() (string, error) {
	switch s {
	case SubjectPolicy:
		return string(subjectPolicy), nil
	case SubjectLifecycle:
		return string(subjectLifecycle), nil
	case SubjectOrg:
		return string(subjectOrg), nil
	case SubjectConsensus:
		return string(subjectConsensus), nil
	case SubjectDomain:
		return string(subjectDomain), nil
	case SubjectZone:
		return string(subjectZone), nil
	default:
		return "", errors.New("unsupported subject " + string(s))
	}
}
