/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

// Package utils provide the common utils.
package utils

import (
	"crypto/rand"
	"crypto/sha256"
	"fmt"
	"math"
	"math/big"
	mrand "math/rand"
	"strconv"
	"sync/atomic"
	"time"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/cryptomgr"
	"huawei.com/huaweichain/common/cryptomgr/cryptoimpl"
	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
)

const (
	parseUintBase    = 10
	parseUintBitSize = 32
)

var debugFlag int32

// Debug sets flag for enable debug switch to show verbose information.
func Debug(enable bool) {
	if enable {
		atomic.StoreInt32(&debugFlag, 1)
	} else {
		atomic.StoreInt32(&debugFlag, 0)
	}

}

// IsDebug returns true if debug flag is enable.
func IsDebug() bool {
	return atomic.LoadInt32(&debugFlag) != 0
}

const (
	// ShardChainIDSep is used to separate chain id and shard id.
	ShardChainIDSep   = "-"
	defTimeoutSeconds = 60 // seconds
)

var timeout = int64(defTimeoutSeconds)

// HashSha256 computes hash by sha256 algorithm.
func HashSha256(data []byte) []byte {
	h := sha256.Sum256(data)
	return h[:]
}

// HashSM3 computes hash by sm3 algorithm.
func HashSM3(data []byte) []byte {
	hashFunc := cryptoimpl.GetHashFunc(cryptomgr.Sm3)
	if hashFunc == nil {
		return nil
	}
	return hashFunc(data)
}

// GenerateTimestamp is used to generate timestamp.
func GenerateTimestamp() uint64 {
	return uint64(time.Now().UTC().UnixNano())
}

// GenerateNonce generates a random number.
func GenerateNonce() uint64 {
	n, err := rand.Int(rand.Reader, big.NewInt(math.MaxInt64))
	if err != nil {
		mrand.Seed(time.Now().UnixNano())
		return mrand.Uint64() // nolint:gosec // this is just a backup operation for secure random
	}
	return n.Uint64()
}

// GetShards used to parse shard id from invoke response.
func GetShards(tx *common.Transaction) ([]uint32, error) {
	var txPayLoad common.TxPayload
	if err := proto.Unmarshal(tx.Payload, &txPayLoad); err != nil {
		return nil, errors.WithMessage(err, "tx payload unmarshal error")
	}

	var shardIds []uint32
	if txPayLoad.Header.Type == common.COMMON_TRANSACTION {
		var txData common.CommonTxData
		if err := proto.Unmarshal(txPayLoad.Data, &txData); err != nil {
			return nil, errors.WithMessage(err, "common tx data unmarshal error")
		}

		for _, shardUpdate := range txData.ShardStateUpdates {
			shardIds = append(shardIds, shardUpdate.ShardId)
		}
	}

	return shardIds, nil
}

// GetShardChainID is used to construct shard chain id with origin chain id and shard id.
func GetShardChainID(chainID string, shardID uint32) string {
	return fmt.Sprintf("%s%s%s", chainID, ShardChainIDSep, strconv.Itoa(int(shardID)))
}

// ShardChainID is used to construct shard chain id with origin chain id and shard id.
func ShardChainID(chainID string, shardID string) (string, error) {
	if err := CheckShardID(shardID); err != nil {
		return "", errors.WithMessage(err, "failed to check shard id")
	}
	shardIDUint32, err := strconv.ParseUint(shardID, parseUintBase, parseUintBitSize)
	if err != nil {
		return "", errors.WithMessage(err, "failed to transfer shardID from string to uint32")
	}
	return GetShardChainID(chainID, uint32(shardIDUint32)), nil
}

// CheckShardID is used to check shard id.
func CheckShardID(shardID string) error {
	_, err := strconv.ParseUint(shardID, parseUintBase, parseUintBitSize)
	if err != nil {
		return errors.WithMessage(err, "failed to transfer shardID from string to uint32")
	}
	return nil
}

// SetTimeout set client timeout (/s)
func SetTimeout(seconds time.Duration) {
	atomic.StoreInt64(&timeout, int64(seconds))
}

// GetTimeout get client timeout (/s)
func GetTimeout() time.Duration {
	return time.Duration(atomic.LoadInt64(&timeout))
}
